/*************************************************************************/
/**      **/
/**            **/
/**                **/
/**              **/
/**                                                                     **/
/**                               **/
/**                                                  **/
/*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <fcntl.h> 
#include <unistd.h>
#include <errno.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <asm/types.h> 

#include <linux/videodev2.h>

#define CLEAR(x) memset (&(x), 0, sizeof (x))

struct ST_BUFFER {
    void * start;
    size_t length;
};

static char g_szDevName[256];
static int  g_nFd = -1;
struct ST_BUFFER *g_pstBuffers = NULL;
static unsigned int g_nBuffers       = 0;

static int xioctl(int fd, int request, void *arg)
{
    int r;

    do {
        r = ioctl (fd, request, arg);
    } while (-1 == r && EINTR == errno);

    return r;
}

static int process_read_frame(char *pcaBuffer, unsigned int *punDataSize)
{
    int rc = 0;
    struct v4l2_buffer buf;
    unsigned int i;

    CLEAR (buf);

    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;

RETRY:

    if (-1 == xioctl (g_nFd, VIDIOC_DQBUF, &buf)) {
        switch (errno) {
        case EAGAIN:
#if 1
            usleep(5000);
            goto RETRY;
#endif
            return 0;

        case EIO:
            /* Could ignore EIO, see spec. */

            /* fall through */

        default:
            //fprintf(stderr, "error in VIDIOC_DQBUF");
            return -1;
        }
    }

    assert (buf.index < g_nBuffers);

    if (*punDataSize >= buf.bytesused) {
        memcpy(pcaBuffer, g_pstBuffers[buf.index].start, buf.bytesused);
        *punDataSize = buf.bytesused; 
    } else {
        fprintf(stderr, "skip this frame, buffer_size(%d), data_length(%d)", *punDataSize, buf.bytesused);
        *punDataSize = 0; 
        rc = -1;
    }

    if (-1 == xioctl (g_nFd, VIDIOC_QBUF, &buf)) {
        fprintf(stderr, "error in VIDIOC_DQBUF");
        rc = -2;
    }

    return rc;
}

int uvc_read_frame(char *pcaBuffer, unsigned int *punDataSize)
{
#if 0
    for (;;) {
        fd_set fds;
        struct timeval tv;
        int r;

        FD_ZERO (&fds);
        FD_SET (g_nFd, &fds);

        /* Timeout. */
        tv.tv_sec = 10;
        tv.tv_usec = 0;

        r = select (g_nFd + 1, &fds, NULL, NULL, &tv);

        if (-1 == r) {
            if (EINTR == errno)
                continue;

            fprintf(stderr, "error in select");
            return -1;
        }

        if (0 == r) {
            fprintf (stderr, "select timeout\n");
            return -1;
        }

        if (0 == process_read_frame(pcaBuffer, punDataSize))
            break;
        
        /* EAGAIN - continue select loop. */
    }
    return 0;
#else
    return process_read_frame(pcaBuffer, punDataSize);
#endif
}

int uvc_stop_capturing(void)
{
    enum v4l2_buf_type type;

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (-1 == xioctl (g_nFd, VIDIOC_STREAMOFF, &type)) {
        fprintf(stderr, "error in VIDIOC_STREAMOFF");
        return -1;
    }

    return 0;
}

int uvc_start_capturing(void)
{
    unsigned int i;
    enum v4l2_buf_type type;

    for (i = 0; i < g_nBuffers; ++i) {
        struct v4l2_buffer buf;

        CLEAR (buf);

        buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory      = V4L2_MEMORY_MMAP;
        buf.index       = i;

        if (-1 == xioctl (g_nFd, VIDIOC_QBUF, &buf)) {
            fprintf(stderr, "error in VIDIOC_QBUF");
            return -1;
        }
    }
                
    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (-1 == xioctl (g_nFd, VIDIOC_STREAMON, &type)) {
        fprintf(stderr, "error in VIDIOC_STREAMON");
        return -1;
    }

    return 0;
}

int uvc_uninit_device(void)
{
    unsigned int i;

    if (g_pstBuffers) {
        for (i = 0; i < g_nBuffers; ++i) {
            if (-1 == munmap (g_pstBuffers[i].start, g_pstBuffers[i].length)) {
                fprintf(stderr, "error in munmap");
                return -1;
            }
        }
        free (g_pstBuffers);
        g_pstBuffers = NULL;
    }

    return 0;
}


int uvc_init_mmap(void)
{
    struct v4l2_requestbuffers req;

    CLEAR (req);

    req.count               = 4;
    req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory              = V4L2_MEMORY_MMAP;

    if (-1 == xioctl (g_nFd, VIDIOC_REQBUFS, &req)) {
        if (EINVAL == errno) {
            fprintf (stderr, "%s does not support "
                     "memory mapping\n", g_szDevName);
            return -1;
        } else {
            fprintf(stderr, "error in VIDIOC_REQBUFS");
            return -1;
        }
    }

    if (req.count < 2) {
        fprintf (stderr, "Insufficient buffer memory on %s\n",
                 g_szDevName);
        return -1;
    }

    g_pstBuffers = calloc (req.count, sizeof (*g_pstBuffers));

    if (!g_pstBuffers) {
        fprintf (stderr, "Out of memory\n");
        return -1;
    }

    for (g_nBuffers = 0; g_nBuffers < req.count; ++g_nBuffers) {
        struct v4l2_buffer buf;

        CLEAR (buf);

        buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory      = V4L2_MEMORY_MMAP;
        buf.index       = g_nBuffers;

        if (-1 == xioctl (g_nFd, VIDIOC_QUERYBUF, &buf)) {
            fprintf(stderr, "error in VIDIOC_QUERYBUF");
            return -1;
        }

        g_pstBuffers[g_nBuffers].length = buf.length;
        g_pstBuffers[g_nBuffers].start =
            mmap (NULL /* start anywhere */,
                  buf.length,
                  PROT_READ | PROT_WRITE /* required */,
                  MAP_SHARED /* recommended */,
                  g_nFd, buf.m.offset);

        if (MAP_FAILED == g_pstBuffers[g_nBuffers].start) {
            fprintf(stderr, "error in mmap");
            return -1;
        }
    }

    return 0;
}


int uvc_init_device (unsigned int unFormat, unsigned int unWidth, unsigned int unHeight, unsigned int unFps)
{
    struct v4l2_capability cap;
    struct v4l2_format fmt;
    struct v4l2_streamparm fps;

    unsigned int min;

    if (-1 == xioctl (g_nFd, VIDIOC_QUERYCAP, &cap)) {
        if (EINVAL == errno) {
            fprintf(stderr, "%s is no V4L2 device\n",
                    g_szDevName);
            return -1;
        } else {
            fprintf(stderr, "error in VIDIOC_QUERYCAP");
            return -1;
        }
    }

    if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
        fprintf (stderr, "%s is no video capture device\n",
                 g_szDevName);
        return -1;
    }

    if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
        fprintf (stderr, "%s does not support streaming i/o\n",
                 g_szDevName);
        return -1;
    }

    /* set format */
    CLEAR (fmt);

    fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fmt.fmt.pix.width       = unWidth; 
    fmt.fmt.pix.height      = unHeight;
    fmt.fmt.pix.pixelformat = unFormat;
    fmt.fmt.pix.field       = V4L2_FIELD_ANY;

    if (-1 == xioctl (g_nFd, VIDIOC_S_FMT, &fmt)) {
        return -1;
    }

    if ((fmt.fmt.pix.width != unWidth) || (fmt.fmt.pix.height != unHeight) || (fmt.fmt.pix.pixelformat != unFormat)) {
        fprintf(stderr, "incorrect format(%dx%d, %c%c%c%c) while requesting(%dx%d, %c%c%c%c)", 
                fmt.fmt.pix.width, 
                fmt.fmt.pix.height,
                (fmt.fmt.pix.pixelformat & 0xff), 
                (fmt.fmt.pix.pixelformat & 0xff00) >> 8, 
                (fmt.fmt.pix.pixelformat & 0xff0000) >> 16, 
                (fmt.fmt.pix.pixelformat & 0xff000000) >> 24,
                unWidth, 
                unHeight,  
                (unFormat & 0xff), 
                (unFormat & 0xff00) >> 8, 
                (unFormat & 0xff0000) >> 16, 
                (unFormat & 0xff000000) >> 24); 
        return -1;
    }

    /* set frame rate */
    CLEAR (fps);

    fps.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fps.parm.capture.timeperframe.numerator=1;
    fps.parm.capture.timeperframe.denominator=unFps;

    if (-1 == xioctl (g_nFd, VIDIOC_S_PARM, &fps)) {
        return -1;
    }

    if (-1 == xioctl (g_nFd, VIDIOC_G_PARM, &fps)) {
        return -1;
    }

    if (fps.parm.capture.timeperframe.numerator != 1 || fps.parm.capture.timeperframe.denominator != unFps) {
        fprintf(stderr, "incorrect frame rate(%u/%u) while requesting(%u)", 
                fps.parm.capture.timeperframe.denominator,
                fps.parm.capture.timeperframe.numerator,
                unFps);
    }

    /* Buggy driver paranoia. */
    min = fmt.fmt.pix.width * 2;
    if (fmt.fmt.pix.bytesperline < min) {
        fmt.fmt.pix.bytesperline = min;
    }
    min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
    if (fmt.fmt.pix.sizeimage < min) {
        fmt.fmt.pix.sizeimage = min;
    }

    return uvc_init_mmap ();
}

void uvc_close_device(void)
{
    if (g_nFd >= 0) {
        close (g_nFd);
    }

    g_nFd = -1;
}

int uvc_open_device(char *pszDevName)
{
    struct stat st; 
     
    strcpy(g_szDevName, pszDevName);

    if (-1 == stat (g_szDevName, &st)) {
        fprintf (stderr, "Cannot identify '%s': %d, %s\n",
                 g_szDevName, errno, strerror (errno));
        return -1;
    }

    if (!S_ISCHR (st.st_mode)) {
        fprintf (stderr, "%s is no device\n", g_szDevName);
        return -1;
    }

    g_nFd = open (pszDevName, O_RDWR /* required */ | O_NONBLOCK, 0);

    if (-1 == g_nFd) {
        fprintf (stderr, "Cannot open '%s': %d, %s\n",
                 g_szDevName, errno, strerror (errno));
        return -1;
    }

    return 0;
}

