///////////////////////////////////////////////////////////////////////////////
//     
//
// Drventry.c
//
// Description:
//
//  PCIe driver entry point
//
// History:
//	Jan 27, 2010	(dt) created
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/string.h>

#include "qptypes.h"
#include "qpdbgmgr.h"
#include "qpmemmgr.h"
#include "qposmgr.h"
#include "qptimemgr.h"
#include "qpmgrspriv.h"

#include "pci_config.h"

#include "plv4ldevversion.h"
#include "cdevice.h"


TQP_DbgMCompID			g_PLV4LDevDBGCompID = 0;
extern TQP_DbgMCompID	g_QPCodecDBGCompID;
extern TQP_DbgMCompID	g_QPPCIeDBGCompID;
extern TQP_DbgMCompID	g_QPPeripheralDBGCompID;


/////////////////////////////////////////////////////////////////////////////
//
//

#define PL330_PCI_VENDER_ID		0x1AF2
#define PL330_PCI_DEVICE_ID		0xA001


/////////////////////////////////////////////////////////////////////////////
//
//

struct semaphore g_module_lock;

static struct pci_device_id	plpci_table[] =
{
	{PCI_DEVICE(PL330_PCI_VENDER_ID, PL330_PCI_DEVICE_ID)},
	{}
};


MODULE_DEVICE_TABLE(pci, plpci_table);


#define MODULE_NAME "PCIe H.264 codec"

static int fwmemmode = 1;   /* =0: let FW default to use static memory mode, >0 let FW default to use dynamically memory mode */
module_param(fwmemmode, int, S_IRUGO);

static unsigned char plpci_get_revision(struct pci_dev *pdev)
{
    TQP_UINT8 revision;

    pci_read_config_byte(pdev, PCI_REVISION_ID, &revision);
    return revision;
}



/////////////////////////////////////////////////////////////////////////////
//
//

static int plpci_probe(struct pci_dev *pdev,
					   const struct pci_device_id *id
					   )
{
	long				lErr = 0;
	PLV4LDEV_TYPE		devType;
    CDevice				*pDevice = QP_NULL;
	PLV4LDEV_PARENT_DEV	parent_dev;
	EQP_Errors			qpStatus;
	TQP_UINT32			i = 0;
	unsigned char		revision;

	// register the component for debug output
	//
	QPDbgMRegComp(&g_PLV4LDevDBGCompID,
				  "QP V4LDev"
				  );
	QPDbgMRegComp(&g_QPCodecDBGCompID,
				  "QP CodecLib"
				  );
	QPDbgMRegComp(&g_QPPCIeDBGCompID,
				  "QP PCIe"
				  );
	QPDbgMRegComp(&g_QPPeripheralDBGCompID,
				  "QP Peripheral"
				  );
	//peter
	/*
	QPDbgMSetCompLevel(g_PLV4LDevDBGCompID,
					   QPDBGM_LEVEL_ERROR
					   );
	QPDbgMSetCompLevel(g_QPCodecDBGCompID,
					   QPDBGM_LEVEL_ERROR
					   );
	QPDbgMSetCompLevel(g_QPPCIeDBGCompID,
					   QPDBGM_LEVEL_ERROR
					   );
	QPDbgMSetCompLevel(g_QPPeripheralDBGCompID,
					   QPDBGM_LEVEL_ERROR
					   );
	*/
	QPDbgMSetCompLevel(g_PLV4LDevDBGCompID,
					   QPDBGM_LEVEL_TRACE
					   );
	QPDbgMSetCompLevel(g_QPCodecDBGCompID,
					   QPDBGM_LEVEL_TRACE
					   );
	QPDbgMSetCompLevel(g_QPPCIeDBGCompID,
					   QPDBGM_LEVEL_TRACE
					   );
	QPDbgMSetCompLevel(g_QPPeripheralDBGCompID,
					   QPDBGM_LEVEL_TRACE
					   );

	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_WARN),
				 "%s() (pci_dev=%08x)\n",
				 __FUNCTION__,
				 pdev
				 ));

	pci_enable_device(pdev);

	revision = plpci_get_revision(pdev);
	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
				 "%s() (revision=%x)\n",
				 __FUNCTION__,
				 revision
				 ));

	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
				 "%s() driver v%d.%d.%d %s\n",
				 __FUNCTION__,
				 PLDRV_VERSION_MAJOR,
				 PLDRV_VERSION_MINOR,
				 PLDRV_VERSION_BRANCH,
				 PLDRV_VERSION_EXPERIMENTAL
				 ));

	if (revision == 0x42)
	{
		return (-ENODEV);
	}

	lErr = pci_set_mwi(pdev);
	if (lErr)
	{
		goto ERROR;
	}

	lErr = pci_request_regions(pdev,
							   MODULE_NAME
							   );
	if (lErr < 0)
	{
		goto ERROR1;
	}

	pci_set_master(pdev);

	parent_dev.p_pciedev = pdev;
    devType = PLV4LDEV_TYPE_QL300_PCIE_EVK;

	switch (devType)
	{
		case PLV4LDEV_TYPE_QL300_PCIE_EVK:

			down(&g_module_lock);
			// find an empty slot for this device
			//
			for (i = 0; i < PLV4LDEV_MAX_DEVICES; i++)
			{
				if (QP_NULL == g_pPlV4lDevices[i])
				{
					// create and int the device
					//
					pDevice = (CDevice *)kmalloc(sizeof(CDevice),
												 GFP_KERNEL
												 );
					if (pDevice)
					{
						memset(pDevice,
							   0,
							   sizeof(CDevice)
							   );

						if (!CDevice_Constructor(pDevice,
												 i,
												 devType,
												 &parent_dev,
												 &pdev->dev,
												 (TQP_PVOID)0x100000,
												 (TQP_PVOID)0,
												 0x100000,
												 0,
												 fwmemmode
												 ))
						{
							kfree((void *)pDevice);
							pDevice = QP_NULL;
							lErr = -ENOMEM;
						}
						else
						{
							qpStatus = CDevice_Create(pDevice);
							if (QP_SUCCESS(qpStatus))
							{
								g_pPlV4lDevices[i] = pDevice;
								pci_set_drvdata(pdev,
												(void *)pDevice
												);
							}
							else
							{
								CDevice_Destructor(pDevice);
								kfree((void *)pDevice);
								pDevice = QP_NULL;
								lErr = -ENOMEM;
							}
						}
					}
					else
					{
						lErr = -ENOMEM;
					}
					break;
				}
			}

			if (PLV4LDEV_MAX_DEVICES == i)
			{
				// no empty slot available
				//
				lErr = -ENOMEM;
			}

			up(&g_module_lock);
			break;

		default:
			lErr = -ENODEV;
			goto ERROR1;
    }

	if (lErr != 0)
	{
		goto ERROR1;
	}

END:
	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_WARN),
				 "%s() lErr(%d)\n",
				 __FUNCTION__,
				 lErr
				 ));
	return (lErr);

ERROR1:
	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
				 "%s() Err1\n",
				 __FUNCTION__
				 ));
    pci_clear_mwi(pdev);

ERROR:
	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
				 "%s() Err\n",
				 __FUNCTION__
				 ));
	pci_disable_device(pdev);
	goto END;
}



static void plpci_remove(struct pci_dev *pdev)
{
    CDevice	*pDevice = (CDevice *)pci_get_drvdata(pdev);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_WARN),
				 "%s() (pdev=%08x) (pDevice=%08x)\n",
				 __FUNCTION__,
				 pdev,
				 pDevice
				 ));

	down(&g_module_lock);

	if (pDevice)
	{
		g_pPlV4lDevices[pDevice->m_id] = QP_NULL;
		CDevice_Close(pDevice);
		CDevice_Destructor(pDevice);
		kfree((void *)pDevice);
    }

	up(&g_module_lock);

	pci_release_regions(pdev);
	pci_disable_device(pdev);

	pci_set_drvdata(pdev,
					NULL
					);
}



int plpci_suspend(struct pci_dev *pdev,
				  pm_message_t state
				  )
{
    CDevice	*pDevice = (CDevice *)pci_get_drvdata(pdev);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_WARN),
				 "%s() \n",
				 __FUNCTION__
				 ));
	if (pDevice)
	{
		CDevice_PowerDown(pDevice);
    }

	pci_save_state(pdev);

	pci_set_power_state(pdev,
						PCI_D3hot
						);

    return 0;
}


int plpci_resume(struct pci_dev *pdev)
{
    CDevice	*pDevice = (CDevice *)pci_get_drvdata(pdev);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_WARN),
				 "%s() \n",
				 __FUNCTION__
				 ));

	pci_restore_state(pdev);

	pci_set_power_state(pdev,
						PCI_D0
						);
	if (pDevice)
	{
		CDevice_PowerUp(pDevice);
	}
	return (0);
}



static struct pci_driver pci_driver =
{
	.name		= MODULE_NAME,
	.id_table	= plpci_table,
	.probe		= plpci_probe,
	.suspend	= plpci_suspend,
	.resume		= plpci_resume,
	.remove		= plpci_remove,
};



static int __init plpci_init_module(void)
{
	long	lErr;

	QPMgrsInit();

	printk("%s() \n",
		   __FUNCTION__
		   );

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37)
	sema_init(&g_module_lock, 1);
#else
	init_MUTEX(&g_module_lock);
#endif

    lErr = pci_register_driver(&pci_driver);

	return (lErr);
}


static void __exit plpci_fini_module(void)
{
	pci_unregister_driver(&pci_driver);

	printk("%s() \n",
		   __FUNCTION__
		   );

	// terminate OS manager
	//
	QPMgrsTerm();
	return;
}

//MODULE_LICENSE("GPL");

module_init(plpci_init_module);
module_exit(plpci_fini_module);



