///////////////////////////////////////////////////////////////////////////////
//     
//
// Cpciecntl.c
//
// Description:
//
//  This is the implementation of the PCIe interface for Linux
//
// History:
//
// Authors: Andy Yu
//
// Revision: (dt) David Tao
//   1). Change CPCIeCntl_StartDMARead() and CPCIeCntl_StartDMAWrite()
//       to non-blocking functions
//   2). Move all the PCIe related functions including ISR from the V4L2 driver
//       to the bus module
//   3). PCIe device extension is created and initialized inside this module,
//       the V4L2 front end only pass the PCI device interface to this module
//   4). Add capability to DMA directly into either driver allocated buffers
//       or user buffers
//
//
///////////////////////////////////////////////////////////////////////////////

#include "qlautoconf.h"
//#include <string.h>
#include "platform.h"
#include "qptypes.h"
#include "cpciecntl.h"
#include "qpdbgmgr.h"
#include "qpmemmgr.h"
#include "qptimemgr.h"
#include "util.h"
#include "qpbusintf.h"
#include "qpmpgcodec.h"
#ifdef CONFIG_QL300_PCIE_EVK
#include "pci_config.h"
#include "qlpciintf.h"
#endif
#ifndef CONFIG_HOST_PLATFORM_USERSPACE
#include "qplinuxkermap.h"
#endif // CONFIG_HOST_PLATFORM_USERSPACE

extern TQP_DbgMCompID	g_QPPCIeDBGCompID;


/////////////////////////////////////////////////////////////////////////////
//
//


// IGenericDevice
//
static EQP_Errors CPCIeCntl_InitDevice(CPCIeCntl *This,
									   QP_DEVICE_CALLBACK pFuncCallback,
									   TQP_PVOID context,
									   TQP_PVOID pInitData
									   )
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
    EQP_Errors            qpStatus;
    PQLPCI_DEVICE         pDeviceExtension = (PQLPCI_DEVICE)This->m_pDeviceExtension;
    PQPMPGCODEC_INITDATA  pCodecInitData = (PQPMPGCODEC_INITDATA)pInitData;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CPCIeCntl_InitDevice()\n"
				 ));

	qpStatus = PciStartDevice(pDeviceExtension);

    if (QP_SUCCESS(qpStatus))
	{
		// copy some of the parsed bus info back to the mpeg codec init data
		//
		pCodecInitData->InterruptNumber = pDeviceExtension->InterruptVector;
        pCodecInitData->RegisterBase = (TQP_PVOID)pDeviceExtension->ioaddr[1];
        pCodecInitData->MemoryBase = (TQP_PVOID)pDeviceExtension->ioaddr[2];

        // CObject Init
        //
        CObject_Init(&This->m_Object);
	}
	else
	{
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CPCIeCntl_InitDevice() PciStartDevice() Failed qpStatus(%d)\n",
					 qpStatus
					 ));
    }

    return (qpStatus);
#endif
}



static EQP_Errors CPCIeCntl_Release(CPCIeCntl *This)
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CPCIeCntl_Release()\n"
				 ));
	if (CObject_IsInitialized(&This->m_Object))
	{
		PciStopDevice((PQLPCI_DEVICE)This->m_pDeviceExtension);

		CObject_Done(&This->m_Object);
	}
    return (QPERR_SUCCESS);
#endif
}



static EQP_Errors CPCIeCntl_Reset(CPCIeCntl *This)
{
    return (QPERR_SUCCESS);
}



static EQP_Errors CPCIeCntl_Set(CPCIeCntl *This,
								TQP_REFGUID guid,
								TQP_OBJECT_HANDLE hIndex,
								TQP_UINT32 dwCode,
								TQP_PVOID pInput,
								TQP_PVOID pOutput,
								TQP_UINT32 dwSize
								)
{
    return (QPERR_NOTIMPL);
}



static EQP_Errors CPCIeCntl_Get(CPCIeCntl *This,
								TQP_REFGUID guid,
								TQP_OBJECT_HANDLE hIndex,
								TQP_UINT32 dwCode,
								TQP_PVOID pInput,
								TQP_PVOID pOutput,
								PTQP_UINT32 pdwSizeGot
								)
{
    return (QPERR_NOTIMPL);
}



/////////////////////////////////////////////////////////////////////////////
//
//

// CPCIeCntl
//
EQP_Errors CPCIeCntl_ReadPCIeRegister(CPCIeCntl *This,
									  TQP_UINT32 dwAddr,
									  PTQP_UINT32 pData
									  )
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
	EQP_Errors        qpStatus;
	PCI_RW_RAM_STRUCT regStruct;

    memset(&regStruct,
		   0,
		   sizeof(regStruct)
		   );

    regStruct.bar = 0;
    regStruct.transferSize = transSize32Bit;
    regStruct.offset = (ULONGLONG)dwAddr;

    qpStatus = PedReadReg((PQLPCI_DEVICE)This->m_pDeviceExtension,
						  &regStruct
						  );
	if (QP_SUCCESS(qpStatus))
	{
		*pData = (TQP_UINT32)regStruct.data;
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE),
					 "CPCIeCntl_ReadPCIeRegister() wAddr(0x%x) data(0x%x)\n",
					 dwAddr,
					 *pData
					 ));
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CPCIeCntl_ReadPCIeRegister() PedReadReg() Failed qpStatus(%d)\n",
					 qpStatus
					 ));
    }
    return (qpStatus);
#endif
}



EQP_Errors CPCIeCntl_WritePCIeRegister(CPCIeCntl *This,
									   TQP_UINT32 dwAddr,
									   TQP_UINT32 dwData
									   )
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
    EQP_Errors        qpStatus;
    PCI_RW_RAM_STRUCT regStruct;

    memset(&regStruct,
		   0,
		   sizeof(regStruct)
		   );

    regStruct.bar = 0;
    regStruct.transferSize = transSize32Bit;
    regStruct.offset = (ULONGLONG)dwAddr;
    regStruct.data = (ULONGLONG)dwData;

    qpStatus = PedWriteReg((PQLPCI_DEVICE)This->m_pDeviceExtension,
						   &regStruct
						   );
    if (QP_SUCCESS(qpStatus))
	{
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CPCIeCntl_WritePCIeRegister() PedWriteReg() Failed qpStatus(%d)\n",
					 qpStatus
					 ));
    }
    return (qpStatus);
#endif
}



EQP_Errors CPCIeCntl_RegisterRead(CPCIeCntl *This,
								  TQP_UINT16 wAddr,
								  PTQP_UINT32 pData
								  )
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
    EQP_Errors        qpStatus;
    PCI_RW_RAM_STRUCT regStruct;

    memset(&regStruct,
		   0,
		   sizeof(regStruct)
		   );

    regStruct.bar = 1;
    regStruct.transferSize = transSize32Bit;
    regStruct.offset = (ULONGLONG)wAddr;

    qpStatus = PedReadReg((PQLPCI_DEVICE)This->m_pDeviceExtension,
						  &regStruct
						  );
    if (QP_SUCCESS(qpStatus))
	{
        *pData = (TQP_UINT32)regStruct.data;
		/*
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_EXTRACE),
					 "CPCIeCntl_RegisterRead() wAddr(0x%x) data(0x%x)\n",
					 wAddr,
					 *pData
					 ));
		*/
    }
	else
	{
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CPCIeCntl_RegisterRead() PedReadReg() Failed qpStatus(%d)\n",
					 qpStatus
					 ));
    }
    return (qpStatus);
#endif
}



EQP_Errors CPCIeCntl_RegisterWrite(CPCIeCntl *This,
								   TQP_UINT16 wAddr,
								   TQP_UINT32 dwData
								   )
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
    EQP_Errors        qpStatus;
    PCI_RW_RAM_STRUCT regStruct;

    memset(&regStruct,
		   0,
		   sizeof(regStruct)
		   );

    regStruct.bar = 1;
    regStruct.transferSize = transSize32Bit;
    regStruct.offset = (ULONGLONG)wAddr;
    regStruct.data = (ULONGLONG)dwData;

    qpStatus = PedWriteReg((PQLPCI_DEVICE)This->m_pDeviceExtension,
						   &regStruct
						   );

    if (QP_SUCCESS(qpStatus))
	{
        /*
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_EXTRACE),
					 "CPCIeCntl_RegisterWrite() PedWriteReg() wAddr(0x%x) data(0x%x)\n",
					 wAddr,
					 dwData
					 ));
		*/
    }
	else
	{
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CPCIeCntl_RegisterWrite() PedWriteReg() Failed qpStatus(0x%x)\n",
					 qpStatus
					 ));

    }
    return (qpStatus);
#endif
}



EQP_Errors CPCIeCntl_RegisterReadEx(CPCIeCntl *This,
									TQP_UINT16 wAddr,
									PTQP_UINT32 pData,
									TQP_UINT32 numReg
									)
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
	PCI_RW_RAM_STRUCT	regStruct;
	ULONG				ulRead = numReg << 2;
	ULONG				ulRet = 0;

	memset(&regStruct,
		   0,
		   sizeof(regStruct)
		   );

	regStruct.bar = 1;
	regStruct.transferSize = transSize32Bit;
	regStruct.offset = (ULONGLONG)wAddr;

	ulRet = PedReadMem((PQLPCI_DEVICE)This->m_pDeviceExtension,
					   &regStruct,
					   (PVOID)pData,
					   ulRead // in bytes
					   );

	if (ulRead != ulRet)
	{
		/*
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CPCIeCntl_RegisterRead() PedReadMem() Failed need(%d) bytes read(%d) bytes\n",
					 ulRead,
					 ulRet
					 ));
		*/
        return (QPERR_FAIL);
    }
	else
	{
        return (QPERR_SUCCESS);
    }
#endif
}



EQP_Errors CPCIeCntl_MemoryRead(CPCIeCntl *This,
								TQP_UINT32 dwAddr,
								PTQP_UINT32 pData
								)
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
	PCI_RW_RAM_STRUCT memStruct;
	ULONG             ulRead = sizeof(TQP_UINT32);
	ULONG             ulRet = 0;

	memset(&memStruct,
		   0,
		   sizeof(memStruct)
		   );

	memStruct.bar = 2;
	memStruct.transferSize = transSize32Bit;
	memStruct.offset = (ULONGLONG)dwAddr;

	ulRet = PedReadMem((PQLPCI_DEVICE)This->m_pDeviceExtension,
					   &memStruct,
					   (PVOID)&memStruct.data,
					   ulRead	// in bytes
					   );

	if (ulRead != ulRet)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CPCIeCntl_MemoryRead() PedReadMem() Failed need(%d) bytes read(%d) bytes\n",
					 ulRead,
					 ulRet
					 ));
        return (QPERR_FAIL);
	}
	else
	{
		*pData = (TQP_UINT32)memStruct.data;
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE),
					 "CPCIeCntl_MemoryRead() addr(0x%x) data(0x%x)\n",
					 dwAddr,
					 *pData
					 ));
		return (QPERR_SUCCESS);
	}
#endif
}



EQP_Errors CPCIeCntl_MemoryWrite(CPCIeCntl *This,
								 TQP_UINT32 dwAddr,
								 TQP_UINT32 dwData
								 )
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
	PCI_RW_RAM_STRUCT memStruct;
	ULONG             ulWrite = sizeof(TQP_UINT32);
	ULONG             ulRet = 0;

	memset(&memStruct,
		   0,
		   sizeof(memStruct)
		   );

	memStruct.bar = 2;
	memStruct.transferSize = transSize32Bit;
	memStruct.offset = (ULONGLONG)dwAddr;

	ulRet = PedWriteMem((PQLPCI_DEVICE)This->m_pDeviceExtension,
						&memStruct,
						(PVOID)&dwData,
						ulWrite	// in bytes
						);

	if (ulWrite != ulRet)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CPCIeCntl_MemoryWrite() PedWriteMem() Failed need(%d) bytes written(%d) bytes\n",
					 ulWrite,
					 ulRet
					 ));
		return (QPERR_FAIL);
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE),
					 "CPCIeCntl_MemoryWrite() PedWriteMem() addr(0x%x) data(0x%x)\n",
					 dwAddr,
					 dwData
					 ));
		return (QPERR_SUCCESS);
	}
#endif
}



EQP_Errors CPCIeCntl_StartDMAWrite(CPCIeCntl *This,
								   TQP_UINT32 ulAddr,	// in WORD(32 bit)
								   TQP_UINT32 ulLength,	// in WORD(32 bit)
								   PTQP_UINT8 pHostAddr,
								   TQP_UINT32 ulXferMode,
								   TQP_UINT32 ulPicWidthMB,
								   TQP_UINT32 ulPicHeightMB,
								   TQP_UINT32 ulSwap,
								   TQP_BOOL IsFrameMode
								   )
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
	PQLPCI_DEVICE	pDeviceExtension = (PQLPCI_DEVICE)This->m_pDeviceExtension;
	long			lErr;

	lErr = PedDMAWrite(pDeviceExtension,
					   ulAddr,	// in WORD(32 bit)
					   ulLength,	// in WORD(32 bit)
					   pHostAddr,
					   ulXferMode,
					   ulPicWidthMB,
					   ulPicHeightMB,
					   ulSwap,
					   IsFrameMode
					   );
	return (LinuxKerToQpStatus(lErr));
#endif
}



EQP_Errors CPCIeCntl_StartDMARead(CPCIeCntl *This,
								  TQP_UINT32 ulAddr,	// in WORD(32 bit)
								  TQP_UINT32 ulLength,	// in WORD(32 bit)
								  PTQP_UINT8 pHostAddr,
								  TQP_UINT32 ulXferMode,
								  TQP_UINT32 ulPicWidthMB,
								  TQP_UINT32 ulPicHeightMB,
								  TQP_UINT32 ulSwap,
								  TQP_BOOL IsFrameMode
								  )
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
	PQLPCI_DEVICE	pDeviceExtension = (PQLPCI_DEVICE)This->m_pDeviceExtension;
	long			lErr;

	lErr = PedDMARead(pDeviceExtension,
					  ulAddr,	// in WORD(32 bit)
					  ulLength,	// in WORD(32 bit)
					  pHostAddr,
					  ulXferMode,
					  ulPicWidthMB,
					  ulPicHeightMB,
					  ulSwap,
					  IsFrameMode
					  );
	return (LinuxKerToQpStatus(lErr));
#endif
}



CPCIeCntl * CPCIeCntl_Constructor(CPCIeCntl *pPCIeCntl,
								  CObject *pParent,
								  TQP_UINT32 dwAttributes,
								  TQP_PVOID pPhysicalDeviceObject,
								  TQP_PVOID pFunctionalDeviceObject,
								  QP_BUSINTF_CALLBACK pBusCallbackFunc,
								  TQP_PVOID pBusCallbackContext
								  )
{
#ifndef CONFIG_QL300_PCIE_EVK
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
				 "CPCIeCntl_Constructor() Driver Not Support PCI!\n"));
	return (QP_NULL);
#else
    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CPCIeCntl_Constructor()\n"
				 ));

    if (pPCIeCntl)
	{
        EQP_Errors	qpStatus;

        // CObject
        //
        CObject_Constructor(&pPCIeCntl->m_Object,
							pParent,
							dwAttributes
							);

        pPCIeCntl->m_pBusCallbackFunc = pBusCallbackFunc;
        pPCIeCntl->m_pBusCallbackContext = pBusCallbackContext;
        pPCIeCntl->m_pDeviceExtension = QP_NULL;

        // fill in the function table
        //

        // IGenericDevice
        pPCIeCntl->InitDevice = CPCIeCntl_InitDevice;
        pPCIeCntl->Release = CPCIeCntl_Release;
        pPCIeCntl->Reset = CPCIeCntl_Reset;
        pPCIeCntl->Set = CPCIeCntl_Set;
        pPCIeCntl->Get = CPCIeCntl_Get;

        qpStatus = PciAddDevice(pPCIeCntl,
								(TQP_PVOID)pFunctionalDeviceObject,
								(TQP_PVOID)pPhysicalDeviceObject
								);

        if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CPCIeCntl_Constructor() PciAddDevice() Failed qpStatus(%d)\n",
						 qpStatus
						 ));
			pPCIeCntl = QP_NULL;
		}
	}

    return (pPCIeCntl);
#endif
}



void CPCIeCntl_Destructor(CPCIeCntl *This)
{
#ifdef CONFIG_QL300_PCIE_EVK
	EQP_Errors	qpStatus;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CPCIeCntl_Destructor()\n"
				 ));

    // remove device
    //
    qpStatus = PciRemoveDevice(This);

	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CPCIeCntl_Destructor() PciRemoveDevice() Failed qpStatus(%d)\n",
					 qpStatus
					 ));
    }

    // CObject
    //
    CObject_Destructor(&This->m_Object);
#endif
}


EQP_Errors CPCIeCntl_EnableInterrupts(CPCIeCntl *This)
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
	PQLPCI_DEVICE	pDeviceExtension = (PQLPCI_DEVICE)This->m_pDeviceExtension;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE),
				 "%s()\n",
				 __FUNCTION__
				 ));
	pDeviceExtension->pRegisters->Common.ControlStatus |= PED_CARD_IRQ_ENABLE;
	return (QPERR_SUCCESS);
#endif
}


EQP_Errors CPCIeCntl_DisableInterrupt(CPCIeCntl *This)
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
	PQLPCI_DEVICE	pDeviceExtension = (PQLPCI_DEVICE)This->m_pDeviceExtension;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE),
				 "%s()\n",
				 __FUNCTION__
				 ));
	pDeviceExtension->pRegisters->Common.ControlStatus &= ~PED_CARD_IRQ_ENABLE;
	return (QPERR_SUCCESS);
#endif
}


EQP_Errors CPCIeCntl_CopyFromCommonBuffer(CPCIeCntl *This,
										  PTQP_UINT8 pHostAddr,
										  TQP_UINT32 ulLength
										  )
{
	return (QPERR_SUCCESS);
}



EQP_Errors CPCIeCntl_DMAReadDone(CPCIeCntl *This)
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_EXTRACE),
				 "%s()\n",
				 __FUNCTION__
				 ));
	PedDMAReadComplete((PQLPCI_DEVICE)This->m_pDeviceExtension);
	return (QPERR_SUCCESS);
#endif
}



EQP_Errors CPCIeCntl_DMAWriteDone(CPCIeCntl *This)
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
	PedDMAWriteComplete((PQLPCI_DEVICE)This->m_pDeviceExtension);
	return (QPERR_SUCCESS);
#endif
}



TQP_UINT32 CPCIeCntl_GetMaxDMASize(CPCIeCntl *This)
{
#ifndef CONFIG_QL300_PCIE_EVK
	return (QPERR_NOTIMPL);
#else
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_EXTRACE),
				 "%s()\n",
				 __FUNCTION__
				 ));
	return (PLPCI_DESCR_MAX_LEN);
#endif
}

