///////////////////////////////////////////////////////////////////////////////
//        
//
// Qlpciintf.c
//
// Description: 
//
//  This is the low level Linux PCIe access functions for codec
// 
// History:
//
// Authors: Andy Yu
//
// Revision: (dt) David Tao
//   1). Change CPCIeCntl_StartDMARead() and CPCIeCntl_StartDMAWrite() 
//       to non-blocking functions
//   2). Move all the PCIe related functions including ISR from the V4L2 driver 
//       to the bus module
//   3). PCIe device extension is created and initialized inside this module,
//       the V4L2 front end only pass the PCI device interface to this module
//   4). Add capability to DMA directly into either driver allocated buffers 
//       or user buffers
//
//
///////////////////////////////////////////////////////////////////////////////

#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/version.h>
#include <linux/signal.h>

#include <linux/pagemap.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include <linux/scatterlist.h>

#include "qpdbgmgr.h"
#include "util.h"
#include "cpciecntl.h"
#include "qptypes.h"
#include "pci_config.h"
#include "qlpciintf.h"

#undef DWG_EXCLUDE_SG_SETUP
#undef DWG_EXCLUDE_DMA_SETUP

extern TQP_DbgMCompID	g_QPPCIeDBGCompID;


/////////////////////////////////////////////////////////////////////////////
//
//

void pldmabuf_init(PPLDMABUF pdma)
{
	memset(pdma,
		   0,
		   sizeof(PLDMABUF)
		   );
}



long pldmabuf_free(PPLDMABUF pdma)
{
	if (pdma->pages) 
	{
		int i;
		for (i = 0; i < pdma->nr_pages; i++)
		{
			//page_cache_release(pdma->pages[i]);
			put_page(pdma->pages[i]);
		}
		kfree(pdma->pages);
		pdma->pages = NULL;
	}

	pdma->commonbuf = 0;

	pdma->vmalloc = NULL;

	pdma->direction = PCI_DMA_NONE;
	pdma->data = NULL;
	pdma->size = 0;
	return (0);
}



long pldmabuf_init_user(PPLDMABUF pdma, 
						int direction,
						unsigned long data, 
						unsigned long size
						)
{
	unsigned long	first,last;
	long			lErr = 0;

	pdma->direction = direction;
	pdma->data = (void *)data;
	pdma->size = size;

	first = (data & PAGE_MASK) >> PAGE_SHIFT;
	last  = ((data + size - 1) & PAGE_MASK) >> PAGE_SHIFT;

	pdma->offset = data & ~PAGE_MASK;
	pdma->nr_pages = last - first + 1;
	pdma->pages = kmalloc(pdma->nr_pages * sizeof(struct page*),
						  GFP_KERNEL
						  );
	if (NULL == pdma->pages)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR), 
					 "%s() Unable to allocate page array\n", 
					 __FUNCTION__
					 ));
		return (-ENOMEM);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE), 
				 "%s() init user [0x%lx + 0x%lx => %d pages]\n", 
				 __FUNCTION__,
				 data,
				 size,
				 pdma->nr_pages
				 ));

	down_read(&current->mm->mmap_sem);
	
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,9,0)
	int flags = 0;

	if(direction == PCI_DMA_FROMDEVICE)
	{
		flags |= FOLL_WRITE;
	}

	flags |= FOLL_FORCE;

	lErr = get_user_pages(data & PAGE_MASK, 
						  pdma->nr_pages,
						  flags,
						  pdma->pages, 
						  NULL
						  );
	#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,6,0)	
	lErr = get_user_pages(data & PAGE_MASK, 
						  pdma->nr_pages,
						  direction == PCI_DMA_FROMDEVICE, 
						  1, /* force */
						  pdma->pages, 
						  NULL
						  );
	#else
	lErr = get_user_pages(current,
						  current->mm,
						  data & PAGE_MASK, 
						  pdma->nr_pages,
						  direction == PCI_DMA_FROMDEVICE, 
						  1, /* force */
						  pdma->pages, 
						  NULL
						  );
	#endif	

	up_read(&current->mm->mmap_sem);
	if (lErr != pdma->nr_pages) 
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR), 
					 "%s() get_user_pages: err=%d nr_pages[%d] data[0x%x] size(%d)\n", 
					 __FUNCTION__,
					 lErr,
					 pdma->nr_pages,
					 data,
					 size
					 ));
		pdma->nr_pages = (lErr >= 0) ? lErr : 0;
		return ((lErr < 0) ? lErr : -EINVAL);
	}
	return (0);
}



long pldmabuf_init_kernel(PPLDMABUF pdma, 
						  int direction,
						  unsigned long data, 
						  unsigned long size
						  )
{
	unsigned long	first,last;

	pdma->direction = direction;
	pdma->data = (void *)data;
	pdma->size = size;

	first = (data & PAGE_MASK) >> PAGE_SHIFT;
	last  = ((data + size - 1) & PAGE_MASK) >> PAGE_SHIFT;
	pdma->nr_pages = last - first + 1;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE), 
				 "%s() init kernel [0x%lx + 0x%lx => %d pages]\n", 
				 __FUNCTION__,
				 data,
				 size,
				 pdma->nr_pages
				 ));

	pdma->vmalloc = (void *)data;
	return (0);
}



long pldma_init_commonbuf(PPLDMABUF pdma, 
						  int direction,
						  dma_addr_t commonbuf,
						  void *buf,
						  unsigned long data, 
						  unsigned long size
						  )
{
	if (size > PLPCIBUFF_SIZE)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR), 
					 "%s() size(%d) exceeding common buffer(%d)\n", 
					 __FUNCTION__,
					 size,
					 PLPCIBUFF_SIZE
					 ));
		return (-EINVAL);
	}
	else
	{
		if (PCI_DMA_TODEVICE == direction)
		{
			// copy from host buffer to common DMA buffer
			memcpy(buf, 
				   (void *)data, 
				   size
				   );
		}
	}

	pdma->direction = direction;
	pdma->data = (void *)data;
	pdma->size = size;

	pdma->nr_pages = 1;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE), 
				 "%s() init common buffer [0x%lx + 0x%lx => %d pages]\n", 
				 __FUNCTION__,
				 data,
				 size,
				 pdma->nr_pages
				 ));

	pdma->commonbuf = commonbuf;
	return 0;
}



struct scatterlist* pldmabuf_vmalloc_to_sg(unsigned char *virt,
										   unsigned long size,
										   int nr_pages
										   )
{
	struct scatterlist	*sglist;
	struct page			*pg;
	int					i;

	sglist = kcalloc(nr_pages, 
					 sizeof(struct scatterlist), 
					 GFP_KERNEL
					 );
	if (NULL == sglist)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR), 
					 "%s() Unable to allocate scatterlist\n", 
					 __FUNCTION__
					 ));
		return (NULL);
	}

	memset(sglist, 
		   0, 
		   nr_pages * sizeof(struct scatterlist)
		   );

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
	sg_init_table(sglist, 
				  nr_pages
				  );
#endif //LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)

	// first page
	if (nr_pages >= 1)
	{
		pg = vmalloc_to_page((void *)((unsigned long)virt & PAGE_MASK));
		if (NULL == pg)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR), 
						 "%s() vmalloc_to_page faile on data(0x%x)\n", 
						 __FUNCTION__,
						 (unsigned long)virt & PAGE_MASK
						 ));
			goto err;
		}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
		sg_assign_page(&sglist[0], pg);
#else //LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
		sglist[0].page = pg;
#endif //LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
		sglist[0].offset = (unsigned long)virt & ~PAGE_MASK;
		sglist[0].length = min(PAGE_SIZE - sglist[0].offset, size);
		virt += sglist[0].length;
		size -= sglist[0].length;
	}

	// the rest of the pages
	for (i = 1; i < nr_pages; i++) 
	{
		pg = vmalloc_to_page(virt);
		if (NULL == pg)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR), 
						 "%s() vmalloc_to_page faile on data(0x%x)\n", 
						 __FUNCTION__,
						 virt
						 ));
			goto err;
		}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
		sg_assign_page(&sglist[i], pg);
#else //LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
		sglist[i].page = pg;
#endif //LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
		sglist[i].offset = 0;
		sglist[i].length = min(PAGE_SIZE, size);
		virt += sglist[i].length;
		size -= sglist[i].length;
	}
	return (sglist);

err:
	kfree(sglist);
	return (NULL);
}



struct scatterlist* pldmabuf_pages_to_sg(struct page **pages, 
										 int nr_pages, 
										 unsigned long size,
										 int offset
										 )
{
	struct scatterlist	*sglist;
	int					i = 0;

	if (NULL == pages[0])
	{
		return NULL;
	}
	sglist = kcalloc(nr_pages, 
					 sizeof(struct scatterlist), 
					 GFP_KERNEL
					 );
	if (NULL == sglist)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR), 
					 "%s() Unable to allocate scatterlist\n", 
					 __FUNCTION__
					 ));
		return NULL;
	}

	memset(sglist, 
		   0, 
		   nr_pages * sizeof(struct scatterlist)
		   );

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
	sg_init_table(sglist, 
				  nr_pages
				  );
#endif //LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
	sg_assign_page(&sglist[0], pages[0]);
#else //LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	sglist[0].page = pages[0];
#endif //LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
	sglist[0].offset = offset;
	sglist[0].length = min(PAGE_SIZE - offset, size);
	size -= sglist[0].length;

	for (i = 1; i < nr_pages; i++) 
	{
		if (NULL == pages[i])
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR), 
						 "%s() no page(%d)\n", 
						 __FUNCTION__,
						 i
						 ));
			goto nopage;
		}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
		sg_assign_page(&sglist[i], pages[i]);
#else //LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
		sglist[i].page = pages[i];
#endif //LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
		sglist[i].offset = 0;
		sglist[i].length = min(PAGE_SIZE, size);
		size -= sglist[i].length;
	}

	return (sglist);

nopage:
	kfree(sglist);
	return (NULL);
}



long pldma_pci_map(struct pci_dev *dev, 
				   PPLDMABUF pdma
				   )
{
	if (pdma->pages) 
	{
		pdma->sglist = pldmabuf_pages_to_sg(pdma->pages, 
											pdma->nr_pages,
											pdma->size,
											pdma->offset
											);
	}
	if (pdma->vmalloc) 
	{
		pdma->sglist = pldmabuf_vmalloc_to_sg(pdma->vmalloc,
											  pdma->size,
											  pdma->nr_pages
											  );
	}
	if (pdma->commonbuf)
	{
		pdma->sglist = kmalloc(sizeof(struct scatterlist), 
							   GFP_KERNEL
							   );
		if (NULL != pdma->sglist) 
		{
			pdma->sglen  = 1;
			sg_dma_address(&pdma->sglist[0]) = pdma->commonbuf;
			pdma->sglist[0].offset = 0;
			pdma->sglist[0].length = pdma->size;
		}
	}

	if (NULL == pdma->sglist) 
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR), 
					 "%s() scatterlist is NULL\n", 
					 __FUNCTION__
					 ));
		return (-ENOMEM);
	}

	if (!pdma->commonbuf)
	{
		pdma->sglen = pci_map_sg(dev,
								 pdma->sglist,
								 pdma->nr_pages,
								 pdma->direction
								 );
		if (pdma->sglen < pdma->nr_pages) 
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR), 
						 "%s() pci_map_sg failed mapped(%d) requested(%d)\n", 
						 __FUNCTION__,
						 pdma->sglen,
						 pdma->nr_pages
						 ));
			kfree(pdma->sglist);
			pdma->sglist = NULL;
			pdma->sglen = 0;
			return (-EIO);
		}
	}
	return (0);
}



long pldma_pci_sync(struct pci_dev *dev, 
					PPLDMABUF pdma
					)
{
	if (!pdma->commonbuf)
	{
		pci_dma_sync_sg_for_cpu(dev,
								pdma->sglist,
								pdma->nr_pages,
								pdma->direction
								);
	}
	return (0);
}



long pldma_pci_unmap(struct pci_dev *dev, 
					 PPLDMABUF pdma
					 )
{
	if (!pdma->commonbuf)
	{
		if (pdma->sglen)
		{
			pci_unmap_sg(dev,
						 pdma->sglist,
						 pdma->nr_pages,
						 pdma->direction
						 );
		}
	}

	pdma->sglen = 0;

	if (pdma->sglist)
	{
		kfree(pdma->sglist);
		pdma->sglist = NULL;
	}
	return (0);
}



long pldma_iolock(PQLPCI_DEVICE pPciDevInfo,
				  unsigned long bufMode,
				  PPLDMABUF pdma,
				  int direction,
				  unsigned long data, 
				  unsigned long size
				  )
{
	long	lErr;

	switch (bufMode) 
	{
		case DMA_BUFFER_MODE_KERNEL:
			lErr = pldmabuf_init_kernel(pdma, 
										direction, 
										data, 
										size
										);
			break;

		case DMA_BUFFER_MODE_USERPTR:
			lErr = pldmabuf_init_user(pdma, 
									  direction, 
									  data, 
									  size
									  );
			break;

		case DMA_BUFFER_MODE_COMMON:
			lErr = pldma_init_commonbuf(pdma, 
										direction, 
										(PCI_DMA_FROMDEVICE == direction) ? pPciDevInfo->dmard_dmaddr : pPciDevInfo->dmawr_dmaddr,
										(PCI_DMA_FROMDEVICE == direction) ? pPciDevInfo->dmard_buffer : pPciDevInfo->dmawr_buffer,
										data, 
										size
										);
			break;

		default:
			lErr = -EPERM;
			break;
	}

	if (!lErr)
	{
		lErr = pldma_pci_map(pPciDevInfo->pcidev, 
							 pdma
							 );
	}

	return (lErr);
}



long PedDMAWrite(PQLPCI_DEVICE pPciDevInfo,	// Address of our qlpci_device.
				 TQP_UINT32 ulAddr,			// in WORD(32 bit)
				 TQP_UINT32 ulLength,		// in WORD(32 bit)
				 PTQP_UINT8 pHostAddr,
				 TQP_UINT32 ulXferMode,
				 TQP_UINT32 ulPicWidthMB,
				 TQP_UINT32 ulPicHeightMB,
				 TQP_UINT32 ulSwap,
				 TQP_BOOL IsFrameMode
				 )
{
	PPED_BAR0_REGISTERS     pDmaRegisters    = pPciDevInfo->pRegisters;
	PPED_BAR0_REGISTERS_EX  pDmaRegistersEx  = pPciDevInfo->pRegistersEx;
	PPED_DMA_DESCRIPTOR     pDmaDescr = (PPED_DMA_DESCRIPTOR)pPciDevInfo->DescrVAddrWr;
	PPLDMABUF				pDmaBuf = &pPciDevInfo->dmaBufWr;
 
	ULONG					Control;
	ULONGLONG				CardOffset;
	ULONGLONG				CardOffsetEx;
	volatile ULONG			ControlStatus;
	long					lErr = 0;
	int						i;

	TQP_UINT32				ulRemainLength = ulLength << 2;

	PED_DMA_COMMON			common;

	CardOffset = ulAddr << 2;

	if (ulRemainLength > 0) 
	{
		ControlStatus = pDmaRegisters->DMA[0x00].ControlStatus;

		if (PED_DMA_IS_BUSY(ControlStatus)) 
		{
            QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
						 "DMA is still busy. ControlStatus = 0x%08X\n",
						 ControlStatus
						 ));
			return (-EBUSY);
		}

		pldmabuf_init(pDmaBuf);

		// lock host buffer
		lErr = pldma_iolock(pPciDevInfo, 
							ulXferMode & DMA_BUFFER_MODE_MASK, 
							pDmaBuf, 
							PCI_DMA_TODEVICE, 
							(unsigned long)pHostAddr, 
							ulRemainLength
							);
		if (0 != lErr)
		{
            QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
						 "%s() pldma_iolock failed = %d\n",
						 __FUNCTION__,
						 lErr
						 ));
			goto DMA_WRITE_ERROR;
		}

		// notify cpu we are about to DMA from this buffer
		pldma_pci_sync(pPciDevInfo->pcidev,
					   pDmaBuf
					   );

		// setup scatter gather list
        Control = 
            PED_DMA_DESCR_CTRL_IRQ_ON_COMPLETION |
            PED_DMA_DESCR_CTRL_IRQ_ON_SHORT_ERR |
            PED_DMA_DESCR_CTRL_IRQ_ON_SHORT_SW |
            PED_DMA_DESCR_CTRL_IRQ_ON_SHORT_HW;

        switch (ulXferMode & DMA_TRANSFER_MODE_MASK) 
		{
			case DMA_TRANSFER_MODE_MB_RAS_LUMA:
				CardOffsetEx = 
					PED_DMA_DESCR_CADDR_X_LEN(2 * ulPicWidthMB) |
					PED_DMA_DESCR_CADDR_Y_LEN(IsFrameMode ? 16 : 8) |
					PED_DMA_DESCR_CADDR_Z_LEN(ulPicHeightMB) |
					PED_DMA_DESCR_CADDR_SWAP(ulSwap);
				Control |= 
					PED_DMA_DESCR_CTRL_X_DELTA_1(1) |
					PED_DMA_DESCR_CTRL_X_DELTA_2(31) |
					PED_DMA_DESCR_CTRL_Y_DELTA(IsFrameMode ? 2 : 4) |
					PED_DMA_DESCR_CTRL_Z_DELTA(IsFrameMode ? 1 : 3);
				break;

			case DMA_TRANSFER_MODE_MB_RAS_CHROMA:
				CardOffsetEx = 
					PED_DMA_DESCR_CADDR_X_LEN(ulPicWidthMB) |
					PED_DMA_DESCR_CADDR_Y_LEN(IsFrameMode ? 8 : 4) |
					PED_DMA_DESCR_CADDR_Z_LEN(ulPicHeightMB) |
					PED_DMA_DESCR_CADDR_SWAP(ulSwap);
				Control |= 
					PED_DMA_DESCR_CTRL_X_DELTA_1(16) |
					PED_DMA_DESCR_CTRL_X_DELTA_2(16) |
					PED_DMA_DESCR_CTRL_Y_DELTA(IsFrameMode ? 1 : 2) |
					PED_DMA_DESCR_CTRL_Z_DELTA(8 + (IsFrameMode ? 1 : 2));
				break;

			case DMA_TRANSFER_MODE_LINEAR:
			default:
				CardOffsetEx = 
					PED_DMA_DESCR_CADDR_X_LEN(16) |
					PED_DMA_DESCR_CADDR_Y_LEN(0) |
					PED_DMA_DESCR_CADDR_Z_LEN(0) |
					PED_DMA_DESCR_CADDR_SWAP(ulSwap);
				Control |= 
					PED_DMA_DESCR_CTRL_X_DELTA_1(1) |
					PED_DMA_DESCR_CTRL_X_DELTA_2(1) |
					PED_DMA_DESCR_CTRL_Y_DELTA(16) |
					PED_DMA_DESCR_CTRL_Z_DELTA(1);
				break;
		}

		for (i = 0; i < pDmaBuf->sglen; i++)
		{
			pDmaDescr->Control        = Control;
			pDmaDescr->SystemAddress  = sg_dma_address(&pDmaBuf->sglist[i]);
			pDmaDescr->CardAddress    = CardOffset | CardOffsetEx;
			pDmaDescr->ByteCount      = pDmaBuf->sglist[i].length; 
			pDmaDescr->NextDescriptor = (ULONGLONG)(pDmaRegisters->DMA[0x00].Descriptor.LowPart) + ((i + 1) * sizeof(*pDmaDescr));

			CardOffset += pDmaDescr->ByteCount;
			pDmaDescr++;
		}

		// add dummy descriptor
		pDmaDescr->Control        = Control;
		pDmaDescr->SystemAddress  = sg_dma_address(&pDmaBuf->sglist[0]);
		pDmaDescr->CardAddress    = PED_DMA_DESCR_CADDR_DUMMY;
		pDmaDescr->ByteCount      = 4;
		pDmaDescr->NextDescriptor = 0x00;

		// make sure interrupt is enabled
		pPciDevInfo->pRegisters->Common.ControlStatus = PED_CARD_IRQ_ENABLE;

		// start DMA
		ControlStatus = (PED_CTRL_IRQ_ENABLE | PED_CTRL_CHAIN_START);
		pDmaRegisters->DMA[0x00].ControlStatus = ControlStatus;
	}

	return (lErr);

DMA_WRITE_ERROR:
	pldmabuf_free(pDmaBuf);
	return (lErr);
}



long PedDMARead(PQLPCI_DEVICE pPciDevInfo,	// Address of our qlpci_device.
				TQP_UINT32 ulAddr,			// in WORD(32 bit)
				TQP_UINT32 ulLength,		// in WORD(32 bit)
				PTQP_UINT8 pHostAddr,
				TQP_UINT32 ulXferMode,
				TQP_UINT32 ulPicWidthMB,
				TQP_UINT32 ulPicHeightMB,
				TQP_UINT32 ulSwap,
				TQP_BOOL IsFrameMode
				)
{
	PPED_BAR0_REGISTERS     pDmaRegisters    = pPciDevInfo->pRegisters;
	PPED_BAR0_REGISTERS_EX  pDmaRegistersEx  = pPciDevInfo->pRegistersEx;
	PPED_DMA_DESCRIPTOR     pDmaDescr = (PPED_DMA_DESCRIPTOR)pPciDevInfo->DescrVAddrRd;
	PPLDMABUF				pDmaBuf = &pPciDevInfo->dmaBufRd;
 
	ULONG					Control;
	ULONGLONG				CardOffset;
	ULONGLONG				CardOffsetEx;
	volatile ULONG			ControlStatus;
	long					lErr = 0;
	int						i;

	TQP_UINT32				ulRemainLength = ulLength << 2;

	PED_DMA_COMMON			common;

	CardOffset = ulAddr << 2;

	if (ulRemainLength > 0) 
	{
		ControlStatus = pDmaRegisters->DMA[0x20].ControlStatus;

		if (PED_DMA_IS_BUSY(ControlStatus)) 
		{
            QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
						 "DMA is still busy. ControlStatus = 0x%08X\n",
						 ControlStatus
						 ));
			return (-EBUSY);
		}

		pldmabuf_init(pDmaBuf);

		// lock host buffer
		lErr = pldma_iolock(pPciDevInfo, 
							ulXferMode & DMA_BUFFER_MODE_MASK, 
							pDmaBuf, 
							PCI_DMA_FROMDEVICE, 
							(unsigned long)pHostAddr, 
							ulRemainLength
							);
		if (0 != lErr)
		{
            QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR),
						 "%s() pldma_iolock failed = %d\n",
						 __FUNCTION__,
						 lErr
						 ));
			goto DMA_READ_ERROR;
		}

		// setup scatter gather list
		Control = 
			PED_DMA_DESCR_CTRL_IRQ_ON_COMPLETION |
			PED_DMA_DESCR_CTRL_IRQ_ON_SHORT_ERR |
			PED_DMA_DESCR_CTRL_IRQ_ON_SHORT_SW |
			PED_DMA_DESCR_CTRL_IRQ_ON_SHORT_HW;

		switch (ulXferMode & DMA_TRANSFER_MODE_MASK) 
		{
			case DMA_TRANSFER_MODE_MB_RAS_LUMA:
				CardOffsetEx = 
					PED_DMA_DESCR_CADDR_X_LEN(2 * ulPicWidthMB) |
					PED_DMA_DESCR_CADDR_Y_LEN(IsFrameMode ? 16 : 8) |
					PED_DMA_DESCR_CADDR_Z_LEN(ulPicHeightMB) |
					PED_DMA_DESCR_CADDR_SWAP(ulSwap);
				Control |= 
					PED_DMA_DESCR_CTRL_X_DELTA_1(1) |
					PED_DMA_DESCR_CTRL_X_DELTA_2(31) |
					PED_DMA_DESCR_CTRL_Y_DELTA(IsFrameMode ? 2 : 4) |
					PED_DMA_DESCR_CTRL_Z_DELTA(IsFrameMode ? 1 : 3);
				break;

			case DMA_TRANSFER_MODE_MB_RAS_CHROMA:
				CardOffsetEx = 
					PED_DMA_DESCR_CADDR_X_LEN(ulPicWidthMB) |
					PED_DMA_DESCR_CADDR_Y_LEN(IsFrameMode ? 8 : 4) |
					PED_DMA_DESCR_CADDR_Z_LEN(ulPicHeightMB) |
					PED_DMA_DESCR_CADDR_SWAP(ulSwap);
				Control |= PED_DMA_DESCR_CTRL_X_DELTA_1(16) |
					PED_DMA_DESCR_CTRL_X_DELTA_2(16) |
					PED_DMA_DESCR_CTRL_Y_DELTA(IsFrameMode ? 1 : 2) |
					PED_DMA_DESCR_CTRL_Z_DELTA(8 + (IsFrameMode ? 1 : 2));
				break;

			case DMA_TRANSFER_MODE_LINEAR:
			default:
				CardOffsetEx = 
					PED_DMA_DESCR_CADDR_X_LEN(16) |
					PED_DMA_DESCR_CADDR_Y_LEN(0) |
					PED_DMA_DESCR_CADDR_Z_LEN(0) |
					PED_DMA_DESCR_CADDR_SWAP(ulSwap);
				Control |= 
					PED_DMA_DESCR_CTRL_X_DELTA_1(1) |
					PED_DMA_DESCR_CTRL_X_DELTA_2(1) |
					PED_DMA_DESCR_CTRL_Y_DELTA(16) |
					PED_DMA_DESCR_CTRL_Z_DELTA(1);
				break;
		}

		for (i = 0; i < pDmaBuf->sglen; i++)
		{
			pDmaDescr->Control        = Control;
			pDmaDescr->SystemAddress  = sg_dma_address(&pDmaBuf->sglist[i]);
			pDmaDescr->CardAddress    = CardOffset | CardOffsetEx;
			pDmaDescr->ByteCount      = pDmaBuf->sglist[i].length; 
			if ((pDmaBuf->sglen - 1) == i)
			{
				pDmaDescr->NextDescriptor = 0;
				pDmaDescr->CardAddress |= PED_DMA_DESCR_CADDR_DUMMY;
			}
			else
			{
				pDmaDescr->NextDescriptor = (ULONGLONG)pDmaRegisters->DMA[0x20].Descriptor.LowPart + ((i + 1) * sizeof(*pDmaDescr));
			}

			CardOffset += pDmaDescr->ByteCount;
			pDmaDescr++;
		}

		// make sure interrupt is enabled
		pPciDevInfo->pRegisters->Common.ControlStatus = PED_CARD_IRQ_ENABLE;

		// start DMA
		ControlStatus = (PED_CTRL_IRQ_ENABLE | PED_CTRL_CHAIN_START);
		pDmaRegisters->DMA[0x20].ControlStatus = ControlStatus;
	}

	return (lErr);

DMA_READ_ERROR:
	pldmabuf_free(pDmaBuf);
	return (lErr);
}



/////////////////////////////////////////////////////////////////////////////
//
//

static void plpcie_get_config(struct pci_dev *pdev)
{
	TQP_UINT8	config;
	TQP_UINT16	count;
	TQP_UINT32	data32;

	printk("config_table:");
	/*
	for (count = 0; count < 0x100; count++) 
	{
		pci_read_config_byte(pdev, count, &config);
		if ((count % 0x10) == 0x00) 
		{
			printk("\n0x%02x:", (count));
		}
		printk(" %02x", config);
	}
	printk("\n\n");
	*/

	pci_read_config_word(pdev, PCI_VENDOR_ID, &count);
	printk("PCI vendorID = %04x\n", count);
	pci_read_config_word(pdev, PCI_DEVICE_ID, &count);
	printk("PCI deviceID = %04x\n", count);

	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_0, &data32);
	printk("PCI BAR0 = %08x\n", data32);
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_1, &data32);
	printk("PCI BAR1 = %08x\n", data32);
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_2, &data32);
	printk("PCI BAR2 = %08x\n\n", data32);

	data32 = pci_resource_start(pdev, 0);
	printk("PCI BAR0 start = %08x\n", data32);
	data32 = pci_resource_start(pdev, 1);
	printk("PCI BAR1 start = %08x\n", data32);
	data32 = pci_resource_start(pdev, 2);
	printk("PCI BAR2 start = %08x\n\n", data32);

	data32 = pci_resource_end(pdev, 0);
	printk("PCI BAR0 end   = %08x\n", data32);
	data32 = pci_resource_end(pdev, 1);
	printk("PCI BAR1 end   = %08x\n", data32);
	data32 = pci_resource_end(pdev, 2);
	printk("PCI BAR2 end   = %08x\n\n", data32);

	data32 = pci_resource_len(pdev, 0);
	printk("PCI BAR0 len   = %08x\n", data32);
	data32 = pci_resource_len(pdev, 1);
	printk("PCI BAR1 len   = %08x\n", data32);
	data32 = pci_resource_len(pdev, 2);
	printk("PCI BAR2 len   = %08x\n\n", data32);

	data32 = pci_resource_flags(pdev, 0);
	printk("PCI BAR0 flags = %08x\n", data32);
	data32 = pci_resource_flags(pdev, 1);
	printk("PCI BAR1 flags = %08x\n", data32);
	data32 = pci_resource_flags(pdev, 2);
	printk("PCI BAR2 flags = %08x\n\n", data32);
}



void PedDMAReadComplete(PQLPCI_DEVICE pPciDevInfo)
{
	if (pPciDevInfo->dmaBufRd.commonbuf)
	{
		if (pPciDevInfo->dmaBufRd.data)
		{
			//peter ?? if m_nAnalogVideoDecoderStatusProperty == 0 but interrupt occurred (signal plug), more than 10 interrupt wil cause this address 4 ??
			if( pPciDevInfo->dmaBufRd.data < 0xFFFF)
			{
				printk( "pPciDevInfo->dmaBufRd.data(0x%lx)\n", pPciDevInfo->dmaBufRd.data );
			}

			memcpy((void *)pPciDevInfo->dmaBufRd.data, 
				   pPciDevInfo->dmard_buffer, 
				   pPciDevInfo->dmaBufRd.size
				   );
		}
	}

	pldma_pci_sync(pPciDevInfo->pcidev, 
				   &pPciDevInfo->dmaBufRd
				   );
	pldma_pci_unmap(pPciDevInfo->pcidev, 
					&pPciDevInfo->dmaBufRd
					);
	pldmabuf_free(&pPciDevInfo->dmaBufRd);
}



void PedDMAWriteComplete(PQLPCI_DEVICE pPciDevInfo)
{
#if 0
	pldma_pci_sync(pPciDevInfo->pcidev, 
				   &pPciDevInfo->dmaBufWr
				   );
#endif
	pldma_pci_unmap(pPciDevInfo->pcidev, 
					&pPciDevInfo->dmaBufWr
					);
	pldmabuf_free(&pPciDevInfo->dmaBufWr);
}


/*
irqreturn_t plpci_irqhandler(int irq, 
							 void *pDevInfo, 
							 struct pt_regs *regs
							 )
*/
irqreturn_t plpci_irqhandler(int irq, 
							 void *pDevInfo 
							 )
{
	PQLPCI_DEVICE	pPciDevInfo = (PQLPCI_DEVICE)pDevInfo;
	volatile ULONG	ControlStatus;
	unsigned long	flags;
	irqreturn_t		irqRet = IRQ_NONE;

	spin_lock_irqsave(&pPciDevInfo->irq_lock, 
					  flags
					  );

	if (!pPciDevInfo->devconnected) 
	{
		irqRet = IRQ_HANDLED;
	} 
	else 
	{
		if (pPciDevInfo->pRegistersEx->UserIntEnableStatus & PED_C2H_SCRATCH_PAD_VALID) 
		{
//peter
#if 0
			printk("received INT 0x%08X: PED_C2H_SCRATCH_PAD_VALID()\n",
				   jiffies
				   );
#endif

			pPciDevInfo->pRegistersEx->UserIntEnableStatus = PED_C2H_SCRATCH_PAD_VALID;
			pPciDevInfo->InterruptStatus |= PED_IRQ_ARM;

			pPciDevInfo->pBusCallbackFunc(pPciDevInfo->pBusCallbackContext, 
										  QPBUSINTF_ARM_INT, 
										  QP_NULL
										  );
			irqRet = IRQ_HANDLED;
		}

		// DMA write
		ControlStatus = pPciDevInfo->pRegisters->DMA[0x00].ControlStatus;
		if ((ControlStatus & PED_CTRL_IRQ_ACTIVE) && 
			(ControlStatus & PED_CTRL_IRQ_ENABLE)
			) 
		{
//peter
#if 0
			printk("received INT 0x%08X: PCIe DMA Write Complete(0x%08X)\n",
				   jiffies, 
				   ControlStatus
				   );
#endif

			ControlStatus = (PED_CTRL_IRQ_ACTIVE | PED_CTRL_IRQ_ENABLE);
			pPciDevInfo->pRegisters->DMA[0x00].ControlStatus = ControlStatus;
			pPciDevInfo->InterruptStatus |= PED_IRQ_DMA_COMPLETE(0x00);

			pPciDevInfo->pBusCallbackFunc(pPciDevInfo->pBusCallbackContext, 
										  QPBUSINTF_DMA_WRITE_DONE, 
										  QP_NULL
										  );

			irqRet = IRQ_HANDLED;
		}

		// DMA read
		ControlStatus = pPciDevInfo->pRegisters->DMA[0x20].ControlStatus;
		if ((ControlStatus & PED_CTRL_IRQ_ACTIVE) && 
			(ControlStatus & PED_CTRL_IRQ_ENABLE)
			) 
		{
//peter
#if 0
			printk("received INT 0x%08X: PCIe DMA Read Complete(0x%08X)\n",
				   jiffies, 
				   ControlStatus
				   );
#endif

			ControlStatus = (PED_CTRL_IRQ_ACTIVE | PED_CTRL_IRQ_ENABLE);
			pPciDevInfo->pRegisters->DMA[0x20].ControlStatus = ControlStatus;
			pPciDevInfo->InterruptStatus |= PED_IRQ_DMA_COMPLETE(0x01);

			pPciDevInfo->pBusCallbackFunc(pPciDevInfo->pBusCallbackContext, 
										  QPBUSINTF_DMA_READ_DONE, 
										  QP_NULL
										  );

			irqRet = IRQ_HANDLED;
		}
	}

	spin_unlock_irqrestore(&pPciDevInfo->irq_lock, 
						   flags
						   );

	return (irqRet);
}



/////////////////////////////////////////////////////////////////////////////
//
//

EQP_Errors PedReadReg(PQLPCI_DEVICE pPciDevInfo,    // Address of qlpci_device
                      PPCI_RW_RAM_STRUCT pRegStruct // Address of PCI_RW_RAM_STRUCT describing read operation.
                     )
{

	CHECK_QLPCI_PRESENCE(pPciDevInfo);

	pRegStruct->data = (ULONGLONG)readl(pPciDevInfo->ioaddr[pRegStruct->bar] + pRegStruct->offset);

	return (QPERR_SUCCESS);
}



EQP_Errors PedWriteReg(PQLPCI_DEVICE pPciDevInfo,   // Address of qlpci_device
                       PPCI_RW_RAM_STRUCT pRegStruct // Address of PCI_RW_RAM_STRUCT describing read operation.
                      )
{
	TQP_UINT32	ulAddr = (TQP_UINT32)pRegStruct->offset;
	TQP_UINT32	ulData = (TQP_UINT32)pRegStruct->data;

	CHECK_QLPCI_PRESENCE(pPciDevInfo);

	writel(ulData, 
		   (pPciDevInfo->ioaddr[pRegStruct->bar] + ulAddr)
		   );

	return (QPERR_SUCCESS);
}



ULONG PedReadMem(PQLPCI_DEVICE pPciDevInfo,     // Address of our qlpci_device.
				 PPCI_RW_RAM_STRUCT pReadInfo,  // Address of PCI_RW_RAM_STRUCT describing read operation.
				 PVOID Buffer,                  // Address of memory where read data is to be copied to.
				 ULONG NumBytes                 // Number of bytes to be read.
				 )           
{
	ULONG	NumBytesRead = NumBytes;
	ULONG	ulCount;
	PVOID	pBaseAddress = pPciDevInfo->ioaddr[pReadInfo->bar];

	CHECK_QLPCI_PRESENCE(pPciDevInfo);

	for (ulCount = 0; ulCount < NumBytesRead; ulCount += 4) 
	{
		pReadInfo->data = (ULONGLONG)readl(pPciDevInfo->ioaddr[pReadInfo->bar] + pReadInfo->offset + ulCount);
		*(ULONG *)(Buffer + ulCount) = (ULONG)pReadInfo->data;
	}

	return (NumBytes);    
}                           

                            
                            
ULONG PedWriteMem(PQLPCI_DEVICE pPciDevInfo,      // Address of our qlpci_device.
				  PPCI_RW_RAM_STRUCT pWriteInfo,  // Address of PCI_RW_RAM_STRUCT describing write operation.
				  PVOID Buffer,                   // Address of memory where write data is to be copied from.
				  ULONG NumBytes                  // Number of bytes to be written.
				  )
{
	ULONG	NumBytesWritten = NumBytes;
	PVOID	pBaseAddress = pPciDevInfo->ioaddr[pWriteInfo->bar];

	CHECK_QLPCI_PRESENCE(pPciDevInfo);

	writel(*(ULONG *)Buffer, 
		   (pPciDevInfo->ioaddr[pWriteInfo->bar] + (ULONG)pWriteInfo->offset)
		   );

	return (NumBytes); 
}                           



EQP_Errors PciRemoveDevice(CPCIeCntl *This)
{
    PQLPCI_DEVICE	pPciDevInfo;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE), 
				 "PciRemoveDevice()\n"
				 ));

	pPciDevInfo = (PQLPCI_DEVICE)This->m_pDeviceExtension;
	if (pPciDevInfo)
	{
		pPciDevInfo->devconnected = 0;

		// free resources
		if (pPciDevInfo->DescrVAddrRd) 
		{
			dma_free_coherent(&pPciDevInfo->pcidev->dev, 
							  PLPCI_NUM_DESCR * sizeof(PED_DMA_DESCRIPTOR), 
							  pPciDevInfo->DescrVAddrRd, 
							  pPciDevInfo->DescrAddrRd
							  );
			pPciDevInfo->DescrVAddrRd = QP_NULL;
		}
		if (pPciDevInfo->DescrVAddrWr) 
		{
			dma_free_coherent(&pPciDevInfo->pcidev->dev, 
							  PLPCI_NUM_DESCR * sizeof(PED_DMA_DESCRIPTOR), 
							  pPciDevInfo->DescrVAddrWr, 
							  pPciDevInfo->DescrAddrWr
							  );
			pPciDevInfo->DescrVAddrWr = QP_NULL;
		}

		pldma_pci_unmap(pPciDevInfo->pcidev, 
						&pPciDevInfo->dmaBufRd
						);
		pldma_pci_unmap(pPciDevInfo->pcidev, 
						&pPciDevInfo->dmaBufWr
						);
		pldmabuf_free(&pPciDevInfo->dmaBufRd);
		pldmabuf_free(&pPciDevInfo->dmaBufWr);

		dma_free_coherent(&pPciDevInfo->pcidev->dev, 
						  PLPCIBUFF_SIZE, 
						  pPciDevInfo->dmard_buffer, 
						  pPciDevInfo->dmard_dmaddr
						  );
		dma_free_coherent(&pPciDevInfo->pcidev->dev, 
						  PLPCIBUFF_SIZE, 
						  pPciDevInfo->dmawr_buffer, 
						  pPciDevInfo->dmard_dmaddr
						  );

		kfree((void *)pPciDevInfo);
		This->m_pDeviceExtension = QP_NULL;
	}

    return (QPERR_SUCCESS);
}



EQP_Errors PciAddDevice(CPCIeCntl *This,
						TQP_PVOID FunctionalDeviceObject,
						TQP_PVOID PhysicalDeviceObject
						)
{
    EQP_Errors		qpStatus = QPERR_SUCCESS;
    PQLPCI_DEVICE	pPciDevInfo;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE), 
                 "PciAddDevice() PDO = %p FDO = %p\n",
                 PhysicalDeviceObject,
                 FunctionalDeviceObject
                ));

	// allocate our device extension
	pPciDevInfo = (PQLPCI_DEVICE)kmalloc(sizeof(QLPCI_DEVICE), 
										 GFP_KERNEL
										 );
	if (QP_NULL == pPciDevInfo)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR), 
					 "PciAddDevice() failed to allocate device extension\n"
					 ));
		return (QPERR_NOMEM);
	}

	memset((void *)pPciDevInfo, 
		   0, 
		   sizeof(QLPCI_DEVICE)
		   );

	// Save the PCIe module object
	pPciDevInfo->pPCIeCntl = This;

    // Save the physical device object for future reference
    pPciDevInfo->pcidev = (struct pci_dev *)PhysicalDeviceObject;

    // Initialize the device state information managed by the driver.
    PciInitializeDeviceState(pPciDevInfo);

	pldmabuf_init(&pPciDevInfo->dmaBufRd);
	pldmabuf_init(&pPciDevInfo->dmaBufWr);

	// allocate common DMA buffers
	pPciDevInfo->dmard_buffer = dma_alloc_coherent(&pPciDevInfo->pcidev->dev, 
												   PLPCIBUFF_SIZE, 
												   &pPciDevInfo->dmard_dmaddr, 
												   GFP_KERNEL
												   );
	pPciDevInfo->dmawr_buffer = dma_alloc_coherent(&pPciDevInfo->pcidev->dev, 
												   PLPCIBUFF_SIZE, 
												   &pPciDevInfo->dmawr_dmaddr, 
												   GFP_KERNEL
												   );
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE), 
				 "dmard_dmaddr = buf(0x%08X) addr(0x%08X), dmawr_dmaddr = buf(0x%08X) addr(0x%08X)\n", 
				 pPciDevInfo->dmard_buffer, 
				 pPciDevInfo->dmard_dmaddr,
				 pPciDevInfo->dmawr_buffer, 
				 pPciDevInfo->dmawr_dmaddr
				 ));

	// allocate DMA descriptor
	pPciDevInfo->DescrVAddrRd = dma_alloc_coherent(&pPciDevInfo->pcidev->dev, 
												   PLPCI_NUM_DESCR * sizeof(PED_DMA_DESCRIPTOR), 
												   &pPciDevInfo->DescrAddrRd, 
												   GFP_KERNEL
												   );
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
				 "%s() read descriptor addr = %x physical addr = %x\n", 
				 __FUNCTION__,
				 pPciDevInfo->DescrVAddrRd,
				 pPciDevInfo->DescrAddrRd
				 ));

	pPciDevInfo->DescrVAddrWr = dma_alloc_coherent(&pPciDevInfo->pcidev->dev, 
												   PLPCI_NUM_DESCR * sizeof(PED_DMA_DESCRIPTOR), 
												   &pPciDevInfo->DescrAddrWr, 
												   GFP_KERNEL
												   );
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
				 "%s() read descriptor addr = %x physical addr = %x\n", 
				 __FUNCTION__,
				 pPciDevInfo->DescrVAddrWr,
				 pPciDevInfo->DescrAddrWr
				 ));

	// save call back function pointer and context
    pPciDevInfo->pBusCallbackFunc = This->m_pBusCallbackFunc;
    pPciDevInfo->pBusCallbackContext = This->m_pBusCallbackContext;

	This->m_pDeviceExtension = (TQP_PVOID)pPciDevInfo;

    return (qpStatus);
}



void PciInitializeDeviceState(PQLPCI_DEVICE pPciDevInfo)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_TRACE), 
                 "PciInitializeDeviceState()\n"
				 ));
	pPciDevInfo->devconnected = 0;
}



EQP_Errors PciStartDevice(PQLPCI_DEVICE pPciDevInfo)
{
	int						dma;		// counter to check PCIe registers
	volatile PED_DMA_ENGINE	*pDMAReg;
	long					lErr = 0;

	// dump pcie config
	plpcie_get_config(pPciDevInfo->pcidev);

	// IO remap of BAR0, BAR1, and BAR2
	pPciDevInfo->ioaddr[0] = ioremap(pci_resource_start(pPciDevInfo->pcidev, 0), 
									 pci_resource_len(pPciDevInfo->pcidev, 0)
									 );	// re-mapped PCIe register address
	pPciDevInfo->ioaddr[1] = ioremap(pci_resource_start(pPciDevInfo->pcidev, 1), 
									 pci_resource_len(pPciDevInfo->pcidev, 1)
									 );	// re-mapped QL300 register address
	pPciDevInfo->ioaddr[2] = ioremap(pci_resource_start(pPciDevInfo->pcidev, 2), 
									 pci_resource_len(pPciDevInfo->pcidev, 2)
									 );	// re-mapped QL300 memory address

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
				 "%s() ioaddr: %x %x %x\n", 
				 __FUNCTION__,
				 pPciDevInfo->ioaddr[0], 
				 pPciDevInfo->ioaddr[1], 
				 pPciDevInfo->ioaddr[2]
				 ));

	// BAR0 maps to device registers
	pPciDevInfo->pRegisters = (PPED_BAR0_REGISTERS)((UCHAR *)pPciDevInfo->ioaddr[0]);
	pPciDevInfo->pRegistersEx = (PPED_BAR0_REGISTERS_EX)((UCHAR *)pPciDevInfo->ioaddr[0] + 0x8000);

	for (dma = 0; dma < PCI_MAX_DMAS; dma++)
	{
		if (pPciDevInfo->pRegisters->DMA[dma].Capabilities & PED_CAP_ENGINE_PRESENT)
		{
			pPciDevInfo->pRegisters->DMA[dma].Descriptor.LowPart = ((0 == dma) ? pPciDevInfo->DescrAddrWr : pPciDevInfo->DescrAddrRd);
			pPciDevInfo->pRegisters->DMA[dma].Descriptor.HighPart = 0x00000000;

			pDMAReg = &(pPciDevInfo->pRegisters->DMA[dma]);
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
						 "\npRegisters : %x\n", 
						 pPciDevInfo->pRegisters
						 ));
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
						 " ->DMA[0x%02X].Capabilities addr(0x%08X)      = 0x%08X\n", 
						 dma, 
						 &pDMAReg->Capabilities, 
						 pDMAReg->Capabilities
						 ));
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
						 " ->DMA[0x%02X].ControlStatus addr(0x%08X)     = 0x%08X\n", 
						 dma, 
						 &pDMAReg->ControlStatus, 
						 pDMAReg->ControlStatus
						 ));
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
						 " ->DMA[0x%02X].Descriptor addr(0x%08X)        = 0x%08X:%08X\n", 
						 dma, 
						 &pDMAReg->Descriptor, 
						 pDMAReg->Descriptor.HighPart, 
						 pDMAReg->Descriptor.LowPart
						 ));
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
						 " ->DMA[0x%02X].HardwareTime       = 0x%08X\n", 
						 dma, 
						 pDMAReg->HardwareTime
						 ));
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
						 " ->DMA[0x%02X].ChainCmplByteCount = 0x%08X\n", 
						 dma, 
						 pDMAReg->ChainCmplByteCount
						 ));
			pPciDevInfo->pRegisters->DMA[dma].ControlStatus = PED_CTRL_IRQ_ENABLE;
		}
	}

	pPciDevInfo->pRegisters->Common.ControlStatus = PED_CARD_IRQ_ENABLE;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
				 "\npRegistersEx: %x\n", 
				 pPciDevInfo->pRegistersEx
				 ));
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
				 " ->UserIntEnableStatus = 0x%08X\n", 
				 pPciDevInfo->pRegistersEx->UserIntEnableStatus
				 ));
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
				 " ->UserDataFlowControl = 0x%08X\n", 
				 pPciDevInfo->pRegistersEx->UserDataFlowControl
				 ));
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
				 " ->UserMisc            = 0x%08X\n", 
				 pPciDevInfo->pRegistersEx->UserMisc));
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
				 " ->UserErrorStatistics = 0x%08X\n", 
				 pPciDevInfo->pRegistersEx->UserErrorStatistics
				 ));

	lErr = pci_read_config_byte(pPciDevInfo->pcidev, 
								PCI_INTERRUPT_LINE, 
								(u8 *)&pPciDevInfo->InterruptVector
								);
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_WARN), 
				 "QL300 PCIe IRQ no = %x\n", 
				 pPciDevInfo->InterruptVector
				 ));

	spin_lock_init(&pPciDevInfo->irq_lock);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
	lErr = request_irq(pPciDevInfo->pcidev->irq, 
					   //(irq_handler_t)plpci_irqhandler, 
					   (irqreturn_t (*)(int, void *))plpci_irqhandler,
					   IRQF_SHARED /*SA_INTERRUPT*/, 
					   "QL300 PCIe ISR", 
					   pPciDevInfo
					   );
#else //LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	lErr = request_irq(pPciDevInfo->pcidev->irq, 
					   //(irq_handler_t)plpci_irqhandler, 
					   (irqreturn_t (*)(int, void *, struct pt_regs *))plpci_irqhandler,
					   SA_SHIRQ /*SA_INTERRUPT*/, 
					   "QL300 PCIe ISR", 
					   pPciDevInfo
					   );
#endif //LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
	if (lErr) 
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPCIeDBGCompID, QPDBGM_LEVEL_ERROR), 
					 "%s() Unable to create ISR, Err = %x\n", 
					 __FUNCTION__,
					 lErr
					 ));
	}

	pPciDevInfo->devconnected = 1;
	return (QPERR_SUCCESS);
}



EQP_Errors PciStopDevice(PQLPCI_DEVICE pPciDevInfo)
{
	// disable device interrupt
	pPciDevInfo->pRegisters->Common.ControlStatus &= ~PED_CARD_IRQ_ENABLE;

	// release interrupt handler
	free_irq(pPciDevInfo->pcidev->irq, 
			 pPciDevInfo
			 );

	// unmap io region
	iounmap(pPciDevInfo->ioaddr[0]);
	iounmap(pPciDevInfo->ioaddr[1]);
	iounmap(pPciDevInfo->ioaddr[2]);

    return (QPERR_SUCCESS);
}





