///////////////////////////////////////////////////////////////////////////////
//        
//
// pci_dma.h
//
// Description: 
//
//  This is declaration of PCIe internal data structures and functions
// 
// History:
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


/*
=============================================================================

    (C) Copyright 2004 TriplePoint, Inc. -- http://www.TriplePoint.com
        This copyright notice must be maintained in all derivative works.

=============================================================================

@doc INTERNAL pci_dma pci_dma_h

@module pci_dma.h |

    This module declares the driver's device interfaces and management
    structures.

@head3 Contents |
@index class,mfunc,func,msg,mdata,struct,enum | pci_dma_h

@end
=============================================================================
*/

#ifndef _PCI_DMA_H_
#define _PCI_DMA_H_

#pragma warning (disable:4013)  // disable undefined warning for prefast
#include <wdm.h>
#pragma warning (default:4013)  // restore undefined warning

#include "pci_config.h"
#include "cpciecntl.h"


// Maximum size of transfer that we support on this device.
#define CARD_MAX_TRANSFER			0x100000    // 1 MB

// The watchdog interval.  Any request pending for this number of seconds is
// cancelled.  This is a one second granularity timer, so the timeout must be
// at least one second more than the minimum allowed.
#define CARD_WATCHDOG_INTERVAL		5

// Device State definitions
//
// State values are used to track the PnP state of the device.
//
// NOTE: The code makes assumptions about the order of these state,
// and the fact that their values are in the high word of a dword.
//
// We also use a set of dummy states to determine the actions we
// take on receipt and completion of an IRP.

// In STATE_REMOVED, we immediately fail any received I/O requests
#define STATE_REMOVED				0x00000000

// In STATE_SURPRISE_REMOVED, we immediately fail all incoming requests
#define STATE_SURPRISE_REMOVED		0x00010000

// In STATE_NEVER_STARTED, we also immediately fail all incoming requests
#define STATE_NEVER_STARTED			0x00020000

// Dummy State -- When the state is < this value, no H/W access is allowed
#define STATE_ALL_BELOW_NO_HW		0x00030000

// In STATE_REMOVE_PENDING, we also immediately fail all incoming requests
#define STATE_REMOVE_PENDING		0x00100000

// Dummy state -- When an IRP arrives at the driver, if the current
// device state is below this value, it is immediately failed
#define STATE_ALL_BELOW_FAIL		0x00FF0000

// In STATE_STARTED, requests are processed and removed from the
// queues normally
#define STATE_STARTED				0X01000000

// Dummy state -- When an IRP arrives at the driver, if the current
// device state is above this value, it is queued, not initiated on
// the device (even if the device is free)
#define STATE_ALL_ABOVE_QUEUE		0x0FFF0000

// Dummy State -- When an IRP is COMPLETED on the device, if
// the device state is below this value, we'll start another
// IRP in an attempt to drain the queue of pending requests.
#define STATE_ALL_BELOW_DRAIN		STATE_ALL_ABOVE_QUEUE

// In STATE_STOP_PENDING, we queue incoming requests
#define STATE_STOP_PENDING			0x10000000

// In STATE_STOPPED, we queue incoming requests
#define STATE_STOPPED				0x10010000



#pragma pack(push,1)

#define PED_MAX_DMA_BYTE_COUNT		(1 << 28)
#define PED_MAX_DMA_TRANSFER		(64 * 1024 * 1024)

#define PED_DMA_ALIGN_LEN			32
#define PED_DMA_ALIGN_VIRT_ADDR(x)	((PPED_DMA_DESCRIPTOR)((((ULONG_PTR)(x)) + (PED_DMA_ALIGN_LEN - 1)) & ~(PED_DMA_ALIGN_LEN - 1)))
#define PED_DMA_ALIGN_PHYS_ADDR(x)	(((x) + (PED_DMA_ALIGN_LEN - 1)) & ~(PED_DMA_ALIGN_LEN - 1))


// Maximum length of the buffer that we can send to IoAllocateMdl()
// according to the MSDN
#define PCI_MAX_MDL_LENGTH			((PAGE_SIZE * (0xFFFF - sizeof(MDL))) / sizeof(ULONG_PTR))



typedef struct PED_DMA_ENGINE
{
#define PED_CAP_ENGINE_PRESENT		(1 << 0)
#define PED_CAP_ENGINE_TYPE(x)		(((x) >> 1) & 0x01)
#define PED_CAP_ENGINE_TYPE_C2S		1
#define PED_CAP_ENGINE_TYPE_S2C		0
#define PED_CAP_ENGINE_NUMBER(x)	(((x) >> 8) & 0xFF)
#define PED_CAP_CARD_RAM_SIZE(x)	(1UL << (((x) >> 16) & 0xFF))

    volatile ULONG				Capabilities;

#define PED_CTRL_IRQ_ENABLE			(1 << 0)
#define PED_CTRL_IRQ_ACTIVE			(1 << 1)
#define PED_CTRL_CHAIN_START		(1 << 8)
#define PED_CTRL_CHAIN_STOP			(1 << 9)
#define PED_STAT_CHAIN_RUNNING		(1 << 10)
#define PED_STAT_CHAIN_COMPLETE		(1 << 11)
#define PED_STAT_CHAIN_SHORT_ERR	(1 << 12)
#define PED_STAT_CHAIN_SHORT_SW		(1 << 13)
#define PED_STAT_CHAIN_SHORT_HW		(1 << 14)

    volatile ULONG				ControlStatus;
    volatile PHYSICAL_ADDRESS	Descriptor;
    volatile ULONG				HardwareTime;
    volatile ULONG				ChainCmplByteCount;
    volatile ULONG				Reserved[(0x0100-0x0018) / sizeof(ULONG)];
} PED_DMA_ENGINE, *PPED_DMA_ENGINE;



typedef struct PED_DMA_COMMON
{
#define PED_CARD_IRQ_ENABLE			(1 << 0)
#define PED_CARD_IRQ_ACTIVE			(1 << 1)
#define PED_CARD_IRQ_PENDING		(1 << 2)
#define PED_CARD_IRQ_MSI			(1 << 3)
#define PED_CARD_MAX_C2S_SIZE(x)	(1UL << ((((x) >> 8) & 0xFF)+7))
#define PED_CARD_MAX_S2C_SIZE(x)	(1UL << ((((x) >> 12) & 0xFF)+7))

    volatile ULONG	ControlStatus;
    volatile ULONG	FpgaVersion;
} PED_DMA_COMMON, *PPED_DMA_COMMON;


// BAR0 Register Layout
typedef struct PED_BAR0_REGISTERS
{
	// DMA Engines
	PED_DMA_ENGINE	DMA[PCI_MAX_DMAS];

	// DMA Common DMA Register Block
	PED_DMA_COMMON	Common;
} PED_BAR0_REGISTERS, *PPED_BAR0_REGISTERS;


// extended BAR0 Registers
typedef struct PED_BAR0_REGISTERS_EX
{
	volatile ULONG	VerCore;			//0x8000
	volatile ULONG	VerBE;				//0x8004
	volatile ULONG	CorePCredits;		//0x8008
	volatile ULONG	CoreNCredits;		//0x800C
	volatile ULONG	CoreCCredits;		//0x8010
	volatile ULONG	RemPCredits;		//0x8014
	volatile ULONG	RemNCredits;		//0x8018
	volatile ULONG	RemCCredits;		//0x801C
	volatile ULONG	H2CScratchPad0;		//0x8020
	volatile ULONG	H2CScratchPad1;		//0x8024
	volatile ULONG	H2CScratchPad2;		//0x8028
	volatile ULONG	H2CScratchPad3;		//0x802C

#define PED_H2C_SCRATCH_PAD_VALID	(1 << 31)
#define PED_C2H_SCRATCH_PAD_VALID	(1 << 30)
	
	volatile ULONG	UserIntEnableStatus;//0x8030
	volatile ULONG	UserDataFlowControl;//0x8034

#define PED_USER_MISC_WR_PAGE_SEL_MASK	(0xFFFFFFF0) 
#define PED_USER_MISC_RD_PAGE_SEL_MASK	(0xFFFFF0FF) 

	volatile ULONG	UserMisc;			//0x8038
	volatile ULONG	UserErrorStatistics;//0x803C
} PED_BAR0_REGISTERS_EX, *PPED_BAR0_REGISTERS_EX;



typedef struct PED_DMA_DESCRIPTOR
{
#define PED_DMA_DESCR_CTRL_IRQ_ON_COMPLETION	(1 << 0)
#define PED_DMA_DESCR_CTRL_IRQ_ON_SHORT_ERR		(1 << 1)
#define PED_DMA_DESCR_CTRL_IRQ_ON_SHORT_SW		(1 << 2)
#define PED_DMA_DESCR_CTRL_IRQ_ON_SHORT_HW		(1 << 3)
#define PED_DMA_DESCR_CTRL_SEQUENCE				(1 << 8)
#define PED_DMA_DESCR_CTRL_CONTINUE				(1 << 9)

#define PED_DMA_DESCR_CADDR_DUMMY				((ULONGLONG)1 << 59)

#define PED_DMA_DESCR_CADDR_SWAP(x)				((ULONGLONG)(x) << 56)

#define PED_DMA_DESCR_CADDR_X_LEN(x)			((ULONGLONG)(x) << 30)
#define PED_DMA_DESCR_CADDR_Y_LEN(x)			((ULONGLONG)(x) << 40)
#define PED_DMA_DESCR_CADDR_Z_LEN(x)			((ULONGLONG)(x) << 48)

#define PED_DMA_DESCR_CTRL_X_DELTA_1(x)			((x) << 12)
#define PED_DMA_DESCR_CTRL_X_DELTA_2(x)			((x) << 17)
#define PED_DMA_DESCR_CTRL_Y_DELTA(x)			((x) << 22)
#define PED_DMA_DESCR_CTRL_Z_DELTA(x)			((x) << 27)

    ULONG		Control;
    ULONG		ByteCount;
    ULONGLONG	SystemAddress;
    ULONGLONG	CardAddress;
    ULONGLONG	NextDescriptor;
} PED_DMA_DESCRIPTOR, *PPED_DMA_DESCRIPTOR;

#pragma pack(pop)

// DMA engine direction
typedef enum
{
    DmaDir_C2S = 0,	// Card to System
    DmaDir_S2C = 1	// System to Card
} tDmaDir;


/* @doc INTERNAL pci_dma pci_dma_h _DEVICE_TRANSFER
=============================================================================

@struct DEVICE_TRANSFER |

    This structure contains the data associated with a DMA transfer.

*/
typedef struct DEVICE_TRANSFER
{
    PDMA_ADAPTER	DmaAdapter;

    tDmaDir			DmaDirection;
    ULONG			DmaChannel;
    ULONG			DmaIndex;

	struct
	{
		PPED_DMA_DESCRIPTOR	VirtAddress;
		PHYSICAL_ADDRESS	PhysAddress;
    } DmaDescriptor;

    ULONG			MapRegsAvailable;
    ULONG			NumDescriptors;
    ULONG			UsedDescriptors;

    ULONG			Timeout;

    PVOID			hDmaRequest; //DWG_FIXUP
	BOOLEAN			bStarted;				
} DEVICE_TRANSFER, *PDEVICE_TRANSFER;



/* @doc INTERNAL pci_dma pci_dma_h PCI_MEMORY_RANGE
=============================================================================

@struct PCI_MEMORY_RANGE |

    This structure contains the data associated with a PCI Memory region.

*/
typedef struct PCI_MEMORY_RANGE
{
	ULONG				Length;
	// The number of bytes of memory the card has in this region.
	// Use the <t CARD_MAX_MEMORY_REGIONS> defininition to set the
	// maximum number of expected memory regions

	PHYSICAL_ADDRESS	PhysicalAddress;
	// System physical address assigned to the card's on board memory.

	PVOID				VirtualAddress;
	// Virtual adress of card's memory area.

} PCI_MEMORY_RANGE, *PPCI_MEMORY_RANGE;



/* @doc INTERNAL pci_dma pci_dma_h PCI_MEMORY_RANGE
=============================================================================

@struct PCI_IO_RANGE |

    This structure contains the data associated with a PCI I/O region.

*/
typedef PCI_MEMORY_RANGE PCI_IO_RANGE, *PPCI_IO_RANGE;


#define _PCI_USE_KE_TIMER


/* @doc INTERNAL pci_dma pci_dma_h _PCIE_DEVICE_EXTENSION
=============================================================================

@struct PCIE_DEVICE_EXTENSION |

    This structure contains the data associated with each device.

*/
typedef struct _PCIE_DEVICE_EXTENSION
{
	// PCIe module object
	CPCIeCntl			*pPCIeCntl;

    // Device information
    PDEVICE_OBJECT      FunctionalDeviceObject;
    PDEVICE_OBJECT      PhysicalDeviceObject;
    PDEVICE_OBJECT      LowerDeviceObject;

    // Interrupt Object
    PKINTERRUPT         InterruptObject;

    // PCI Configuration infomration; once set at configuration time,
    // this information remains the same.
    INTERFACE_TYPE      InterfaceType;
    ULONG               BusType;
    ULONG               BusNumber;
    UCHAR               InterruptLevel;
    ULONG               InterruptVector;
    KAFFINITY           InterruptAffinity;
    ULONG               InterruptMode;

    // Flags to keep track of resource allocation.
    BOOLEAN             TimerStarted;
#ifdef _PCI_USE_KE_TIMER
	PRKDPC				pTimerDpc;
	PKTIMER				pTimer;
#endif //_PCI_USE_KE_TIMER

    // PNP Events
    KEVENT              StartEvent;
    KEVENT              StopEvent;
    KEVENT              RemoveEvent;

    // Device State for PnP Purposes
    ULONG               State;

	// Outstanding I/O counters
	ULONG				OutstandingIO;

    // InterruptStatus register read by PciInterruptService ISR.
	volatile ULONG		InterruptStatus;

    // InterruptEnable register written by PciDpcForIsr DPC.
	ULONG				InterruptEnable;

	// DPC object for defered produce call
	KDPC				Dpc;

	DEVICE_TRANSFER		DmaDevices[PCI_MAX_DMAS];
	int					ReadDmaChannel;
	int					WriteDmaChannel;

	LIST_ENTRY			DmaBufferList;
	KSPIN_LOCK			DmaBufferListLock;

    KSPIN_LOCK			DmaQueueLock;
    KIRQL				DmaQueueLockIrql;

	USHORT				m_dmaBar;
	USHORT				m_numBars;
	USHORT				m_CardConfigId;
	USHORT				m_NumDmaAvailable;
	USHORT				m_InterruptAvailable;
	BOOLEAN				m_64BitAddress;
	BOOLEAN				m_64BitCapable;
	ULONG				m_DmaMaxTransfer;

	PCI_MEMORY_RANGE	m_MemoryRange[PCI_TYPE0_ADDRESSES];
	PCI_IO_RANGE		m_IoRange[PCI_TYPE0_ADDRESSES];

	PCI_COMMON_CONFIG	m_pciConfig;
	PCI_CONFIG_HEADER	m_pciHeader;
	PCI_RW_RAM_STRUCT	m_ReadInfo;
	PCI_RW_RAM_STRUCT	m_WriteInfo;
	PPED_BAR0_REGISTERS	pRegisters;
	PPED_BAR0_REGISTERS_EX	pRegistersEx;
	LARGE_INTEGER		DmaStartTime;
	LARGE_INTEGER		IrpStartTime;


	// bus callback
	//
	QP_BUSINTF_CALLBACK	pBusCallbackFunc;
	TQP_PVOID			pBusCallbackContext;

} PCIE_DEVICE_EXTENSION, *PPCIE_DEVICE_EXTENSION;


#define DMA_BUFFER_SIGNATURE    0xFEEDBEEF


typedef VOID (*PCI_DMA_GET_SGLIST_CB)(struct PCI_DMA_BUFFER *pDmaBuffer, NTSTATUS status);

typedef struct PCI_DMA_BUFFER
{
	ULONG					Signature;
	LIST_ENTRY				ListEntry;

	PCIE_DEVICE_EXTENSION	*pDeviceExtension;
	DMA_ADAPTER				*pDmaAdapter;

	MDL						*pMdlAddressCaller;
	MDL						*pMdl;
	SCATTER_GATHER_LIST		*pScatterGatherList;

	BOOLEAN					IsWrite;
	PCI_DMA_GET_SGLIST_CB	pGetSgListCB;
	PVOID					pGetSgListUD;

	LIST_ENTRY				DmaRequestList;
	KSPIN_LOCK				DmaRequestListLock;
} PCI_DMA_BUFFER, *PPCI_DMA_BUFFER;


typedef VOID (*PCI_DMA_REQUEST_CB)(struct PCI_DMA_REQUEST *pDmaRequest);


typedef struct PCI_DMA_REQUEST
{
	LIST_ENTRY				ListEntry;

	PCI_DMA_BUFFER			*pDmaBuffer;
	ULONG					NumBytes;
	ULONG					NumDescriptors;
	MDL						*pPerformanceMdl;

	PCI_DMA_REQUEST_CB		pRequestCB;
} PCI_DMA_REQUEST, *PPCI_DMA_REQUEST;



// Protoytpes
//
NTSTATUS PciAddDevice(CPCIeCntl *This,
					  PDEVICE_OBJECT FunctionalDeviceObject,
					  PDEVICE_OBJECT PhysicalDeviceObject
					  );
NTSTATUS PciStartDevice(PPCIE_DEVICE_EXTENSION DeviceExtension,
						PCM_RESOURCE_LIST pResourceListTranslated
						);
NTSTATUS PciReadConfig(PPCIE_DEVICE_EXTENSION DeviceExtension);
NTSTATUS PciValidateConfig(PPCIE_DEVICE_EXTENSION DeviceExtension);

NTSTATUS PciRemoveDevice(CPCIeCntl *This);
NTSTATUS PciStopDevice(PPCIE_DEVICE_EXTENSION DeviceExtension);

BOOLEAN  PciInterruptService(PKINTERRUPT Interrupt, 
							 PVOID ServiceContext
							 );

NTSTATUS PciStartWatchdogTimer(PPCIE_DEVICE_EXTENSION DeviceExtension);
VOID PciStopWatchdogTimer(PPCIE_DEVICE_EXTENSION DeviceExtension);

VOID PciDpcForIsr(IN PKDPC Dpc,						// Address of our DPC object.
				  IN PDEVICE_OBJECT DeviceObject,	// Address of our DEVICE_OBJECT.
				  IN PIRP UnusedIrp,				// Unused.
				  IN PVOID Context					// Address of our PCIE_DEVICE_EXTENSION.
				  );
VOID PciDpcForIsrArmMsg(IN PKDPC Dpc,				// Address of our DPC object.
						IN PVOID DeferredContext,
						IN PVOID SystemArgument1,
						IN PVOID SystemArgument2
						);

VOID PciInitializeDeviceState(PPCIE_DEVICE_EXTENSION DeviceExtension);
VOID PciProcessQueuedRequests(PPCIE_DEVICE_EXTENSION DeviceExtension);
VOID PciFlushQueuedRequests(PPCIE_DEVICE_EXTENSION DeviceExtension);
VOID PciIncrementIOCount(PPCIE_DEVICE_EXTENSION DeviceExtension);
VOID PciDecrementIOCount(PPCIE_DEVICE_EXTENSION DeviceExtension);
VOID PciWaitForStopEvent(PPCIE_DEVICE_EXTENSION DeviceExtension);
VOID PciWaitForRemoveEvent(PPCIE_DEVICE_EXTENSION DeviceExtension);
VOID PciReleaseResources(PPCIE_DEVICE_EXTENSION DeviceExtension);

PSTR DbgStateName(ULONG State);
VOID DbgPrintResourceList(PCM_RESOURCE_LIST);
VOID DbgPrintConfig(PPCI_COMMON_CONFIG  configInfo);

#include "ped_dma.h"
//#include "msglog.h"

// The following inline functions are included here to support 64-bit operations
// using the same methods used by Windows for 8,16&32 bit operations.
#if defined(_M_AMD64)

__forceinline VOID WRITE_REGISTER_BUFFER_ULONGLONG(PULONGLONG Register,
												   PULONGLONG Buffer,
												   ULONG Count
												   )
{
	LONG	Synch;

	__movsq(Register, Buffer, Count);
	InterlockedOr(&Synch, 1);
}

__forceinline VOID WRITE_REGISTER_ULONGLONG(PULONGLONG Register,
											ULONGLONG Value
											)
{
    LONG	Synch;

    *Register = Value;
    InterlockedOr(&Synch, 1);
}

__forceinline VOID READ_REGISTER_BUFFER_ULONGLONG(PULONGLONG Register,
												  PULONGLONG Buffer,
												  ULONG Count
												  )
{
	__movsq(Register, Buffer, Count);
}

__forceinline ULONGLONG READ_REGISTER_ULONGLONG(volatile ULONGLONG *Register)
{
	return (*Register);
}

#endif // defined(_M_AMD64)

#if defined(_M_IA64)

extern void __cdecl __mf(void);

__forceinline VOID WRITE_REGISTER_BUFFER_ULONGLONG(PULONGLONG registerBuffer,
												   PULONGLONG writeBuffer,
												   ULONG writeCount
												   )
{
	for (; writeCount--; writeBuffer++, registerBuffer++)
		*(volatile ULONGLONG * const)(registerBuffer) = *writeBuffer;
	KeFlushWriteBuffer();
}

__forceinline VOID WRITE_REGISTER_ULONGLONG(PULONGLONG Register,
											ULONGLONG Value
											)
{
	*(volatile ULONGLONG * const)(Register) = Value;
	KeFlushWriteBuffer();
}

__forceinline VOID READ_REGISTER_BUFFER_ULONGLONG(PULONGLONG registerBuffer,
												  PULONGLONG readBuffer,
												  ULONG readCount
												  )
{
	__mf();
	for (; readCount--; readBuffer++, registerBuffer++)
		*readBuffer = *(volatile ULONGLONG * const)(registerBuffer);
}

__forceinline ULONGLONG READ_REGISTER_ULONGLONG(volatile ULONGLONG *Register)
{
	__mf();
	return (*(volatile ULONGLONG * const)(Register));
}

#endif // defined(_M_IA64)

#endif /* _PCI_DMA_H_ */

