///////////////////////////////////////////////////////////////////////////////
//     
//
// Cqpusbintf.c
//
// Description:
//
//  This is the Linux implementation of the USB bus interface
//
// History:
//	Jan 19, 2010 (dt) created
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


#include "qlautoconf.h"

#include <linux/version.h>
#include <linux/string.h>
#include <linux/usb.h>
#include <linux/mm.h>
#include <linux/module.h>
#include <asm/io.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/slab.h>

#include "cqpusbintf.h"
#include "util.h"
#include "qpdbgmgr.h"
#include "qpmemmgr.h"


extern TQP_DbgMCompID	g_QPUsbDBGCompID;


/////////////////////////////////////////////////////////////////////////////
//
//

typedef struct _qpusb_pipe_info
{
	int	pipenum;
	int	endpointaddr;
	int	max_pkt_size;
	int attribute;
} qpusb_pipe_info, *pqpusb_pipe_info;


// private data
//

typedef struct _QPUSBINTF_PRIV_DATA
{
	struct usb_device	*usbdev;
	unsigned int		numEndpoints;
	pqpusb_pipe_info	p_pipeinfo;
    struct urb			*dmardurb;
    struct urb			*dmawrurb;
    unsigned char		*dmard_buffer;
    unsigned char		*dmawr_buffer;
	dma_addr_t			dmard_dmaddr;
	dma_addr_t			dmawr_dmaddr;
	// error
	TQP_BOOL			bDeviceIoErrorOccured;
} QPUSBINTF_PRIV_DATA, *PQPUSBINTF_PRIV_DATA;


#ifdef USE_BIG_USB_DMA_SIZE
#define QLUSB_DMA_BUFF_SIZE		(512 * 1024)
#else
#define QLUSB_DMA_BUFF_SIZE		(128 * 1024)
#endif

/////////////////////////////////////////////////////////////////////////////
//
//

static EQP_Errors _usb_memalloc(unsigned long *physaddr,
								unsigned long *virtaddr,
								unsigned long size
								)
{
    EQP_Errors	qpStatus = QPERR_SUCCESS;
    int			pagecnt;

	if (physaddr &&
		virtaddr &&
		size
		)
	{
		*virtaddr = (unsigned long)kmalloc(size,
										   GFP_KERNEL | __GFP_DMA
										   );
		if (*virtaddr)
		{
			*physaddr = virt_to_phys((void*)(*virtaddr));
			pagecnt = (size + PAGE_SIZE - 1) / PAGE_SIZE;
            while (pagecnt--)
			{
				SetPageReserved(virt_to_page((*virtaddr) + (PAGE_SIZE * pagecnt)));
			}
        }
		else
		{
            qpStatus = QPERR_NOMEM;
        }
	}
	else
	{
		qpStatus = QPERR_PARMS;
	}

	return (qpStatus);
}



static EQP_Errors _usb_memfree(unsigned long *physaddr,
							   unsigned long *virtaddr,
							   unsigned long size
							   )
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;
	int			pagecnt;

	if (physaddr &&
		virtaddr &&
		size
		)
	{
		pagecnt = (size + PAGE_SIZE - 1) / PAGE_SIZE;
        while (pagecnt--)
		{
			ClearPageReserved(virt_to_page((*virtaddr) + (PAGE_SIZE*pagecnt)));
		}
		kfree((void*)*virtaddr);
	}
	else
	{
		qpStatus = QPERR_PARMS;
    }

	return (qpStatus);
}



#ifndef __LINUX24__
void BulkTransferComplete(struct urb *urb)
#else
void BulkTransferComplete(struct urb *urb,
						  struct pt_regs *regs
						  )
#endif //!__LINUX24__
{
	PQPUSB_IO_CONTEXT		pIoContext = (PQPUSB_IO_CONTEXT)urb->context;
	CUsbInterface			*This =	GET_INHERITED_OBJECT(CUsbInterface, m_iUsbInterface, pIoContext->pUsb);
	PQPUSBINTF_PRIV_DATA	pPrivData = (PQPUSBINTF_PRIV_DATA)This->m_pPrivateData;
	EQP_Errors				qpStatus = QPERR_SUCCESS;
	TQP_UINT32				transferSize = urb->actual_length;
	QPUSBCOMPLETION			pCompletionRoutine = QP_NULL;
	TQP_PVOID				pUserContext = QP_NULL;

	if (urb->status)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPUsbDBGCompID, QPDBGM_LEVEL_ERROR),
					 "%s() Failed: pipe(%d) urb status(%d) transferSize(%d)\n",
					 __FUNCTION__,
					 pIoContext->Pipe,
					 urb->status,
					 transferSize
					 ));
		qpStatus = QPERR_FAIL;
		pPrivData->bDeviceIoErrorOccured = QP_TRUE;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPUsbDBGCompID, QPDBGM_LEVEL_TRACE),
					 "%s() Pipe(%d) transferSize(%d)\n",
					 __FUNCTION__,
					 pIoContext->Pipe,
					 transferSize
					 ));
	}

	// save status and actual size transferred
	//
	pIoContext->qpStatus = qpStatus;
	pIoContext->transferSize = transferSize;

	if (pIoContext->pCallerBuffer)
	{
        memcpy(pIoContext->pCallerBuffer,
			   pPrivData->dmard_buffer,
			   transferSize
			   );
		pIoContext->pCallerBuffer = QP_NULL;
	}

	pCompletionRoutine = pIoContext->pCompletion;
	pUserContext = pIoContext->UserContext;

	// Callback client
	//
	if (pCompletionRoutine)
	{
		(*pCompletionRoutine)(pUserContext,
							  pIoContext
							  );
	}
}



/////////////////////////////////////////////////////////////////////////////
//
//

static EQP_Errors CUsbInterface_UsbCallBusDriver(IUsbInterface *This,
												 TQP_PVOID pUrb
												 )
{
	return (QPERR_SUCCESS);
}



static EQP_Errors CUsbInterface_UsbSendCmd(IUsbInterface *This_p,
										   TQP_UINT32 dwPipeNum,
										   TQP_PVOID pCmd,
										   TQP_UINT32 dwSize,
										   TQP_UINT32* pdwXfered
										   )
{
    EQP_Errors				qpStatus = QPERR_SUCCESS;
    CUsbInterface			*This = GET_INHERITED_OBJECT(CUsbInterface, m_iUsbInterface, This_p);
	PQPUSBINTF_PRIV_DATA	pPrivData = (PQPUSBINTF_PRIV_DATA)This->m_pPrivateData;
	struct usb_device		*usbdev = pPrivData->usbdev;
	int						endp = 0;
    long					lErr = 0;

	if (pPrivData->bDeviceIoErrorOccured)
	{
		return (QPERR_FAIL);
	}

	if (pPrivData->numEndpoints > dwPipeNum)
	{
		endp = pPrivData->p_pipeinfo[dwPipeNum].endpointaddr;
		if ((endp & USB_ENDPOINT_DIR_MASK) == USB_DIR_OUT)
		{
			if (USB_ENDPOINT_XFER_INT == (pPrivData->p_pipeinfo[dwPipeNum].attribute & USB_ENDPOINT_XFERTYPE_MASK))
			{
				lErr = usb_bulk_msg(usbdev,
									usb_sndintpipe(usbdev, endp),
									pCmd,
									dwSize,
									(int *)pdwXfered,
									50 * HZ
									);
			}
			else
			{
				lErr = usb_bulk_msg(usbdev,
									usb_sndbulkpipe(usbdev, endp),
									pCmd,
									dwSize,
									(int *)pdwXfered,
									50 * HZ
									);
			}
		}
		else
		{
			if (USB_ENDPOINT_XFER_INT == (pPrivData->p_pipeinfo[dwPipeNum].attribute & USB_ENDPOINT_XFERTYPE_MASK))
			{
				lErr = usb_bulk_msg(usbdev,
									usb_rcvintpipe(usbdev, endp),
									pCmd,
									dwSize,
									(int *)pdwXfered,
									50 * HZ
									);
			}
			else
			{
				lErr = usb_bulk_msg(usbdev,
									usb_rcvbulkpipe(usbdev, endp),
									pCmd,
									dwSize,
									(int *)pdwXfered,
									50 * HZ
									);
			}
		}

		if (lErr)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPUsbDBGCompID, QPDBGM_LEVEL_ERROR),
						 "%s() usb_bulk_msg send failed lErr(%d)\n",
						 __FUNCTION__,
						 lErr
						 ));
			pPrivData->bDeviceIoErrorOccured = QP_TRUE;
			qpStatus = QPERR_FAIL;
		}
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPUsbDBGCompID, QPDBGM_LEVEL_ERROR),
					 "%s() Invalid pipe number(%d)\n",
					 __FUNCTION__,
					 dwPipeNum
					 ));
		qpStatus = QPERR_PARMS;
	}

	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPUsbDBGCompID, QPDBGM_LEVEL_ERROR),
					 "%s() pipe(%d) data(%d) len(%d) xfered(%d) status(0x%x) ]\n",
					 __FUNCTION__,
					 dwPipeNum,
					 *((PTQP_UINT8)pCmd),
					 dwSize,
					 *pdwXfered,
					 qpStatus
					 ));
	}

	return (qpStatus);
}



static EQP_Errors CUsbInterface_UsbAsyncIo(IUsbInterface *This_p,
										   TQP_UINT32 dwPipeNum,
										   PTQP_UINT8 pBuff,
										   TQP_UINT32 dwSize,
										   PQPUSB_IO_CONTEXT pIoContext
										   )
{
	EQP_Errors				qpStatus = QPERR_SUCCESS;
    CUsbInterface			*This = GET_INHERITED_OBJECT(CUsbInterface, m_iUsbInterface, This_p);
	PQPUSBINTF_PRIV_DATA	pPrivData = (PQPUSBINTF_PRIV_DATA)This->m_pPrivateData;
	struct usb_device		*usbdev = pPrivData->usbdev;
	int						endp = 0;
	long					lErr = 0;

	if (pPrivData->bDeviceIoErrorOccured)
	{
		return (QPERR_FAIL);
	}

	if (pPrivData->numEndpoints > dwPipeNum)
	{
		// Save context information.
		//
		pIoContext->pUsb = This_p;
		pIoContext->Pipe = dwPipeNum;
		pIoContext->dwXferSize = dwSize;
		pIoContext->qpStatus = QPERR_PENDING;
		pIoContext->transferSize = 0;

		endp = pPrivData->p_pipeinfo[dwPipeNum].endpointaddr;
		if ((endp & USB_ENDPOINT_DIR_MASK) == USB_DIR_OUT)
		{
			pIoContext->pCallerBuffer = QP_NULL;
            memcpy(pPrivData->dmawr_buffer,
				   pBuff,
				   dwSize
				   );
            usb_fill_bulk_urb(pPrivData->dmawrurb,
							  usbdev,
							  usb_sndbulkpipe(usbdev, endp),
							  pPrivData->dmawr_buffer,
							  QLUSB_DMA_BUFF_SIZE,
							  BulkTransferComplete,
							  pIoContext
							  );
			pPrivData->dmawrurb->transfer_buffer_length = dwSize;
			pPrivData->dmawrurb->actual_length = 0;
#ifndef __LINUX24__
            pPrivData->dmawrurb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
            pPrivData->dmawrurb->transfer_dma = pPrivData->dmawr_dmaddr;

			lErr = usb_submit_urb(pPrivData->dmawrurb,
								  GFP_KERNEL | __GFP_COMP
								  );
#else //__LINUX24__
			lErr = usb_submit_urb(pPrivData->dmawrurb);
#endif //!__LINUX24__
		}
		else
		{
			pIoContext->pCallerBuffer = pBuff;

			usb_fill_bulk_urb(pPrivData->dmardurb,
							  usbdev,
							  usb_rcvbulkpipe(usbdev, endp),
							  pPrivData->dmard_buffer,
							  QLUSB_DMA_BUFF_SIZE,
							  BulkTransferComplete,
							  pIoContext
							  );
			pPrivData->dmardurb->transfer_buffer_length = dwSize;
			pPrivData->dmardurb->actual_length = 0;
#ifndef __LINUX24__
            pPrivData->dmardurb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
            pPrivData->dmardurb->transfer_dma = pPrivData->dmard_dmaddr;

			lErr = usb_submit_urb(pPrivData->dmardurb,
								  GFP_KERNEL | __GFP_COMP
								  );
#else //__LINUX24__
			lErr = usb_submit_urb(pPrivData->dmardurb);
#endif //!__LINUX24__
		}
		if (lErr)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPUsbDBGCompID, QPDBGM_LEVEL_ERROR),
						 "%s() usb_submit_urb failed lErr(%d)\n",
						 __FUNCTION__,
						 lErr
						 ));
			pPrivData->bDeviceIoErrorOccured = QP_TRUE;
			qpStatus = QPERR_FAIL;
		}
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPUsbDBGCompID, QPDBGM_LEVEL_ERROR),
					 "%s() Invalid pipe number(%d)\n",
					 __FUNCTION__,
					 dwPipeNum
					 ));
		qpStatus = QPERR_PARMS;
	}

	return (qpStatus);
}



static EQP_Errors CUsbInterface_UsbResetPipe(IUsbInterface *This_p,
											 TQP_UINT32 dwPipeNum
											 )
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;

	return (qpStatus);
}



static EQP_Errors CUsbInterface_UsbAbortPipe(IUsbInterface *This_p,
											 TQP_UINT32 dwPipeNum
											 )
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;
	return (qpStatus);
}



static TQP_BOOL CUsbInterface_UsbIsDataPipeInput(IUsbInterface *This_p,
												 TQP_UINT32 dwPipeNum
												 )
{
	CUsbInterface			*This =	GET_INHERITED_OBJECT(CUsbInterface, m_iUsbInterface, This_p);
	PQPUSBINTF_PRIV_DATA	pPrivData = (PQPUSBINTF_PRIV_DATA)This->m_pPrivateData;

	if (pPrivData->numEndpoints > dwPipeNum)
	{
		return ((pPrivData->p_pipeinfo[dwPipeNum].endpointaddr & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN);
	}
	else
	{
		return (QP_FALSE);
	}
}



CUsbInterface * CUsbInterface_Constructor(CUsbInterface *pUsbInterface,
										  CObject *pParent,
										  TQP_UINT32 dwAttributes,
										  TQP_PVOID pDO,
										  TQP_PVOID pDOLayered
										  )
{
	PQPUSBINTF_PRIV_DATA			pPrivData;
#ifndef __LINUX24__
	struct usb_interface			*intf = (struct usb_interface *)pDO;
    struct usb_device				*usbdev = interface_to_usbdev(intf);
	struct usb_endpoint_descriptor	endpointdesc;
	struct usb_host_endpoint		*endpoint;
#else //__LINUX24__
    struct usb_device				*usbdev = (struct usb_device *)pDO;
    struct usb_interface_descriptor	*qlinterface;
    struct usb_endpoint_descriptor	*qlendpoint;
#endif //!__LINUX24__
	unsigned int					i;

	if (pUsbInterface)
	{
		// CObject
		//
		CObject_Constructor(&pUsbInterface->m_Object,
							pParent,
							dwAttributes
		  					);

		// fill in the function table
		//
        pUsbInterface->m_iUsbInterface.UsbCallBusDriver = CUsbInterface_UsbCallBusDriver;
        pUsbInterface->m_iUsbInterface.UsbSendCmd = CUsbInterface_UsbSendCmd;
        pUsbInterface->m_iUsbInterface.UsbAsyncIo = CUsbInterface_UsbAsyncIo;
        pUsbInterface->m_iUsbInterface.UsbResetPipe = CUsbInterface_UsbResetPipe;
        pUsbInterface->m_iUsbInterface.UsbAbortPipe = CUsbInterface_UsbAbortPipe;
        pUsbInterface->m_iUsbInterface.UsbIsDataPipeInput = CUsbInterface_UsbIsDataPipeInput;

		// allocate storage for private data
		//
        pUsbInterface->m_pPrivateData = (PQPUSBINTF_PRIV_DATA)kmalloc(sizeof(QPUSBINTF_PRIV_DATA),
																	  GFP_KERNEL
																	  );
		if (QP_NULL == pUsbInterface->m_pPrivateData)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPUsbDBGCompID, QPDBGM_LEVEL_ERROR),
						 "%s Unable to allocate private data!\n",
						 __FUNCTION__
						 ));
			CUsbInterface_Destructor(pUsbInterface);
			return (QP_NULL);
		}

		memset(pUsbInterface->m_pPrivateData,
			   0,
			   sizeof(QPUSBINTF_PRIV_DATA)
			   );

		pPrivData = (PQPUSBINTF_PRIV_DATA)pUsbInterface->m_pPrivateData;

		// clear error
		pPrivData->bDeviceIoErrorOccured = QP_FALSE;

		// save USB device
		pPrivData->usbdev = usbdev;

#ifndef __LINUX24__
        // usb_reset_configuration(usbdev);
		// usb_reset_device(usbdev);
#endif //!__LINUX24__

		// allocate DMA buffers and URBs
		//
#ifndef __LINUX24__
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
        pPrivData->dmard_buffer = (unsigned char *)usb_alloc_coherent(usbdev,
																	QLUSB_DMA_BUFF_SIZE,
																	GFP_KERNEL | __GFP_COMP,
																	&pPrivData->dmard_dmaddr
																	);
        pPrivData->dmawr_buffer = (unsigned char *)usb_alloc_coherent(usbdev,
																	QLUSB_DMA_BUFF_SIZE,
																	GFP_KERNEL|__GFP_COMP,
																	&pPrivData->dmawr_dmaddr
																	);
#else
        pPrivData->dmard_buffer = (unsigned char *)usb_buffer_alloc(usbdev,
																	QLUSB_DMA_BUFF_SIZE,
																	GFP_KERNEL | __GFP_COMP,
																	&pPrivData->dmard_dmaddr
																	);
        pPrivData->dmawr_buffer = (unsigned char *)usb_buffer_alloc(usbdev,
																	QLUSB_DMA_BUFF_SIZE,
																	GFP_KERNEL|__GFP_COMP,
																	&pPrivData->dmawr_dmaddr
																	);
#endif
        pPrivData->dmardurb = usb_alloc_urb(0,
											GFP_KERNEL | __GFP_COMP
											);
        pPrivData->dmawrurb = usb_alloc_urb(0,
											GFP_KERNEL | __GFP_COMP
											);
#else //__LINUX24_
        qpStatus = _usb_memalloc((void*)&pPrivData->dmard_dmaddr,
								 (void*)&pPrivData->dmard_buffer,
								 QLUSB_DMA_BUFF_SIZE
								 );
        qpStatus = _usb_memalloc((void*)&pPrivData->dmawr_dmaddr,
								 (void*)&pPrivData->dmawr_buffer,
								 QLUSB_DMA_BUFF_SIZE
								 );
        pPrivData->dmardurb = usb_alloc_urb(0);
        pPrivData->dmawrurb = usb_alloc_urb(0);
#endif //!__LINUX24_

		if (!pPrivData->dmard_buffer ||
			!pPrivData->dmawr_buffer ||
			!pPrivData->dmardurb ||
			!pPrivData->dmawrurb
			)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPUsbDBGCompID, QPDBGM_LEVEL_ERROR),
						 "%s Unable to allocate memory or urb!\n",
						 __FUNCTION__
						 ));
			CUsbInterface_Destructor(pUsbInterface);
			return (QP_NULL);
		}

		// save pipe information
		//
#ifndef __LINUX24__
		pPrivData->numEndpoints = (unsigned int)intf->cur_altsetting->desc.bNumEndpoints;
#else //__LINUX24__
		pPrivData->numEndpoints = (unsigned int)usbdev->actconfig->interface->altsetting->bNumEndpoints;
#endif //!__LINUX24__

		if (pPrivData->numEndpoints)
		{
			pPrivData->p_pipeinfo = (pqpusb_pipe_info)kmalloc(sizeof(qpusb_pipe_info) * pPrivData->numEndpoints,
															  GFP_KERNEL
															  );
		}

		if (!pPrivData->p_pipeinfo)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPUsbDBGCompID, QPDBGM_LEVEL_ERROR),
						 "%s Unable to allocate memory for usb pipe info!\n",
						 __FUNCTION__
						 ));
			CUsbInterface_Destructor(pUsbInterface);
			return (QP_NULL);
		}

#ifndef __LINUX24__
		endpoint = intf->cur_altsetting->endpoint;
		for (i = 0; i < pPrivData->numEndpoints; i++)
		{
			endpointdesc = endpoint->desc;
			pPrivData->p_pipeinfo[i].pipenum = i;
			pPrivData->p_pipeinfo[i].endpointaddr = endpointdesc.bEndpointAddress;
			pPrivData->p_pipeinfo[i].max_pkt_size = endpointdesc.wMaxPacketSize;
			pPrivData->p_pipeinfo[i].attribute = endpointdesc.bmAttributes;

			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPUsbDBGCompID, QPDBGM_LEVEL_WARN),
						 "%s: pipe=%d, endpoint=0x%02x, maxpktsize=%d\n",
						 __FUNCTION__,
						 i,
						 endpointdesc.bEndpointAddress,
						 endpointdesc.wMaxPacketSize
						 ));
			endpoint++;
		}
#else //__LINUX24__
		qlinterface= &usbdev->actconfig->interface[0].altsetting[0];
		for (i = 0; i < pPrivData->numEndpoints; i++)
		{
			qlendpoint = &qlinterface->endpoint[i];
			pPrivData->p_pipeinfo[i].pipenum = i;
			pPrivData->p_pipeinfo[i].endpointaddr = qlendpoint->bEndpointAddress;
			pPrivData->p_pipeinfo[i].max_pkt_size = qlendpoint->wMaxPacketSize;
			pPrivData->p_pipeinfo[i].attribute = qlendpoint->bmAttributes;
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPUsbDBGCompID, QPDBGM_LEVEL_WARN),
						 "%s: pipe=%d, endpoint=0x%02x, maxpktsize=%d\n",
						 __FUNCTION__,
						 i,
						 qlendpoint->bEndpointAddress,
						 qlendpoint->wMaxPacketSize
						 ));
		}
#endif //!__LINUX24__
    }

	return (pUsbInterface);
}



void CUsbInterface_Destructor(CUsbInterface *This)
{
	PQPUSBINTF_PRIV_DATA	pPrivData;

	pPrivData = (PQPUSBINTF_PRIV_DATA)This->m_pPrivateData;

	if (pPrivData)
	{
		if (pPrivData->dmard_buffer)
		{
#ifndef __LINUX24__
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
			usb_free_coherent(pPrivData->usbdev,
							QLUSB_DMA_BUFF_SIZE,
							pPrivData->dmard_buffer,
							pPrivData->dmard_dmaddr
							);
#else
			usb_buffer_free(pPrivData->usbdev,
							QLUSB_DMA_BUFF_SIZE,
							pPrivData->dmard_buffer,
							pPrivData->dmard_dmaddr
							);
#endif
#else //__LINUX24__
			_usb_memfree((void*)&pPrivData->dmard_dmaddr,
						 (void*)&pPrivData->dmard_buffer,
						 QLUSB_DMA_BUFF_SIZE
						 );
#endif //!__LINUX24__
			pPrivData->dmard_buffer = QP_NULL;
			pPrivData->dmard_dmaddr = QP_NULL;
		}

        if (pPrivData->dmawr_buffer)
		{
#ifndef __LINUX24__
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
            usb_free_coherent(pPrivData->usbdev,
							QLUSB_DMA_BUFF_SIZE,
							pPrivData->dmawr_buffer,
							pPrivData->dmawr_dmaddr
							);
#else
            usb_buffer_free(pPrivData->usbdev,
							QLUSB_DMA_BUFF_SIZE,
							pPrivData->dmawr_buffer,
							pPrivData->dmawr_dmaddr
							);
#endif
#else //__LINUX24__
            _usb_memfree((void*)&pPrivData->dmawr_dmaddr,
						 (void*)&pPrivData->dmawr_buffer,
						 QLUSB_DMA_BUFF_SIZE
						 );
#endif //!__LINUX24__
			pPrivData->dmawr_buffer = QP_NULL;
			pPrivData->dmawr_dmaddr = QP_NULL;
		}

		if (pPrivData->dmardurb)
		{
			usb_free_urb(pPrivData->dmardurb);
			pPrivData->dmardurb = 0;
		}
		if (pPrivData->dmawrurb)
		{
			usb_free_urb(pPrivData->dmawrurb);
			pPrivData->dmawrurb = 0;
		}

		if (pPrivData->p_pipeinfo)
		{
			kfree(pPrivData->p_pipeinfo);
			pPrivData->p_pipeinfo = QP_NULL;
		}

        kfree(pPrivData);
		This->m_pPrivateData = QP_NULL;
	}

	// CObject
	//
    CObject_Destructor(&This->m_Object);
}

