///////////////////////////////////////////////////////////////////////////////
//     
//
// Cchannel.c
//
// Description:
//
//	channels represent the data streaming on MPEG chipsets
//
// History:
//	Aug 4, 2006	(dt) initial check in
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


#ifdef QP_LINXKER
#include <asm/div64.h>
#endif
#include "cchannel.h"
#include "cqlcodec.h"
#include "qpdbgmgr.h"
#include "qpmemmgr.h"
#include "qptimemgr.h"
#include "cqueue.h"
#include "util.h"
#ifndef QP_LINXKER
#include <string.h>
#endif //!QP_LINXKER


extern TQP_DbgMCompID	g_QPCodecDBGCompID;



/////////////////////////////////////////////////////////////////////////////
//
//


TQP_UINT64 CChannel_GetStreamTime(CChannel *This)
{
	return (GetSystemTimeUtil());
}

static TQP_BOOL CChannel_RecycleEntry(CChannel *This,
									  QUEUE_ENTRY *pEntry,
									  TQP_BOOL bCancel
									  );


/////////////////////////////////////////////////////////////////////////////
//
//

// CChannel
//
static EQP_Errors CChannel_Open(CChannel *This,
								TQP_OBJECT_HANDLE hChannel,
								TQP_UINT32 dwFlags,
								TQP_PVOID pDataFormat,
								QP_DEVICE_CALLBACK pFuncCallback,
								TQP_PVOID context
								)
{
	int			i;
	EQP_Errors	qpStatus;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CChannel_Open()\n"
				 ));

	// save handle
	//
	This->m_hChannel = hChannel;

	// save the open flags
	//
	This->m_dwOpenFlags = dwFlags;

	// save callback address and context
	//
	This->m_pDeviceCallback = pFuncCallback;
	This->m_callbackContext = context;

	if (!This->m_pFreeQueue ||
		!This->m_pDataRequestQueue ||
		!This->m_pDataPendingQueue
		)
	{
		return (QPERR_NOMEM);
	}

	// make sure there is no residule in the queues
	//
	CQueue_FlushQueue(This->m_pFreeQueue);
	CQueue_FlushQueue(This->m_pDataRequestQueue);
	CQueue_FlushQueue(This->m_pDataPendingQueue);

	// initialize free queue
	//
	for (i = 0; i < QPCODECLIB_MAX_DATA_ENTRIES; i++)
	{
		This->m_Entries[i].Data = QP_NULL;
		CQueue_AddEntry(This->m_pFreeQueue,
						&This->m_Entries[i]
						);
	}

#ifdef _QPCODEC_CALLBACK_IN_THREAD

	// make sure there is no residule in the fifos
	//
	CFifo_FlushFifo(This->m_pPendingCallbackFifo);

#endif //_QPCODEC_CALLBACK_IN_THREAD


	// reset channel state
	//
	This->m_State = QPSTATE_STOP;
	This->m_bPaused = QP_FALSE;

	// clear flush flag
	//
	This->m_bFlushing = QP_FALSE;

	// reset counters and flags
	//
	This->m_ullCntBytes = 0;

	// reset saved dword counter
	//
	This->m_savedNumberBytes = 0;

	// reset timestamp
	//
	This->m_llBasePts = 0;
	This->m_llPrevPts = 0;
	This->m_llAddPts = 0;

	// yes, we are open
	//
	This->m_bOpened = QP_TRUE;

	// Task open
	//
	qpStatus = This->m_pTask->Open(This->m_pTask,
								   This->m_hTask,
								   (TASK_DATA_TYPE)This->m_dataType,
								   This->m_ChannelDirection,
								   (FIRMWARE_DATA_TYPE)This->m_FWDataType,
								   This
								   );

	if (!QP_SUCCESS(qpStatus))
		This->Close(This);

	return (qpStatus);
}



static EQP_Errors CChannel_Close(CChannel *This)
{
	if (This->m_bOpened)
	{
		if (QPSTATE_RUN == This->m_State)
		{
			This->Pause(This);
		}

		if (QPSTATE_STOP != This->m_State)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CChannel_Close() !!! Not Stopped State (%d)!!!!\n",
						 This->m_State
						 ));

			This->Stop(This);
		}

		This->Flush(This);

		// task close
		//
		This->m_pTask->Close(This->m_pTask,
							 This->m_hTask,
							 (TASK_DATA_TYPE)This->m_dataType,
							 QP_TRUE
							 );

		This->m_bOpened = QP_FALSE;
	}

	return (QPERR_SUCCESS);
}



static EQP_Errors CChannel_Start(CChannel *This)
{
	This->m_State = QPSTATE_RUN;
	if (This->m_bPaused)
		This->m_bPaused = QP_FALSE;
	return (QPERR_SUCCESS);
}



static EQP_Errors CChannel_Stop(CChannel *This)
{
	This->m_bPaused = QP_FALSE;
	This->m_State = QPSTATE_STOP;
	return (QPERR_SUCCESS);
}



static EQP_Errors CChannel_Acquire(CChannel *This)
{
	EQP_Errors	qpStatus;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CChannel_Acquire()\n"
				 ));

	qpStatus = This->m_pTask->Acquire(This->m_pTask,
									  This->m_hTask,
									  (TASK_DATA_TYPE)This->m_dataType
									  );
	if (QP_SUCCESS(qpStatus))
	{
		This->m_State = QPSTATE_ACQUIRE;
	}

	return (qpStatus);
}



static EQP_Errors CChannel_Pause(CChannel *This)
{
	if (QPSTATE_RUN == This->m_State)
		This->m_bPaused = QP_TRUE;
	This->m_State = QPSTATE_PAUSE;
	return (QPERR_SUCCESS);
}



static EQP_Errors CChannel_Step(CChannel *This,
								TQP_UINT32 nb
								)
{
	return (QPERR_SUCCESS);
}



static EQP_Errors CChannel_SetRate(CChannel *This,
								   TQP_INT32 lNewRate
								   )
{
	return (QPERR_NOTIMPL);
}



static EQP_Errors CChannel_GetRate(CChannel *This,
								   PTQP_INT32 plRate
								   )
{
	return (QPERR_NOTIMPL);
}



static EQP_Errors CChannel_BeginFlush(CChannel *This)
{
	This->m_bFlushing = QP_TRUE;
	return (QPERR_SUCCESS);
}



static EQP_Errors CChannel_Flush(CChannel *This)
{
	QUEUE_ENTRY	*pEntry;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, /*QPDBGM_LEVEL_WARN*/QPDBGM_LEVEL_TRACE),
				 "CChannel_Flush()[\n"
				 ));

	// let the task know we are flushing
	//
	This->m_pTask->Flush(This->m_pTask,
						 This->m_hTask,
						 (TASK_DATA_TYPE)This->m_dataType
						 );

	while (QP_NULL != (pEntry = CQueue_GetOneEntry(This->m_pDataPendingQueue)))
	{
		if (!CChannel_RecycleEntry(This,
								   pEntry,
								   QP_TRUE
								   ))
		{
			CQueue_AddEntry(This->m_pDataPendingQueue,
							pEntry
							);

			return (QPERR_INVALID);
		}
	}

	while (QP_NULL != (pEntry = CQueue_GetOneEntry(This->m_pDataRequestQueue)))
	{
		CChannel_RecycleEntry(This,
							  pEntry,
							  QP_FALSE
							  );
	}

	return (QPERR_SUCCESS);
}



static EQP_Errors CChannel_EndFlush(CChannel *This)
{
	This->m_bFlushing = QP_FALSE;
	return (QPERR_SUCCESS);
}



static EQP_Errors CChannel_AddBuffer(CChannel *This,
									 PQP_BUFFER_DESCRIPTOR pBufDesc
									 )
{
	EQP_Errors	qpStatus = QPERR_FAIL;
	QUEUE_ENTRY	*pEntry;

	if (pBufDesc)
	{
		if (This->m_bFlushing)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CChannel_AddBuffer() FLUSHING!"
						 ));
			pBufDesc->Status = QPERR_SUCCESS/*QPERR_CANCELLED*/;
			return (QPERR_SUCCESS/*QPERR_CANCELLED*/);
		}

		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_EXTRACE),
					 "CChannel_AddBuffer() (%d)",
					 This->m_hChannel
					 ));

		pEntry = CQueue_GetOneEntry(This->m_pFreeQueue);

		if (!pEntry)
		{
			pBufDesc->Status = QPERR_NOMEM;
			qpStatus = QPERR_NOMEM;

			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CChannel_AddBuffer() Failed No entry available for this buffer!"
						 ));
		}
		else
		{
			pBufDesc->Status = QPERR_PENDING;
			qpStatus = QPERR_SUCCESS;

			pEntry->Data = (TQP_PVOID)pBufDesc;
			CQueue_AddEntry(This->m_pDataRequestQueue,
							pEntry
							);

			// inform task of the arrival of this buffer
			//
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_EXTRACE),
						 "CChannel_AddBuffer() NewBuffer"
						 ));

			// make sure we don't touch the buffer descriptor after set
			// the signal. this is to prevent a super high priority thread
			// to complete the buffer before the addBuffer() call return.
			//
			This->m_pTask->NewBuffer(This->m_pTask,
									 This->m_hTask
									 );
		}
	}

	return (qpStatus);
}



static EQP_Errors CChannel_CancelBuffer(CChannel *This,
										PQP_BUFFER_DESCRIPTOR pBufDesc,
										TQP_BOOL bTimeOut
										)
{
	QUEUE_ENTRY	*pEntry;


	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
				 "CChannel_CancelBuffer() %s (0x%x)\n",
				 bTimeOut ? "Time Out" : "Cancel",
				 pBufDesc
				 ));

	pEntry = CQueue_GetEntryByData(This->m_pDataRequestQueue,
								   pBufDesc
								   );

	if (!pEntry)
	{
		if (CQueue_PeekEntryByData(This->m_pDataPendingQueue,
								   pBufDesc
								   ))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CChannel_CancelBuffer()(id : %d) ::%s - in pending queue try cancel\n",
						 This->m_hChannel,
						 bTimeOut ? "Time Out" : "Cancel"
						 ));
			if (!This->m_pTask->CancelBuffer(This->m_pTask,
											 This->m_hTask,
											 (TASK_DATA_TYPE)This->m_dataType,
											 pBufDesc
											 ))
			{
				return (QPERR_FAIL);
			}
			else
			{
				return (QPERR_SUCCESS);
			}
		}
	}

	if (pEntry)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
					 "CChannel_CancelBuffer()(id : %d) %s - GOT IT\n",
					  This->m_hChannel,
					  bTimeOut ? "Time Out" : "Cancel"
					  ));
		CChannel_RecycleEntry(This,
							  pEntry,
							  QP_FALSE
							  );
		return (QPERR_SUCCESS);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
				 "CChannel_CancelBuffer()(id : %d) %s - NOT FOUND\n",
				 This->m_hChannel,
				 bTimeOut ? "Time Out" : "Cancel"
				 ));

	return (QPERR_NOTFOUND);
}



static EQP_Errors CChannel_TimeoutBuffer(CChannel *This,
										 PQP_BUFFER_DESCRIPTOR pBufDesc
										 )
{
	return (CChannel_CancelBuffer(This,
								  pBufDesc,
								  QP_TRUE
								  ));
}



static EQP_Errors CChannel_XferData(CChannel *This,
									PQP_BUFFER_DESCRIPTOR pBufDesc
									)
{
	return (QPERR_SUCCESS);
}


// called from streaming task
//
static TQP_BOOL CChannel_GetBuffer(CChannel *This,
								   PQP_BUFFER_DESCRIPTOR *ppBufDesc,
								   PTQP_UINT8 *ppBuffer,
								   PTQP_UINT32 pSize
								   )
{
	return (QP_FALSE);
}



static TQP_BOOL CChannel_GetBufferYUV(CChannel *This,
									  PQP_BUFFER_DESCRIPTOR *ppBufDesc,
									  PTQP_UINT8 *ppYBuffer,
									  PTQP_UINT32 pYSize,
									  PTQP_UINT8 *ppUVBuffer,
									  PTQP_UINT32 pUVSize
									  )
{
	return (QP_FALSE);
}



static TQP_BOOL CChannel_GetBufferYUVRAS(CChannel *This,
										 PQP_BUFFER_DESCRIPTOR *ppBufDesc,
										 PTQP_UINT8 *ppYBuffer,
										 PTQP_UINT32 pYSize,
										 PTQP_UINT8 *ppUBuffer,
										 PTQP_UINT32 pUSize,
										 PTQP_UINT8 *ppVBuffer,
										 PTQP_UINT32 pVSize
										 )
{
	return (QP_FALSE);
}



static void CChannel_CompleteBuffer(CChannel *This,
									PQP_BUFFER_DESCRIPTOR pBufDesc
									)
{
}


// call back helper to the interface driver
//
EQP_Errors CChannel_DeviceCallback(CChannel *This,
 								   TQP_UINT32 dwCode,
 				   				   TQP_PVOID pParam
 								   )
{
	if (This->m_bOpened &&
		This->m_pDeviceCallback &&
		This->m_callbackContext
		)
	{
#ifdef _QPCODEC_CALLBACK_IN_THREAD

		CALLBACK_ENTRY	Entry;

		// save to pending fifo
		//
		Entry.pDeviceCallback = This->m_pDeviceCallback;
		Entry.callbackContext = This->m_callbackContext;
		Entry.dwCode = dwCode;
		Entry.pParam = pParam;

		if (CFifo_SetFifo(This->m_pPendingCallbackFifo,
						  (TQP_PVOID)&Entry
						  ))
		{
			CQLCodec	*pQlCodec = GET_INHERITED_OBJECT(CQLCodec, m_Object, This->m_Object.m_pParent);

			QPOSMSetEvtgrp(pQlCodec->m_EvtCallbackThread,
						   QLCODEC_EVENT_CALLBACK_THREAD_DATA
						   );
			return (QPERR_SUCCESS);
		}
		else
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CChannel_DeviceCallback() CFifo_SetFifo() failed\n"
						 ));
		}

#else //!_QPCODEC_CALLBACK_IN_THREAD
		return (This->m_pDeviceCallback(This->m_callbackContext,
										dwCode,
										pParam
										));
#endif //_QPCODEC_CALLBACK_IN_THREAD
	}

	return (QPERR_FAIL);
}



#ifdef _QPCODEC_CALLBACK_IN_THREAD

// call back to the interface driver
//
TQP_BOOL CChannel_GetDeviceCallback(CChannel *This)
{
	CALLBACK_ENTRY	Entry;

	// get free fifo entry
	//
	if (This->m_bOpened &&
		This->m_pPendingCallbackFifo &&
		This->m_pDeviceCallback &&
		This->m_callbackContext
		)
	{
		if (CFifo_GetFifo(This->m_pPendingCallbackFifo,
						  (TQP_PVOID *)&Entry
						  ))
		{
			QP_DEVICE_CALLBACK	pDeviceCallback;	// codec client call back function
			TQP_PVOID			callbackContext;	// codec client call back function context
 			TQP_UINT32			dwCode;
 			TQP_PVOID			pParam;

			pDeviceCallback = Entry.pDeviceCallback;
			callbackContext = Entry.callbackContext;
			dwCode = Entry.dwCode;
			pParam = Entry.pParam;

			pDeviceCallback(callbackContext,
							dwCode,
							pParam
							);
			return (QP_TRUE);
		}
	}

	return (QP_FALSE);
}



TQP_BOOL CChannel_CheckDeviceCallback(CChannel *This)
{
	TQP_BOOL	bFifoInUse = QP_FALSE;

	// get free fifo entry
	//
	if (This->m_bOpened &&
		This->m_pPendingCallbackFifo &&
		This->m_pDeviceCallback &&
		This->m_callbackContext
		)
	{
		bFifoInUse = !CFifo_IsEmpty(This->m_pPendingCallbackFifo);
	}

	return (bFifoInUse);
}

#endif //_QPCODEC_CALLBACK_IN_THREAD



static EQP_Errors CChannel_BeginChannelChange(CChannel *This)
{
	return (QPERR_NOTIMPL);
}



static EQP_Errors CChannel_EndChannelChange(CChannel *This)
{
	return (QPERR_NOTIMPL);
}



static EQP_Errors CChannel_GetResolution(CChannel *This,
										 PTQP_UINT32 pWidth,
										 PTQP_UINT32 pHeight
										 )
{
	return (QPERR_NOTSUPP);
}



static EQP_Errors CChannel_GetYUVFormat(CChannel *This,
										PTQP_UINT32 pYUVFormat
										)
{
	return (QPERR_NOTSUPP);
}



static TQP_BOOL CChannel_RecycleEntry(CChannel *This,
									  QUEUE_ENTRY *pEntry,
									  TQP_BOOL bCancel
									  )
{
	PQP_BUFFER_DESCRIPTOR	pBufDesc;

	if (pEntry)
	{
		if (bCancel)
		{
			if (!This->m_pTask->CancelBuffer(This->m_pTask,
											 This->m_hTask,
											 (TASK_DATA_TYPE)This->m_dataType,
											 (PQP_BUFFER_DESCRIPTOR)pEntry->Data
											 ))
			{
				return (QP_FALSE);
			}
		}

		pBufDesc = (PQP_BUFFER_DESCRIPTOR)pEntry->Data;
		pEntry->Data = QP_NULL;
		pBufDesc->Status = QPERR_SUCCESS/*QPERR_CANCELLED*/;

		// recycle buffer entry
		//
		CQueue_AddEntry(This->m_pFreeQueue,
						pEntry
						);

		// call back the interface driver
		//
		CChannel_DeviceCallback(This,
								QPMPGCODEC_CMD_DONE_DATA,
								pBufDesc
								);
	}

	return (QP_TRUE);
}



static void CChannel_TimeStamp(CChannel *This,
							   PQP_BUFFER_DESCRIPTOR pBufDesc,
							   PQP_KSSTREAM_HEADER pDataPacket
							   )
{
	if (pBufDesc &&
		pDataPacket &&
		(pBufDesc->ulFlags & QP_BUFDESC_FLAG_PTS) &&
		!(pBufDesc->ulFlags & QP_BUFDESC_FLAG_USE_USER_DEFINED_PTS)
		)
	{
#ifdef QP_LINXKER /* Use absolute PTS */
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN /*QPDBGM_LEVEL_TRACE*/),
					 "CChannel_TimeStamp() PTS(0x%lx)\n",
					 pBufDesc->ulPTS
					 ));
#else /* Use related PTS */
        TQP_INT64	llPts;

		// get PTS for this frame
		//
		llPts = pBufDesc->ulPTS;

		// set base PTS if we are at the begining of the stream
		//
		if ((0 == This->m_llBasePts) &&
			(0 == This->m_llAddPts)
			)
		{
			This->m_llBasePts = llPts;
		}

		// check for PTS wrap around
		//
		if ((0 != This->m_llPrevPts) &&
			(This->m_llPrevPts > llPts)
			)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CChannel_TimeStamp() PTS wrap around(0x%x : 0x%x) (0x%x : 0x%x)\n",
						 (TQP_UINT32)((This->m_llPrevPts >> 32) & 0xFFFFFFFF),
						 (TQP_UINT32)(This->m_llPrevPts & 0xFFFFFFFF),
						 (TQP_UINT32)((llPts >> 32) & 0xFFFFFFFF),
						 (TQP_UINT32)(llPts & 0xFFFFFFFF)
						 ));
			This->m_llAddPts += 0x1FFFFFFFF;
		}

		This->m_llPrevPts = llPts;
		llPts += This->m_llAddPts;
		pBufDesc->ulPTS = llPts - This->m_llBasePts;

		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
					 "CChannel_TimeStamp() PTS(%lu:%lu)",
					 (TQP_UINT32)((llPts >> 32) & 0xFFFFFFFF),
					 (TQP_UINT32)(llPts & 0xFFFFFFFF)
					 ));
#endif
	}

}



// helper
//
QPSTATE CChannel_GetState(CChannel *This)
{
	return (This->m_State);
}



TQP_BOOL CChannel_isOpened(CChannel *This)
{
	return (This->m_bOpened);
}



TQP_BOOL CChannel_needByteSwapping(CChannel *This)
{
	return (This->m_bByteSwap);
}



CChannel * CChannel_Constructor(CChannel *pChannel,
								CObject *pParent,
								TQP_UINT32 dwAttributes,
								TQP_OBJECT_HANDLE hTask,
								TQP_INT32 dataType,
								TQP_INT32 FWDataType,
								CHANNEL_DIRECTION ChannelDirection,
								QPMPGCODEC_OPEN_TYPE ChannelType,
								CTask *pTask
								)
{
	if (pChannel)
	{
		CQueue	*pQueue;

		// CObject
		//
		CObject_Constructor(&pChannel->m_Object,
							pParent,
		  					dwAttributes
		  					);
		// fill in the function table
		//
		pChannel->Open = CChannel_Open;
		pChannel->Close = CChannel_Close;
		pChannel->Start = CChannel_Start;
		pChannel->Stop = CChannel_Stop;
		pChannel->Acquire = CChannel_Acquire;
		pChannel->Pause = CChannel_Pause;
		pChannel->Step = CChannel_Step;
		pChannel->SetRate = CChannel_SetRate;
		pChannel->GetRate = CChannel_GetRate;
		pChannel->BeginFlush = CChannel_BeginFlush;
		pChannel->Flush = CChannel_Flush;
		pChannel->EndFlush = CChannel_EndFlush;
		pChannel->AddBuffer = CChannel_AddBuffer;
		pChannel->CancelBuffer = CChannel_CancelBuffer;
		pChannel->TimeoutBuffer = CChannel_TimeoutBuffer;
		pChannel->XferData = CChannel_XferData;
		pChannel->GetBuffer = CChannel_GetBuffer;
		pChannel->GetBufferYUV = CChannel_GetBufferYUV;
		pChannel->GetBufferYUVRAS = CChannel_GetBufferYUVRAS;
		pChannel->CompleteBuffer = CChannel_CompleteBuffer;
		pChannel->BeginChannelChange = CChannel_BeginChannelChange;
		pChannel->EndChannelChange = CChannel_EndChannelChange;
		pChannel->GetResolution = CChannel_GetResolution;
		pChannel->GetYUVFormat = CChannel_GetYUVFormat;

		// initialize members
		//
		pChannel->m_dwOpenFlags = 0;

		pChannel->m_pDeviceCallback = QP_NULL;
		pChannel->m_callbackContext = QP_NULL;

		pChannel->m_hTask = hTask;
		pChannel->m_hChannel = TQP_NULL_HANDLE;
		pChannel->m_dataType = dataType;
		pChannel->m_FWDataType = FWDataType;
		pChannel->m_ChannelDirection = ChannelDirection;
		pChannel->m_ChannelType = ChannelType;
		pChannel->m_pTask = pTask;

		pChannel->m_bOpened = QP_FALSE;
		pChannel->m_State = QPSTATE_STOP;	// channel state
		pChannel->m_bPaused = QP_FALSE;
		pChannel->m_bFlushing = QP_FALSE;

		pChannel->m_bByteSwap = QP_FALSE;

		pChannel->m_ullCntBytes = 0;

		// buffer management stuffs
		//
		pChannel->m_pFreeQueue = QP_NULL;
		pChannel->m_pDataRequestQueue = QP_NULL;
		pChannel->m_pDataPendingQueue = QP_NULL;
#ifdef _QPCODEC_CALLBACK_IN_THREAD
		pChannel->m_pPendingCallbackFifo = QP_NULL;
#endif //_QPCODEC_CALLBACK_IN_THREAD

		pChannel->m_llStartTime = 0;

		// create queues
		//
		if (TASK_DATA_TYPE_VIRTUAL != dataType)
		{
			// free queue
			//
			pQueue = (CQueue *)QPMMMalloc2Ex(sizeof(CQueue));
			if (!pQueue)
			{
				CChannel_Destructor(pChannel);
				return (QP_NULL);
			}

			pChannel->m_pFreeQueue = CQueue_Constructor(pQueue,
														&pChannel->m_Object,
														OBJECT_CRITICAL_LIGHT
														);
			if (!pChannel->m_pFreeQueue)
			{
				QPMMFree2Ex((void *)pQueue);
				CChannel_Destructor(pChannel);
				return (QP_NULL);
			}

			// request queue
			//
			pQueue = (CQueue *)QPMMMalloc2Ex(sizeof(CQueue));
			if (!pQueue)
			{
				CChannel_Destructor(pChannel);
				return (QP_NULL);
			}

			pChannel->m_pDataRequestQueue = CQueue_Constructor(pQueue,
															   &pChannel->m_Object,
															   OBJECT_CRITICAL_LIGHT
															   );
			if (!pChannel->m_pDataRequestQueue)
			{
				QPMMFree2Ex((void *)pQueue);
				CChannel_Destructor(pChannel);
				return (QP_NULL);
			}

			// pending queue
			//
			pQueue = (CQueue *)QPMMMalloc2Ex(sizeof(CQueue));
			if (!pQueue)
			{
				CChannel_Destructor(pChannel);
				return (QP_NULL);
			}

			pChannel->m_pDataPendingQueue = CQueue_Constructor(pQueue,
															   &pChannel->m_Object,
															   OBJECT_CRITICAL_LIGHT
															   );
			if (!pChannel->m_pDataPendingQueue)
			{
				QPMMFree2Ex((void *)pQueue);
				CChannel_Destructor(pChannel);
				return (QP_NULL);
			}

#ifdef _QPCODEC_CALLBACK_IN_THREAD
			{
				// create fifo
				//
				CFifo	*pFifo;

				// pending fifo
				//
				pFifo = (CFifo *)QPMMMalloc2Ex(sizeof(CFifo));
				if (!pFifo)
				{
					CChannel_Destructor(pChannel);
					return (QP_NULL);
				}

				pChannel->m_pPendingCallbackFifo = CFifo_Constructor(pFifo,
																	 &pChannel->m_Object,
																	 OBJECT_CRITICAL_LIGHT,
																	 ////OBJECT_DEFAULT,
																	 QPCODECLIB_MAX_DATA_ENTRIES + 1,
																	 sizeof(CALLBACK_ENTRY)
																	 );
				if (!pChannel->m_pPendingCallbackFifo)
				{
					QPMMFree2Ex((void *)pFifo);
					CChannel_Destructor(pChannel);
					return (QP_NULL);
				}
			}
#endif //_QPCODEC_CALLBACK_IN_THREAD
		}
	}

	return (pChannel);
}



void CChannel_Destructor(CChannel *This)
{
	if (This->m_bOpened)
		This->Close(This);

	if (This->m_pDataPendingQueue)
	{
		CQueue_Destructor(This->m_pDataPendingQueue);
		QPMMFree2Ex((void *)This->m_pDataPendingQueue);
		This->m_pDataPendingQueue = QP_NULL;
	}

	if (This->m_pDataRequestQueue)
	{
		CQueue_Destructor(This->m_pDataRequestQueue);
		QPMMFree2Ex((void *)This->m_pDataRequestQueue);
		This->m_pDataRequestQueue = QP_NULL;
	}

	if (This->m_pFreeQueue)
	{
		CQueue_Destructor(This->m_pFreeQueue);
		QPMMFree2Ex((void *)This->m_pFreeQueue);
		This->m_pFreeQueue = QP_NULL;
	}

#ifdef _QPCODEC_CALLBACK_IN_THREAD

	if (This->m_pPendingCallbackFifo)
	{
		CFifo_Destructor(This->m_pPendingCallbackFifo);
		QPMMFree2Ex((void *)This->m_pPendingCallbackFifo);
		This->m_pPendingCallbackFifo = QP_NULL;
	}

#endif //_QPCODEC_CALLBACK_IN_THREAD

	CObject_Destructor(&This->m_Object);
}




/////////////////////////////////////////////////////////////////////////////
//
//

// CNullChannel
//

// CChannel
//
static EQP_Errors CNullChannel_Open(CChannel *This,
									TQP_OBJECT_HANDLE hChannel,
									TQP_UINT32 dwFlags,
									TQP_PVOID pDataFormat,
									QP_DEVICE_CALLBACK pFuncCallback,
									TQP_PVOID context
									)
{
	EQP_Errors	qpStatus;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CNullChannel_Open()\n"
				 ));

	// save handle
	//
	This->m_hChannel = hChannel;

	// save the open flags
	//
	This->m_dwOpenFlags = dwFlags;

	// save callback address and context
	//
	This->m_pDeviceCallback = pFuncCallback;
	This->m_callbackContext = context;

	// reset channel state
	//
	This->m_State = QPSTATE_STOP;
	This->m_bPaused = QP_FALSE;

	// clear flush flag
	//
	This->m_bFlushing = QP_FALSE;

	// reset counters and flags
	//
	This->m_ullCntBytes = 0;

	// yes, we are open
	//
	This->m_bOpened = QP_TRUE;

	// Task open
	//
	qpStatus = This->m_pTask->Open(This->m_pTask,
								   This->m_hTask,
								   (TASK_DATA_TYPE)This->m_dataType,
								   This->m_ChannelDirection,
								   (FIRMWARE_DATA_TYPE)This->m_FWDataType,
								   This
								   );

	if (!QP_SUCCESS(qpStatus))
		This->Close(This);

	return (qpStatus);
}


static EQP_Errors CNullChannel_Close(CChannel *This)
{
	if (This->m_bOpened)
	{
		if (QPSTATE_STOP != This->m_State)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CNullChannel_Close() !!! Not Stopped State (%d)!!!!\n",
						 This->m_State
						 ));

			This->Stop(This);
		}

		// task close
		//
		This->m_pTask->Close(This->m_pTask,
							 This->m_hTask,
							 (TASK_DATA_TYPE)This->m_dataType,
							 QP_TRUE
							 );

		This->m_bOpened = QP_FALSE;
	}

	return (QPERR_SUCCESS);
}


static EQP_Errors CNullChannel_Start(CChannel *This_p)
{
	EQP_Errors	qpStatus;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CNullChannel_Start() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 This_p->m_dataType
				 ));

	if (This_p->m_bPaused)
	{
		qpStatus = This_p->m_pTask->Resume(This_p->m_pTask,
										   This_p->m_hTask,
										   (TASK_DATA_TYPE)This_p->m_dataType
										   );

		if (QP_SUCCESS(qpStatus))
			This_p->m_bPaused = QP_FALSE;
	}
	else
	{
		// start the task
		//
		qpStatus = This_p->m_pTask->Start(This_p->m_pTask,
										  This_p->m_hTask,
										  (TASK_DATA_TYPE)This_p->m_dataType
										  );
	}

	if (QP_SUCCESS(qpStatus))
	{
		This_p->m_State = QPSTATE_RUN;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CNullChannel_Start() Failed Status(%d)\n",
					 qpStatus
					 ));
	}

	return (qpStatus);
}


static EQP_Errors CNullChannel_Stop(CChannel *This_p)
{
	EQP_Errors	qpStatus;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CNullChannel_Stop() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 (TASK_DATA_TYPE)This_p->m_dataType
				 ));

	qpStatus = This_p->m_pTask->Stop(This_p->m_pTask,
									 This_p->m_hTask,
									 (TASK_DATA_TYPE)This_p->m_dataType
									 );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CNullChannel_Stop() m_pTask->Stop failed(%d)\n",
					 qpStatus
					 ));
	}

	// clear falgs
	//
	This_p->m_bPaused = QP_FALSE;

	// change state to stop
	//
	This_p->m_State = QPSTATE_STOP;

	return (QPERR_SUCCESS);
}


static EQP_Errors CNullChannel_Pause(CChannel *This_p)
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CNullChannel_Pause() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 (TASK_DATA_TYPE)This_p->m_dataType
				 ));

	if (QPSTATE_RUN == This_p->m_State)
	{
		qpStatus = This_p->m_pTask->Pause(This_p->m_pTask,
										  This_p->m_hTask,
										  (TASK_DATA_TYPE)This_p->m_dataType
										  );
		This_p->m_bPaused = QP_SUCCESS(qpStatus);
	}

	if (QP_SUCCESS(qpStatus))
	{
		This_p->m_State = QPSTATE_PAUSE;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CNullChannel_Pause() m_pTask->Pause failed(%d)\n",
					 qpStatus
					 ));
	}

	return (qpStatus);
}



static EQP_Errors CNullChannel_BeginFlush(CChannel *This_p)
{
	return (QPERR_SUCCESS);
}


static EQP_Errors CNullChannel_Flush(CChannel *This_p)
{
	return (QPERR_SUCCESS);
}


static EQP_Errors CNullChannel_EndFlush(CChannel *This_p)
{
	return (QPERR_SUCCESS);
}


static EQP_Errors CNullChannel_AddBuffer(CChannel *This_p,
										 PQP_BUFFER_DESCRIPTOR pBufDesc
										 )
{
	return (QPERR_SUCCESS);
}


static EQP_Errors CNullChannel_CancelBuffer(CChannel *This_p,
											PQP_BUFFER_DESCRIPTOR pBufDesc,
											TQP_BOOL bTimeOut
											)
{
	return (QPERR_SUCCESS);
}


static EQP_Errors CNullChannel_TimeoutBuffer(CChannel *This_p,
											 PQP_BUFFER_DESCRIPTOR pBufDesc
											 )
{
	return (QPERR_SUCCESS);
}


static EQP_Errors CNullChannel_XferData(CChannel *This_p,
										PQP_BUFFER_DESCRIPTOR pBufDesc
										)
{
	return (QPERR_SUCCESS);
}


// constructor
//
CNullChannel * CNullChannel_Constructor(CNullChannel *pNullChannel,
										CObject *pParent,
										TQP_OBJECT_HANDLE hTask,
										CTask *pTask
										)
{
	if (pNullChannel)
	{
		CChannel	*pChannel;

		pChannel = CChannel_Constructor(&pNullChannel->m_Channel,
										pParent,
										OBJECT_DEFAULT,
										hTask,
										TASK_DATA_TYPE_VIRTUAL,
										0,
										CHANNEL_DIR_NONE,
										QPMPGCODEC_CODEC_VIRTUAL,
										pTask
										);
		if (!pChannel)
		{
			pNullChannel = QP_NULL;
		}
		else
		{

			// fill in the functions
			//
			pNullChannel->m_Channel.Open = CNullChannel_Open;
			pNullChannel->m_Channel.Close = CNullChannel_Close;
			pNullChannel->m_Channel.Start = CNullChannel_Start;
			pNullChannel->m_Channel.Stop = CNullChannel_Stop;
			pNullChannel->m_Channel.Pause = CNullChannel_Pause;
			pNullChannel->m_Channel.BeginFlush = CNullChannel_BeginFlush;
			pNullChannel->m_Channel.Flush = CNullChannel_Flush;
			pNullChannel->m_Channel.EndFlush = CNullChannel_EndFlush;
			pNullChannel->m_Channel.AddBuffer = CNullChannel_AddBuffer;
			pNullChannel->m_Channel.CancelBuffer = CNullChannel_CancelBuffer;
			pNullChannel->m_Channel.TimeoutBuffer = CNullChannel_TimeoutBuffer;
			pNullChannel->m_Channel.XferData = CNullChannel_XferData;
		}
	}

	return (pNullChannel);
}



// destructor
//
void CNullChannel_Destructor(CNullChannel *This)
{
	CChannel_Destructor(&This->m_Channel);
}


/////////////////////////////////////////////////////////////////////////////
//
//

// CMPEGOutChannel
//


// CChannel
//
static EQP_Errors CMPEGOutChannel_Start(CChannel *This_p)
{
	EQP_Errors	qpStatus;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMPEGOutChannel_Start() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 This_p->m_dataType
				 ));

	if (This_p->m_bPaused)
	{
		qpStatus = This_p->m_pTask->Resume(This_p->m_pTask,
										   This_p->m_hTask,
										   (TASK_DATA_TYPE)This_p->m_dataType
										   );

		if (QP_SUCCESS(qpStatus))
			This_p->m_bPaused = QP_FALSE;
	}
	else
	{
		// reset counter and flags
		//
		This_p->m_ullCntBytes = 0;

		This_p->m_llPrevPts = 0;
		This_p->m_llAddPts = 0;
		This_p->m_llBasePts = 0;

		// start the encoder
		//
		qpStatus = This_p->m_pTask->Start(This_p->m_pTask,
										  This_p->m_hTask,
										  (TASK_DATA_TYPE)This_p->m_dataType
										  );
		// remember the start time
		//
		This_p->m_llStartTime = QPTMGetSysTimeMilliS();
	}

	This_p->m_bFlushing = QP_FALSE;

	if (QP_SUCCESS(qpStatus))
	{
		This_p->m_State = QPSTATE_RUN;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CMPEGOutChannel_Start() Failed Status(%d)\n",
					 qpStatus
					 ));
	}

	return (qpStatus);
}



static EQP_Errors CMPEGOutChannel_Stop(CChannel *This_p)
{
	EQP_Errors	qpStatus;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMPEGOutChannel_Stop() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 (TASK_DATA_TYPE)This_p->m_dataType
				 ));

	qpStatus = This_p->m_pTask->Stop(This_p->m_pTask,
									 This_p->m_hTask,
									 (TASK_DATA_TYPE)This_p->m_dataType
									 );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CMPEGOutChannel_Stop() m_pTask->Stop failed(%d)\n",
					 qpStatus
					 ));
	}

	// clear falgs
	//
	This_p->m_bPaused = QP_FALSE;

	// change state to stop
	//
	This_p->m_State = QPSTATE_STOP;

	// flush the data
	//
	This_p->BeginFlush(This_p);
	This_p->Flush(This_p);
	This_p->EndFlush(This_p);

	return (QPERR_SUCCESS);
}



static EQP_Errors CMPEGOutChannel_Pause(CChannel *This_p)
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMPEGOutChannel_Pause() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 This_p->m_dataType
				 ));

	if (QPSTATE_RUN == This_p->m_State)
	{
		qpStatus = This_p->m_pTask->Pause(This_p->m_pTask,
										  This_p->m_hTask,
										  (TASK_DATA_TYPE)This_p->m_dataType
										  );
		This_p->m_bPaused = QP_SUCCESS(qpStatus);
	}

	if (QP_SUCCESS(qpStatus))
	{
		This_p->m_State = QPSTATE_PAUSE;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CMPEGOutChannel_Pause() m_pTask->Pause failed(%d)\n",
					 qpStatus
					 ));
	}

	return (qpStatus);
}



// called from streaming task
//
static TQP_BOOL CMPEGOutChannel_GetBuffer(CChannel *This_p,
										  PQP_BUFFER_DESCRIPTOR *ppBufDesc,
										  PTQP_UINT8 *ppBuffer,
										  PTQP_UINT32 pSize
										  )
{

	QUEUE_ENTRY				*pEntry;
	PQP_BUFFER_DESCRIPTOR	pBufDesc;
	PQP_KSSTREAM_HEADER		pDataPacket;

	pEntry = CQueue_GetOneEntry(This_p->m_pDataRequestQueue);

	if (QP_NULL != pEntry)
	{
		pBufDesc = (PQP_BUFFER_DESCRIPTOR)pEntry->Data;
		pDataPacket = pBufDesc->DataBufferArray;

		pBufDesc->ulFlags &= ~QP_BUFDESC_FLAG_MASK_PRIVATE;
		pBufDesc->ulBufferIndex = 0;
		pBufDesc->ulBufferOffset = 0;
		pBufDesc->ulTotalUsed = 0;

		pDataPacket->DataUsed = 0;

		*ppBuffer = (PTQP_UINT8)pDataPacket->Data;
		*pSize = pBufDesc->ulBufferSize;
		*ppBufDesc = pBufDesc;

		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, /*QPDBGM_LEVEL_WARN*/QPDBGM_LEVEL_TRACE),
					 "GetBuffer(%d) buffer(0x%x) size(%d) Desc(0x%x)\n",
					 This_p->m_hChannel,
					 *ppBuffer,
					 *pSize,
					 *ppBufDesc
					 ));

		// move this entry to the pending queue
		//
		CQueue_AddEntry(This_p->m_pDataPendingQueue,
						pEntry
						);
	}

	return (QP_NULL != pEntry);
}



static void CMPEGOutChannel_CompleteBuffer(CChannel *This_p,
										   PQP_BUFFER_DESCRIPTOR pBufDesc
										   )
{
	QUEUE_ENTRY			*pEntry;
	PQP_KSSTREAM_HEADER	pDataPacket;
	CQLCodec			*pQlCodec = GET_INHERITED_OBJECT(CQLCodec, m_Object, This_p->m_Object.m_pParent);


	////QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
	////			 "CMPEGOutChannel_CompleteBuffer()\n"
	////			 ));

	// get the buffer entry
	//
	pEntry = CQueue_GetEntryByData(This_p->m_pDataPendingQueue,
								   (TQP_PVOID)pBufDesc
								   );
	if (pEntry)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, /*QPDBGM_LEVEL_WARN*/QPDBGM_LEVEL_TRACE),
					 "CMPEGOutChannel_CompleteBuffer() bufDesc(0x%x) Status(%d), totalUsed(%d)\n",
					 pBufDesc,
					 pBufDesc->Status,
					 pBufDesc->ulTotalUsed
					 ));

		pDataPacket = pBufDesc->DataBufferArray;
		pDataPacket->OptionsFlags = 0;

		if (This_p->m_bByteSwap &&
			!QPHCI_CanSwapData(&pQlCodec->m_hci) &&
			(0 != pBufDesc->ulTotalUsed)
			)
		{
			BytesSwap((PTQP_UINT8)pDataPacket->Data,
					  (PTQP_UINT8)pDataPacket->Data,
					  pBufDesc->ulTotalUsed
					  );
		}

		// set these flags if we are at the begining of the stream
		//
		if (0 == This_p->m_ullCntBytes)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
						 "CMPEGOutChannel_CompleteBuffer() (0 == m_ullCntBytes) QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY\n"
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY |
										 QP_KSSTREAM_HEADER_OPTIONSF_TIMEDISCONTINUITY
										 ;
		}

		// set time stamp
		//
		CChannel_TimeStamp(This_p,
						   pBufDesc,
						   pDataPacket
						   );

		This_p->m_ullCntBytes += pBufDesc->ulTotalUsed;
		pDataPacket->DataUsed = pBufDesc->ulTotalUsed;

		if (pBufDesc->ulFlags & QP_BUFDESC_FLAG_EOS)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CMPEGOutChannel_CompleteBuffer() used(%d) total(%d) QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM\n",
						 pBufDesc->ulTotalUsed,
						 This_p->m_ullCntBytes
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM;
		}

		if (pBufDesc->ulFlags & QP_BUFDESC_FLAG_INCOMPLETE_FRAME)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CMPEGOutChannel_CompleteBuffer(%d) QP_KSSTREAM_HEADER_OPTIONSF_PREROLL\n",
						 pBufDesc->ulTotalUsed
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_PREROLL;
		}

#if 1
		if (pBufDesc->ulFlags & QP_BUFDESC_FLAG_RAW_AUDIO_MASK)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "%s() compressed audio type (0x%x)\n",
						 __FUNCTION__,
						 pBufDesc->ulFlags & QP_BUFDESC_FLAG_RAW_AUDIO_MASK));

		    switch (pBufDesc->ulFlags & QP_BUFDESC_FLAG_RAW_AUDIO_MASK) {
                case QP_BUFDESC_FLAG_RAW_AUDIO_AC3:
                    pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_RAW_AUDIO_AC3;
				    break;
				case QP_BUFDESC_FLAG_RAW_AUDIO_AAC:
				    pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_RAW_AUDIO_AAC;
				    break;
				case QP_BUFDESC_FLAG_RAW_AUDIO_AAC2:
				    pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_RAW_AUDIO_AAC2;
				    break;
				case QP_BUFDESC_FLAG_RAW_AUDIO_DTS:
				    pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_RAW_AUDIO_DTS;
				    break;
		    }

		}
#endif

		// clear the entry content
		//
		pEntry->Data = QP_NULL;

		// move to free queue
		//
		CQueue_AddEntry(This_p->m_pFreeQueue,
						pEntry
						);

		// call back to the api
		//
		CChannel_DeviceCallback(This_p,
								QPMPGCODEC_CMD_DONE_DATA,
								pBufDesc
								);
	}
}



// constructor
//
CMPEGOutChannel * CMPEGOutChannel_Constructor(CMPEGOutChannel *pMpgOutChannel,
											  CObject *pParent,
											  TQP_UINT32 dwAttributes,
											  TQP_OBJECT_HANDLE hTask,
											  CTask *pTask
											  )
{
	if (pMpgOutChannel)
	{
		CChannel	*pChannel;

		pChannel = CChannel_Constructor(&pMpgOutChannel->m_Channel,
										pParent,
										dwAttributes,
										hTask,
										TASK_DATA_TYPE_COMP_VID,
										FWDataTypeCmprSysVidOutput,
										CHANNEL_DIR_READ,
										QPMPGCODEC_ENC_MPEG_OUT,
										pTask
										);
		if (!pChannel)
		{
			pMpgOutChannel = QP_NULL;
		}
		else
		{
			if (BYTES_SWAP_BY_DRIVER == pTask->m_TaskData[hTask].m_inputControl.RegArtesa.bytes_swap)
			{
				// BYTES_SWAP_BY_DRIVER
				pMpgOutChannel->m_Channel.m_bByteSwap = QP_TRUE;
			}
			else
			{
				// BYTES_SWAP_BY_FW
				pMpgOutChannel->m_Channel.m_bByteSwap = QP_FALSE;
			}

			// fill in the functions table
			//
			pMpgOutChannel->m_Channel.Start = CMPEGOutChannel_Start;
			pMpgOutChannel->m_Channel.Stop = CMPEGOutChannel_Stop;
			pMpgOutChannel->m_Channel.Pause = CMPEGOutChannel_Pause;
			pMpgOutChannel->m_Channel.GetBuffer = CMPEGOutChannel_GetBuffer;
			pMpgOutChannel->m_Channel.CompleteBuffer = CMPEGOutChannel_CompleteBuffer;
		}
	}

	return (pMpgOutChannel);
}



// destructor
//
void CMPEGOutChannel_Destructor(CMPEGOutChannel *This)
{
	CChannel_Destructor(&This->m_Channel);
}



/////////////////////////////////////////////////////////////////////////////
//
//

// CAESOutChannel
//


// CChannel
//


// called from streaming task
//
static void CAESOutChannel_CompleteBuffer(CChannel *This_p,
										  PQP_BUFFER_DESCRIPTOR pBufDesc
										  )
{
	QUEUE_ENTRY			*pEntry;
	PQP_KSSTREAM_HEADER	pDataPacket;
	CQLCodec			*pQlCodec = GET_INHERITED_OBJECT(CQLCodec, m_Object, This_p->m_Object.m_pParent);


	////QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
	////			 "CAESOutChannel_CompleteBuffer()\n"
	////			 ));

	// get the buffer entry
	//
	pEntry = CQueue_GetEntryByData(This_p->m_pDataPendingQueue,
								   (TQP_PVOID)pBufDesc
								   );
	if (pEntry)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, /*QPDBGM_LEVEL_WARN*/QPDBGM_LEVEL_TRACE),
					 "CAESOutChannel_CompleteBuffer() bufDesc(0x%x) Status(%d), totalUsed(%d)\n",
					 pBufDesc,
					 pBufDesc->Status,
					 pBufDesc->ulTotalUsed
					 ));

		pDataPacket = pBufDesc->DataBufferArray;
		pDataPacket->OptionsFlags = 0;

		if (This_p->m_bByteSwap &&
			!QPHCI_CanSwapData(&pQlCodec->m_hci) &&
			(0 != pBufDesc->ulTotalUsed)
			)
		{
			BytesSwap((PTQP_UINT8)pDataPacket->Data,
					  (PTQP_UINT8)pDataPacket->Data,
					  pBufDesc->ulTotalUsed
					  );
		}

		// set these flags if we are at the begining of the stream
		//
		if (0 == This_p->m_ullCntBytes)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
						 "CAESOutChannel_CompleteBuffer() (0 == m_ullCntBytes) QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY\n"
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY |
										 QP_KSSTREAM_HEADER_OPTIONSF_TIMEDISCONTINUITY
										 ;
		}

		// set time stamp
		//
		CChannel_TimeStamp(This_p,
						   pBufDesc,
						   pDataPacket
						   );


#ifndef QP_LINXKER
		/*
		** to remove silk padding bytes
		** 1. the leading frame has extra 9 bytes file header "#!SILK_V3"
		** 2. the actual length is record in leading two bytes
		** 3. algorighm: padding bytes = buffer_length - actual_length
		*/
		if (AUDIO_TYPE_SILK == This_p->m_pTask->m_TaskData[This_p->m_hTask].m_audioControlParam.Reg.audio_type)
		{
			int nHeaderLen = 0;
			int nPaddingBytes = 0;
			int nFrameLen = 0;

			do {
				if (pDataPacket->OptionsFlags & QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY) {
					nHeaderLen = 9;
				}

				if (pBufDesc->ulTotalUsed < (unsigned long)(nHeaderLen + 2)) {
					QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID,  QPDBGM_LEVEL_ERROR),
						"%s() frame length (%d) less than SILK header length (%d+2), skip padding byte checking",
						__FUNCTION__,
						pBufDesc->ulTotalUsed,
						nHeaderLen));
					break;
				}

				nFrameLen = *(unsigned short *)((unsigned char *)pDataPacket->Data + nHeaderLen);

				nPaddingBytes = pBufDesc->ulTotalUsed - nHeaderLen - nFrameLen - 2;

				if ((nPaddingBytes < 0) || (nPaddingBytes > 3)) {
				    #ifdef QP_LINXKER
				    /* in linux, it is possible there have two or more frames sent at the same time */
					QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID,  QPDBGM_LEVEL_EXTRACE),
						"%s() incorrect padding bytes(%d), ulTotalUsed(%d) nHeaderLen(%d) pusFrameLen(%d)",
						__FUNCTION__,
						nPaddingBytes,
						pBufDesc->ulTotalUsed,
						nHeaderLen,
						nFrameLen));
				    #else
					QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID,  QPDBGM_LEVEL_ERROR),
						"%s() incorrect padding bytes(%d), ulTotalUsed(%d) nHeaderLen(%d) pusFrameLen(%d)",
						__FUNCTION__,
						nPaddingBytes,
						pBufDesc->ulTotalUsed,
						nHeaderLen,
						nFrameLen));
					#endif
					break;
				}

				if (pBufDesc->ulTotalUsed < (unsigned long)(nHeaderLen + 2 + nPaddingBytes)) {
					QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID,  QPDBGM_LEVEL_ERROR),
						"%s() frame length (%d) less than SILK header length (%d+2) + padding length (%d), skip padding byte removing",
						__FUNCTION__,
						pBufDesc->ulTotalUsed,
						nHeaderLen,
						nPaddingBytes));
					break;
				}

				pBufDesc->ulTotalUsed -= nPaddingBytes;

				break;
			} while(0);
		}
#endif

		This_p->m_ullCntBytes += pBufDesc->ulTotalUsed;
		pDataPacket->DataUsed = pBufDesc->ulTotalUsed;

		if (pBufDesc->ulFlags & QP_BUFDESC_FLAG_EOS)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CAESOutChannel_CompleteBuffer() used(%d) total(%d) QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM\n",
						 pBufDesc->ulTotalUsed,
						 This_p->m_ullCntBytes
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM;
		}

		if (pBufDesc->ulFlags & QP_BUFDESC_FLAG_INCOMPLETE_FRAME)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CAESOutChannel_CompleteBuffer(%d) QP_KSSTREAM_HEADER_OPTIONSF_PREROLL\n",
						 pBufDesc->ulTotalUsed
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_PREROLL;
		}

		// clear the entry content
		//
		pEntry->Data = QP_NULL;

		// move to free queue
		//
		CQueue_AddEntry(This_p->m_pFreeQueue,
						pEntry
						);

		// call back the api
		//
		CChannel_DeviceCallback(This_p,
								QPMPGCODEC_CMD_DONE_DATA,
								pBufDesc
								);
	}
}



// constructor
//
CAESOutChannel * CAESOutChannel_Constructor(CAESOutChannel *pAesOutChannel,
											CObject *pParent,
											TQP_UINT32 dwAttributes,
											TQP_OBJECT_HANDLE hTask,
											CTask *pTask
											)
{
	if (pAesOutChannel)
	{
		CChannel	*pChannel;

		pChannel = CChannel_Constructor(&pAesOutChannel->m_Channel,
										pParent,
										dwAttributes,
										hTask,
										TASK_DATA_TYPE_COMP_AUD,
										FWDataTypeCmprAudOutput,
										CHANNEL_DIR_READ,
										QPMPGCODEC_ENC_AES_OUT,
										pTask
										);
		if (!pChannel)
		{
			pAesOutChannel = QP_NULL;
		}
		else
		{
			pAesOutChannel->m_Channel.m_bByteSwap = QP_TRUE;

			// fill in the functions table
			//
			pAesOutChannel->m_Channel.Start = CMPEGOutChannel_Start;
			pAesOutChannel->m_Channel.Stop = CMPEGOutChannel_Stop;		// same as MpegOut
			pAesOutChannel->m_Channel.Pause = CMPEGOutChannel_Pause;	// same as MpegOut
			pAesOutChannel->m_Channel.GetBuffer = CMPEGOutChannel_GetBuffer;	// same as MpegOut
			pAesOutChannel->m_Channel.CompleteBuffer = CAESOutChannel_CompleteBuffer;
		}
	}

	return (pAesOutChannel);
}



// destructor
//
void CAESOutChannel_Destructor(CAESOutChannel *This)
{
	CChannel_Destructor(&This->m_Channel);
}




/////////////////////////////////////////////////////////////////////////////
//
//

// CYUVOutChannel
//


// CChannel
//

static EQP_Errors CYUVOutChannel_Open(CChannel *This_p,
									  TQP_OBJECT_HANDLE hChannel,
									  TQP_UINT32 dwFlags,
									  TQP_PVOID pDataFormat,
									  QP_DEVICE_CALLBACK pFuncCallback,
									  TQP_PVOID context
									  )
{
	EQP_Errors			qpStatus;
	PQP_YUV_DATAFORMAT	pYUVFormat = (PQP_YUV_DATAFORMAT)pDataFormat;
	CYUVOutChannel		*This =	GET_INHERITED_OBJECT(CYUVOutChannel, m_Channel, This_p);


	// Gather information we need for this stream
	//
	switch (pYUVFormat->nFrameRate)
	{
		case 30:
			This->m_ulAvgPTSPerFrame = PTS_NTSC_INCREMENT;
			This->m_ulAvgPTSPerFrame1000X = PTS_NTSC_INCREMENT_1000X;
			break;

		case 25:
			This->m_ulAvgPTSPerFrame = PTS_PAL_INCREMENT;
			This->m_ulAvgPTSPerFrame1000X = PTS_PAL_INCREMENT_1000X;
			break;

		default:
			This->m_ulAvgPTSPerFrame = (90000 / pYUVFormat->nFrameRate);
			This->m_ulAvgPTSPerFrame1000X = (90000000 / pYUVFormat->nFrameRate);
			break;
	}

#ifdef QP_LINXKER
	{
		unsigned long long ullTemp = (unsigned long long)This->m_ulAvgTimePerFrame;
		ullTemp *= 10000;
		do_div(ullTemp, 90);
		This->m_ulAvgTimePerFrame = (unsigned long)ullTemp;
	}
#else
	This->m_ulAvgTimePerFrame = (TQP_UINT32)PTS_TO_KSPTS(This->m_ulAvgPTSPerFrame);
#endif

	This->m_lWidth = pYUVFormat->nWidth;
	This->m_lHeight = pYUVFormat->nHeight;
	This->m_nBitCount = pYUVFormat->nBitCount;
	This->m_nDataType = pYUVFormat->nDataType;

	This->m_dwFrameSize = (This->m_lWidth * This->m_lHeight * This->m_nBitCount) / 8;

#if 0
	// for FPGA test
	This->m_Channel.m_bByteSwap = (12 == This->m_nBitCount);
#else
	This->m_Channel.m_bByteSwap = ((12 == This->m_nBitCount) &&
								  ((QP_YUV_DATA_TYPE_YV12 == This->m_nDataType) ||
								  (QP_YUV_DATA_TYPE_I420 == This->m_nDataType))
								  );
#endif

	if (QP_YUV_DATA_TYPE_YUYV == This->m_nDataType)
	{
		This->m_Channel.m_FWDataType = FWDataTypeRawYUYVOutput;
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CYUVOutChannel_Open() m_ulAvgTimePerFrame(%d), m_dwFrameSize(%d) m_lWidth(%d) m_lHeight(%d) swap(%d) m_nDataType(%d)\n",
				 This->m_ulAvgTimePerFrame,
				 This->m_dwFrameSize,
				 This->m_lWidth,
				 This->m_lHeight,
				 This->m_Channel.m_bByteSwap,
				 This->m_nDataType
				 ));

	// reset counters and flags
	//
	This->m_ulVideoFramesCount = 0;
	This->m_ulDroppedFramesCount = 0;

	// generic open
	//
	qpStatus = CChannel_Open(This_p,
							 hChannel,
							 dwFlags,
							 pDataFormat,
							 pFuncCallback,
							 context
							 );

	return (qpStatus);
}



static EQP_Errors CYUVOutChannel_Start(CChannel *This_p)
{
	EQP_Errors		qpStatus;
	CYUVOutChannel	*This =	GET_INHERITED_OBJECT(CYUVOutChannel, m_Channel, This_p);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CYUVOutChannel_Start() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 (TASK_DATA_TYPE)This_p->m_dataType
				 ));

	if (This_p->m_bPaused)
	{
		qpStatus = This_p->m_pTask->Resume(This_p->m_pTask,
										   This_p->m_hTask,
										   (TASK_DATA_TYPE)This_p->m_dataType
										   );

		if (QP_SUCCESS(qpStatus))
			This_p->m_bPaused = QP_FALSE;
	}
	else
	{
		// reset counter and flags
		//
		This_p->m_ullCntBytes = 0;

		This->m_ulVideoFramesCount = 0;
		This->m_ulDroppedFramesCount = 0;

		This_p->m_llPrevPts = 0;
		This_p->m_llAddPts = 0;
		This_p->m_llBasePts = 0;

		// start the encoder
		//
		qpStatus = This_p->m_pTask->Start(This_p->m_pTask,
										  This_p->m_hTask,
										  (TASK_DATA_TYPE)This_p->m_dataType
										  );

		This_p->m_llStartTime = QPTMGetSysTimeMilliS();
	}

	This_p->m_bFlushing = QP_FALSE;

	if (QP_SUCCESS(qpStatus))
	{
		This_p->m_State = QPSTATE_RUN;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CYUVOutChannel_Start() Failed Status(%d)\n",
					 qpStatus
					 ));
	}

	return (qpStatus);
}



// called from streaming task
//
static TQP_BOOL CYUVOutChannel_GetBuffer(CChannel *This_p,
										 PQP_BUFFER_DESCRIPTOR *ppBufDesc,
										 PTQP_UINT8 *ppYBuffer,
										 PTQP_UINT32 pYSize,
										 PTQP_UINT8 *ppUVBuffer,
										 PTQP_UINT32 pUVSize
										 )
{
	QUEUE_ENTRY				*pEntry;
	PQP_BUFFER_DESCRIPTOR	pBufDesc;
	PQP_KSSTREAM_HEADER		pDataPacket;
	CYUVOutChannel			*This =	GET_INHERITED_OBJECT(CYUVOutChannel, m_Channel, This_p);

	pEntry = CQueue_GetOneEntry(This_p->m_pDataRequestQueue);

	if (QP_NULL != pEntry)
	{
		pBufDesc = (PQP_BUFFER_DESCRIPTOR)pEntry->Data;
		pDataPacket = pBufDesc->DataBufferArray;

		pBufDesc->ulFlags &= ~QP_BUFDESC_FLAG_MASK_PRIVATE;
		pBufDesc->ulBufferIndex = 0;
		pBufDesc->ulBufferOffset = 0;
		pBufDesc->ulTotalUsed = 0;

		pDataPacket->DataUsed = 0;

		*ppYBuffer = (PTQP_UINT8)pDataPacket->Data;
		*pYSize = This->m_lWidth * This->m_lHeight;
		*ppUVBuffer = ((PTQP_UINT8)pDataPacket->Data) + (*pYSize);
		*pUVSize = (*pYSize) >> 1;
		*ppBufDesc = pBufDesc;

		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, /*QPDBGM_LEVEL_WARN*/QPDBGM_LEVEL_TRACE),
					 "CYUVOutChannel_GetBuffer(%d) buffer(0x%x) size(%d) Desc(0x%x)\n",
					 This_p->m_hChannel,
					 *ppYBuffer,
					 *pYSize + *pUVSize,
					 *ppBufDesc
					 ));

		// move this entry to the pending queue
		//
		CQueue_AddEntry(This_p->m_pDataPendingQueue,
						pEntry
						);
	}

	return (QP_NULL != pEntry);
}



static TQP_BOOL CYUVOutChannel_GetBufferRas(CChannel *This_p,
											PQP_BUFFER_DESCRIPTOR *ppBufDesc,
											PTQP_UINT8 *ppYBuffer,
											PTQP_UINT32 pYSize,
											PTQP_UINT8 *ppUBuffer,
											PTQP_UINT32 pUSize,
											PTQP_UINT8 *ppVBuffer,
											PTQP_UINT32 pVSize
											)
{
	QUEUE_ENTRY				*pEntry;
	PQP_BUFFER_DESCRIPTOR	pBufDesc;
	PQP_KSSTREAM_HEADER		pDataPacket;
	CYUVOutChannel			*This =	GET_INHERITED_OBJECT(CYUVOutChannel, m_Channel, This_p);

	pEntry = CQueue_GetOneEntry(This_p->m_pDataRequestQueue);

	if (QP_NULL != pEntry)
	{
		pBufDesc = (PQP_BUFFER_DESCRIPTOR)pEntry->Data;
		pDataPacket = pBufDesc->DataBufferArray;

		pBufDesc->ulFlags &= ~QP_BUFDESC_FLAG_MASK_PRIVATE;

		if (pBufDesc->ulBufferSize < This->m_dwFrameSize)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CYUVOutChannel_GetBufferRas() buffer size(%d) < frame size(%d)\n",
						 pBufDesc->ulBufferSize,
						 This->m_dwFrameSize
						 ));
			// clear the entry content
			//
			pEntry->Data = QP_NULL;

			// move to free queue
			//
			CQueue_AddEntry(This_p->m_pFreeQueue,
							pEntry
							);

			// nil pEntry so the function will return FALSE
			//
			pEntry = QP_NULL;

			// default success
			//
			pBufDesc->Status = QPERR_SUCCESS;

			// call back the api
			//
			CChannel_DeviceCallback(This_p,
									QPMPGCODEC_CMD_DONE_DATA,
									pBufDesc
									);
		}
		else
		{
			pBufDesc->ulBufferIndex = 0;
			pBufDesc->ulBufferOffset = 0;
			pBufDesc->ulTotalUsed = 0;

			pDataPacket->DataUsed = 0;

			*ppYBuffer = (PTQP_UINT8)pDataPacket->Data;
			*pYSize = This->m_lWidth * This->m_lHeight;
			*pUSize = *pVSize = ((*pYSize) >> 2);
			if (QP_YUV_DATA_TYPE_YV12 == This->m_nDataType)
			{
				*ppVBuffer = ((PTQP_UINT8)pDataPacket->Data) + (*pYSize);
				*ppUBuffer = (*ppVBuffer) + (*pVSize);
			}
			else
			{
				*ppUBuffer = ((PTQP_UINT8)pDataPacket->Data) + (*pYSize);
				*ppVBuffer = (*ppUBuffer) + (*pUSize);
			}
			*ppBufDesc = pBufDesc;

			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, /*QPDBGM_LEVEL_WARN*/QPDBGM_LEVEL_TRACE),
						 "CYUVOutChannel_GetBufferRas(%d) Y(0x%x) size(%d) U(0x%x) size(%d) V(0x%x) size(%d) frame(%d)\n",
						 *ppYBuffer,
						 *pYSize,
						 *ppUBuffer,
						 *pUSize,
						 *ppVBuffer,
						 *pVSize,
						 This->m_ulVideoFramesCount
						 ));

			// move this entry to the pending queue
			//
			CQueue_AddEntry(This_p->m_pDataPendingQueue,
							pEntry
							);
		}
	}

	return (QP_NULL != pEntry);
}



static void CYUVOutChannel_CompleteBuffer(CChannel *This_p,
										  PQP_BUFFER_DESCRIPTOR pBufDesc
										  )
{
	QUEUE_ENTRY			*pEntry;
	PQP_KSSTREAM_HEADER	pDataPacket;
	CYUVOutChannel		*This =	GET_INHERITED_OBJECT(CYUVOutChannel, m_Channel, This_p);

	////QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
	////			 "CYUVOutChannel_CompleteBuffer()\n"
	////			 ));

	// get the buffer entry
	//
	pEntry = CQueue_GetEntryByData(This_p->m_pDataPendingQueue,
								   (TQP_PVOID)pBufDesc
								   );
	if (pEntry)
	{
		pDataPacket = pBufDesc->DataBufferArray;
		pDataPacket->OptionsFlags = 0;

		// set these flags if we are at the begining of the stream
		//
		if (0 == This->m_ulVideoFramesCount)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
						 "CYUVOutChannel_CompleteBuffer() (0 == m_ulVideoFramesCount) QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY\n"
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY |
										 QP_KSSTREAM_HEADER_OPTIONSF_TIMEDISCONTINUITY
										 ;
		}

		// increment total frame captured
		//
		This->m_ulVideoFramesCount++;

		// update frame info
		//
		pBufDesc->dwFrameFlags = QP_VIDEO_FLAG_FRAME;
		pBufDesc->PictureNumber = This->m_ulVideoFramesCount;
		pBufDesc->DropCount = 0;

		// set the timestamp and flags
		//
		CChannel_TimeStamp(This_p,
						   pBufDesc,
						   pDataPacket
						   );

		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
					 "CYUVOutChannel_CompleteBuffer() flgs(0x%x) Cap(%d) size(%d)",
					 pBufDesc->ulFlags,
					 This->m_ulVideoFramesCount,
					 pBufDesc->ulTotalUsed
					 ));

		This_p->m_ullCntBytes += pBufDesc->ulTotalUsed;
		pDataPacket->DataUsed = pBufDesc->ulTotalUsed;

		// clear the entry content
		//
		pEntry->Data = QP_NULL;

		// move to free queue
		//
		CQueue_AddEntry(This_p->m_pFreeQueue,
						pEntry
						);

		// call back to the api
		//
		CChannel_DeviceCallback(This_p,
								QPMPGCODEC_CMD_DONE_DATA,
								pBufDesc
								);
	}
}



static EQP_Errors CYUVOutChannel_GetResolution(CChannel *This_p,
											   PTQP_UINT32 pWidth,
											   PTQP_UINT32 pHeight
											   )
{
	CYUVOutChannel *This = GET_INHERITED_OBJECT(CYUVOutChannel, m_Channel, This_p);

	if (pWidth &&
		pHeight
		)
	{
		*pWidth = (TQP_UINT32)This->m_lWidth;
		*pHeight = (TQP_UINT32)This->m_lHeight;
		return(QPERR_SUCCESS);
	}
	else
	{
		return (QPERR_PARMS);
	}
}



static EQP_Errors CYUVOutChannel_GetYUVFormat(CChannel *This_p,
											  PTQP_UINT32 pYUVFormat
											  )
{
	CYUVOutChannel *This = GET_INHERITED_OBJECT(CYUVOutChannel, m_Channel, This_p);

	if (pYUVFormat)
	{
		*pYUVFormat = This->m_nDataType;
		return(QPERR_SUCCESS);
	}
	else
	{
		return (QPERR_PARMS);
	}
}



// constructor
//
CYUVOutChannel * CYUVOutChannel_Constructor(CYUVOutChannel *pYUVOutChannel,
											CObject *pParent,
											TQP_UINT32 dwAttributes,
											TQP_OBJECT_HANDLE hTask,
											QPMPGCODEC_OPEN_TYPE ChannelType,
											CTask *pTask
											)
{
	if (pYUVOutChannel)
	{
		CChannel	*pChannel;

		pChannel = CChannel_Constructor(&pYUVOutChannel->m_Channel,
										pParent,
										dwAttributes,
										hTask,
										TASK_DATA_TYPE_YUV,
										FWDataTypeRawVidOutput,
										CHANNEL_DIR_READ,
										ChannelType,
										pTask
										);
		if (!pChannel)
		{
			pYUVOutChannel = QP_NULL;
		}
		else
		{
			// fill in the functions table
			//
			pYUVOutChannel->m_Channel.Open = CYUVOutChannel_Open;
			pYUVOutChannel->m_Channel.Start = CYUVOutChannel_Start;
			pYUVOutChannel->m_Channel.Stop = CMPEGOutChannel_Stop;		// same as MpegOut
			pYUVOutChannel->m_Channel.Pause = CMPEGOutChannel_Pause;	// same as MpegOut
			pYUVOutChannel->m_Channel.GetBuffer = CMPEGOutChannel_GetBuffer;	// same as MpegOut
			pYUVOutChannel->m_Channel.GetBufferYUV = CYUVOutChannel_GetBuffer;
			pYUVOutChannel->m_Channel.GetBufferYUVRAS = CYUVOutChannel_GetBufferRas;
			pYUVOutChannel->m_Channel.CompleteBuffer = CYUVOutChannel_CompleteBuffer;
			pYUVOutChannel->m_Channel.GetResolution = CYUVOutChannel_GetResolution;
			pYUVOutChannel->m_Channel.GetYUVFormat = CYUVOutChannel_GetYUVFormat;


			// CYUVOutChannel
			//
			pYUVOutChannel->m_dwFrameSize = 0;
			pYUVOutChannel->m_ulAvgTimePerFrame = KS_NTSC_INCREMENT;
			pYUVOutChannel->m_ulAvgPTSPerFrame = PTS_NTSC_INCREMENT;
			pYUVOutChannel->m_ulAvgPTSPerFrame1000X = PTS_NTSC_INCREMENT_1000X;
			pYUVOutChannel->m_ulVideoFramesCount = 0;
			pYUVOutChannel->m_ulDroppedFramesCount = 0;

			pYUVOutChannel->m_lWidth = 0;
			pYUVOutChannel->m_lHeight = 0;
			pYUVOutChannel->m_nDataType = QP_YUV_DATA_TYPE_QYUV;
		}
	}

	return (pYUVOutChannel);
}



// destructor
//
void CYUVOutChannel_Destructor(CYUVOutChannel *This)
{
	CChannel_Destructor(&This->m_Channel);
}



/////////////////////////////////////////////////////////////////////////////
//
//

// CPCMOutChannel
//


// CChannel
//

static EQP_Errors CPCMOutChannel_Open(CChannel *This_p,
									  TQP_OBJECT_HANDLE hChannel,
									  TQP_UINT32 dwFlags,
									  TQP_PVOID pDataFormat,
									  QP_DEVICE_CALLBACK pFuncCallback,
									  TQP_PVOID context
									  )
{
	EQP_Errors			qpStatus;
	PQP_PCM_DATAFORMAT	pPCMFormat = (PQP_PCM_DATAFORMAT)pDataFormat;
//	CPCMOutChannel		*This =	GET_INHERITED_OBJECT(CPCMOutChannel, m_Channel, This_p);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CPCMOutChannel_Open() AvgBytesPerSec(%d)\n  BlockAlign(%d)\n    Channels(%d)\n  SamplesPerSec(%d)\n  BitsPerSample(%d)\n",
				 pPCMFormat->nAvgBytesPerSec,
				 pPCMFormat->nBlockAlign,
				 pPCMFormat->nChannels,
				 pPCMFormat->nSamplesPerSec,
				 pPCMFormat->nBitsPerSample
				 ));

	// Gather information we need for this stream
	//

	// generic open
	//
	qpStatus = CChannel_Open(This_p,
							 hChannel,
							 dwFlags,
							 pDataFormat,
							 pFuncCallback,
							 context
							 );

	return (qpStatus);
}



static EQP_Errors CPCMOutChannel_Start(CChannel *This_p)
{
	EQP_Errors		qpStatus;
//	CPCMOutChannel	*This =	GET_INHERITED_OBJECT(CPCMOutChannel, m_Channel, This_p);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CPCMOutChannel_Start() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 This_p->m_dataType
				 ));

	if (This_p->m_bPaused)
	{
		qpStatus = This_p->m_pTask->Resume(This_p->m_pTask,
										   This_p->m_hTask,
										   (TASK_DATA_TYPE)This_p->m_dataType
										   );

		if (QP_SUCCESS(qpStatus))
			This_p->m_bPaused = QP_FALSE;
	}
	else
	{
		// reset counter and flags
		//
		This_p->m_ullCntBytes = 0;

		This_p->m_llPrevPts = 0;
		This_p->m_llAddPts = 0;
		This_p->m_llBasePts = 0;

		// start the encoder
		//
		qpStatus = This_p->m_pTask->Start(This_p->m_pTask,
										  This_p->m_hTask,
										  (TASK_DATA_TYPE)This_p->m_dataType
										  );

		This_p->m_llStartTime = QPTMGetSysTimeMilliS();
	}

	This_p->m_bFlushing = QP_FALSE;

	if (QP_SUCCESS(qpStatus))
	{
		This_p->m_State = QPSTATE_RUN;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CPCMOutChannel_Start() Failed Status(%d)\n",
					 qpStatus
					 ));
	}

	return (qpStatus);
}



// called from streaming task
//
static void CPCMOutChannel_CompleteBuffer(CChannel *This_p,
										  PQP_BUFFER_DESCRIPTOR pBufDesc
										  )
{
	QUEUE_ENTRY			*pEntry;
	PQP_KSSTREAM_HEADER	pDataPacket;
//	CPCMOutChannel		*This =	GET_INHERITED_OBJECT(CPCMOutChannel, m_Channel, This_p);
	CQLCodec			*pQlCodec = GET_INHERITED_OBJECT(CQLCodec, m_Object, This_p->m_Object.m_pParent);


	////QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
	////			 "CPCMOutChannel_CompleteBuffer()\n"
	////			 ));

	// get the buffer entry
	//
	pEntry = CQueue_GetEntryByData(This_p->m_pDataPendingQueue,
								   (TQP_PVOID)pBufDesc
								   );
	if (pEntry)
	{
		pDataPacket = pBufDesc->DataBufferArray;
		pDataPacket->OptionsFlags = 0;

		if (This_p->m_bByteSwap &&
			!QPHCI_CanSwapData(&pQlCodec->m_hci) &&
			(0 != pBufDesc->ulTotalUsed)
			)
		{
			BytesSwap((PTQP_UINT8)pDataPacket->Data,
					  (PTQP_UINT8)pDataPacket->Data,
					  pBufDesc->ulTotalUsed
					  );
		}

		// set these flags if we are at the begining of the stream
		//
		if (0 == This_p->m_ullCntBytes)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
						 "CPCMOutChannel_CompleteBuffer() (0 == m_ullCntBytes) QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY\n"
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY |
										 QP_KSSTREAM_HEADER_OPTIONSF_TIMEDISCONTINUITY
										 ;
		}

		// set time stamp
		//
		CChannel_TimeStamp(This_p,
						   pBufDesc,
						   pDataPacket
						   );

		This_p->m_ullCntBytes += pBufDesc->ulTotalUsed;
		pDataPacket->DataUsed = pBufDesc->ulTotalUsed;

		// clear the entry content
		//
		pEntry->Data = QP_NULL;

		// recycle buffer entry
		//
		CQueue_AddEntry(This_p->m_pFreeQueue,
						pEntry
						);

		// call back to the api
		//
		CChannel_DeviceCallback(This_p,
								QPMPGCODEC_CMD_DONE_DATA,
								pBufDesc
								);
	}
}




// constructor
//
CPCMOutChannel * CPCMOutChannel_Constructor(CPCMOutChannel *pPCMOutChannel,
											CObject *pParent,
											TQP_UINT32 dwAttributes,
											TQP_OBJECT_HANDLE hTask,
											QPMPGCODEC_OPEN_TYPE ChannelType,
											CTask *pTask
											)
{
	if (pPCMOutChannel)
	{
		CChannel	*pChannel;

		pChannel = CChannel_Constructor(&pPCMOutChannel->m_Channel,
										pParent,
										dwAttributes,
										hTask,
										TASK_DATA_TYPE_PCM,
										FWDataTypeRawAudOutput,
										CHANNEL_DIR_READ,
										ChannelType,
										pTask
										);
		if (!pChannel)
		{
			pPCMOutChannel = QP_NULL;
		}
		else
		{
			// fill in the functions table
			//
			pPCMOutChannel->m_Channel.Open = CPCMOutChannel_Open;
			pPCMOutChannel->m_Channel.Start = CPCMOutChannel_Start;
			pPCMOutChannel->m_Channel.Stop = CMPEGOutChannel_Stop;		// same as MpegOut
			pPCMOutChannel->m_Channel.Pause = CMPEGOutChannel_Pause;	// same as MpegOut
			pPCMOutChannel->m_Channel.GetBuffer = CMPEGOutChannel_GetBuffer;	// same as MpegOut
			pPCMOutChannel->m_Channel.CompleteBuffer = CPCMOutChannel_CompleteBuffer;
		}
	}

	return (pPCMOutChannel);
}



// destructor
//
void CPCMOutChannel_Destructor(CPCMOutChannel *This)
{
	CChannel_Destructor(&This->m_Channel);
}



/////////////////////////////////////////////////////////////////////////////
//
//

// CVBIOutChannel
//


// CChannel
//

static EQP_Errors CVBIOutChannel_Open(CChannel *This_p,
									  TQP_OBJECT_HANDLE hChannel,
									  TQP_UINT32 dwFlags,
									  TQP_PVOID pDataFormat,
									  QP_DEVICE_CALLBACK pFuncCallback,
									  TQP_PVOID context
									  )
{
	EQP_Errors			qpStatus;
	PQP_VBI_DATAFORMAT	pVBIFormat = (PQP_VBI_DATAFORMAT)pDataFormat;
	CVBIOutChannel		*This =	GET_INHERITED_OBJECT(CVBIOutChannel, m_Channel, This_p);

	// Gather information we need for this stream
	//
	switch (pVBIFormat->nFrameRate)
	{
		case 30:
			This->m_ulAvgPTSPerFrame = PTS_NTSC_INCREMENT;
			This->m_ulAvgPTSPerFrame1000X = PTS_NTSC_INCREMENT_1000X;
			break;

		case 25:
			This->m_ulAvgPTSPerFrame = PTS_PAL_INCREMENT;
			This->m_ulAvgPTSPerFrame1000X = PTS_PAL_INCREMENT_1000X;
			break;

		default:
			This->m_ulAvgPTSPerFrame = (90000 / pVBIFormat->nFrameRate);
			This->m_ulAvgPTSPerFrame1000X = (90000000 / pVBIFormat->nFrameRate);
			break;
	}

#ifdef QP_LINXKER
	{
		unsigned long long ullTemp = (unsigned long long)(This->m_ulAvgPTSPerFrame);
		ullTemp *= 10000;
		do_div(ullTemp, 90);
		This->m_ulAvgTimePerFrame = (unsigned long)ullTemp;
	}
#else
	This->m_ulAvgTimePerFrame = (TQP_UINT32)PTS_TO_KSPTS(This->m_ulAvgPTSPerFrame);
#endif
	This->m_ulAvgTimePerField = (This->m_ulAvgTimePerFrame >> 1);
	This->m_ulAvgPTSPerField = (This->m_ulAvgPTSPerFrame >> 1);

    This->m_Top_StartLine = pVBIFormat->Top_StartLine;			// inclusive
    This->m_Top_EndLine = pVBIFormat->Top_EndLine;				// inclusive
  	This->m_Top_StartPixel = pVBIFormat->Top_StartPixel;		// starting from 0
    This->m_Top_SamplesPerLine = pVBIFormat->Top_SamplesPerLine;// maximum (1440 - StartPixel)

    This->m_Bottom_StartLine = pVBIFormat->Bottom_StartLine;	// inclusive
    This->m_Bottom_EndLine = pVBIFormat->Bottom_EndLine;		// inclusive
  	This->m_Bottom_StartPixel = pVBIFormat->Bottom_StartPixel;	// starting from 0
    This->m_Bottom_SamplesPerLine = pVBIFormat->Bottom_SamplesPerLine;	// maximum (1440 - StartPixel)

	This->m_BufferSize = (TQP_UINT32)pVBIFormat->BufferSize;
	This->m_nVBIDataCaptureSize = ((This->m_Top_EndLine - This->m_Top_StartLine + 1) * This->m_Top_SamplesPerLine) +
								  ((This->m_Bottom_EndLine - This->m_Bottom_StartLine + 1) * This->m_Bottom_SamplesPerLine)
								  ;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
				 "CVBIOutChannel::Open()\n"
				 ));
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
				 "  StartLine(%d : %d) EndLine(%d : %d) StartPixel(%d : %d) Samples(%d : %d) buf_size(%d) capture Size(%d)\n",
				 This->m_Top_StartLine,
				 This->m_Bottom_StartLine,
				 This->m_Top_EndLine,
				 This->m_Bottom_EndLine,
  				 This->m_Top_StartPixel,
  				 This->m_Bottom_StartPixel,
				 This->m_Top_SamplesPerLine,
				 This->m_Bottom_SamplesPerLine,
				 This->m_BufferSize,
				 This->m_nVBIDataCaptureSize
				 ));

	// reset counters and flags
	//
	This->m_ullVBIFramesCount = 0;
	This->m_ulDroppedFramesCount = 0;

	// generic open
	//
	qpStatus = CChannel_Open(This_p,
							 hChannel,
							 dwFlags,
							 pDataFormat,
							 pFuncCallback,
							 context
							 );

	return (qpStatus);
}



static EQP_Errors CVBIOutChannel_Start(CChannel *This_p)
{
	EQP_Errors		qpStatus;
	CVBIOutChannel	*This =	GET_INHERITED_OBJECT(CVBIOutChannel, m_Channel, This_p);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CVBIOutChannel_Start() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 This_p->m_dataType
				 ));

	if (This_p->m_bPaused)
	{
		qpStatus = This_p->m_pTask->Resume(This_p->m_pTask,
										   This_p->m_hTask,
										   (TASK_DATA_TYPE)This_p->m_dataType
										   );

		if (QP_SUCCESS(qpStatus))
			This_p->m_bPaused = QP_FALSE;
	}
	else
	{
		// reset counter and flags
		//
		This_p->m_ullCntBytes = 0;

		This->m_ullVBIFramesCount = 0;
		This->m_ulDroppedFramesCount = 0;

		This_p->m_llBasePts = 0;
		This_p->m_llPrevPts = 0;
		This_p->m_llAddPts = 0;

		// start the encoder
		//
		qpStatus = This_p->m_pTask->Start(This_p->m_pTask,
										  This_p->m_hTask,
										  (TASK_DATA_TYPE)This_p->m_dataType
										  );

		This_p->m_llStartTime = QPTMGetSysTimeMilliS();
	}

	This_p->m_bFlushing = QP_FALSE;

	if (QP_SUCCESS(qpStatus))
	{
		This_p->m_State = QPSTATE_RUN;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CVBIOutChannel_Start() Failed Status(%d)\n",
					 qpStatus
					 ));
	}

	return (qpStatus);
}



// called from streaming task
//
static TQP_BOOL CVBIOutChannel_GetBuffer(CChannel *This_p,
										 PQP_BUFFER_DESCRIPTOR *ppBufDesc,
										 PTQP_UINT8 *ppBuffer,
										 PTQP_UINT32 pSize
										 )
{

	QUEUE_ENTRY				*pEntry;
	PQP_BUFFER_DESCRIPTOR	pBufDesc;
	PQP_KSSTREAM_HEADER		pDataPacket;
	CVBIOutChannel			*This =	GET_INHERITED_OBJECT(CVBIOutChannel, m_Channel, This_p);

	pEntry = CQueue_GetOneEntry(This_p->m_pDataRequestQueue);

	if (QP_NULL != pEntry)
	{
		pBufDesc = (PQP_BUFFER_DESCRIPTOR)pEntry->Data;
		pDataPacket = pBufDesc->DataBufferArray;

		pBufDesc->ulFlags &= ~QP_BUFDESC_FLAG_MASK_PRIVATE;
		pBufDesc->ulBufferIndex = 0;
		pBufDesc->ulBufferOffset = 0;
		pBufDesc->ulTotalUsed = 0;

		pDataPacket->DataUsed = 0;

		*ppBuffer = (PTQP_UINT8)pDataPacket->Data;
		*pSize = This->m_nVBIDataCaptureSize;
		*ppBufDesc = pBufDesc;

		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, /*QPDBGM_LEVEL_WARN*/QPDBGM_LEVEL_TRACE),
					 "CVBIOutChannel_GetBuffer(&d) buffer(0x%x) size(%d) Desc(0x%x)\n",
					 This_p->m_hChannel,
					 *ppBuffer,
					 *pSize,
					 *ppBufDesc
					 ));

		// move this entry to the pending queue
		//
		CQueue_AddEntry(This_p->m_pDataPendingQueue,
						pEntry
						);
	}

	return (QP_NULL != pEntry);
}



static void CVBIOutChannel_CompleteBuffer(CChannel *This_p,
										  PQP_BUFFER_DESCRIPTOR pBufDesc
										  )
{
	QUEUE_ENTRY			*pEntry;
	PQP_KSSTREAM_HEADER	pDataPacket;
	//TQP_UINT64			ullVBIFramesCountPTS = 0;
	//TQP_UINT64			ullDroppedFramesCount = 0;
	CVBIOutChannel		*This =	GET_INHERITED_OBJECT(CVBIOutChannel, m_Channel, This_p);

	////QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
	////			 "CVBIOutChannel()\n"
	////			 ));

	// get the buffer entry
	//
	pEntry = CQueue_GetEntryByData(This_p->m_pDataPendingQueue,
								   (TQP_PVOID)pBufDesc
								   );
	if (pEntry)
	{
		pDataPacket = pBufDesc->DataBufferArray;
		pDataPacket->OptionsFlags = 0;

		// set these flags if we are at the begining of the stream
		//
		if (0 == This->m_ullVBIFramesCount)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
						 "CVBIOutChannel_CompleteBuffer() (0 == m_ullVBIFramesCount) QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY\n"
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY |
										 QP_KSSTREAM_HEADER_OPTIONSF_TIMEDISCONTINUITY
										 ;
		}

		// increment total frame captured
		//
		This->m_ullVBIFramesCount++;

		// update frame info
		//
		pBufDesc->dwFrameFlags = QP_VIDEO_FLAG_FRAME;
		pBufDesc->PictureNumber = This->m_ullVBIFramesCount;
		pBufDesc->DropCount = 0;

		// set time stamp
		//
		CChannel_TimeStamp(This_p,
						   pBufDesc,
						   pDataPacket
						   );

		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, /*QPDBGM_LEVEL_WARN*/QPDBGM_LEVEL_TRACE),
					 "CVBIOutChannel_CompleteBuffer() Cap(%lu:%lu)",
					 (TQP_UINT32)((This->m_ullVBIFramesCount >> 32) & 0xFFFFFFFF),
					 (TQP_UINT32)(This->m_ullVBIFramesCount & 0xFFFFFFFF)
					 ));

		This_p->m_ullCntBytes += pBufDesc->ulTotalUsed;
		pDataPacket->DataUsed = pBufDesc->ulTotalUsed;

		// clear the entry content
		//
		pEntry->Data = QP_NULL;

		// move to free queue
		//
		CQueue_AddEntry(This_p->m_pFreeQueue,
						pEntry
						);

		// call back to the api
		//
		CChannel_DeviceCallback(This_p,
								QPMPGCODEC_CMD_DONE_DATA,
								pBufDesc
								);
	}
}



// constructor
//
CVBIOutChannel * CVBIOutChannel_Constructor(CVBIOutChannel *pVBIOutChannel,
											CObject *pParent,
											TQP_UINT32 dwAttributes,
											TQP_OBJECT_HANDLE hTask,
											CTask *pTask
											)
{
	if (pVBIOutChannel)
	{
		CChannel	*pChannel;

		pChannel = CChannel_Constructor(&pVBIOutChannel->m_Channel,
										pParent,
										dwAttributes,
										hTask,
										TASK_DATA_TYPE_VBI,
										FWDataTypeVBIOutput,
										CHANNEL_DIR_READ,
										QPMPGCODEC_ENC_VBI_OUT,
										pTask
										);
		if (!pChannel)
		{
			pVBIOutChannel = QP_NULL;
		}
		else
		{
			// fill in the functions table
			//
			pVBIOutChannel->m_Channel.Open = CVBIOutChannel_Open;
			pVBIOutChannel->m_Channel.Start = CVBIOutChannel_Start;
			pVBIOutChannel->m_Channel.Stop = CMPEGOutChannel_Stop;		// same as MpegOut
			pVBIOutChannel->m_Channel.Pause = CMPEGOutChannel_Pause;	// same as MpegOut
			pVBIOutChannel->m_Channel.GetBuffer = CVBIOutChannel_GetBuffer;
			pVBIOutChannel->m_Channel.CompleteBuffer = CVBIOutChannel_CompleteBuffer;

			// CVBIOutChannel
			//
			pVBIOutChannel->m_Top_StartLine = 0;		// inclusive
			pVBIOutChannel->m_Top_EndLine = 0;			// inclusive
  			pVBIOutChannel->m_Top_StartPixel = 0;		// starting from 0
			pVBIOutChannel->m_Top_SamplesPerLine = 0;	// maximum (1440 - StartPixel)

			pVBIOutChannel->m_Bottom_StartLine = 0;		// inclusive
			pVBIOutChannel->m_Bottom_EndLine = 0;		// inclusive
  			pVBIOutChannel->m_Bottom_StartPixel = 0;	// starting from 0
			pVBIOutChannel->m_Bottom_SamplesPerLine = 0;// maximum (1440 - StartPixel)

			pVBIOutChannel->m_BufferSize = 0;
			pVBIOutChannel->m_nVBIDataCaptureSize = 0;
			pVBIOutChannel->m_nVBILines = 0;

			pVBIOutChannel->m_ulAvgTimePerFrame = KS_NTSC_INCREMENT;
			pVBIOutChannel->m_ulAvgPTSPerFrame = PTS_NTSC_INCREMENT;
			pVBIOutChannel->m_ulAvgPTSPerFrame1000X = PTS_NTSC_INCREMENT_1000X;
			pVBIOutChannel->m_ulAvgTimePerField = (KS_NTSC_INCREMENT >> 1);
			pVBIOutChannel->m_ulAvgPTSPerField = (PTS_NTSC_INCREMENT >> 1);

			pVBIOutChannel->m_ullVBIFramesCount = 0;
			pVBIOutChannel->m_ulDroppedFramesCount = 0;
		}
	}

	return (pVBIOutChannel);
}



// destructor
//
void CVBIOutChannel_Destructor(CVBIOutChannel *This)
{
	CChannel_Destructor(&This->m_Channel);
}



// helper
//
TQP_BOOL CVBIOutChannel_GetVBIFormat(CVBIOutChannel *This,
									 PQP_VBI_DATAFORMAT pVBIFormat
									 )
{
	if (This &&
		pVBIFormat &&
		This->m_Channel.m_bOpened
		)
	{
		pVBIFormat->Top_StartLine = This->m_Top_StartLine;			// inclusive
		pVBIFormat->Top_EndLine = This->m_Top_EndLine;				// inclusive
  		pVBIFormat->Top_StartPixel = This->m_Top_StartPixel;		// starting from 0
		pVBIFormat->Top_SamplesPerLine = This->m_Top_SamplesPerLine;// maximum (1440 - StartPixel)

		pVBIFormat->Bottom_StartLine = This->m_Bottom_StartLine;	// inclusive
		pVBIFormat->Bottom_EndLine = This->m_Bottom_EndLine;		// inclusive
  		pVBIFormat->Bottom_StartPixel = This->m_Bottom_StartPixel;	// starting from 0
		pVBIFormat->Bottom_SamplesPerLine = This->m_Bottom_SamplesPerLine;	// maximum (1440 - StartPixel)

		return (QP_TRUE);
	}

	return (QP_FALSE);
}



/////////////////////////////////////////////////////////////////////////////
//
//

// CIndexOutChannel
//


// CChannel
//

// called from streaming task
//
static void CIndexOutChannel_CompleteBuffer(CChannel *This_p,
											PQP_BUFFER_DESCRIPTOR pBufDesc
											)
{
	QUEUE_ENTRY			*pEntry;
	PQP_KSSTREAM_HEADER	pDataPacket;

	////QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
	////			 "CIndexOutChannel_CompleteBuffer()\n"
	////			 ));

	// get the buffer entry
	//
	pEntry = CQueue_GetEntryByData(This_p->m_pDataPendingQueue,
								   (TQP_PVOID)pBufDesc
								   );
	if (pEntry)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, /*QPDBGM_LEVEL_WARN*/QPDBGM_LEVEL_TRACE),
					 "CIndexOutChannel_CompleteBuffer() bufDesc(0x%x) Status(%d), totalUsed(%d)\n",
					 pBufDesc,
					 pBufDesc->Status,
					 pBufDesc->ulTotalUsed
					 ));

		pDataPacket = pBufDesc->DataBufferArray;
		pDataPacket->OptionsFlags = 0;

		// set these flags if we are at the begining of the stream
		//
		if (0 == This_p->m_ullCntBytes)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
						 "CIndexOutChannel_CompleteBuffer() (0 == m_ullCntBytes) QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY\n"
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY |
										 QP_KSSTREAM_HEADER_OPTIONSF_TIMEDISCONTINUITY
										 ;
		}

		// set time stamp
		//
		CChannel_TimeStamp(This_p,
						   pBufDesc,
						   pDataPacket
						   );

		This_p->m_ullCntBytes += pBufDesc->ulTotalUsed;
		pDataPacket->DataUsed = pBufDesc->ulTotalUsed;

		if (pBufDesc->ulFlags & QP_BUFDESC_FLAG_EOS)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CIndexOutChannel_CompleteBuffer() used(%d) QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM\n",
						 pBufDesc->ulTotalUsed
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM;
		}

		// clear the entry content
		//
		pEntry->Data = QP_NULL;

		// move to free queue
		//
		CQueue_AddEntry(This_p->m_pFreeQueue,
						pEntry
						);

		// call back the api
		//
		CChannel_DeviceCallback(This_p,
								QPMPGCODEC_CMD_DONE_DATA,
								pBufDesc
								);
	}
}



// constructor
//
CIndexOutChannel * CIndexOutChannel_Constructor(CIndexOutChannel *pIndexOutChannel,
												CObject *pParent,
												TQP_UINT32 dwAttributes,
												TQP_OBJECT_HANDLE hTask,
												CTask *pTask
												)
{
	if (pIndexOutChannel)
	{
		CChannel	*pChannel;

		pChannel = CChannel_Constructor(&pIndexOutChannel->m_Channel,
										pParent,
										dwAttributes,
										hTask,
										TASK_DATA_TYPE_INDEX,
										FWDataTypeIndexOutput,
										CHANNEL_DIR_READ,
										QPMPGCODEC_ENC_INDEX_OUT,
										pTask
										);
		if (!pChannel)
		{
			pIndexOutChannel = QP_NULL;
		}
		else
		{
			// fill in the functions table
			//
			pIndexOutChannel->m_Channel.Start = CMPEGOutChannel_Start;
			pIndexOutChannel->m_Channel.Stop = CMPEGOutChannel_Stop;	// same as MpegOut
			pIndexOutChannel->m_Channel.Pause = CMPEGOutChannel_Pause;	// same as MpegOut
			pIndexOutChannel->m_Channel.GetBuffer = CMPEGOutChannel_GetBuffer;	// same as MpegOut
			pIndexOutChannel->m_Channel.CompleteBuffer = CIndexOutChannel_CompleteBuffer;
		}
	}

	return (pIndexOutChannel);
}



// destructor
//
void CIndexOutChannel_Destructor(CIndexOutChannel *This)
{
	CChannel_Destructor(&This->m_Channel);
}



/////////////////////////////////////////////////////////////////////////////
//
//

// CMPEGInChannel
//

static EQP_Errors CMPEGInChannel_AddBufferMjpeg(CChannel *This,
									 PQP_BUFFER_DESCRIPTOR pBufDesc
									 );

static TQP_BOOL CMPEGInChannel_GetBuffer(CChannel *This_p,
										 PQP_BUFFER_DESCRIPTOR *ppBufDesc,
										 PTQP_UINT8 *ppBuffer,
										 PTQP_UINT32 pSize
										 );
static TQP_BOOL CMPEGInChannel_GetBufferMjpeg(CChannel *This_p,
											  PQP_BUFFER_DESCRIPTOR *ppBufDesc,
											  PTQP_UINT8 *ppBuffer,
											  PTQP_UINT32 pSize
											  );

// CChannel
//
static EQP_Errors CMPEGInChannel_Open(CChannel *This_p,
									  TQP_OBJECT_HANDLE hChannel,
									  TQP_UINT32 dwFlags,
									  TQP_PVOID pDataFormat,
									  QP_DEVICE_CALLBACK pFuncCallback,
									  TQP_PVOID context
									  )
{
	EQP_Errors						qpStatus;
	PQP_COMPRESSED_VID_DATAFORMAT	pMpegInFormat = (PQP_COMPRESSED_VID_DATAFORMAT)pDataFormat;
	CMPEGInChannel					*This =	GET_INHERITED_OBJECT(CMPEGInChannel, m_Channel, This_p);

	// Gather information we need for this stream
	//
	This->m_nDataType = pMpegInFormat ? pMpegInFormat->nDataType : QP_COMP_VID_DATA_TYPE_H264;

	if (QP_COMP_VID_DATA_TYPE_MJPEG == This->m_nDataType)
	{
		This_p->GetBuffer = CMPEGInChannel_GetBufferMjpeg;
		//This_p->AddBuffer = CMPEGInChannel_AddBufferMjpeg;
	}
	else
	{
		This_p->GetBuffer = CMPEGInChannel_GetBuffer;
	}

	// generic open
	//
	qpStatus = CChannel_Open(This_p,
							 hChannel,
							 dwFlags,
							 pDataFormat,
							 pFuncCallback,
							 context
							 );

	return (qpStatus);
}



static EQP_Errors CMPEGInChannel_Start(CChannel *This_p)
{
	EQP_Errors	qpStatus;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMPEGInChannel_Start() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 This_p->m_dataType
				 ));

	if (This_p->m_bPaused)
	{
		qpStatus = This_p->m_pTask->Resume(This_p->m_pTask,
										   This_p->m_hTask,
										   (TASK_DATA_TYPE)This_p->m_dataType
										   );

		if (QP_SUCCESS(qpStatus))
			This_p->m_bPaused = QP_FALSE;
	}
	else
	{
		// reset counter and flags
		//
		This_p->m_ullCntBytes = 0;
		This_p->m_savedNumberBytes = 0;

		qpStatus = This_p->m_pTask->Start(This_p->m_pTask,
										  This_p->m_hTask,
										  (TASK_DATA_TYPE)This_p->m_dataType
										  );
		This_p->m_llStartTime = QPTMGetSysTimeMilliS();
	}

	This_p->m_bFlushing = QP_FALSE;

	if (QP_SUCCESS(qpStatus))
	{
		This_p->m_State = QPSTATE_RUN;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CMPEGInChannel_Start() Failed Status(%d)\n",
					 qpStatus
					 ));
	}

	return (qpStatus);
}



static EQP_Errors CMPEGInChannel_Stop(CChannel *This_p)
{
	EQP_Errors	qpStatus;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMPEGInChannel_Stop() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 This_p->m_dataType
				 ));

	// stop the decoder
	//
	qpStatus = This_p->m_pTask->Stop(This_p->m_pTask,
									 This_p->m_hTask,
									 (TASK_DATA_TYPE)This_p->m_dataType
									 );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
					 "CMPEGInChannel_Stop() m_pTask->Stop failed(%d)\n",
					 qpStatus
					 ));
	}

	// clear falgs
	//
	This_p->m_bPaused = QP_FALSE;

	// change state to stop
	//
	This_p->m_State = QPSTATE_STOP;

	// flush the data
	//
	This_p->BeginFlush(This_p);
	This_p->Flush(This_p);
	This_p->EndFlush(This_p);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
				 "CMPEGInChannel_Stop() data received(%d)\n",
				 This_p->m_ullCntBytes
				 ));

	return (QPERR_SUCCESS);
}



static EQP_Errors CMPEGInChannel_Pause(CChannel *This_p)
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMPEGInChannel_Pause() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 This_p->m_dataType
				 ));

	if (QPSTATE_RUN == This_p->m_State)
	{
		qpStatus = This_p->m_pTask->Pause(This_p->m_pTask,
										  This_p->m_hTask,
										  (TASK_DATA_TYPE)This_p->m_dataType
										  );
		This_p->m_bPaused = QP_SUCCESS(qpStatus);
	}

	if (QP_SUCCESS(qpStatus))
	{
		This_p->m_State = QPSTATE_PAUSE;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CMPEGInChannel_Pause() m_pTask->Pause failed(%d)\n",
					 qpStatus
					 ));
	}

	return (qpStatus);
}



static EQP_Errors CMPEGInChannel_SetRate(CChannel *This_p,
										 TQP_INT32 lNewRate
										 )
{
	return (QPERR_SUCCESS);
}



static EQP_Errors CMPEGInChannel_GetRate(CChannel *This_p,
										 PTQP_INT32 plRate
										 )
{
	return (QPERR_SUCCESS);
}



// called from streaming task
//
static TQP_BOOL CMPEGInChannel_GetBuffer(CChannel *This_p,
										 PQP_BUFFER_DESCRIPTOR *ppBufDesc,
										 PTQP_UINT8 *ppBuffer,
										 PTQP_UINT32 pSize
										 )
{
	QUEUE_ENTRY				*pEntry;
	PQP_BUFFER_DESCRIPTOR	pBufDesc;
	PQP_KSSTREAM_HEADER		pDataPacket;
	CQLCodec				*pQlCodec = GET_INHERITED_OBJECT(CQLCodec, m_Object, This_p->m_Object.m_pParent);
	TQP_UINT32				pktSize;
	PTQP_UINT8				pBuffer;


	// take the buffer from the request queue
	//
	pEntry = CQueue_GetOneEntry(This_p->m_pDataRequestQueue);

	if (QP_NULL != pEntry)
	{
		pBufDesc = (PQP_BUFFER_DESCRIPTOR)pEntry->Data;
		pDataPacket = pBufDesc->DataBufferArray;
		pktSize = (pDataPacket->DataUsed + This_p->m_savedNumberBytes) & 0xFFFFFFFC;

		// clear the private portion of the flags
		//
		pBufDesc->ulFlags &= ~QP_BUFDESC_FLAG_MASK_PRIVATE;

		// find end of stream indication
		//
		if (pDataPacket->OptionsFlags & QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CMPEGInChannel_GetBuffer() QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM\n"
						 ));
			pBufDesc->ulFlags |= QP_BUFDESC_FLAG_EOS;
		}

		if ((0 != This_p->m_savedNumberBytes) &&
			(0 != pktSize)
			)
		{
			// need to allocate dword aligned buffer
			//
			pBufDesc->pBuffer = PLvmalloc(pktSize);
			if (QP_NULL == pBufDesc->pBuffer)
			{
				QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
							 "CMPEGInChannel_GetBuffer() can not allocate dword aligned buffer for (%d) bytes\n",
							 pktSize
							 ));
				return (QP_FALSE);
			}

			pBuffer = (PTQP_UINT8) pBufDesc->pBuffer;

			// copy left over bytes to the dword aligned buffer
			memcpy(pBuffer,
				   &This_p->m_savedDword[0],
				   This_p->m_savedNumberBytes
				   );

			// copy user data to the dowrd aligned buffer
			memcpy(pBuffer + This_p->m_savedNumberBytes,
				   (PTQP_UINT8)pDataPacket->Data,
				   pktSize - This_p->m_savedNumberBytes
				   );
		}
		else
		{
			pBuffer = (PTQP_UINT8)pDataPacket->Data;
		}

		if ((0 == pktSize) &&
			(0 != pDataPacket->DataUsed)
			)
		{
			memcpy(&This_p->m_savedDword[This_p->m_savedNumberBytes],
				   (PTQP_UINT8)pDataPacket->Data,
				   pDataPacket->DataUsed
				   );
			This_p->m_savedNumberBytes += pDataPacket->DataUsed;

			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CMPEGInChannel_GetBuffer() buffer to small(%d) to send saved(%d) bytes\n",
						 pDataPacket->DataUsed,
						 This_p->m_savedNumberBytes
						 ));
		}

		if (This_p->m_bByteSwap &&
			!QPHCI_CanSwapData(&pQlCodec->m_hci) &&
			(0 != pktSize)
			)
		{
			BytesSwap(pBuffer,
					  pBuffer,
					  pktSize
					  );
		}

		pBufDesc->ulBufferIndex = 0;
		pBufDesc->ulBufferOffset = 0;
		pBufDesc->ulTotalUsed = 0;

		*ppBuffer = pBuffer;
		*pSize = pktSize;
		*ppBufDesc = pBufDesc;

		// remember left over bytes
		//
		if (pktSize)
		{
			pktSize -= This_p->m_savedNumberBytes;
			This_p->m_savedNumberBytes = pDataPacket->DataUsed - pktSize;
			if (This_p->m_savedNumberBytes)
			{
				memcpy(&This_p->m_savedDword[0],
					   ((PTQP_UINT8)pDataPacket->Data) + pktSize,
					   This_p->m_savedNumberBytes
					   );
			}
		}

		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
					 "CMPEGInChannel_GetBuffer() orignal_size(%d) buffer(0x%x) size(%d) Desc(0x%x) saved(%d)\n",
					 pDataPacket->DataUsed,
					 *ppBuffer,
					 *pSize,
					 *ppBufDesc,
					 This_p->m_savedNumberBytes
					 ));

		// move this entry to the pending queue
		//
		CQueue_AddEntry(This_p->m_pDataPendingQueue,
						pEntry
						);
	}

	return (QP_NULL != pEntry);
}



static EQP_Errors CMPEGInChannel_AddBufferMjpeg(CChannel *This,
												PQP_BUFFER_DESCRIPTOR pBufDesc
												)
{
	PQP_KSSTREAM_HEADER		pDataPacket;
	unsigned char 			*pucaBuffer;
#ifndef QP_LINXKER
	unsigned int			unFrameEndOffset = 0;
	unsigned int			unFoundFrameEnd = 0;
#endif

	if (pBufDesc && !This->m_bFlushing)
	{
		pDataPacket = pBufDesc->DataBufferArray;
		pucaBuffer = (unsigned char *)pDataPacket->Data;

		if (pDataPacket->DataUsed >= 4)
		{
			/* check frame start */
			if ((0xff != *pucaBuffer) || (0xd8 != *(pucaBuffer+1)))
			{
				pBufDesc->Status = QPERR_SUCCESS;

				QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
							 "CMPEGInChannel_AddBufferMjpeg() Discard incompleted jpeg frame! BuffAddr(0x%08x) nLen(%d) Header(0x%02x 0x%02x)",
							 pucaBuffer,
							 pDataPacket->DataUsed,
							 *pucaBuffer,
							 *(pucaBuffer+1)
							 ));

				// discard the incomplted jpeg frame
				return (CChannel_DeviceCallback(This, QPMPGCODEC_CMD_DONE_DATA, pBufDesc));
			}

#ifndef QP_LINXKER
			/* 2009.0929, temporary disable the implmentation in Linux environment for performance concern */

			/* check frame end,
			 * QC4/QC8 outputs jpeg data in 8 bytes alignment.
			 * and could append some garbage after frame end (till 8 bytes alignment) */
			unFrameEndOffset = 0;
			while ((pDataPacket->DataUsed > unFrameEndOffset) && (unFrameEndOffset < 8))
			{
				if ((0xff == *(pucaBuffer + pDataPacket->DataUsed-unFrameEndOffset - 2)) &&
					(0xd9 == *(pucaBuffer + pDataPacket->DataUsed-unFrameEndOffset - 1))
					)
				{
					unFoundFrameEnd = 1;
					break;
				}
				else
				{
					unFrameEndOffset++;
				}
			}

			if (!unFoundFrameEnd)
			{
				pBufDesc->Status = QPERR_SUCCESS;

				QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
							 "CMPEGInChannel_AddBufferMjpeg() Discard incompleted jpeg frame! BuffAddr(0x%08x) nLen(%d) Trailer(0x%02x 0x%02x)",
							 pucaBuffer,
							 pDataPacket->DataUsed,
							 *(pucaBuffer+pDataPacket->DataUsed-2),
							 *(pucaBuffer+pDataPacket->DataUsed-1)
							 ));

				/* discard the incomplted jpeg frame */
				return (CChannel_DeviceCallback(This, QPMPGCODEC_CMD_DONE_DATA, pBufDesc));
			}
			else
			{
				pDataPacket->DataUsed -= unFrameEndOffset;
			}
#endif //!QP_LINXKER
		}
	}
	return (CChannel_AddBuffer(This,
							   pBufDesc
							   ));
}



static TQP_BOOL CMPEGInChannel_GetBufferMjpeg(CChannel *This_p,
											  PQP_BUFFER_DESCRIPTOR *ppBufDesc,
											  PTQP_UINT8 *ppBuffer,
											  PTQP_UINT32 pSize
											  )
{
	QUEUE_ENTRY				*pEntry;
	PQP_BUFFER_DESCRIPTOR	pBufDesc;
	PQP_KSSTREAM_HEADER		pDataPacket;
	CQLCodec				*pQlCodec = GET_INHERITED_OBJECT(CQLCodec, m_Object, This_p->m_Object.m_pParent);
	TQP_UINT32				pktSize;

	// take the buffer from the request queue
	//
	pEntry = CQueue_GetOneEntry(This_p->m_pDataRequestQueue);

	if (QP_NULL != pEntry)
	{
		pBufDesc = (PQP_BUFFER_DESCRIPTOR)pEntry->Data;
		pDataPacket = pBufDesc->DataBufferArray;
#if 1
		pBufDesc->unAlignedNumberBytes = 0;
		pktSize = (pDataPacket->DataUsed + 3) & 0xFFFFFFFC;
#else
		pBufDesc->unAlignedNumberBytes = pDataPacket->DataUsed & 0x0000003;
		pktSize = pDataPacket->DataUsed - pBufDesc->unAlignedNumberBytes;
#endif

		// clear the private portion of the flags
		//
		pBufDesc->ulFlags &= ~QP_BUFDESC_FLAG_MASK_PRIVATE;

		// find end of stream indication
		//
		if (pDataPacket->OptionsFlags & QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CMPEGInChannel_GetBufferMjpeg() QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM\n"
						 ));
			pBufDesc->ulFlags |= QP_BUFDESC_FLAG_EOS;
		}

		pBufDesc->ulFlags |= (QP_BUFDESC_FLAG_FRAME_START | QP_BUFDESC_FLAG_FRAME_END);

		pBufDesc->ulBufferIndex = 0;
		pBufDesc->ulBufferOffset = 0;
		pBufDesc->ulTotalUsed = 0;

		*ppBuffer = (PTQP_UINT8)pDataPacket->Data;
		*pSize = pktSize;
		*ppBufDesc = pBufDesc;

		if (This_p->m_bByteSwap &&
			!QPHCI_CanSwapData(&pQlCodec->m_hci) &&
			(0 != pktSize)
			)
		{
			PTQP_UINT8	pBuffer = (PTQP_UINT8)pDataPacket->Data;

			BytesSwap(pBuffer,
					  pBuffer,
					  pktSize
					  );
		}

		// remember odd bytes
		//
		if (pBufDesc->unAlignedNumberBytes)
		{
			pBufDesc->unAlignedData = 0;
			memcpy(&pBufDesc->unAlignedData,
				   ((PTQP_UINT8)pDataPacket->Data) + pktSize,
				   pBufDesc->unAlignedNumberBytes
				   );
			if (This_p->m_bByteSwap)
			{
				BytesSwap((PTQP_UINT8)&pBufDesc->unAlignedData,
						  (PTQP_UINT8)&pBufDesc->unAlignedData,
						  sizeof(TQP_UINT32)
						  );
			}
		}

		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
					 "CMPEGInChannel_GetBufferMjpeg() original_size(%d) data(0x%x) buffer(0x%x) size(%d) Desc(0x%x) save (%d) bytes (0x%x)\n",
					 pDataPacket->DataUsed,
					 (pDataPacket->DataUsed >= 4) ? *((PTQP_UINT32)pDataPacket->Data) : 0,
					 *ppBuffer,
					 *pSize,
					 *ppBufDesc,
					 pBufDesc->unAlignedNumberBytes,
					 pBufDesc->unAlignedData
					 ));

		// move this entry to the pending queue
		//
		CQueue_AddEntry(This_p->m_pDataPendingQueue,
						pEntry
						);
	}

	return (QP_NULL != pEntry);
}



static void CMPEGInChannel_CompleteBuffer(CChannel *This_p,
										  PQP_BUFFER_DESCRIPTOR pBufDesc
										  )
{
	QUEUE_ENTRY	*pEntry;

	////QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
	////			 "CMPEGInChannel_CompleteBuffer()\n"
	////			 ));

	// get the buffer entry
	//
	pEntry = CQueue_GetEntryByData(This_p->m_pDataPendingQueue,
								   (TQP_PVOID)pBufDesc
								   );
	if (pEntry)
	{
		if (!QP_SUCCESS(pBufDesc->Status))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CMPEGInChannel_CompleteBuffer(%d) Status(%d)\n",
						 This_p->m_hChannel,
						 pBufDesc->Status
						 ));
		}

		This_p->m_ullCntBytes += pBufDesc->ulTotalUsed;

		if (pBufDesc->DataBufferArray->OptionsFlags & QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
						 "CMPEGInChannel_CompleteBuffer(%d) QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM\n",
						 This_p->m_hChannel
						 ));
		}

		if (QP_NULL != pBufDesc->pBuffer)
		{
			PLvfree(pBufDesc->pBuffer);
			pBufDesc->pBuffer = QP_NULL;
		}

		// clear the entry content
		//
		pEntry->Data = QP_NULL;

		// move to the free queue
		//
		CQueue_AddEntry(This_p->m_pFreeQueue,
						pEntry
						);

		// call back the api
		//
		CChannel_DeviceCallback(This_p,
								QPMPGCODEC_CMD_DONE_DATA,
								pBufDesc
								);
	}
}


// constructor
//
CMPEGInChannel * CMPEGInChannel_Constructor(CMPEGInChannel *pMpgInChannel,
											CObject *pParent,
											TQP_UINT32 dwAttributes,
											TQP_OBJECT_HANDLE hTask,
											CTask *pTask
											)
{
	if (pMpgInChannel)
	{
		CChannel	*pChannel;

		pChannel = CChannel_Constructor(&pMpgInChannel->m_Channel,
										pParent,
										dwAttributes,
										hTask,
										TASK_DATA_TYPE_COMP_VID,
										FWDataTypeCmprSysVidInput,
										CHANNEL_DIR_WRITE,
										QPMPGCODEC_DEC_MPEG_IN,
										pTask
										);
		if (!pChannel)
		{
			pMpgInChannel = QP_NULL;
		}
		else
		{
			if (BYTES_SWAP_BY_DRIVER == pTask->m_TaskData[hTask].m_streamID.RegEx.bytes_swap)
			{
				// BYTES_SWAP_BY_DRIVER
				pMpgInChannel->m_Channel.m_bByteSwap = QP_TRUE;
			}
			else
			{
				// BYTES_SWAP_BY_FW
				pMpgInChannel->m_Channel.m_bByteSwap = QP_FALSE;
			}

			// fill in the functions
			//
			pMpgInChannel->m_Channel.Open = CMPEGInChannel_Open;
			pMpgInChannel->m_Channel.Start = CMPEGInChannel_Start;
			pMpgInChannel->m_Channel.Stop = CMPEGInChannel_Stop;
			pMpgInChannel->m_Channel.Pause = CMPEGInChannel_Pause;
			pMpgInChannel->m_Channel.SetRate = CMPEGInChannel_SetRate;
			pMpgInChannel->m_Channel.GetRate = CMPEGInChannel_GetRate;
			pMpgInChannel->m_Channel.GetBuffer = CMPEGInChannel_GetBuffer;
			pMpgInChannel->m_Channel.CompleteBuffer = CMPEGInChannel_CompleteBuffer;
		}
	}

	return (pMpgInChannel);
}



// destructor
//
void CMPEGInChannel_Destructor(CMPEGInChannel *This)
{
	CChannel_Destructor(&This->m_Channel);
}



/////////////////////////////////////////////////////////////////////////////
//
//

// CAESInChannel
//

// CChannel
//

// constructor
//
CAESInChannel * CAESInChannel_Constructor(CAESInChannel *pAesInChannel,
										  CObject *pParent,
										  TQP_UINT32 dwAttributes,
										  TQP_OBJECT_HANDLE hTask,
										  CTask *pTask
										  )
{
	if (pAesInChannel)
	{
		CChannel	*pChannel;

		pChannel = CChannel_Constructor(&pAesInChannel->m_Channel,
										pParent,
										dwAttributes,
										hTask,
										TASK_DATA_TYPE_COMP_AUD,
										FWDataTypeCmprAudInput,
										CHANNEL_DIR_WRITE,
										QPMPGCODEC_DEC_MPEG_IN,
										pTask
										);
		if (!pChannel)
		{
			pAesInChannel = QP_NULL;
		}
		else
		{
			pAesInChannel->m_Channel.m_bByteSwap = QP_TRUE;

			// fill in the functions
			//
			pAesInChannel->m_Channel.Start = CMPEGInChannel_Start;
			pAesInChannel->m_Channel.Stop = CMPEGInChannel_Stop;
			pAesInChannel->m_Channel.Pause = CMPEGInChannel_Pause;
			pAesInChannel->m_Channel.GetBuffer = CMPEGInChannel_GetBuffer;	// same as MpegIn
			pAesInChannel->m_Channel.CompleteBuffer = CMPEGInChannel_CompleteBuffer;
		}
	}

	return (pAesInChannel);
}



// destructor
//
void CAESInChannel_Destructor(CAESInChannel *This)
{
	CChannel_Destructor(&This->m_Channel);
}



/////////////////////////////////////////////////////////////////////////////
//
//

// CYUVInChannel
//

// CChannel
//

static EQP_Errors CYUVInChannel_Open(CChannel *This_p,
									 TQP_OBJECT_HANDLE hChannel,
									 TQP_UINT32 dwFlags,
									 TQP_PVOID pDataFormat,
									 QP_DEVICE_CALLBACK pFuncCallback,
									 TQP_PVOID context
									 )
{
	EQP_Errors			qpStatus;
	PQP_YUV_DATAFORMAT	pYUVFormat = (PQP_YUV_DATAFORMAT)pDataFormat;
	CYUVInChannel		*This =	GET_INHERITED_OBJECT(CYUVInChannel, m_Channel, This_p);

	// Gather information we need for this stream
	//
	This->m_lWidth = pYUVFormat->nWidth;
	This->m_lHeight = pYUVFormat->nHeight;
	This->m_nBitCount = pYUVFormat->nBitCount;
	This->m_nDataType = pYUVFormat->nDataType;

	This->m_dwFrameSize = (This->m_lWidth * This->m_lHeight * This->m_nBitCount) / 8;

	This->m_Channel.m_bByteSwap = ((12 == This->m_nBitCount) &&
								  ((QP_YUV_DATA_TYPE_YV12 == This->m_nDataType) ||
								  (QP_YUV_DATA_TYPE_I420 == This->m_nDataType))
								  );
////	This->m_Channel.m_bByteSwap = (12 == This->m_nBitCount);

	if (16 == This->m_nBitCount)
	{
		This->m_Channel.m_FWDataType = FWDataTypeRawYUYVInput;
	}
	else
	{
		This->m_Channel.m_FWDataType = FWDataTypeRawVidInput;
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CYUVInChannel_Open() m_dwFrameSize(%d) m_lWidth(%d) m_lHeight(%d) bitCnt(%d) swap(%d) m_nDataType(%d)\n",
				 This->m_dwFrameSize,
				 This->m_lWidth,
				 This->m_lHeight,
				 This->m_nBitCount,
				 This->m_Channel.m_bByteSwap,
				 This->m_nDataType
				 ));

	// reset counters and flags
	//
	This->m_ulVideoFramesCount = 0;

	// generic open
	//
	qpStatus = CChannel_Open(This_p,
							 hChannel,
							 dwFlags,
							 pDataFormat,
							 pFuncCallback,
							 context
							 );

	return (qpStatus);
}



static EQP_Errors CYUVInChannel_Start(CChannel *This_p)
{
	EQP_Errors		qpStatus;
	CYUVInChannel	*This =	GET_INHERITED_OBJECT(CYUVInChannel, m_Channel, This_p);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CYUVInChannel_Start() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 This_p->m_dataType
				 ));

	if (This_p->m_bPaused)
	{
		qpStatus = This_p->m_pTask->Resume(This_p->m_pTask,
										   This_p->m_hTask,
										   (TASK_DATA_TYPE)This_p->m_dataType
										   );

		if (QP_SUCCESS(qpStatus))
			This_p->m_bPaused = QP_FALSE;
	}
	else
	{
		// reset counter and flags
		//
		This_p->m_ullCntBytes = 0;
		This->m_ulVideoFramesCount = 0;

		// start the encoder
		//
		qpStatus = This_p->m_pTask->Start(This_p->m_pTask,
										  This_p->m_hTask,
										  (TASK_DATA_TYPE)This_p->m_dataType
										  );

		This_p->m_llStartTime = QPTMGetSysTimeMilliS();
	}

	This_p->m_bFlushing = QP_FALSE;

	if (QP_SUCCESS(qpStatus))
	{
		This_p->m_State = QPSTATE_RUN;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CYUVInChannel_Start() Failed Status(%d)\n",
					 qpStatus
					 ));
	}

	return (qpStatus);
}



// called from streaming task
//
static TQP_BOOL CYUVInChannel_GetBuffer(CChannel *This_p,
										PQP_BUFFER_DESCRIPTOR *ppBufDesc,
										PTQP_UINT8 *ppBuffer,
										PTQP_UINT32 pSize
										)
{
	QUEUE_ENTRY				*pEntry;
	PQP_BUFFER_DESCRIPTOR	pBufDesc;
	PQP_KSSTREAM_HEADER		pDataPacket;
	CYUVInChannel			*This =	GET_INHERITED_OBJECT(CYUVInChannel, m_Channel, This_p);

	pEntry = CQueue_GetOneEntry(This_p->m_pDataRequestQueue);

	if (QP_NULL != pEntry)
	{
		pBufDesc = (PQP_BUFFER_DESCRIPTOR)pEntry->Data;
		pDataPacket = pBufDesc->DataBufferArray;

		// clear the private portion of the flags
		//
		pBufDesc->ulFlags &= ~QP_BUFDESC_FLAG_MASK_PRIVATE;

		if ((pDataPacket->OptionsFlags & QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM) ||
			(0 == pBufDesc->ulBufferSize)
			)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
						 "CYUVInChannel_GetBuffer() QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM\n"
						 ));
			pBufDesc->ulFlags |= QP_BUFDESC_FLAG_EOS;
		}

		if (pBufDesc->ulBufferSize < This->m_dwFrameSize)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CYUVInChannel_GetBuffer() buffer size(%d) < frame size(%d)\n",
						 pBufDesc->ulBufferSize,
						 This->m_dwFrameSize
						 ));
			// clear the entry content
			//
			pEntry->Data = QP_NULL;

			// move to free queue
			//
			CQueue_AddEntry(This_p->m_pFreeQueue,
							pEntry
							);

			// nil pEntry so the function will return FALSE
			//
			pEntry = QP_NULL;

			// default success
			//
			pBufDesc->Status = QPERR_SUCCESS;

			// call back the api
			//
			CChannel_DeviceCallback(This_p,
									QPMPGCODEC_CMD_DONE_DATA,
									pBufDesc
									);
		}
		else
		{
			if (pBufDesc->ulFlags & QP_BUFDESC_FLAG_PTS)
			{
				QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_EXTRACE),
							 "CYUVInChannel_GetBuffer() PTS(%u) KSTime(%lu:%lu) Numerator(%d) Denominator(%d)\n",
							 pBufDesc->ulPTS,
							 (TQP_UINT32)((pDataPacket->PresentationTime.Time >> 32) & 0xFFFFFFFF),
							 (TQP_UINT32)(pDataPacket->PresentationTime.Time & 0xFFFFFFFF),
							 pDataPacket->PresentationTime.Numerator,
							 pDataPacket->PresentationTime.Denominator
							 ));
			}

			pBufDesc->ulBufferIndex = 0;
			pBufDesc->ulBufferOffset = 0;
			pBufDesc->ulTotalUsed = 0;

			*ppBuffer = (PTQP_UINT8)pDataPacket->Data;
			*pSize = pBufDesc->ulBufferSize;
			*ppBufDesc = pBufDesc;

			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
						 "CYUVInChannel_GetBuffer() buffer(0x%x) size(%d) Desc(0x%x) frame(%d)\n",
						 *ppBuffer,
						 *pSize,
						 *ppBufDesc,
						 This->m_ulVideoFramesCount
						 ));

			// move this entry to the pending queue
			//
			CQueue_AddEntry(This_p->m_pDataPendingQueue,
							pEntry
							);
		}
	}

	return (QP_NULL != pEntry);
}



static TQP_BOOL CYUVInChannel_GetBufferYUV(CChannel *This_p,
										   PQP_BUFFER_DESCRIPTOR *ppBufDesc,
										   PTQP_UINT8 *ppYBuffer,
										   PTQP_UINT32 pYSize,
										   PTQP_UINT8 *ppUVBuffer,
										   PTQP_UINT32 pUVSize
										   )
{
	QUEUE_ENTRY				*pEntry;
	PQP_BUFFER_DESCRIPTOR	pBufDesc;
	PQP_KSSTREAM_HEADER		pDataPacket;
	CYUVInChannel			*This =	GET_INHERITED_OBJECT(CYUVInChannel, m_Channel, This_p);

	pEntry = CQueue_GetOneEntry(This_p->m_pDataRequestQueue);

	if (QP_NULL != pEntry)
	{
		pBufDesc = (PQP_BUFFER_DESCRIPTOR)pEntry->Data;
		pDataPacket = pBufDesc->DataBufferArray;

		// clear the private portion of the flags
		//
		pBufDesc->ulFlags &= ~QP_BUFDESC_FLAG_MASK_PRIVATE;

		if ((pDataPacket->OptionsFlags & QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM) ||
			(0 == pBufDesc->ulBufferSize)
			)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
						 "CYUVInChannel_GetBufferYUV() QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM\n"
						 ));
			pBufDesc->ulFlags |= QP_BUFDESC_FLAG_EOS;
		}

		if (pBufDesc->ulBufferSize < This->m_dwFrameSize)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CYUVInChannel_GetBufferYUV() buffer size(%d) < frame size(%d)\n",
						 pBufDesc->ulBufferSize,
						 This->m_dwFrameSize
						 ));
			// clear the entry content
			//
			pEntry->Data = QP_NULL;

			// move to free queue
			//
			CQueue_AddEntry(This_p->m_pFreeQueue,
							pEntry
							);

			// nil pEntry so the function will return FALSE
			//
			pEntry = QP_NULL;

			// default success
			//
			pBufDesc->Status = QPERR_SUCCESS;

			// call back the api
			//
			CChannel_DeviceCallback(This_p,
									QPMPGCODEC_CMD_DONE_DATA,
									pBufDesc
									);
		}
		else
		{
			if (pBufDesc->ulFlags & QP_BUFDESC_FLAG_PTS)
			{
				QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_EXTRACE),
							 "CYUVInChannel_GetBufferYUV() PTS(%u) KSTime(%lu:%lu) Numerator(%d) Denominator(%d)\n",
							 pBufDesc->ulPTS,
							 (TQP_UINT32)((pDataPacket->PresentationTime.Time >> 32) & 0xFFFFFFFF),
							 (TQP_UINT32)(pDataPacket->PresentationTime.Time & 0xFFFFFFFF),
							 pDataPacket->PresentationTime.Numerator,
							 pDataPacket->PresentationTime.Denominator
							 ));
			}

			pBufDesc->ulBufferIndex = 0;
			pBufDesc->ulBufferOffset = 0;
			pBufDesc->ulTotalUsed = 0;

			*ppYBuffer = (PTQP_UINT8)pDataPacket->Data;
			*pYSize = This->m_lWidth * This->m_lHeight;
			*ppUVBuffer = ((PTQP_UINT8)pDataPacket->Data) + (*pYSize);
			*pUVSize = (*pYSize) >> 1;
			*ppBufDesc = pBufDesc;

			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_EXTRACE),
						 "CYUVInChannel_GetBufferYUV() Y(0x%x) size(%d) UV(0x%x) size(%d) frame(%d)\n",
						 *ppYBuffer,
						 *pYSize,
						 *ppUVBuffer,
						 *pUVSize,
						 This->m_ulVideoFramesCount
						 ));

			// move this entry to the pending queue
			//
			CQueue_AddEntry(This_p->m_pDataPendingQueue,
							pEntry
							);
		}
	}

	return (QP_NULL != pEntry);
}



static TQP_BOOL CYUVInChannel_GetBufferRas(CChannel *This_p,
										   PQP_BUFFER_DESCRIPTOR *ppBufDesc,
										   PTQP_UINT8 *ppYBuffer,
										   PTQP_UINT32 pYSize,
										   PTQP_UINT8 *ppUBuffer,
										   PTQP_UINT32 pUSize,
										   PTQP_UINT8 *ppVBuffer,
										   PTQP_UINT32 pVSize
										   )
{
	QUEUE_ENTRY				*pEntry;
	PQP_BUFFER_DESCRIPTOR	pBufDesc;
	PQP_KSSTREAM_HEADER		pDataPacket;
	CYUVInChannel			*This =	GET_INHERITED_OBJECT(CYUVInChannel, m_Channel, This_p);
	CQLCodec				*pQlCodec = GET_INHERITED_OBJECT(CQLCodec, m_Object, This_p->m_Object.m_pParent);

	pEntry = CQueue_GetOneEntry(This_p->m_pDataRequestQueue);

	if (QP_NULL != pEntry)
	{
		pBufDesc = (PQP_BUFFER_DESCRIPTOR)pEntry->Data;
		pDataPacket = pBufDesc->DataBufferArray;

		// clear the private portion of the flags
		//
		pBufDesc->ulFlags &= ~QP_BUFDESC_FLAG_MASK_PRIVATE;

		if ((pDataPacket->OptionsFlags & QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM) ||
			(0 == pBufDesc->ulBufferSize)
			)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
						 "CYUVInChannel_GetBufferRas() QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM\n"
						 ));
			pBufDesc->ulFlags |= QP_BUFDESC_FLAG_EOS;
		}

		if (pBufDesc->ulBufferSize < This->m_dwFrameSize)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CYUVInChannel_GetBufferRas() buffer size(%d) < frame size(%d)\n",
						 pBufDesc->ulBufferSize,
						 This->m_dwFrameSize
						 ));
			// clear the entry content
			//
			pEntry->Data = QP_NULL;

			// move to free queue
			//
			CQueue_AddEntry(This_p->m_pFreeQueue,
							pEntry
							);

			// nil pEntry so the function will return FALSE
			//
			pEntry = QP_NULL;

			// default success
			//
			pBufDesc->Status = QPERR_SUCCESS;

			// call back the api
			//
			CChannel_DeviceCallback(This_p,
									QPMPGCODEC_CMD_DONE_DATA,
									pBufDesc
									);
		}
		else
		{
			if (pBufDesc->ulFlags & QP_BUFDESC_FLAG_PTS)
			{
				QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_EXTRACE),
							 "CYUVInChannel_GetBufferRas() PTS(%u) KSTime(%lu:%lu) Numerator(%d) Denominator(%d)\n",
							 pBufDesc->ulPTS,
							 (TQP_UINT32)((pDataPacket->PresentationTime.Time >> 32) & 0xFFFFFFFF),
							 (TQP_UINT32)(pDataPacket->PresentationTime.Time & 0xFFFFFFFF),
							 pDataPacket->PresentationTime.Numerator,
							 pDataPacket->PresentationTime.Denominator
							 ));
			}

			if (This_p->m_bByteSwap &&
				!QPHCI_CanSwapData(&pQlCodec->m_hci) &&
				(0 != pBufDesc->ulBufferSize)
				)
			{
				BytesSwap((PTQP_UINT8)pDataPacket->Data,
						  (PTQP_UINT8)pDataPacket->Data,
						  pBufDesc->ulBufferSize
						  );
			}

			pBufDesc->ulBufferIndex = 0;
			pBufDesc->ulBufferOffset = 0;
			pBufDesc->ulTotalUsed = 0;

			*ppYBuffer = (PTQP_UINT8)pDataPacket->Data;
			*pYSize = This->m_lWidth * This->m_lHeight;
			*pUSize = *pVSize = ((*pYSize) >> 2);
			if (QP_YUV_DATA_TYPE_YV12 == This->m_nDataType)
			{
				*ppVBuffer = ((PTQP_UINT8)pDataPacket->Data) + (*pYSize);
				*ppUBuffer = (*ppVBuffer) + (*pVSize);
			}
			else
			{
				*ppUBuffer = ((PTQP_UINT8)pDataPacket->Data) + (*pYSize);
				*ppVBuffer = (*ppUBuffer) + (*pUSize);
			}
			*ppBufDesc = pBufDesc;

			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_EXTRACE),
						 "CYUVInChannel_GetBufferRas() Y(0x%x) size(%d) U(0x%x) size(%d) V(0x%x) size(%d) frame(%d)\n",
						 *ppYBuffer,
						 *pYSize,
						 *ppUBuffer,
						 *pUSize,
						 *ppVBuffer,
						 *pVSize,
						 This->m_ulVideoFramesCount
						 ));

			// move this entry to the pending queue
			//
			CQueue_AddEntry(This_p->m_pDataPendingQueue,
							pEntry
							);
		}
	}

	return (QP_NULL != pEntry);
}



static void CYUVInChannel_CompleteBuffer(CChannel *This_p,
										 PQP_BUFFER_DESCRIPTOR pBufDesc
										 )
{
	QUEUE_ENTRY			*pEntry;
	PQP_KSSTREAM_HEADER	pDataPacket;
	CYUVInChannel		*This =	GET_INHERITED_OBJECT(CYUVInChannel, m_Channel, This_p);


	////QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
	////			 "CYUVInChannel_CompleteBuffer()\n"
	////			 ));

	// get the buffer entry
	//
	pEntry = CQueue_GetEntryByData(This_p->m_pDataPendingQueue,
								   (TQP_PVOID)pBufDesc
								   );
	if (pEntry)
	{
		pDataPacket = pBufDesc->DataBufferArray;

		if (!QP_SUCCESS(pBufDesc->Status))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CYUVInChannel_CompleteBuffer() Status(%d)\n",
						 pBufDesc->Status
						 ));
		}
		else
		{
			// increment total frame consumed
			//
			This->m_ulVideoFramesCount++;
		}

		// clear the entry content
		//
		pEntry->Data = QP_NULL;

		// move to free queue
		//
		CQueue_AddEntry(This_p->m_pFreeQueue,
						pEntry
						);

		// call back to the api
		//
		CChannel_DeviceCallback(This_p,
								QPMPGCODEC_CMD_DONE_DATA,
								pBufDesc
								);
	}
}



static EQP_Errors CYUVInChannel_GetResolution(CChannel *This_p,
											  PTQP_UINT32 pWidth,
											  PTQP_UINT32 pHeight
											  )
{
	CYUVInChannel *This = GET_INHERITED_OBJECT(CYUVInChannel, m_Channel, This_p);

	if (pWidth &&
		pHeight
		)
	{
		*pWidth = (TQP_UINT32)This->m_lWidth;
		*pHeight = (TQP_UINT32)This->m_lHeight;
		return(QPERR_SUCCESS);
	}
	else
	{
		return (QPERR_PARMS);
	}
}



static EQP_Errors CYUVInChannel_GetYUVFormat(CChannel *This_p,
											 PTQP_UINT32 pYUVFormat
											 )
{
	CYUVInChannel *This = GET_INHERITED_OBJECT(CYUVInChannel, m_Channel, This_p);

	if (pYUVFormat)
	{
		*pYUVFormat = This->m_nDataType;
		return(QPERR_SUCCESS);
	}
	else
	{
		return (QPERR_PARMS);
	}
}



// constructor
//
CYUVInChannel * CYUVInChannel_Constructor(CYUVInChannel *pYUVInChannel,
										  CObject *pParent,
										  TQP_UINT32 dwAttributes,
										  TQP_OBJECT_HANDLE hTask,
										  CTask *pTask
										  )
{
	if (pYUVInChannel)
	{
		CChannel	*pChannel;

		pChannel = CChannel_Constructor(&pYUVInChannel->m_Channel,
										pParent,
										dwAttributes,
										hTask,
										TASK_DATA_TYPE_YUV,
										FWDataTypeRawVidInput,
										CHANNEL_DIR_WRITE,
										QPMPGCODEC_ENC_YUV_IN,
										pTask
										);
		if (!pChannel)
		{
			pYUVInChannel = QP_NULL;
		}
		else
		{
			pYUVInChannel->m_Channel.m_bByteSwap = QP_TRUE;

			// fill in the functions table
			//
			pYUVInChannel->m_Channel.Open = CYUVInChannel_Open;
			pYUVInChannel->m_Channel.Start = CYUVInChannel_Start;
			pYUVInChannel->m_Channel.Stop = CMPEGOutChannel_Stop;	// same as MpegOut
			pYUVInChannel->m_Channel.Pause = CMPEGOutChannel_Pause;	// same as MpegOut
			pYUVInChannel->m_Channel.GetBuffer = CYUVInChannel_GetBuffer;
			pYUVInChannel->m_Channel.GetBufferYUV = CYUVInChannel_GetBufferYUV;
			pYUVInChannel->m_Channel.GetBufferYUVRAS = CYUVInChannel_GetBufferRas;
			pYUVInChannel->m_Channel.CompleteBuffer = CYUVInChannel_CompleteBuffer;
			pYUVInChannel->m_Channel.GetResolution = CYUVInChannel_GetResolution;
			pYUVInChannel->m_Channel.GetYUVFormat = CYUVInChannel_GetYUVFormat;

			// CYUVInChannel
			//
			pYUVInChannel->m_dwFrameSize = 0;
			pYUVInChannel->m_ulVideoFramesCount = 0;
			pYUVInChannel->m_lWidth = 0;
			pYUVInChannel->m_lHeight = 0;
			pYUVInChannel->m_nBitCount = 0;
			pYUVInChannel->m_nDataType = QP_YUV_DATA_TYPE_QYUV;
		}
	}

	return (pYUVInChannel);
}



// destructor
//
void CYUVInChannel_Destructor(CYUVInChannel *This)
{
	CChannel_Destructor(&This->m_Channel);
}



/////////////////////////////////////////////////////////////////////////////
//
//

// CPCMInChannel
//

// CChannel
//


// called from streaming task
//
static TQP_BOOL CPCMInChannel_GetBuffer(CChannel *This_p,
										PQP_BUFFER_DESCRIPTOR *ppBufDesc,
										PTQP_UINT8 *ppBuffer,
										PTQP_UINT32 pSize
										)
{
	QUEUE_ENTRY				*pEntry;
	PQP_BUFFER_DESCRIPTOR	pBufDesc;
	PQP_KSSTREAM_HEADER		pDataPacket;

	pEntry = CQueue_GetOneEntry(This_p->m_pDataRequestQueue);

	if (QP_NULL != pEntry)
	{
		pBufDesc = (PQP_BUFFER_DESCRIPTOR)pEntry->Data;
		pDataPacket = pBufDesc->DataBufferArray;

		// clear the private portion of the flags
		//
		pBufDesc->ulFlags &= ~QP_BUFDESC_FLAG_MASK_PRIVATE;

		if (pDataPacket->OptionsFlags & QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
						 "CPCMInChannel_GetBuffer() QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM\n"
						 ));
			pBufDesc->ulFlags |= QP_BUFDESC_FLAG_EOS;
		}

		if (pBufDesc->ulFlags & QP_BUFDESC_FLAG_PTS)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_EXTRACE),
						 "CPCMInChannel_GetBuffer() PTS(%u) KSTime(%lu:%lu) Numerator(%d) Denominator(%d)\n",
						 pBufDesc->ulPTS,
						 (TQP_UINT32)((pDataPacket->PresentationTime.Time >> 32) & 0xFFFFFFFF),
						 (TQP_UINT32)(pDataPacket->PresentationTime.Time & 0xFFFFFFFF),
						 pDataPacket->PresentationTime.Numerator,
						 pDataPacket->PresentationTime.Denominator
						 ));
		}

		pBufDesc->ulBufferIndex = 0;
		pBufDesc->ulBufferOffset = 0;
		pBufDesc->ulTotalUsed = 0;

		*ppBuffer = (PTQP_UINT8)pDataPacket->Data;
		*pSize = pDataPacket->DataUsed/*pBufDesc->ulBufferSize*/;
		*ppBufDesc = pBufDesc;

		// move this entry to the pending queue
		//
		CQueue_AddEntry(This_p->m_pDataPendingQueue,
						pEntry
						);
	}

	return (QP_NULL != pEntry);
}



// constructor
//
CPCMInChannel * CPCMInChannel_Constructor(CPCMInChannel *pPCMInChannel,
										  CObject *pParent,
										  TQP_UINT32 dwAttributes,
										  TQP_OBJECT_HANDLE hTask,
										  CTask *pTask
										  )
{
	if (pPCMInChannel)
	{
		CChannel	*pChannel;

		pChannel = CChannel_Constructor(&pPCMInChannel->m_Channel,
										pParent,
										dwAttributes,
										hTask,
										TASK_DATA_TYPE_PCM,
										FWDataTypeRawAudInput,
										CHANNEL_DIR_WRITE,
										QPMPGCODEC_ENC_PCM_IN,
										pTask
										);
		if (!pChannel)
		{
			pPCMInChannel = QP_NULL;
		}
		else
		{
			// fill in the functions table
			//
			pPCMInChannel->m_Channel.Start = CMPEGOutChannel_Start;	// same as MpegOut
			pPCMInChannel->m_Channel.Stop = CMPEGOutChannel_Stop;	// same as MpegOut
			pPCMInChannel->m_Channel.Pause = CMPEGOutChannel_Pause;	// same as MpegOut
			pPCMInChannel->m_Channel.GetBuffer = CPCMInChannel_GetBuffer;
			pPCMInChannel->m_Channel.CompleteBuffer = CMPEGInChannel_CompleteBuffer;	// same as MpegIn
		}
	}

	return (pPCMInChannel);
}



// destructor
//
void CPCMInChannel_Destructor(CPCMInChannel *This)
{
	CChannel_Destructor(&This->m_Channel);
}



/////////////////////////////////////////////////////////////////////////////
//
//

// CVBIInChannel
//


// CChannel
//

static EQP_Errors CVBIInChannel_Open(CChannel *This_p,
									 TQP_OBJECT_HANDLE hChannel,
									 TQP_UINT32 dwFlags,
									 TQP_PVOID pDataFormat,
									 QP_DEVICE_CALLBACK pFuncCallback,
									 TQP_PVOID context
									 )
{
	EQP_Errors			qpStatus;
	PQP_VBI_DATAFORMAT	pVBIFormat = (PQP_VBI_DATAFORMAT)pDataFormat;
	CVBIInChannel		*This =	GET_INHERITED_OBJECT(CVBIInChannel, m_Channel, This_p);

	// Gather information we need for this stream
	//
    This->m_Top_StartLine = pVBIFormat->Top_StartLine;			// inclusive
    This->m_Top_EndLine = pVBIFormat->Top_EndLine;				// inclusive
  	This->m_Top_StartPixel = pVBIFormat->Top_StartPixel;		// starting from 0
    This->m_Top_SamplesPerLine = pVBIFormat->Top_SamplesPerLine;// maximum (1440 - StartPixel)

    This->m_Bottom_StartLine = pVBIFormat->Bottom_StartLine;	// inclusive
    This->m_Bottom_EndLine = pVBIFormat->Bottom_EndLine;		// inclusive
  	This->m_Bottom_StartPixel = pVBIFormat->Bottom_StartPixel;	// starting from 0
    This->m_Bottom_SamplesPerLine = pVBIFormat->Bottom_SamplesPerLine;	// maximum (1440 - StartPixel)

    This->m_BufferSize = (TQP_UINT32)pVBIFormat->BufferSize;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
				 "CVBIInChannel_Open()  StartLine(%d : %d) EndLine(%d : %d) StartPixel(%d : %d) Samples(%d : %d) buf_size(%d)\n",
				 This->m_Top_StartLine,
				 This->m_Bottom_StartLine,
				 This->m_Top_EndLine,
				 This->m_Bottom_EndLine,
  				 This->m_Top_StartPixel,
  				 This->m_Bottom_StartPixel,
				 This->m_Top_SamplesPerLine,
				 This->m_Bottom_SamplesPerLine,
				 This->m_BufferSize
				 ));


	// reset counters and flags
	//
	This->m_ullVBIFramesCount = 0;

	// generic open
	//
	qpStatus = CChannel_Open(This_p,
							 hChannel,
							 dwFlags,
							 pDataFormat,
							 pFuncCallback,
							 context
							 );

	return (qpStatus);
}



static EQP_Errors CVBIInChannel_Start(CChannel *This_p)
{
	EQP_Errors		qpStatus;
	CVBIInChannel	*This =	GET_INHERITED_OBJECT(CVBIInChannel, m_Channel, This_p);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CVBIInChannel_Start() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 This_p->m_dataType
				 ));

	if (This_p->m_bPaused)
	{
		qpStatus = This_p->m_pTask->Resume(This_p->m_pTask,
										   This_p->m_hTask,
										   (TASK_DATA_TYPE)This_p->m_dataType
										   );

		if (QP_SUCCESS(qpStatus))
			This_p->m_bPaused = QP_FALSE;
	}
	else
	{
		// reset counter and flags
		//
		This_p->m_ullCntBytes = 0;
		This->m_ullVBIFramesCount = 0;

		// start the encoder
		//
		qpStatus = This_p->m_pTask->Start(This_p->m_pTask,
										  This_p->m_hTask,
										  (TASK_DATA_TYPE)This_p->m_dataType
										  );
		This_p->m_llStartTime = QPTMGetSysTimeMilliS();
	}

	This_p->m_bFlushing = QP_FALSE;

	if (QP_SUCCESS(qpStatus))
	{
		This_p->m_State = QPSTATE_RUN;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CVBIInChannel_Start() Failed Status(%d)\n",
					 qpStatus
					 ));
	}

	return (qpStatus);
}



// called from streaming task
//
static TQP_BOOL CVBIInChannel_GetBuffer(CChannel *This_p,
										PQP_BUFFER_DESCRIPTOR *ppBufDesc,
										PTQP_UINT8 *ppBuffer,
										PTQP_UINT32 pSize
										)
{
	QUEUE_ENTRY				*pEntry;
	PQP_BUFFER_DESCRIPTOR	pBufDesc;
	PQP_KSSTREAM_HEADER		pDataPacket;
	CVBIInChannel			*This =	GET_INHERITED_OBJECT(CVBIInChannel, m_Channel, This_p);

	pEntry = CQueue_GetOneEntry(This_p->m_pDataRequestQueue);

	if (QP_NULL != pEntry)
	{
		pBufDesc = (PQP_BUFFER_DESCRIPTOR)pEntry->Data;
		pDataPacket = pBufDesc->DataBufferArray;

		// clear the private portion of the flags
		//
		pBufDesc->ulFlags &= ~QP_BUFDESC_FLAG_MASK_PRIVATE;

		if ((pDataPacket->OptionsFlags & QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM) ||
			(0 == pBufDesc->ulBufferSize)
			)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
						 "CVBIInChannel_GetBuffer() QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM\n"
						 ));
			pBufDesc->ulFlags |= QP_BUFDESC_FLAG_EOS;
		}

		if (pBufDesc->ulBufferSize < This->m_BufferSize)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CVBIInChannel_GetBuffer() buffer size(%d) < frame size(%d)\n",
						 pBufDesc->ulBufferSize,
						 This->m_BufferSize
						 ));
			// clear the entry content
			//
			pEntry->Data = QP_NULL;

			// move to free queue
			//
			CQueue_AddEntry(This_p->m_pFreeQueue,
							pEntry
							);

			// nil pEntry so the function will return FALSE
			//
			pEntry = QP_NULL;

			// default success
			//
			pBufDesc->Status = QPERR_SUCCESS;

			// call back the api
			//
			CChannel_DeviceCallback(This_p,
									QPMPGCODEC_CMD_DONE_DATA,
									pBufDesc
									);
		}
		else
		{
			if (pBufDesc->ulFlags & QP_BUFDESC_FLAG_PTS)
			{
				QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_EXTRACE),
							 "CVBIInChannel_GetBuffer() PTS(%u) KSTime(%lu:%lu) Numerator(%d) Denominator(%d)\n",
							 pBufDesc->ulPTS,
							 (TQP_UINT32)((pDataPacket->PresentationTime.Time >> 32) & 0xFFFFFFFF),
							 (TQP_UINT32)(pDataPacket->PresentationTime.Time & 0xFFFFFFFF),
							 pDataPacket->PresentationTime.Numerator,
							 pDataPacket->PresentationTime.Denominator
							 ));
			}

			pBufDesc->ulBufferIndex = 0;
			pBufDesc->ulBufferOffset = 0;
			pBufDesc->ulTotalUsed = 0;

			*ppBuffer = (PTQP_UINT8)pDataPacket->Data;
			*pSize = pDataPacket->DataUsed/*pBufDesc->ulBufferSize*/;
			*ppBufDesc = pBufDesc;

			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, /*QPDBGM_LEVEL_WARN*/QPDBGM_LEVEL_TRACE),
						 "CVBIInChannel_GetBuffer() buffer(0x%x) size(%d) Desc(0x%x)\n",
						 *ppBuffer,
						 *pSize,
						 *ppBufDesc
						 ));

			// move this entry to the pending queue
			//
			CQueue_AddEntry(This_p->m_pDataPendingQueue,
							pEntry
							);
		}
	}

	return (QP_NULL != pEntry);
}



static void CVBIInChannel_CompleteBuffer(CChannel *This_p,
										 PQP_BUFFER_DESCRIPTOR pBufDesc
										 )
{
	QUEUE_ENTRY			*pEntry;
	PQP_KSSTREAM_HEADER	pDataPacket;
	CVBIInChannel		*This =	GET_INHERITED_OBJECT(CVBIInChannel, m_Channel, This_p);


	////QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
	////			 "CVBIInChannel_CompleteBuffer()\n"
	////			 ));

	// get the buffer entry
	//
	pEntry = CQueue_GetEntryByData(This_p->m_pDataPendingQueue,
								   (TQP_PVOID)pBufDesc
								   );
	if (pEntry)
	{
		pDataPacket = pBufDesc->DataBufferArray;

		if (!QP_SUCCESS(pBufDesc->Status))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CVBIInChannel_CompleteBuffer() Status(%d)\n",
						 pBufDesc->Status
						 ));
		}
		else
		{
			// increment total frame consumed
			//
			This->m_ullVBIFramesCount++;
		}

		// clear the entry content
		//
		pEntry->Data = QP_NULL;

		// move to free queue
		//
		CQueue_AddEntry(This_p->m_pFreeQueue,
						pEntry
						);

		// call back to the api
		//
		CChannel_DeviceCallback(This_p,
								QPMPGCODEC_CMD_DONE_DATA,
								pBufDesc
								);
	}
}



// constructor
//
CVBIInChannel * CVBIInChannel_Constructor(CVBIInChannel *pVBIInChannel,
										  CObject *pParent,
										  TQP_UINT32 dwAttributes,
										  TQP_OBJECT_HANDLE hTask,
										  CTask *pTask
										  )
{
	if (pVBIInChannel)
	{
		CChannel	*pChannel;

		pChannel = CChannel_Constructor(&pVBIInChannel->m_Channel,
										pParent,
										dwAttributes,
										hTask,
										TASK_DATA_TYPE_VBI,
										FWDataTypeVBIInput,
										CHANNEL_DIR_WRITE,
										QPMPGCODEC_DEC_VBI_IN,
										pTask
										);
		if (!pChannel)
		{
			pVBIInChannel = QP_NULL;
		}
		else
		{
			// fill in the functions
			//
			pVBIInChannel->m_Channel.Open = CVBIInChannel_Open;
			pVBIInChannel->m_Channel.Start = CVBIInChannel_Start;
			pVBIInChannel->m_Channel.Stop = CMPEGInChannel_Stop;
			pVBIInChannel->m_Channel.Pause = CMPEGInChannel_Pause;
			pVBIInChannel->m_Channel.GetBuffer = CVBIInChannel_GetBuffer;
			pVBIInChannel->m_Channel.CompleteBuffer = CVBIInChannel_CompleteBuffer;

			pVBIInChannel->m_Top_StartLine = 0;		// inclusive
			pVBIInChannel->m_Top_EndLine = 0;	  	// inclusive
  			pVBIInChannel->m_Top_StartPixel = 0;	// starting from 0
			pVBIInChannel->m_Top_SamplesPerLine = 0;// maximum (1440 - StartPixel)

			pVBIInChannel->m_Bottom_StartLine = 0;	// inclusive
			pVBIInChannel->m_Bottom_EndLine = 0;	// inclusive
  			pVBIInChannel->m_Bottom_StartPixel = 0;	// starting from 0
			pVBIInChannel->m_Bottom_SamplesPerLine = 0;	// maximum (1440 - StartPixel)

			pVBIInChannel->m_BufferSize = 0;

			pVBIInChannel->m_ullVBIFramesCount = 0;
		}
	}

	return (pVBIInChannel);
}



// destructor
//
void CVBIInChannel_Destructor(CVBIInChannel *This)
{
	CChannel_Destructor(&This->m_Channel);
}




/////////////////////////////////////////////////////////////////////////////
//
//

// CActivityOutChannel
//


// CChannel
//
static EQP_Errors CActivityOutChannel_Start(CChannel *This_p)
{
	EQP_Errors	qpStatus;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CActivityOutChannel_Start() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 This_p->m_dataType
				 ));

	if (This_p->m_bPaused)
	{
		qpStatus = This_p->m_pTask->Resume(This_p->m_pTask,
										   This_p->m_hTask,
										   (TASK_DATA_TYPE)This_p->m_dataType
										   );

		if (QP_SUCCESS(qpStatus))
			This_p->m_bPaused = QP_FALSE;
	}
	else
	{
		// reset counter and flags
		//
		This_p->m_ullCntBytes = 0;

		This_p->m_llPrevPts = 0;
		This_p->m_llAddPts = 0;
		This_p->m_llBasePts = 0;

		// start the encoder
		//
		qpStatus = This_p->m_pTask->Start(This_p->m_pTask,
										  This_p->m_hTask,
										  (TASK_DATA_TYPE)This_p->m_dataType
										  );
		// remember the start time
		//
		This_p->m_llStartTime = QPTMGetSysTimeMilliS();
	}

	This_p->m_bFlushing = QP_FALSE;

	if (QP_SUCCESS(qpStatus))
	{
		This_p->m_State = QPSTATE_RUN;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CActivityOutChannel_Start() Failed Status(%d)\n",
					 qpStatus
					 ));
	}

	return (qpStatus);
}



static EQP_Errors CActivityOutChannel_Stop(CChannel *This_p)
{
	EQP_Errors	qpStatus;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CActivityOutChannel_Stop() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 (TASK_DATA_TYPE)This_p->m_dataType
				 ));

	qpStatus = This_p->m_pTask->Stop(This_p->m_pTask,
									 This_p->m_hTask,
									 (TASK_DATA_TYPE)This_p->m_dataType
									 );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CActivityOutChannel_Stop() m_pTask->Stop failed(%d)\n",
					 qpStatus
					 ));
	}

	// clear falgs
	//
	This_p->m_bPaused = QP_FALSE;

	// change state to stop
	//
	This_p->m_State = QPSTATE_STOP;

	// flush the data
	//
	This_p->BeginFlush(This_p);
	This_p->Flush(This_p);
	This_p->EndFlush(This_p);

	return (QPERR_SUCCESS);
}



static EQP_Errors CActivityOutChannel_Pause(CChannel *This_p)
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CActivityOutChannel_Pause() hChannel(%d) hTask(%d) dataType(%d)\n",
				 This_p->m_hChannel,
				 This_p->m_hTask,
				 This_p->m_dataType
				 ));

	if (QPSTATE_RUN == This_p->m_State)
	{
		qpStatus = This_p->m_pTask->Pause(This_p->m_pTask,
										  This_p->m_hTask,
										  (TASK_DATA_TYPE)This_p->m_dataType
										  );
		This_p->m_bPaused = QP_SUCCESS(qpStatus);
	}

	if (QP_SUCCESS(qpStatus))
	{
		This_p->m_State = QPSTATE_PAUSE;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CActivityOutChannel_Pause() m_pTask->Pause failed(%d)\n",
					 qpStatus
					 ));
	}

	return (qpStatus);
}



// called from streaming task
//
static TQP_BOOL CActivityOutChannel_GetBuffer(CChannel *This_p,
										  PQP_BUFFER_DESCRIPTOR *ppBufDesc,
										  PTQP_UINT8 *ppBuffer,
										  PTQP_UINT32 pSize
										  )
{

	QUEUE_ENTRY				*pEntry;
	PQP_BUFFER_DESCRIPTOR	pBufDesc;
	PQP_KSSTREAM_HEADER		pDataPacket;

	pEntry = CQueue_GetOneEntry(This_p->m_pDataRequestQueue);

	if (QP_NULL != pEntry)
	{
		pBufDesc = (PQP_BUFFER_DESCRIPTOR)pEntry->Data;
		pDataPacket = pBufDesc->DataBufferArray;

		pBufDesc->ulFlags &= ~QP_BUFDESC_FLAG_MASK_PRIVATE;
		pBufDesc->ulBufferIndex = 0;
		pBufDesc->ulBufferOffset = 0;
		pBufDesc->ulTotalUsed = 0;

		pDataPacket->DataUsed = 0;

		*ppBuffer = (PTQP_UINT8)pDataPacket->Data;
		*pSize = pBufDesc->ulBufferSize;
		*ppBufDesc = pBufDesc;

		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, /*QPDBGM_LEVEL_WARN*/QPDBGM_LEVEL_TRACE),
					 "CActivityOutChannel_GetBuffer(%d) buffer(0x%x) size(%d) Desc(0x%x)\n",
					 This_p->m_hChannel,
					 *ppBuffer,
					 *pSize,
					 *ppBufDesc
					 ));

		// move this entry to the pending queue
		//
		CQueue_AddEntry(This_p->m_pDataPendingQueue,
						pEntry
						);
	}

	return (QP_NULL != pEntry);
}



static void CActivityOutChannel_CompleteBuffer(CChannel *This_p,
										   PQP_BUFFER_DESCRIPTOR pBufDesc
										   )
{
	QUEUE_ENTRY			*pEntry;
	PQP_KSSTREAM_HEADER	pDataPacket;
	CQLCodec			*pQlCodec = GET_INHERITED_OBJECT(CQLCodec, m_Object, This_p->m_Object.m_pParent);


	////QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
	////			 "CActivityOutChannel_CompleteBuffer()\n"
	////			 ));

	// get the buffer entry
	//
	pEntry = CQueue_GetEntryByData(This_p->m_pDataPendingQueue,
								   (TQP_PVOID)pBufDesc
								   );
	if (pEntry)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, /*QPDBGM_LEVEL_WARN*/QPDBGM_LEVEL_TRACE),
					 "CActivityOutChannel_CompleteBuffer() bufDesc(0x%x) Status(%d), totalUsed(%d)\n",
					 pBufDesc,
					 pBufDesc->Status,
					 pBufDesc->ulTotalUsed
					 ));

		pDataPacket = pBufDesc->DataBufferArray;
		pDataPacket->OptionsFlags = 0;

		if (This_p->m_bByteSwap &&
			!QPHCI_CanSwapData(&pQlCodec->m_hci) &&
			(0 != pBufDesc->ulTotalUsed)
			)
		{
			BytesSwap((PTQP_UINT8)pDataPacket->Data,
					  (PTQP_UINT8)pDataPacket->Data,
					  pBufDesc->ulTotalUsed
					  );
		}

		// set these flags if we are at the begining of the stream
		//
		if (0 == This_p->m_ullCntBytes)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN/*QPDBGM_LEVEL_TRACE*/),
						 "CActivityOutChannel_CompleteBuffer() (0 == m_ullCntBytes) QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY\n"
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_DATADISCONTINUITY |
										 QP_KSSTREAM_HEADER_OPTIONSF_TIMEDISCONTINUITY
										 ;
		}

		// set time stamp
		//
		CChannel_TimeStamp(This_p,
						   pBufDesc,
						   pDataPacket
						   );

		This_p->m_ullCntBytes += pBufDesc->ulTotalUsed;
		pDataPacket->DataUsed = pBufDesc->ulTotalUsed;

		if (pBufDesc->ulFlags & QP_BUFDESC_FLAG_EOS)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CActivityOutChannel_CompleteBuffer() used(%d) total(%d) QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM\n",
						 pBufDesc->ulTotalUsed,
						 This_p->m_ullCntBytes
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_ENDOFSTREAM;
		}

		if (pBufDesc->ulFlags & QP_BUFDESC_FLAG_INCOMPLETE_FRAME)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CActivityOutChannel_CompleteBuffer(%d) QP_KSSTREAM_HEADER_OPTIONSF_PREROLL\n",
						 pBufDesc->ulTotalUsed
						 ));
			pDataPacket->OptionsFlags |= QP_KSSTREAM_HEADER_OPTIONSF_PREROLL;
		}

		// clear the entry content
		//
		pEntry->Data = QP_NULL;

		// move to free queue
		//
		CQueue_AddEntry(This_p->m_pFreeQueue,
						pEntry
						);

		// call back to the api
		//
		CChannel_DeviceCallback(This_p,
								QPMPGCODEC_CMD_DONE_DATA,
								pBufDesc
								);
	}
}



// constructor
//
CActivityOutChannel * CActivityOutChannel_Constructor(CActivityOutChannel *pActivityOutChannel,
											  CObject *pParent,
											  TQP_UINT32 dwAttributes,
											  TQP_OBJECT_HANDLE hTask,
											  CTask *pTask
											  )
{
	if (pActivityOutChannel)
	{
		CChannel	*pChannel;

		pChannel = CChannel_Constructor(&pActivityOutChannel->m_Channel,
										pParent,
										dwAttributes,
										hTask,
										TASK_DATA_TYPE_COMP_VID,
										FWDataTypeCmprSysVidOutput,
										CHANNEL_DIR_READ,
										QPMPGCODEC_ENC_ACTIVITY_OUT,
										pTask
										);
		if (!pChannel)
		{
			pActivityOutChannel = QP_NULL;
		}
		else
		{
			if (BYTES_SWAP_BY_DRIVER == pTask->m_TaskData[hTask].m_inputControl.RegArtesa.bytes_swap)
			{
				// BYTES_SWAP_BY_DRIVER
				pActivityOutChannel->m_Channel.m_bByteSwap = QP_TRUE;
			}
			else
			{
				// BYTES_SWAP_BY_FW
				pActivityOutChannel->m_Channel.m_bByteSwap = QP_FALSE;
			}

			// fill in the functions table
			//
			pActivityOutChannel->m_Channel.Start = CActivityOutChannel_Start;
			pActivityOutChannel->m_Channel.Stop = CActivityOutChannel_Stop;
			pActivityOutChannel->m_Channel.Pause = CActivityOutChannel_Pause;
			pActivityOutChannel->m_Channel.GetBuffer = CActivityOutChannel_GetBuffer;
			pActivityOutChannel->m_Channel.CompleteBuffer = CActivityOutChannel_CompleteBuffer;
		}
	}

	return (pActivityOutChannel);
}



// destructor
//
void CActivityOutChannel_Destructor(CActivityOutChannel *This)
{
	CChannel_Destructor(&This->m_Channel);
}
