///////////////////////////////////////////////////////////////////////////////
//        
//
// Cfifo.h
//
// Description: 
//
//	Header file for the fifo.
// 
// History:
//	Aug 4, 2006	(dt) initial check in
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////

#ifndef __CFIFO_H__
#define __CFIFO_H__


#include "qptypes.h"
#include "cobject.h"
#include "util.h"

#ifdef __cplusplus
extern "C" {
#endif //__cplusplus

/////////////////////////////////////////////////////////////////////////////
//
//

// C defination, visible to both C and C++ code
//

#ifndef __CFIFO_FWD_DEFINED__
#define __CFIFO_FWD_DEFINED__
typedef struct CFifo CFifo;
#endif //!__CFIFO_FWD_DEFINED__


/////////////////////////////////////////////////////////////////////////////
//
//

struct CFifo
{
	CObject			m_Object;

	TQP_UINT32		m_dwReadPtr;			// read pointer (initial 0)
	TQP_UINT32		m_dwWritePtr;			// write pointer (initial 0)
	PTQP_UINT8		m_Fifo;
	TQP_UINT32		m_dwFifoLevel;
	TQP_UINT32		m_size;
	TQP_UINT32		m_sizeEntry;
};


/////////////////////////////////////////////////////////////////////////////
//
//

CFifo * CFifo_Constructor(CFifo *pFifo,
						  CObject *pParent, 
						  TQP_UINT32 dwAttributes,
						  TQP_UINT32 size,
						  TQP_UINT32 sizeEntry
						  );
void CFifo_Destructor(CFifo *This);
TQP_BOOL CFifo_PeekFifo(CFifo *This, 
						TQP_PVOID val
						);
TQP_BOOL CFifo_GetFifo(CFifo *This, 
					   TQP_PVOID val
					   );
TQP_BOOL CFifo_SetFifo(CFifo *This, 
					   TQP_PVOID Data
					   );
void CFifo_FlushFifo(CFifo *This);
void CFifo_UpdateReadPtr(CFifo *This);
TQP_BOOL CFifo_UpdateEntry(CFifo *This, 
						   TQP_PVOID val
						   );
TQP_UINT32 CFifo_GetFifoLevel(CFifo *This);
TQP_BOOL CFifo_IsEmpty(CFifo *This);

#ifdef __cplusplus
}
#endif //__cplusplus


/////////////////////////////////////////////////////////////////////////////
//
//

// C++ defination, visible only to C++ code
//

#ifdef __cplusplus

template <class T, TQP_UINT32 size> class CppFifo : public CppObject
{
public:
	CppFifo(CppObject *pParent)
		: CppObject(pParent, 
					OBJECT_FIFO, 
					OBJECT_CRITICAL
					)
		, m_dwReadPtr(0)
		, m_dwWritePtr(0)
		, m_dwFifoLevel(0)
		, m_size(size)
	{
	}

	CppFifo(CppObject *pParent, 
			TQP_UINT32 dwWhoAmI, 
			TQP_UINT32 dwAttributes
			)
		: CppObject(pParent, 
					dwWhoAmI, 
					dwAttributes
					)
		, m_dwReadPtr(0)
		, m_dwWritePtr(0)
		, m_dwFifoLevel(0)
		, m_size(size)
	{
	}

	virtual ~CppFifo() {}

	TQP_BOOL PeekFifo(T& val)
	{
		ENTER_CRITICAL_CPP

		TQP_BOOL ret = (m_dwReadPtr != m_dwWritePtr);
		if (ret)
			val = m_Fifo[m_dwReadPtr];

		LEAVE_CRITICAL_CPP
		return (ret);
	}

	TQP_BOOL GetFifo(T& val)
	{
		ENTER_CRITICAL_CPP

		TQP_BOOL ret = (m_dwReadPtr != m_dwWritePtr);
		if (ret)
		{
			val = m_Fifo[m_dwReadPtr];

			if (m_dwReadPtr >= m_dwWritePtr)
			{
				if (m_dwReadPtr == (size - 1))
					m_dwReadPtr = 0; // wrap around
				else
					m_dwReadPtr++;
			}
			else
				m_dwReadPtr++;

			// decrease number in fifo
			m_dwFifoLevel--;
		}

		LEAVE_CRITICAL_CPP
		return (ret);
	}

	TQP_BOOL SetFifo(T& Data)
	{
		ENTER_CRITICAL_CPP

		// sacrifice one fifo entry for efficency
		if (m_dwFifoLevel == size - 1)
		{
			LEAVE_CRITICAL_CPP

			return (QP_FALSE);
		}

		// save it in the fifo
		m_Fifo[m_dwWritePtr] = Data;

		// increase number of fifo
		m_dwFifoLevel++;

		// update write pointer
		if (m_dwWritePtr >= m_dwReadPtr)
		{
			if (m_dwWritePtr == (size - 1))
				m_dwWritePtr = 0; // wrap around
			else
				m_dwWritePtr++;
		}
		else
			m_dwWritePtr++;

		LEAVE_CRITICAL_CPP
		return (QP_TRUE);
	}

	void FlushFifo() 
	{
		ENTER_CRITICAL_CPP

		m_dwReadPtr = m_dwWritePtr = m_dwFifoLevel = 0;

		LEAVE_CRITICAL_CPP
	}

	void UpdateReadPtr()
	{
		ENTER_CRITICAL_CPP

		// m_dwReadPtr == m_dwWritePtr should never happen in this case
		if (m_dwReadPtr >= m_dwWritePtr)
		{
			if (m_dwReadPtr == (size - 1))
				m_dwReadPtr = 0; // wrap around
			else
				m_dwReadPtr++;
		}
		else
			m_dwReadPtr++;

		// decrease number in fifo
		m_dwFifoLevel--;

		LEAVE_CRITICAL_CPP
	}

	TQP_BOOL UpdateEntry(T& val)
	{
		ENTER_CRITICAL_CPP

		TQP_BOOL ret = (m_dwReadPtr != m_dwWritePtr);
		if (ret)
		{
			m_Fifo[m_dwReadPtr] = val;
		}

		LEAVE_CRITICAL_CPP
		return (ret);
	}

	TQP_UINT32 GetFifoLevel() {return (m_dwFifoLevel);}

	TQP_BOOL IsEmpty() {return (0 == m_dwFifoLevel);}

protected:
	TQP_UINT32		m_dwReadPtr;			// read pointer (initial 0)
	TQP_UINT32		m_dwWritePtr;			// write pointer (initial 0)
	T				m_Fifo[size];
	TQP_UINT32		m_dwFifoLevel;
	TQP_UINT32		m_size;
};

#endif //__cplusplus

#endif //__CFIFO_H__

