///////////////////////////////////////////////////////////////////////////////
//        
//
// Cqueue.h
//
// Description: 
//
//	Header that defines the queue.
// 
// History:
//	Aug 4, 2006	(dt) initial check in
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////

#ifndef __CQUEUE_H__
#define __CQUEUE_H__


#include "qptypes.h"
#include "cobject.h"

/////////////////////////////////////////////////////////////////////////////
//
//

// C defination, visible to both C and C++ code
//

#ifdef __cplusplus
extern "C" {
#endif //__cplusplus


#ifndef __CQUEUE_FWD_DEFINED__
#define __CQUEUE_FWD_DEFINED__
typedef struct CQueue CQueue;
typedef struct QUEUE_ENTRY QUEUE_ENTRY;
typedef struct QUEUE_LIST QUEUE_LIST;
#endif //!__CQUEUE_FWD_DEFINED__


/////////////////////////////////////////////////////////////////////////////
//
//

struct QUEUE_ENTRY
{
	QUEUE_ENTRY	*pNext;
	TQP_PVOID	Data;
};


struct QUEUE_LIST
{
	QUEUE_ENTRY	*pHead;					// head of the list
	QUEUE_ENTRY	*pTail;					// tail of the list
};


struct CQueue
{
	// CObject
	//
	CObject				m_Object;

	// CQueue
	//
	QUEUE_LIST			m_Queue;		// queue list
	TQP_UINT32			m_dwNbInQueue;	// nb of items in the queue
};



/////////////////////////////////////////////////////////////////////////////
//
//

CQueue * CQueue_Constructor(CQueue *pQueue,
							CObject *pParent, 
							TQP_UINT32 dwAttributes
							);
void CQueue_Destructor(CQueue *This);
// add one entry to the end of the queue
//
void CQueue_AddEntry(CQueue *This, 
					 QUEUE_ENTRY *pEntry
					 );
// add one entry to the top of the queue
//
void CQueue_AddEntryToTop(CQueue *This, 
						  QUEUE_ENTRY *pEntry
						  );
// remove the specified entry from the queue
//
QUEUE_ENTRY* CQueue_RemoveEntry(CQueue *This, 
								QUEUE_ENTRY *pEntry
								);
// get one entry actually remove the first entry from the queue.
//
QUEUE_ENTRY* CQueue_GetOneEntry(CQueue *This);
// return the entry without remove it from the queue
//
QUEUE_ENTRY* CQueue_PeekEntryByData(CQueue *This, 
									TQP_PVOID Data
									);
// remove the specified entry from the queue
//
QUEUE_ENTRY* CQueue_GetEntryByData(CQueue *This, 
								   TQP_PVOID Data
								   );
QUEUE_ENTRY* CQueue_PeekOneEntry(CQueue *This);
QUEUE_ENTRY* CQueue_PeekEntryByIndex(CQueue *This, 
									 TQP_UINT32 index
									 );
void CQueue_ResetQueue(CQueue *This);
// flush the queue
//
void CQueue_FlushQueue(CQueue *This);
TQP_BOOL CQueue_IsEmpty(CQueue *This);
TQP_UINT32 CQueue_GetQueueLevel(CQueue *This);
// is data already in queue?
//
TQP_BOOL CQueue_InQueue(CQueue *This, 
						TQP_PVOID Data
						);
void CQueue_WalkQueue(CQueue *This);

#ifdef __cplusplus
}
#endif //__cplusplus



/////////////////////////////////////////////////////////////////////////////
//
//

// C++ defination, visible only to C++ code
//

#ifdef __cplusplus

typedef EQP_Errors (*QUEUE_CALLBACK)(TQP_PVOID pParam);

/////////////////////////////////////////////////////////////////////////////
//
//

template <class T> class QUEUE_ENTRY_CPP
{
public:
	QUEUE_ENTRY_CPP() 
		: pNext(QP_NULL) 
	{
	}
	virtual ~QUEUE_ENTRY_CPP() {}

	QUEUE_ENTRY_CPP	*pNext;
	T				Data;
};



/////////////////////////////////////////////////////////////////////////////
//
//

template <class T> class CppQueue : public CppObject
{
public:
	CppQueue(CppObject *pParent, 
			 QUEUE_CALLBACK pFuncCallback
			 );
	CppQueue(CppObject *pParent, 
			 TQP_UINT32 dwWhoAmI, 
			 TQP_UINT32 dwAttributes, 
			 QUEUE_CALLBACK pFuncCallback
			 );
	virtual ~CppQueue();

	virtual void AddEntry(QUEUE_ENTRY_CPP<T> *pEntry);
	virtual void AddEntryToTop(QUEUE_ENTRY_CPP<T> *pEntry);
	QUEUE_ENTRY_CPP<T>* RemoveEntry(QUEUE_ENTRY_CPP<T> *pEntry);
	QUEUE_ENTRY_CPP<T>* GetOneEntry();
	QUEUE_ENTRY_CPP<T>* PeekEntryByData(T& Data);
	QUEUE_ENTRY_CPP<T>* GetEntryByData(T& Data);
	QUEUE_ENTRY_CPP<T>* PeekOneEntry();
	QUEUE_ENTRY_CPP<T>* PeekEntryByIndex(TQP_UINT32 index);
	void ResetQueue() 
	{
		ENTER_CRITICAL_CPP
		m_Queue.pHead = m_Queue.pTail = QP_NULL;
		m_dwNbInQueue = 0;
		LEAVE_CRITICAL_CPP
	}
	virtual void FlushQueue();
	TQP_BOOL IsEmpty() 
	{
		return (m_Queue.pHead ? QP_FALSE : QP_TRUE);
	}
	TQP_UINT32 GetQueueLevel() 
	{
		return (m_dwNbInQueue);
	}
	TQP_BOOL InQueue(T& Data);

protected:

	class QUEUE_LIST_CPP
	{
		public:
		QUEUE_ENTRY_CPP<T>	*pHead;				// head of the list
		QUEUE_ENTRY_CPP<T>	*pTail;				// tail of the list
	};

	QUEUE_LIST_CPP		m_Queue;				// queue list
	TQP_UINT32			m_dwNbInQueue;			// nb of items in the queue
	QUEUE_CALLBACK		m_pFuncCallback;		// callback
};


/////////////////////////////////////////////////////////////////////////////
//
//

template <class T> 
CppQueue<T>::CppQueue(CppObject *pParent, 
					  QUEUE_CALLBACK pFuncCallback
					  )
	: CppObject(pParent, 
				OBJECT_QUEUE,
				OBJECT_CRITICAL
				)
	, m_dwNbInQueue(0)
	, m_pFuncCallback(pFuncCallback)
{
	ResetQueue();
}



template <class T> 
CppQueue<T>::CppQueue(CppObject *pParent, 
					  TQP_UINT32 dwWhoAmI, 
					  TQP_UINT32 dwAttributes, 
					  QUEUE_CALLBACK pFuncCallback
					  )
	: CppObject(pParent, 
				dwWhoAmI, 
				dwAttributes
				)
	, m_dwNbInQueue(0)
	, m_pFuncCallback(pFuncCallback)
{
	ResetQueue();
}



template <class T> 
CppQueue<T>::~CppQueue()
{
	FlushQueue();
}



// add one entry to the end of the queue
//
template <class T> 
void CppQueue<T>::AddEntry(QUEUE_ENTRY_CPP<T> *pEntry)
{
	ENTER_CRITICAL_CPP

	if (pEntry)
	{
		if (m_Queue.pTail)
		{
			m_Queue.pTail->pNext = pEntry;
			m_Queue.pTail = pEntry;
		}
		else
		{
			m_Queue.pHead = pEntry;
			m_Queue.pTail = pEntry;
		}
		pEntry->pNext = QP_NULL;
		m_dwNbInQueue++;
	}

	LEAVE_CRITICAL_CPP
}



// add one entry to the top of the queue
//
template <class T> 
void CppQueue<T>::AddEntryToTop(QUEUE_ENTRY_CPP<T> *pEntry)
{
	ENTER_CRITICAL_CPP

	if (pEntry)
	{
        pEntry->pNext = m_Queue.pHead;
        m_Queue.pHead = pEntry;

	    if (!m_Queue.pTail)
	    {
		    m_Queue.pTail = pEntry;
	    }
	    m_dwNbInQueue++;
	}

	LEAVE_CRITICAL_CPP
}



// remove the specified entry from the queue
//
template <class T> 
QUEUE_ENTRY_CPP<T>* CppQueue<T>::RemoveEntry(QUEUE_ENTRY_CPP<T> *pEntry)
{
	QUEUE_ENTRY_CPP<T>	*pthis_queue = m_Queue.pHead;
	QUEUE_ENTRY_CPP<T>	*pprev_queue = QP_NULL;

	ENTER_CRITICAL_CPP

	while (pthis_queue)
	{
		if (pthis_queue == pEntry)
		{
			if (!pprev_queue)
			{
				m_Queue.pHead = pthis_queue->pNext;
				if (!m_Queue.pHead)
					m_Queue.pTail = m_Queue.pHead;
			}
			else
			{
				pprev_queue->pNext = pthis_queue->pNext;
				if (!pthis_queue->pNext)
					m_Queue.pTail = pprev_queue;
			}
			m_dwNbInQueue--;
			break;
		}
		pprev_queue = pthis_queue;
		pthis_queue = pthis_queue->pNext;
	}

	LEAVE_CRITICAL_CPP

	return (pthis_queue);
}



// get one entry actually remove the first entry from the queue.
//
template <class T> 
QUEUE_ENTRY_CPP<T>* CppQueue<T>::GetOneEntry()
{
	QUEUE_ENTRY_CPP<T>	*pEntry = QP_NULL;

	ENTER_CRITICAL_CPP

	if (m_Queue.pHead)
	{
		pEntry = m_Queue.pHead;
		m_Queue.pHead = m_Queue.pHead->pNext;
		if (!m_Queue.pHead)
			m_Queue.pTail = m_Queue.pHead;
		m_dwNbInQueue--;
	}

	LEAVE_CRITICAL_CPP

	return (pEntry);
}



// return the entry without remove it from the queue
//
template <class T> 
QUEUE_ENTRY_CPP<T>* CppQueue<T>::PeekEntryByData(T& Data)
{
	QUEUE_ENTRY_CPP<T>	*pthis_queue = m_Queue.pHead;
	QUEUE_ENTRY_CPP<T>	*pRet = QP_NULL;
		  
	ENTER_CRITICAL_CPP

	while (pthis_queue)
	{
		if (pthis_queue->Data == Data)
		{
			pRet = pthis_queue;
			break;
		}
		pthis_queue = pthis_queue->pNext;
	}

	LEAVE_CRITICAL_CPP

	return (pRet);
}



// remove the specified entry from the queue
//
template <class T> 
QUEUE_ENTRY_CPP<T>* CppQueue<T>::GetEntryByData(T& Data)
{
	QUEUE_ENTRY_CPP<T>	*pthis_queue = m_Queue.pHead;
	QUEUE_ENTRY_CPP<T>	*pprev_queue = QP_NULL;

	ENTER_CRITICAL_CPP

	while (pthis_queue)
	{
		if (pthis_queue->Data == Data)
		{
			if (!pprev_queue)
			{
				m_Queue.pHead = pthis_queue->pNext;
				if (!m_Queue.pHead)
					m_Queue.pTail = m_Queue.pHead;
			}
			else
			{
				pprev_queue->pNext = pthis_queue->pNext;
				if (!pthis_queue->pNext)
					m_Queue.pTail = pprev_queue;
			}
			m_dwNbInQueue--;
			break;
		}
		pprev_queue = pthis_queue;
		pthis_queue = pthis_queue->pNext;
	}

	LEAVE_CRITICAL_CPP

	return (pthis_queue);
}



template <class T> 
QUEUE_ENTRY_CPP<T>* CppQueue<T>::PeekOneEntry()
{
	QUEUE_ENTRY_CPP<T>	*pEntry;

	ENTER_CRITICAL_CPP

	pEntry = m_Queue.pHead;

	LEAVE_CRITICAL_CPP

	return (pEntry);
}



template <class T>
QUEUE_ENTRY_CPP<T>* CppQueue<T>::PeekEntryByIndex(TQP_UINT32 index)
{
	QUEUE_ENTRY_CPP<T>	*pthis_queue = m_Queue.pHead;
	QUEUE_ENTRY_CPP<T>	*pRet = QP_NULL;
	TQP_UINT32           count = 0;
		  
	ENTER_CRITICAL_CPP

	while (pthis_queue)
	{
		if (count++ == index)
		{
			pRet = pthis_queue;
			break;
		}
		pthis_queue = pthis_queue->pNext;
	}

	LEAVE_CRITICAL_CPP

	return (pRet);
}



// flush the queue
//
template <class T> 
void CppQueue<T>::FlushQueue()
{
	QUEUE_ENTRY_CPP<T>	*pthis_queue;

	ENTER_CRITICAL_CPP

	while (m_Queue.pHead)
	{
		pthis_queue = m_Queue.pHead;
		m_Queue.pHead = pthis_queue->pNext;
		m_dwNbInQueue--;

		if (m_pFuncCallback && pthis_queue->Data)
		{
			m_pFuncCallback((TQP_PVOID)pthis_queue->Data);
		}
	}

	m_Queue.pTail = QP_NULL;

	LEAVE_CRITICAL_CPP
}



// is data already in queue?
//
template <class T> 
TQP_BOOL CppQueue<T>::InQueue(T& Data)
{
	QUEUE_ENTRY_CPP<T>	*pthis_queue = m_Queue.pHead;
	TQP_BOOL			bRet = QP_FALSE;

	ENTER_CRITICAL_CPP

	while (pthis_queue)
	{
		if (pthis_queue->Data == Data)
		{
			bRet = QP_TRUE;
			break;
		}
		pthis_queue = pthis_queue->pNext;
	}

	LEAVE_CRITICAL_CPP

	return (bRet);
}



/////////////////////////////////////////////////////////////////////////////
//
//

template <class T> class CppStack : public CppQueue<T>
{
public:
	CppStack(CppObject *pParent, 
			 QUEUE_CALLBACK pFuncCallback
			 )
        : CppQueue<T>(pParent, 
					  pFuncCallback
					  )
    {
    }
	
	CppStack(CppObject *pParent, 
			 TQP_UINT32 dwWhoAmI, 
			 TQP_UINT32 dwAttributes, 
			 QUEUE_CALLBACK pFuncCallback
			 )
        : CppQueue<T>(pParent, 
					  dwWhoAmI, 
					  dwAttributes, 
					  pFuncCallback
					  )
    {
    }

	void PushEntry(QUEUE_ENTRY_CPP<T> *pEntry)
	{

	    ENTER_CRITICAL_CPP

	    if (pEntry)
	    {
            pEntry->pNext = m_Queue.pHead;
            m_Queue.pHead = pEntry;

		    if (!m_Queue.pTail)
		    {
			    m_Queue.pTail = pEntry;
		    }
		    m_dwNbInQueue++;
	    }

	    LEAVE_CRITICAL_CPP
	}

	// pop one entry out of the stack
	//
	QUEUE_ENTRY_CPP<T>* PopEntry()
	{
		return (GetOneEntry());
	}

	virtual void AddEntry(QUEUE_ENTRY_CPP<T> *pEntry)
    {
		PushEntry(pEntry);
    }
};

#endif //__cplusplus

#endif //__CQUEUE_H__

