///////////////////////////////////////////////////////////////////////////////
//        
//
// Cthread.h
//
// Description: 
//
//	Header that defines the base thread.
// 
// History:
//	Aug 4, 2006	(dt) initial check in
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////

#ifndef __CTHREAD_H__
#define __CTHREAD_H__


#include "qptypes.h"
#include "qposmgr.h"
#include "qptimemgr.h"


enum ThreadEventsWait
{   
    THREAD_EVENT_EXIT					= BIT(0),
    THREAD_EVENT_NEW_CMD				= BIT(1),

    THREAD_EVENT_MASK					= THREAD_EVENT_EXIT | 
										  THREAD_EVENT_NEW_CMD
};

#define THREAD_EVENTS_WAIT_END			1

enum ThreadEventsReply
{   
    THREAD_EVENT_EXIT_DONE				= BIT(0),

    THREAD_EVENT_REPLY_MASK				= THREAD_EVENT_EXIT_DONE
};

#define THREAD_EVENTS_REPLY_END			0


#ifdef __cplusplus
extern "C" {
#endif //__cplusplus


/////////////////////////////////////////////////////////////////////////////
//
//

// C defination, visible to both C and C++ code
//

#ifndef __CTHREAD_FWD_DEFINED__
#define __CTHREAD_FWD_DEFINED__
typedef struct CThread CThread;
#endif //!__CTHREAD_FWD_DEFINED__


/////////////////////////////////////////////////////////////////////////////
//
//

struct CThread
{
	// thread function
	//
	int (*m_ThreadProc)(CThread *This);

	// wait thread exit
	//
	TQP_BOOL (*ThreadExit)(CThread *This);

	TQP_OSMThreadID	m_threadID;

	TQP_PVOID		m_context;
	TQP_UINT32		m_priority;
	TQP_OSMEvtgrpID	m_EvtWait;
	TQP_OSMEvtgrpID	m_EvtReply;
	TQP_BOOL		m_bRemoveUserSpaceMapping;

	char			m_szThreadName[QPOS_THRNAME_LEN];
};


/////////////////////////////////////////////////////////////////////////////
//
//

CThread	* CThread_Constructor(CThread *pThread,
							  char *Name,
							  TQP_PVOID context,
							  TQP_UINT32 priority,
							  TQP_OSMEvtgrpID evtWait,
							  TQP_OSMEvtgrpID evtReply,
							  TQP_BOOL bRemoveUserSpaceMapping
							  );
void CThread_Destructor(CThread *This);
// Thread creation and clean up
//
TQP_BOOL CThread_ThreadInit(CThread *This);
TQP_BOOL CThread_ThreadDone(CThread *This);
// helper
//
TQP_BOOL CThread_IsThreadInited(CThread *This);


#ifdef __cplusplus
}
#endif //__cplusplus


/////////////////////////////////////////////////////////////////////////////
//
//

// C++ defination, visible only to C++ code
//

#ifdef __cplusplus

#include "qpdbgmgr.h"

extern "C" TQP_DbgMCompID	g_QPCodecDBGCompID;

#include "cqueue.h"

/////////////////////////////////////////////////////////////////////////////
//
//

typedef EQP_Errors (*THREAD_CMD_CALLBACK_FUNC)(TQP_PVOID pParam);


// thread commands
//
#define THREAD_CMD_CALLBACK				0x00000001


// thread command structure
//
typedef struct _THREAD_CMD
{
	TQP_UINT32			dwCmdCode;		// command code
	TQP_OSMEvtgrpID		pEvent;			// command acknowledge event
	TQP_UINT32			dwEventMask;	// mask for the event
	EQP_Errors			dwError;		// error code
	THREAD_CMD_CALLBACK_FUNC	pFuncCallback;	// cmd callback function pointer
	TQP_PVOID			pParam;			// cmd parameter
} THREAD_CMD, *PTHREAD_CMD;



#define MAX_THREAD_CMD_ENTRIES			24

// hide the template syntax
//
typedef QUEUE_ENTRY_CPP<PTHREAD_CMD>	CMD_ENTRY;
typedef CppQueue<PTHREAD_CMD>			CCmdQueue;


/////////////////////////////////////////////////////////////////////////////
//
//


class CppThread
{
public :
	CppThread(char *Name,
			  TQP_UINT32 priority,
			  TQP_OSMEvtgrpID evtWait,
			  TQP_OSMEvtgrpID evtReply
			  )
		: m_threadID(QP_NULL)
		, m_priority(priority)
		, m_EvtWait(evtWait)
		, m_EvtReply(evtReply)

		, m_pFreeCmdQueue(QP_NULL)
		, m_pCmdQueue(QP_NULL)
	{
	}
	CppThread(char *Name,
			  TQP_UINT32 priority
			  )
		: m_threadID(QP_NULL)
		, m_priority(priority)
		, m_EvtWait(QP_NULL)
		, m_EvtReply(QP_NULL)

		, m_pFreeCmdQueue(QP_NULL)
		, m_pCmdQueue(QP_NULL)
	{
	}
	virtual ~CppThread()
	{
	}


	// thread command interface
	//

	TQP_BOOL SetCommand(PTHREAD_CMD pCmd)
	{
		if (!m_pFreeCmdQueue || 
			!m_threadID
			)
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_WARN),
						 "CppThread::SetCommand() NO THREAD YET!!!\n"
						 ));
			return (QP_FALSE);
		}

		// get a free queue entry from the free queue
		//
		CMD_ENTRY *pEntry = m_pFreeCmdQueue->GetOneEntry();

		// save it to the command queue
		//
		if (pEntry)
		{
			TQP_BOOL		fRet = QP_TRUE;
			TQP_OSMEvtgrpID	pEvent;
			TQP_UINT32		dwEventMask;	// mask for the event


			// clear command error
			//
			pCmd->dwError = QPERR_SUCCESS;

			// save the command to the queue entry
			//
			pEntry->Data = pCmd;
			pEvent = pCmd->pEvent;
			dwEventMask = pCmd->dwEventMask;

			// reset command ack event
			//
			if (pEvent && 
				dwEventMask
				)
			{
				QPOSMClearEvtgrp(pEvent, 
								 dwEventMask
								 );
			}
		
			// addd the command to the device command queue
			//
			m_pCmdQueue->AddEntry(pEntry);

			// signal new command
			//
			QPOSMSetEvtgrp(m_EvtWait, 
						   THREAD_EVENT_NEW_CMD
						   );

			// wait on command ack
			//
			if (pEvent)
			{
				EQP_Errors	qpStatus;
				TQP_UINT32	ulEvents;

				qpStatus = QPOSMWaitEvtgrp(pEvent,
										   dwEventMask, 
										   &ulEvents, 
										   2000
										   );
				QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
							 "CppThread::SetCommand() waiting for cmd(0x%08lx) Status(%d)\n", 
							 pCmd->dwCmdCode,
							 qpStatus
							 ));

				fRet = QP_SUCCESS(qpStatus) && QP_SUCCESS(pCmd->dwError);
			}

			return (fRet);
		}	
		else
		{
			return (QP_FALSE);
		}
	}

	inline TQP_UINT32 getThreadPriority() {return (m_priority);}

protected:

	// Thread creation and clean up
	//
	virtual TQP_BOOL ThreadInit()
	{
		EQP_Errors	qpStatus = QPERR_SUCCESS;
		int			i;

		if (!m_threadID)
		{
			// create command queues
			//
			m_pFreeCmdQueue = new CCmdQueue(QP_NULL,
											OBJECT_QUEUE, 
											OBJECT_CRITICAL_HEAVY,
											(QUEUE_CALLBACK)QP_NULL
											);
			if (!m_pFreeCmdQueue)
			{
				QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
							 "CppThread::ThreadInit() - Unable to create CCmdQueue!\n"
							 ));
				return (QP_FALSE);
			}

			m_pCmdQueue = new CCmdQueue(QP_NULL, 
										OBJECT_QUEUE, 
										OBJECT_CRITICAL_HEAVY,
										(QUEUE_CALLBACK)QP_NULL
										);
			if (!m_pCmdQueue)
			{
				QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
							 "CppThread::ThreadInit() - Unable to create CCmdQueue!\n"
							 ));
				return (QP_FALSE);
			}

			// initialize command queues
			//
			for (i = 0; i < MAX_THREAD_CMD_ENTRIES; i++)
			{
				m_Entries[i].pNext = QP_NULL;
				m_Entries[i].Data = QP_NULL;
				m_pFreeCmdQueue->AddEntry(&m_Entries[i]);
			}

			// create thread
			//
			TQP_OSMThreadParms	threadParam;

			threadParam.procP = CppThread::StaticThreadProc;
			threadParam.parm = (void *)this;
			threadParam.stackP = QP_NULL;
			threadParam.stackSize = 8192;
			threadParam.priority = m_priority;

			qpStatus = QPOSMCreateThread(&m_threadID, 
										 &threadParam
										 );
		}

		return (QP_SUCCESS(qpStatus));
	}

	virtual TQP_BOOL ThreadDone()
	{
		EQP_Errors	qpStatus;

		// terminate thread
		//
		if (m_threadID)
		{
			// Flush all the outstanding commands
			//
			FlushCommands();

			// signal the thread that we are done
			//
			ThreadExit();

			// put minimum delay here just to be on the safe side.
			// OS manager already synchronize thread exit and 
			// thread deletion as of Nov 13th, 2006
			//
			QPTMDelayMilliS(15);

			// call OS manager delete thread function
			//
			qpStatus = QPOSMDeleteThread(m_threadID);

			// nil the thread id regardless
			m_threadID = QP_NULL;

			if (!QP_SUCCESS(qpStatus))
			{
				QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_ERROR),
							 "CppThread::ThreadDone() QPOSMDeleteThread failed status(%d)!!!\n", 
							 qpStatus
							 ));
			}
		}

		// destroy the device queues
		//
		if (m_pFreeCmdQueue)
		{
			delete m_pFreeCmdQueue;
			m_pFreeCmdQueue = QP_NULL;
		}

		if (m_pCmdQueue)
		{
			delete m_pCmdQueue;
			m_pCmdQueue = QP_NULL;
		}

		return (QP_TRUE);
	}

	virtual TQP_BOOL ThreadExit()
	{
		EQP_Errors	qpStatus;
		TQP_UINT32	ulEvents;

		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
					 "CppThread::ThreadExit()\n"
					 ));

		QPOSMClearEvtgrp(m_EvtReply, 
						 THREAD_EVENT_EXIT_DONE
						 );
		QPOSMSetEvtgrp(m_EvtWait, 
					   THREAD_EVENT_EXIT
					   );
		qpStatus = QPOSMWaitEvtgrp(m_EvtReply, 
								   THREAD_EVENT_EXIT_DONE, 
								   &ulEvents, 
								   2000
								   );
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
					 "CppThread::ThreadExit() Status(%d)\n", 
					 qpStatus
					));

		return (QP_SUCCESS(qpStatus));
	}

	// thread command related functions
	//
	virtual TQP_BOOL ProcessCommand()
	{
		CMD_ENTRY		*pEntry = QP_NULL;
		THREAD_CMD		*pCmd = QP_NULL;
		TQP_OSMEvtgrpID	pEvent;
		TQP_UINT32		dwEventMask;

		if (!m_threadID)
			return (QP_FALSE);

		// get and process command
		//
		pEntry = GetCommand();

		if (pEntry)
		{
			pCmd = pEntry->Data;

			// in case of an asychronous callback command, the storage 
			// for the device command may no longer be valid after the 
			// callback returned. save the event now for later reference
			//
			pEvent = pCmd->pEvent;
			dwEventMask = pCmd->dwEventMask;

			if (THREAD_CMD_CALLBACK == pCmd->dwCmdCode)
			{
				if (pCmd->pFuncCallback)
				{
					pCmd->dwError = QPERR_SUCCESS;
					pCmd->pFuncCallback(pCmd->pParam);
				}
				else
				{
					pCmd->dwError = QPERR_PARMS;
				}
			}
			else
			{
				DoCommand(pCmd);
			}

			// signal caller we have processed the command
			if (pEvent && 
				dwEventMask
				)
			{
				QPDbgMPrintf((QPDBGMCOMPLVL(g_QPCodecDBGCompID, QPDBGM_LEVEL_TRACE),
							 "CppThread::ProcessCommand() -- Signal Caller --\n"
							 ));
				QPOSMSetEvtgrp(pEvent, 
							   dwEventMask
							   );
			}

			FreeCommand(pEntry);

			return (QP_TRUE);
		}
		else
		{
			return (QP_FALSE);
		}
	}

	inline CMD_ENTRY * GetCommand()
	{
		return (m_pCmdQueue->GetOneEntry());
	}
	
	inline void FreeCommand(CMD_ENTRY *pEntry)
	{
		pEntry->pNext = QP_NULL;
		pEntry->Data = QP_NULL;
		if (m_pFreeCmdQueue)
			m_pFreeCmdQueue->AddEntry(pEntry);
	}

	void FlushCommands()
	{
		m_pFreeCmdQueue->FlushQueue();

		while (ProcessCommand());
	}

	virtual void DoCommand(PTHREAD_CMD pCmd)
	{
		pCmd->dwError = QPERR_SUCCESS;
	}


	// thread proc
	//
	virtual int ThreadProc() = 0;
	static void StaticThreadProc(void *Context) {((CppThread *)Context)->ThreadProc();}


	// thread management
	//
	TQP_OSMThreadID	m_threadID;

	TQP_UINT32		m_priority;
	TQP_OSMEvtgrpID	m_EvtWait;
	TQP_OSMEvtgrpID	m_EvtReply;


	// thread commands
	//
	CMD_ENTRY		m_Entries[MAX_THREAD_CMD_ENTRIES];	// device command queue entries
	CCmdQueue		*m_pFreeCmdQueue;					// thread command entry pool
	CCmdQueue		*m_pCmdQueue;						// thread command queue

};

#endif //__cplusplus


#endif //__CTHREAD_H__




