///////////////////////////////////////////////////////////////////////////////
//        
//
//		File Manager interface
//
//	Platform-independent OS interface
//
// History:
//	May 24, 2007	(EC) Created
//
// Authors: (EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __QPFILEMGR_H__
#define __QPFILEMGR_H__

#include "qptypes.h"
#include "qptimemgr.h"

#ifdef __cplusplus
extern "C" {
#endif

// File Manager ID (identifies clients)
typedef void * TQP_FMID;

// Volume ID
typedef void * TQP_FMVolID;

// Open file ID
typedef void * TQP_FMFileID;

typedef enum _EQP_FMEvents {
	QPFM_EVENT_NONE,
	QPFM_EVENT_VOLMOUNT,
	QPFM_EVENT_VOLUNMOUNT,
	QPFM_EVENT_VOLNAMECHANGE
} EQP_FMEvents;

// Callback for client notification
typedef void (*TQP_FMCallBack) (void * refCon, EQP_FMEvents code, void * param);

//
// Capabilities structure
typedef struct _TQP_FMCaps {
	void *		dummy;
} TQP_FMCaps, *PTQP_FMCaps;

//
// Returns the list of capabilities for this implementation
//
void QPFileMGetCaps (PTQP_FMCaps capsP);


//
// Registers a client with the File Manager; returns a File Manager ID to be used in other calls
// to identify this client.
// The client registers a callback function (cbP) and the corresponding context recovery
// value (refCon).
//
TQP_FMID QPFileMRegister (void * refCon, TQP_FMCallBack cbP);

//
// Unregisters a registered client by ID
//
EQP_Errors QPFileMUnregister (TQP_FMID fmID);


///////////////////////////////////////////////////////////////////////////////
//
// Volume handling
//

//
// Get volume IDs
//
EQP_Errors QPFileMGetVolList (TQP_FMID fmID, TQP_FMVolID * volIDsP, unsigned long * volNumP);

//
// Get current volume
//
EQP_Errors QPFileMGetCurVol (TQP_FMID fmID, TQP_FMVolID * volIDP);

//
// Set current volume
//
EQP_Errors QPFileMSetCurVol (TQP_FMID fmID, TQP_FMVolID volID);

//
// Get the ID of a Volume from a path name
//
TQP_FMVolID QPFileMGetVolID (TQP_FMID fmID, const PTQP_CHAR);

// Static Volume information

typedef enum _EQP_FMFSTypes {
	QPFM_FSTYPE_FAT12,
	QPFM_FSTYPE_FAT16,
	QPFM_FSTYPE_FAT32,
	QPFM_FSTYPE_NTFS
} EQP_FMFSTypes;

typedef struct _TQP_FMVolInfo {
	EQP_FMFSTypes	type;				// File System type
	TQP_UINT64		size;				// number of bytes in this volume
	TQP_UINT64		maxFileSize;		// max file size in bytes
	TQP_UINT32		maxFileNameLen;		// max file name length in bytes
	TQP_BOOL		unicode;			// true if can handle Unicode file names
	TQP_BOOL		removable;			// true if can be dynamically mounted/unmounted
	TQP_BOOL		persistent;			// true if preserves contents across power cycles
	TQP_BOOL		remote;				// true if it is a remote (network) drive
	TQP_BOOL		readOnly;			// true if it is read only (e.g. CD-ROM)
} TQP_FMVolInfo, *PTQP_FMVolInfo;

//
// Returns information about an available Volume
//
EQP_Errors QPFileMGetVolInfo (TQP_FMVolID volID, PTQP_FMVolInfo volInfoP);

//
// Returns a Volume's name
//
EQP_Errors QPFileMGetVolName (TQP_FMVolID volID, PTQP_CHAR nameP, unsigned long * lenP);

//
// Sets a volume's name
//
EQP_Errors QPFileMSetVolName (TQP_FMVolID volID, const PTQP_CHAR nameP);

//
// Returns the size in bytes of the free space on a Volume
//
EQP_Errors QPFileMGetVolFree (TQP_FMVolID volID, PTQP_UINT64 freeP);

//
// Returns TRUE if the volume (volID) is mounted
//
TQP_BOOL QPFileMIsVolAvailable (TQP_FMVolID volID);


///////////////////////////////////////////////////////////////////////////////
//
// File/Directory management
//

typedef struct _TQP_FMFileInfo {
	TQP_UINT64		size;
	TQP_BOOL		isDirectory;
	TQP_BOOL		readOnly;
	TQPTM_Time		creTime;		// creation date and time
	TQPTM_Time		modTime;		// last modifictaion (write) date and time
	TQPTM_Time		accTime;		// last access (read or write) time
	TQP_CHAR *		nameP;			// caller supplied storage for file name
	TQP_UINT32		nameLen;		// size of name or of the caller's buffer
} TQP_FMFileInfo, *PTQP_FMFileInfo;


//
// Returns the pathname of the current directory.  Every volume ID object
// maintains a current directory.  When the volume is first mounted, the
// current directory is the root directory.
// pnP is a caller's string pointer of length *lenP.  The actual length
// is returned in *lenP. If pnP is NULL the length is still returned in *lenP.
// 
EQP_Errors QPFileMGetCurDir (TQP_FMVolID volID,
							PTQP_CHAR pnP,
							unsigned long * lenP);

//
// Sets the current directory on a given volume ID.  .
// pnP is a NULL-terminated string with  the new directory path name: it can be
// absolute or relative to the previous current directory.
//
EQP_Errors QPFileMSetCurDir (TQP_FMVolID volID, PTQP_CHAR pnP);


//
// Starts listing a named directory (pnP).
// Only one listing operation can be active at any time on a volume ID.
// When this function is called, a previous listing operation (if any) is
// cancelled and a new one started. If this function is not called before
// calling QPFileMDirListNext the directory searched is the current directory.
// The directory name in pnP is either absolute or relative to the current directory.
//
EQP_Errors QPFileMDirListStart (TQP_FMVolID volID, const PTQP_CHAR pnP);

//
// Recursively retrieve an item's name from a directory on which listing was
// started using QPFileMDirListStart.  The item info is returned in
// *fileInfoP.  The listing operation is over when QPERR_NOTFOUND is returned.
//
EQP_Errors QPFileMDirListNext (TQP_FMVolID volID, PTQP_FMFileInfo fileInfoP);

//
// Returns the info for a named file or directory.
// The path name in pnP is either absolute or relative to the current directory.
//
EQP_Errors QPFileMGetNamedInfo (TQP_FMVolID volID, const PTQP_CHAR pnP, PTQP_FMFileInfo fileInfoP);

//
// Sets the read-only flag of a named file
//
EQP_Errors QPFileMSetReadOnly (TQP_FMVolID volID, const PTQP_CHAR pnP, TQP_BOOL readOnly);

//
// Removes (delete) a named file or directory (pnP).
// The path name in pnP is either absolute or relative to the current directory.
//
EQP_Errors QPFileMRemove (TQP_FMVolID volID, const PTQP_CHAR pnP);

//
// Renames a file or a directory.  The old pathname is (pnP); the new pathname
// is (newPnP).
// The path names in pnP and newPnP are either absolute or relative to the
// current directory.
//
EQP_Errors QPFileMRename (TQP_FMVolID volID, const PTQP_CHAR pnP, const PTQP_CHAR newPnP);

//
// Creates a new empty directory.  The new directory pathname in pnP is either
// absolute or relative to the current directory.
//
EQP_Errors QPFileMMakeDir (TQP_FMVolID volID, const PTQP_CHAR pnP);


///////////////////////////////////////////////////////////////////////////////
//
// File access
//

//
// Opens a named file (pnP) on a given Volume (volID).  flags describes various
// sharing and access modes; the ID of the open file is returned at (fIDP).
//
EQP_Errors QPFileMOpen (TQP_FMVolID volID, const PTQP_CHAR pnP,
 					TQP_UINT32 flags, TQP_FMFileID * fIDP);

#define QPFM_ACPRM_READ (1<<0)		// open file for reading
#define QPFM_ACPRM_WRITE (1<<1)		// open file for writing
#define QPFM_ACPRM_TRUNC (1<<2)		// truncates to 0 length an existing file
#define QPFM_ACPRM_EXCL (1<<3)		// open exclusive: only one client can do so
#define QPFM_ACPRM_CREATE (1<<4)	// create new file if it does not exist

//
// Returns information about an open file (fID).  File information is returned in the structure
// pointed by fiP.
//
EQP_Errors QPFileMGetFileInfo (TQP_FMFileID fID, PTQP_FMFileInfo fiP);

//
// Reads from an open file (fID) into a local buffer.  bufP points to a caller-allocated
// buffer to read data into.  The number of bytes to read is at lenP on entry.
// The number of bytes actually read is at lenP on exit.  The file pointer is advanced
// by this amount.
//
EQP_Errors QPFileMRead (TQP_FMFileID fID, void * bufP, TQP_UINT64 * lenP);

//
// Writes data into an open file (fID).  bufP points to a caller-allocated buffer
// with the data to write.  The number of bytes to write is at lenP on entry.  The
// number of bytes actually written is at lenP on exit.  The file pointer is advanced
// by this amount.
//
EQP_Errors QPFileMWrite (TQP_FMFileID fID, const void * bufP, TQP_UINT64 * lenP);

//
// Returns the current pointer in an open file (fID).  The current pointer position
// is returned at posP.
//
EQP_Errors QPFileMGetPos (TQP_FMFileID fID, TQP_UINT64 * posP);

//
// Sets the current pointer in an open file (fID).  The desired pointer position
// is at posP on entry.  The actual pointer position is returned at posP.  It can
// differ from the desired position if attempting to go past the end of file.
//
EQP_Errors QPFileMSetPos (TQP_FMFileID fID, TQP_UINT64 * posP);

//
// Returns the length in bytes of an open file (fID).
//
EQP_Errors QPFileMGetLength (TQP_FMFileID fID, TQP_UINT64 * lenP);

//
// Change the length of an open file (fID).
// The file will be truncated (i.e. data will be lost) if the requested length
// is smaller than the actual file size.
//
EQP_Errors QPFileMSetLength (TQP_FMFileID fID, TQP_UINT64 len);

//
// Closes an open file (fID).
//
EQP_Errors QPFileMClose (TQP_FMFileID fID);


#ifdef __cplusplus
}
#endif

#endif //__QPFILEMGR_H__
