///////////////////////////////////////////////////////////////////////////////
//        
//
//		Definitions for Flash usage
//
//
// History:
//	Aug 7, 2009	(EC) Created
//
// Authors: (EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __QPFLASH_H__
#define __QPFLASH_H__

#include "qptypes.h"

//
// The Flash is organized in sectors; currently, sectors have 64kB.
// A sector is the smallest unit that can be erased.
// The first sector (sector #0) is reserved for the boot code, i.e. the code
// that first executes when the system is powered up; this code must be
// extremely simple and reliable since it has to fit in a single sector
// and cannot be replaced easily (risk of bricking the device if done
// on the field).
// Every other sector contains an 8-word header (32 bytes) at its beginning.
// The header starts with a tag that identifies the type of sector.  The
// following tags are defined:
//
//	Image Start: this sector is the first sector of a firmware image
//	Image Mid: this sector is a continuation sector of a firmware image
//	Configuration: this sector contains configuration data.
//
// Image sectors have additional information in the header.
// All sectors have a checksum as the last word: in a valid sector, this
// checksum must be equal to the sum of all other words in the header.
//

// Sector size is set to 64kB (dictated by our chips' limitation in
// Flash erase commands).
// Must be a word multiple.
#define QPFLASH_SECT_SIZE 0x10000

#define RAWFLASH_SECT_WORDS (QPFLASH_SECT_SIZE/4)
#define RAWFLASH_SECT_MAXBYTES (QPFLASH_SECT_SIZE)
#define RAWFLASH_SECT_MAXWORDS ((QPFLASH_SECT_SIZE)/4)

#define QPFLASH_SECT_WORDS (QPFLASH_SECT_SIZE/4)
#define QPFLASH_SECT_MAXBYTES (QPFLASH_SECT_SIZE - sizeof(TQPFlashHeader))
#define QPFLASH_SECT_MAXWORDS ((QPFLASH_SECT_SIZE - sizeof(TQPFlashHeader))/4)
#define QPFLASH_HEADER_BYTES (sizeof(TQPFlashHeader))
#define QPFLASH_HEADER_WORDS (sizeof(TQPFlashHeader)/4)

// For the tag field of the header
#define QPFLASH_IMAGE_START_TAG QP_MAKEFOURCC('i','m','g','s')
#define QPFLASH_IMAGE_MID_TAG QP_MAKEFOURCC('i','m','g','m')
#define QPFLASH_CONFIG_START_TAG QP_MAKEFOURCC('c','n','f','s')
#define QPFLASH_CONFIG_MID_TAG QP_MAKEFOURCC('c','n','f','m')

#define QPFLASH_IS_STARTTAG(tag) ((tag == QPFLASH_IMAGE_START_TAG) || (tag == QPFLASH_CONFIG_START_TAG))


// For the tagType field of the header
#define QPFLASH_IMGTYPE_LOADER QP_MAKEFOURCC('l','d','F','W')
#define QPFLASH_IMGTYPE_MAIN QP_MAKEFOURCC('m','a','F','W')
#define QPFLASH_IMGTYPE_AUDIO QP_MAKEFOURCC('a','u','F','W')
#define QPFLASH_IMGTYPE_DIAG QP_MAKEFOURCC('d','i','a','g')
#define QPFLASH_IMGTYPE_UBOOT QP_MAKEFOURCC('u','b','o','t')

//
// Every sector starts with a header
// Must be a word multiple. 
//
typedef struct _TQPFlashHeader {
	TQP_UINT32		tag;			// identify the type of sector using the QPFLASH_xyz_TAG constants
	TQP_UINT32		words;			// number of valid words in this sector: <= (FLASH_SECT_SIZE-32)/4
	TQP_UINT32		tagType;		// boot sector: reserved (0)
									// image sectors: kind of image  (QPFLASH_IMGTYPE_xyz)
									// config sectors: names the settings block
	TQP_UINT32		dateTime;		// a date-time stamp used for safe update (higher -> more recent)
	TQP_UINT32		nextSector;		// 0 if there is no next sector, otherwise is the next sector #
	TQP_UINT32		loadAddress;	// boot sector: memory address where the minibooter is loaded and run
									// image sectors: memory address where to load the image data (for every image sector)
									// config sectors: reserved (0)
	TQP_UINT32		dataChecksum;	// checksum over the sector data
	TQP_UINT32		hdrChecksum;	// header checksum
} TQPFlashHeader, *PTQPFlashHeader;

typedef struct {
	TQPFlashHeader	header;
	TQP_UINT32		data[QPFLASH_SECT_MAXWORDS];
} TQPFlashSector, *PTQPFlashSector;


// Macros to extract date and time info from the dateTime field
#define QPFLASH_GETYEAR(dateTime) (2000 + ((dateTime >> 26) & 0x3F))
#define QPFLASH_GETMONTH(dateTime) QPFLASH_CLIPVALUE(((dateTime >> 22) & 0xF),1,12)
#define QPFLASH_GETDAY(dateTime) QPFLASH_CLIPVALUE(((dateTime >> 17) & 0x1F),1,31)
#define QPFLASH_GETHOUR(dateTime) QPFLASH_CLIPVALUE(((dateTime >> 12) & 0x1F),0,23)
#define QPFLASH_GETMINUTE(dateTime) QPFLASH_CLIPVALUE(((dateTime >> 6) & 0x3F),0,59)
#define QPFLASH_GETSECOND(dateTime) QPFLASH_CLIPVALUE((dateTime & 0x3F),0,59)

// Macro to construct the dateTime field
#define QPFLASH_MAKEDATETIME(year,month,day,hour,minute,second) \
	((TQP_UINT32) (										\
	((QPFLASH_CLIPVALUE((year-2000),0,127) & 0x3F) << 26)	\
	| ((QPFLASH_CLIPVALUE(month,1,12) & 0xF) << 22)		\
	| ((QPFLASH_CLIPVALUE(day,1,31) & 0x1F) << 17)		\
	| ((QPFLASH_CLIPVALUE(hour,0,23) & 0x1F) << 12)		\
	| ((QPFLASH_CLIPVALUE(minute,0,59) & 0x3F) << 6)	\
	| (QPFLASH_CLIPVALUE(second,0,59) & 0x3F)			\
	))

// Macro to add 1 second to a dateTime value
// WARNING: does not handle the month length
// WARNING: will fail at 23:59:59 on Dec 31, 2127
#define QPFLASH_ADD1SEC(dateTime) 	\
{													\
	TQP_UINT32	temp = dateTime;					\
	TQP_UINT32	second = QPFLASH_GETSECOND(temp);	\
	TQP_UINT32	minute = QPFLASH_GETMINUTE(temp);	\
	TQP_UINT32	hour = QPFLASH_GETHOUR(temp);		\
	TQP_UINT32	day = QPFLASH_GETDAY(temp);			\
	TQP_UINT32	month = QPFLASH_GETMONTH(temp);		\
	TQP_UINT32	year = QPFLASH_GETYEAR(temp);		\
													\
	second += 1;									\
	if (second > 59) {								\
		second = 0;									\
		minute += 1;								\
		if (minute > 59) {							\
			minute = 0;								\
			hour += 1;								\
			if (hour > 23) {						\
				hour = 0;							\
				day += 1;							\
				if (day > 31) {						\
					day = 1;						\
					month += 1;						\
					if (month > 12) {				\
						month = 1;					\
						year += 1;					\
					}								\
				}									\
			}										\
		}											\
	}												\
													\
	dateTime = QPFLASH_MAKEDATETIME(year,month,day,hour,minute,second);	\
}



#define QPFLASH_CLIPVALUE(val,min,max) \
	((TQP_UINT32) \
	((((TQP_UINT32) (val)) <= ((TQP_UINT32) (min))) ? (min)	\
	: ((((TQP_UINT32) (val)) >= ((TQP_UINT32) (max))) ? (max) : (val)))	\
	)

#endif //__QPFLASH_H__

