///////////////////////////////////////////////////////////////////////////////
//        
//
//		GPIO Manager interface
//
// History:
//	Sep 17, 2009	(EC) Created
//
// Authors: (EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __QPGPIOMGR_H__
#define __QPGPIOMGR_H__

#include "qptypes.h"

#ifdef __cplusplus
extern "C" {
#endif

//
// All functions take as first parameter the number of the GPIO bit
// being controlled.
//
// All functions return an error code. Common error codes:
//	QPERR_SUCCESS	completed successfully
//	QPERR_PARMS		invalid GPIO bit number
//
// One ISR can be registered per each GPIO bit that is set for input.


//
// Set a GPIO bit as output.
// Return QPERR_INVALID if the bit is already in use as an input and an
// ISR is registered on it.
//
EQP_Errors QPGPIOSetOutput (unsigned int bitNum);

//
// Set a GPIO bit as input 
//
EQP_Errors QPGPIOSetInput (unsigned int bitNum);

//
// Get the direction of a GPIO bit.
// Return QPERR_PARMS if isInputP is NULL or an invalid pointer.
//
EQP_Errors QPGPIOGetDirection (unsigned int bitNum, TQP_BOOL * isInputP);

//
// Read the value of a GPIO bit. (*valueP: 0 => 0, non-0 => 1)
// Return QPERR_PARMS if valueP is NULL or an invalid pointer.
//
EQP_Errors QPGPIOGetBit (unsigned int bitNum, unsigned int * valueP);

//
// Sets the value of a GPIO bit. (value: 0 => 0, non-0 => 1)
// Return QPERR_INVALID if the GPIO bit is set for input.
//
EQP_Errors QPGPIOSetBit (unsigned int bitNum, unsigned int value);


//
// Proc type of the GPIO ISR.
// The GPIO ISR is very similar to a generic ISR: it is executed at IRQ
// level.  One difference is that the user does not need to clear the
// interrupt as it is done by the GPIO manager.
// Only 0->1 transitions of a GPIO bit will generate an interrupt.
//
typedef void (*TQP_GPIOISR) (void * refCon, unsigned int bitNum);

//
// Register an interrupt service routine on a GPIO input bit
// Return QPERR_PARMS if the isr pointer is NULL.
// Return QPERR_INVALID if the GPIO bit is set for output.
// The refCon value is the first parameter to the ISR function
//
EQP_Errors QPGPIORegisterISR (unsigned int	bitNum,
							  TQP_GPIOISR	isr,
							  void *		refCon);

//
// Unregister an interrupt service routine
// Return QPERR_INVALID if the GPIO bit is set for output.
// Return QPERR_FAIL if there was no ISR registered on this bit.
//
EQP_Errors QPGPIOUnregisterISR (unsigned int bitNum);

//
// Change GPIO interrupt enable state (TRUE means enabled).
// Return QPERR_PARMS if the isrStatusP pointer is NULL.
// Return QPERR_INVALID if the GPIO bit is set for output.
//
EQP_Errors QPGPIOSetISRStatus (unsigned int bitNum, TQP_BOOL * isrStatusP);

//
// Retunr GPIO interrupt enable state (TRUE means enabled).
// Return QPERR_PARMS if the isrStatusP pointer is NULL.
// Return QPERR_INVALID if the GPIO bit is set for output.
//
EQP_Errors QPGPIOGetISRStatus (unsigned int bitNum, TQP_BOOL * isrStatusP);


//
// Helper functions: simply enable or disable; return FALSE if they fail.
//
TQP_BOOL QPGPIOEnableISR (unsigned int bitNum); 
TQP_BOOL QPGPIODisableISR (unsigned int bitNum); 


#ifdef __cplusplus
}
#endif

#endif //__QPGPIOMGR_H__
