///////////////////////////////////////////////////////////////////////////////
//        
//
//		Memory Manager interface
//
//	Platform-independent memory management interface
//
// History:
//	Aug 4, 2006	(EC) Created
//
// Authors: (EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __QPMEMMGR_H__
#define __QPMEMMGR_H__

#include "qptypes.h"

#ifdef __cplusplus
extern "C" {
#endif

// ID for a pool and its associated memory manager
typedef void * TQP_MMID;

//
// Capabilities structure
typedef struct _TQP_MMCaps {
	TQP_BOOL	stdMM;			// true if standard memory manager is supported
	TQP_BOOL	clientPools;	// true if client pools are supported
	TQP_BOOL	poolTypes;		// true if pool types are supported (an alternative memory manager interface)
	TQP_BOOL	dualMapping;	// true if supports double mapping of memory (both cached and non-cached)
	TQP_UINT32	pageAlignment;	// log2 of the virtual memory page size 
	TQP_UINT32	cacheAlignment;	// log2 of the cache line size
	TQP_UINT32	minAlignment;	// minimum alignment for the pool types interface
} TQP_MMCaps, *PTQP_MMCaps;

//
// Returns the list of capabilities for this implementation
//
void
QPMMGetCaps (PTQP_MMCaps capsP);


///////////////////////////////////////////////////////////////////////////////
//
// Cache handling
//

//
// Flush cache to memory
//
TQP_BOOL QPMMFlush (void * virAddr, unsigned long size);

//
// Invalidate cache
//
TQP_BOOL QPMMInvalidate (void * virAddr, unsigned long size);


///////////////////////////////////////////////////////////////////////////////
//
// Virtual/Physyical translation
//

// Returns NULL if the Phy address is not mapped
void * QPMMPhy2Vir (void * phyAddr);

// Returns (-1) if the Virtual address is not mapped
void * QPMMVir2Phy (void * virAddr);


///////////////////////////////////////////////////////////////////////////////
//
// Dual address handling
//

// Return FALSE if the Physical address is not mapped
TQP_BOOL QPMMPhy2VirCached (void ** addrP);
TQP_BOOL QPMMPhy2VirNonCached (void ** addrP);

// Returns TRUE if the virtual address is cached
TQP_BOOL QPMMIsCached (void * virAddr);

// Returns TRUE if the virtual address is non-cached
TQP_BOOL QPMMIsNonCached (void * virAddr);

// Return FALSE if the Virtual address is not mapped
TQP_BOOL QPMMVirMap2Cached (void ** virAddrP);
TQP_BOOL QPMMVirMap2NonCached (void ** virAddrP);


typedef enum _EQP_MMPhyAddrTypes {
	QPMM_PHYADDR_RAM,			// read/write RAM
	QPMM_PHYADDR_ROM,			// read only memory
	QPMM_PHYADDR_FLASH,			// flash memory
	QPMM_PHYADDR_REG,			// register space
	QPMM_PHYADDR_INVAL,			// invalid memory access
} EQP_MMPhyAddrTypes;

//
// Return the type of memory at a given physical address.
//
EQP_MMPhyAddrTypes QPMMGetPhyAddrType (void * phyAddr);


///////////////////////////////////////////////////////////////////////////////
//
// Client-pools memory management functions
//

//
// Initializes the memory manager on a client memory pool;
// If no memory pool is passed in (ptr = NULL), it will use the standard memory
// management functions (if supported: stdMM cap).
// If a memory pool is passed in (ptr != NULL), it will return NULL if client
// pools are not supported (clientPools cap) or if the pool is not big enough.
// The non-NULL id returned must be used in all other calls as a context
// reference for the memory manager.
// 
TQP_MMID QPMMInit (void * ptr, unsigned long bytes);

//
// malloc function; uses the standard memory manager if it is supported and
// mmID is NULL; otherwise it will use the memory pool identified by mmID and
// previously initialized.
//
void * QPMMMalloc (TQP_MMID mmID, unsigned long bytes);

//
// free function; uses the standard memory manager if it is supported and
// mmID is NULL; otherwise it will use the memory pool identified by mmID and
// previously initialized.
//
void QPMMFree (TQP_MMID mmID, void * ptr);

//
// Terminates a previously allocated memory pool identified by mmID.
//
void QPMMTerm (TQP_MMID mmID);


///////////////////////////////////////////////////////////////////////////////
//
// Pool-less memory management functions
//

enum {
	QPMM_ATTR_ANY = 0,			// any
	QPMM_ATTR_TRUE,				// the attribute must be satisified
	QPMM_ATTR_FALSE				// the attribute must not be satisfied
};

enum {
	QPMM_ALIGN_ANY = 0,
	QPMM_ALIGN_CACHE = 1
};

#define SetMemType(paging,caching,contiguous,alignment)\
	((paging & 0x3) | ((caching & 0x3) << 2)	\
	| ((contiguous & 0x3) << 4) | ((alignment & 0xFF) << 24))

#define GetPagingAttr(memType) (memType & 0x3)
#define GetCachingAttr(memType) ((memType >> 2) & 0x3)
#define GetContiguousAttr(memType) ((memType >> 4) & 0x3)
#define GetAlignment(memType) ((memType >> 24) & 0xFF)

void * QPMMMalloc2 (PTQP_UINT32 memTypeP, unsigned long bytes);
void * QPMMMalloc2Cached (unsigned long bytes);
void * QPMMMalloc2NonCached (unsigned long bytes);

void QPMMFree2 (void * memP);



///////////////////////////////////////////////////////////////////////////////
//
// Memory to memory DMA service
//

// ID for an M2M request
typedef void * TQP_MMM2MReqID;


typedef enum {
	QPMM_M2MSTATUS_ERR = -1,		// completed with errors
	QPMM_M2MSTATUS_SUCCESS = 0,		// completed successfully
	QPMM_M2MSTATUS_PENDING = 1,		// request in progress
	QPMM_M2MSTATUS_NOTSTARTED = 2	// request not started yet
} EQP_MMM2MReqStates;

// Completion routine; called when the request is completed 
typedef void (*TQP_MMM2MReqCompl) (void * refCon, EQP_MMM2MReqStates state);

typedef enum {
	QPMM_M2MBYTESWAP_NONE = 0,
	QPMM_M2MBYTESWAP_16,
	QPMM_M2MBYTESWAP_32
} EQP_MMM2MByteSwap;


typedef struct _TQP_MMM2MReqParms {
	void *				srcAddr;	// source memory address
	void *				dstAddr;	// destination memory address
	TQP_UINT32			bytes;		// size in bytes
	EQP_MMM2MByteSwap	byteSwap;	// byte swapping option
} TQP_MMM2MReqParms, *PTQP_MMM2MReqParms;


//
// Start an M2M request asynchronously; the complProc function is called when
// the request completes. The completion routine is executed at interrupt level.
// The request ID is returned in *reqIDP.  This ID can be used to kill the
// request later on and/or to get status information.
//
EQP_Errors QPMMStartM2MRequest (PTQP_MMM2MReqParms reqParmsP,
					TQP_MMM2MReqCompl complProc, void * refCon,
					TQP_MMM2MReqID  * reqIDP);

//
// Return the status of an M2M request.
//
EQP_Errors QPMMM2MRequestState (TQP_MMM2MReqID reqID, EQP_MMM2MReqStates * stateP);

//
// Kill a pending M2M request
//
EQP_Errors QPMMKillM2MRequest (TQP_MMM2MReqID reqID);


#ifdef __cplusplus
}
#endif

#endif
