///////////////////////////////////////////////////////////////////////////////
//        
//
//		OS Manager interface
//
//	Platform-independent OS interface
//
// History:
//	Aug 9, 2006	(EC) New interrupt handling interface
//	Aug 4, 2006	(EC) Created
//
// Authors: (EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __QPOSMGR_H__
#define __QPOSMGR_H__

#include "qptypes.h"

#ifdef __cplusplus
extern "C" {
#endif


typedef enum _EQP_OSMOSTypes {
	QPOS_TYPE_UNKNOWN,
	QPOS_TYPE_WIN32,
	QPOS_TYPE_WINKERNEL,
	QPOS_TYPE_ITRON,
	QPOS_TYPE_LINUX,
	QPOS_TYPE_LINUXKERNEL,
	QPOS_TYPE_WINCE,
	QPOS_TYPE_SOS,
	QPOS_TYPE_MACOSX,
	QPOS_TYPE_MACOSXKERNEL
} EQP_OSMOSTypes;


// Version types for ITRON
enum _OSVersionTypes {
	QPOS_ITRON_SANYO = 0x00,
	QPOS_ITRON_NUCORE = 0x01
};
#define QP_ITRONosVersion(type,number) ((unsigned long) (((type & 0xFF) << 24) | (number & 0xFFFFFF)))

//
// Capabilities structure
typedef struct _TQP_OSMCaps {
	EQP_OSMOSTypes	osType;			// what OS is running?
	unsigned long	osVersion;		// OS version (OS specific)
	unsigned long	defStackSz;		// default stack size in bytes
	long			maxMsgQueues;	// max number of message queues per thread
									// 0 means messaging is not supported
									// negative number means unlimited (< 2^31-1)
} TQP_OSMCaps, *PTQP_OSMCaps;

//
// Returns the list of capabilities for this implementation
//
void QPOSMGetCaps (PTQP_OSMCaps capsP);


///////////////////////////////////////////////////////////////////////////////
//
//	THREADS
//
//
typedef void (*TQP_ThrProc) (void *);

#define QPOS_THRNAME_LEN 32

typedef struct _TQP_OSMThreadParms {
	TQP_ThrProc		procP;
	void *			parm;
	void *			stackP;
	TQP_UINT32		stackSize;
	TQP_UINT32		priority;
	TQP_CHAR		name[QPOS_THRNAME_LEN];
} TQP_OSMThreadParms, *PTQP_OSMThreadParms;

// The priority field of the TQP_OSMThreadParms structure is also used to
// set the number of message queues that should be activated when the thread
// is created.
// The most significant 16 bits contain the number of message queues.
// The lower 16 bits contain the priority level.
// Priority levels definitions
#define QPOS_THRPRI_DEFAULT	0
#define QPOS_THRPRI_LOWEST	1
#define QPOS_THRPRI_HIGHEST	16

// Use the following macro to set the priority field
#define QPOSMSetPriority(queues,pri)\
	(((queues & 0xFFFF) << 16) | (pri & 0xFFFF))


typedef void * TQP_OSMThreadID;

//
// Create a new thread; returns the ID of the new thread in tIDP.
// The stack memory can be provided by the caller or allocated by the system
// if stackP is NULL.  If also stackSize is 0, the default stack size will be
// allocated.
//
EQP_Errors QPOSMCreateThread (TQP_OSMThreadID * tIDP,
 				PTQP_OSMThreadParms parmsP);

//
// Delete an existing thread; use caution since it does not free any memory or
// other resources allocated by the thread.
//
EQP_Errors QPOSMDeleteThread (TQP_OSMThreadID tID);

//
// Return a thread's name.  nameP must point to an array of QPOS_THRNAME_LEN
// chars.  If the thread ID is not valid, it returns QPERR_INVALID; in this case
// the nameP array will still be filled with the data found and might not
// contain proper printable ASCII characters.
//
EQP_Errors QPOSMGetThreadName (TQP_OSMThreadID tID, PTQP_CHAR nameP);

//
// Get own thread ID
// Returns QP_NULL if called from a non-thread context.
//
TQP_OSMThreadID QPOSMGetThreadID (void);


///////////////////////////////////////////////////////////////////////////////
//
// SEMAPHORES
//
// We use the counting semaphore paradigm: a semaphore has a max count
// corresponding to the number of resources it controls.  If maxCount is 1,
// the semaphore is equivalent to a mutex: signal is equivalent to unlocking
// and wait is equivalent to locking the mutex.
//
typedef void * TQP_OSMSemID;

//
// Create a new semaphore object; returns the ID in sIDP.
//
EQP_Errors QPOSMCreateSem (TQP_OSMSemID * sIDP,
 				unsigned long	maxCount,
				unsigned long	initCount);

//
// Delete an existing semaphore.
//
EQP_Errors QPOSMDeleteSem (TQP_OSMSemID sID);

//
// Signal a semaphore.
//
EQP_Errors QPOSMSignalSem (TQP_OSMSemID sID);

//
// Wait on a semaphore.
//
// if timeoutMS < 0 => wait for ever
// if timeoutMS = 0 => poll
// if timeoutMS > 0 => timeout after timeoutMS milliseconds
//
// returns QPERR_SUCCESS if the semaphore was acquired successfully
// returns QPERR_TIMEOUT if timed out (timeoutMS >= 0 only)
//
EQP_Errors QPOSMWaitSem (TQP_OSMSemID sID, long timeoutMS);

//
// Helper functions
TQP_OSMSemID QPOSMCreateMutex (void);
TQP_BOOL QPOSMLockMutex (TQP_OSMSemID mID);
void QPOSMUnlockMutex (TQP_OSMSemID mID);
void QPOSMDeleteMutex (TQP_OSMSemID mID);

///////////////////////////////////////////////////////////////////////////////
//
// EVENT GROUPS
//
// Event groups are sets of up to 32 bits in a single word that can be used to
// trigger actions
//
typedef void * TQP_OSMEvtgrpID;

//
// Create a new Event Group object; the ID is returned in eIDP
//
EQP_Errors QPOSMCreateEvtgrp (TQP_OSMEvtgrpID * eIDP,
 				TQP_UINT32 initState);

//
// Delete an existing Event Group object
//
EQP_Errors QPOSMDeleteEvtgrp (TQP_OSMEvtgrpID eID);

//
// Wait for a given set of events to occur
//
// if timeoutMS < 0 => wait for ever
// if timeoutMS = 0 => poll
// if timeoutMS > 0 => timeout after timeoutMS milliseconds
//
// returns QPERR_SUCCESS if the semaphore was acquired successfully
// returns QPERR_TIMEOUT if timed out (timeoutMS >= 0 only)
//
EQP_Errors QPOSMWaitEvtgrp (TQP_OSMEvtgrpID eID,
 							TQP_UINT32 mask,
							TQP_UINT32 * evtStatusP,
							long timeoutMS);

//
// Set a set of events
//
EQP_Errors QPOSMSetEvtgrp (TQP_OSMEvtgrpID eID,
 				TQP_UINT32 mask);

//
// Clear a set of events
//
EQP_Errors QPOSMClearEvtgrp (TQP_OSMEvtgrpID eID,
 				TQP_UINT32 mask);


///////////////////////////////////////////////////////////////////////////////
//
// MESSAGING
//
// A message is a small block of data that can be sent to a thread.
// A thread can wait on message reception. The same waiting options as for
// semaphores and event groups are available when receiving a message.
// A message can be sent from an interrupt service routine.
// Every thread has a number (NQ) of message queues that can be dynamically
// changed. Message queues are numbered 0 thru NQ - 1. Lower queue indexes
// have higher priority. Messages are extracted from the highest priority queue
// first.  A thread cannot have more than 32 queues: a bitmap is used to
// address a queue or a set of queues.
// A message is sent to a thread using its thread ID (the one returned when the
// thread is created) and an index to indicate in which queue the message must
// be put.
// Messages are owned by the system and must be released when not used any more
// in order to allow other threads to use them; for efficiency reasons, an
// implementation could have a fixed number of messages globally available.
//

//
// The message structure.
// The fields in this structure are for a typical usage:
//	- type indicates the meaning of the message;
//		usually, even numbers are used for message types so the reply message
//		can be identified by adding 1 to the message type
//	- ptr is a generic pointer parameter to be used if more than a few
//		parameters need to be passed with the message: it is up to the user to
//		make sure that the data pointed by ptr is valid when the receiving
//		thread gets the message.
//	- parms is an array of QPOS_MSGPARMS_NUM 32-bit parameters to be defined
//		by the user.
//
#define QPOS_MSGPARMS_NUM 8
typedef struct _TQP_OSMMsg {
	TQP_UINT32		type;		// message type (user defined)
	void *			ptr;		// pointer parameter (user defined)
	TQP_UINT32		parms[QPOS_MSGPARMS_NUM];	// message parameters (user defined)
} TQP_OSMMsg, *PTQP_OSMMsg;

//
// Get a message block
//
// Return QP_NULL if there are no messages available.
//
PTQP_OSMMsg QPOSMGetMsg (void);

//
// Free a message block
//
// Return QPERR_PARMS if the message pointer is not valid
// Return QPERR_INVALID if the message cannot be freed (e.g. has been sent or
// has already been freed)
//
EQP_Errors QPOSMFreeMsg (PTQP_OSMMsg msgP);

//
// Send a message to a thread.
// qIdx selects the queue to send the message to.
//
// Return QPERR_PARMS if the message pointer is not valid
// Return QPERR_PARMS if the thread ID is not valid (thread terminated ...)
// Return QPERR_INVALID if the requested queue is not available at the
//	receiving thread,
// Return QPERR_INVALID if the message cannot be sent (e.g. has been sent or
//	freed)
//
EQP_Errors QPOSMSendMsg (PTQP_OSMMsg msgP, TQP_OSMThreadID toID, TQP_UINT32 qIdx);


//
// Receive a message
//
// If timeoutMS < 0 => wait for ever
// If timeoutMS = 0 => poll
// If timeoutMS > 0 => timeout after timeoutMS milliseconds
//
// fromIDP has the ID of the sender if a message is successfuly received
// On entry, *qMaskP contains the set of queues (one bit per queue) on which to
// perform the receive operation; when more than one queue is selected this
// function will return a message from the lowest numbered queue (highest
// priority) in the subset.
// On successful exit, *qMaskP returns the index of the queue the
// message came from (i.e. not a bitmap).
// qMaskP set to NULL is equivalent to selecting all queues; however, there will
// be no indication on the queue the message came from.
//
// returns QPERR_SUCCESS if a message was received
// returns QPERR_TIMEOUT if timed out (timeoutMS >= 0 only)
// returns QPERR_INVALID if the call is not issued from a thread context
// returns QPERR_INVALID if the requested queue is not available
//
EQP_Errors QPOSMRcvMsg (PTQP_OSMMsg * msgPP, TQP_OSMThreadID * fromIDP,
						PTQP_UINT32 qMaskP, long timeoutMS);

//
// Manages the number of receive queues.
// On entry *queueNumP has the number of message queues requested.
// On exit *queueNumP returns the actual number of queues. This number can
// be different from the request if there isn't enough memory or if the
// implementation has limitations or if the requested number is larger than 32.
// If *queueNumP is set to a negative number on entry, this function will
// simply return the number of message queues currently active in *queueNumP.
// If the number of queues is reduced, any messages present in the queues that
// are being eliminated will be freed.
// The number of queues can be set to 0, which is equivalent to disabling
// message reception.
// 
EQP_Errors QPOSMSetMsgQueues (PTQP_INT8 queueNumP);

//
// flush(& free) messages related to ch_n in message queue of ThrdId
//
//
// Some threads, ex. DTM/M2M, can recieve messages sent from many other threads belong to different channels.
// When ENC_STOP/DEC_STOP for some ch_n, there maybe still some messages related ch_n queued
// in DTM's/M2M's message queue. This routine help to flush/free those messages. 
//
// returns QPERR_SUCCESS if ch_n's message(s) flushed or no ch_n's messages found
// returns QPERR_INVALID if the call is not issued from a thread context
// returns QPERR_INVALID if the requested queue is not available
//
EQP_Errors QPOSMFlushChMsg (TQP_OSMThreadID ThrId, TQP_UINT32 ch_n, TQP_UINT32 FlushMask);

//
// flush(& free) all messages in message queue of ThrdId
//
//
// returns QPERR_SUCCESS if all message(s) flushed or no messages found
// returns QPERR_INVALID if the call is not issued from a thread context
// returns QPERR_INVALID if the requested queue is not available
//
EQP_Errors QPOSMFlushAllMsg (TQP_OSMThreadID ThrId);


///////////////////////////////////////////////////////////////////////////////
//
//	INTERRUPT HANDLING
//

// Interrupt Service Routines (ISRs) can be registered and unregistered on
// any of the available interrupt numbers; interrupt numbers availability
// and definition depend on the platform.
// An ISR is executed in a non-thread environement, therefore no calls that
// make a thread yield are allowed (wait..., create...); signal calls are
// allowed so a thread can be woken up.
//
// Generic interrupt service routine prototype
//	refCon is a context reference parameter for the caller's convenience
//
typedef void (*TQP_IntSR) (void * refCon);

//
// The system will call the following entry point whenever the interrupt associated
// to a certain number occurs.  This function will call all ISRs registered with that number.
//
void QPOSMISREntry (TQP_UINT32 isrNumber);

typedef void * TQP_OSMIntSRID;

enum _EISRNumbers {
	QPOS_ISR_CODEC = 0,
	QPOS_ISR_VIU,
	QPOS_ISR_IIS,
	QPOS_ISR_DSP,
	QPOS_ISR_SSU,
	QPOS_ISR_HIU,
	QPOS_ISR_M2M0,
	QPOS_ISR_M2M1,
	QPOS_ISR_MSU,
	QPOS_ISR_PPU,
	QPOS_ISR_DMAC,
	QPOS_ISR_VCU, QPOS_ISR_VDU = QPOS_ISR_VCU,
	QPOS_ISR_VOU,
	QPOS_ISR_RTC,
	QPOS_ISR_JPG,
	QPOS_ISR_TMR1,
	QPOS_ISR_TMR2,
	QPOS_ISR_VLD,
	QPOS_ISR_UART,
	QPOS_ISR_SDIO,
	QPOS_ISR_USB,
	QPOS_ISR_PCIE,
	QPOS_ISR_IRU,
	QPOS_ISR_PWM,
	QPOS_ISR_SW0,
	QPOS_ISR_SW1,
	QPOS_ISR_SW2,
	QPOS_ISR_SW3,
	QPOS_ISR_SW4,
	QPOS_ISR_SW5,
	QPOS_ISR_SW6,
	QPOS_ISR_SW7,
	QPOS_ISR_VPU,
	QPOS_ISR_DEC,
	QPOS_ISR_MSU_S,

	QPOS_ISR_MAXNUMBER
};

//
// Register an interrupt service routine
//
EQP_Errors QPOSMRegisterISR (TQP_OSMIntSRID *	isrIDP,
							 TQP_IntSR			isr,
							 void *				refCon,
							 TQP_UINT32			isrNumber);

//
// Unregister an interrupt service routine
//
EQP_Errors QPOSMUnregisterISR (TQP_OSMIntSRID isrID);

//
// Change ISR enable state (TRUE means ISRs are enabled)
//
EQP_Errors QPOSMSetISRStatus (TQP_UINT32 isrNumber, TQP_BOOL * isrStatusP);

// Helper functions: simply enable or disable; return FALSE if they fail.

TQP_BOOL QPOSMEnableISR (TQP_UINT32 isrNumber); 
TQP_BOOL QPOSMDisableISR (TQP_UINT32 isrNumber); 

#ifdef __cplusplus
}
#endif

#endif //__QPOSMGR_H__
