///////////////////////////////////////////////////////////////////////////////
//        
//
//		Definitions for persistent platform-independent settings.
//
//
// History:
//	Aug 17, 2009	(EC) Created
//
// Authors: (EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __QPSETTINGS_H__
#define __QPSETTINGS_H__

#include "qptypes.h"

//
// This API provides persistent platform-independent application settings.
// These settings can be saved in different storage systems depending on the
// platform.
// On the SOC the settings are stored in the Flash. On Windows they can be
// stored in the registry or in a file.
//
// Settings information is organized in blocks marked by a 32-bit identifier.
// When a non-existing block ID is accessed for writing, the block is created
// (if there is room available); on the Flash the block corresponds to a sector;
// in the Windows registry, the block is equivalent to a Section. The
// block might have a limited size (~64 KB on the Flash).
// Several settings can be stored within a block, each identified by a string
// called key (equivalent to entries in Windows registry).  The amount of space
// taken by a setting depends also on the key length.
//
// Functions are provided to access settings as unsigned 32-bit integers,
// character strings or generic arrays of binary data.
//
//

#ifdef __cplusplus
extern "C" {
#endif


///////////////////////////////////////////////////////////////////////////////
//
// Predefined block tags and keys
//

//
// 'plat'  Reserved to system
//
#define QPSTNGS_SYSTEM_BLOCKTAG QP_MAKEFOURCC('p','l','a','t')
#define QPSTNGS_SYSTEM_I2C_KEY "i2c"
#define QPSTNGS_SYSTEM_I2CDATABIT_KEY "i2c_data_bit"
#define QPSTNGS_SYSTEM_I2CCLOCKBIT_KEY "i2c_clock_bit"
#define QPSTNGS_SYSTEM_GPIODIRECTIONS_KEY "gpio_directions"
#define QPSTNGS_SYSTEM_GPIOINITVALUES_KEY "gpio_init_values"
#define QPSTNGS_SYSTEM_USEMODE_KEY "use_mode"
#define QPSTNGS_SYSTEM_CODECLIBMEMALLOC_KEY "codeclib_mem_alloc"
#define QPSTNGS_SYSTEM_VIDEOINPUTDEVICE_KEY "video_input_device"
#define QPSTNGS_SYSTEM_PLL4_KEY "pll4"
#define QPSTNGS_SYSTEM_PLL5_KEY "pll5"
#define QPSTNGS_SYSTEM_VID_KEY "vid"
#define QPSTNGS_SYSTEM_PID_Q_KEY "pid_qcam"
#define QPSTNGS_SYSTEM_PID_U_KEY "pid_uvc"
#define QPSTNGS_SYSTEM_DEVICENAME_KEY "DeviceName"
#define QPSTNGS_SYSTEM_VENDOR_KEY "vendor"
#define QPSTNGS_SYSTEM_QL300_KEY "conf_intf_qcam"
#define QPSTNGS_SYSTEM_UVC_GADGET_KEY "conf_intf_uvc"
#define QPSTNGS_SYSTEM_OFC_ENABLE_KEY "OFC_en"
#define QPSTNGS_SYSTEM_PLL3_KEY "pll3"
#define QPSTNGS_SYSTEM_CMOS_RESET_ENABLE_KEY "cmos_reset_en"
#define QPSTNGS_SYSTEM_OFC_SOURCE_KEY "OFC_source"
#define QPSTNGS_SYSTEM_OFC_SOURCE_PLL_MASTER_CONFIG_KEY "OFC_pll_master_config"
#define QPSTNGS_SYSTEM_OFC_SOURCE_PLL_SLAVE_CONFIG_KEY "OFC_pll_slave_config"
#define QPSTNGS_SYSTEM_BOOT_CONFIGURE_KEY "boot_configure"	// 0: slect by HW (GPIO_4), 1: select by SW (boot_sw_select), 2: auto detect (USB host exist?)
#define QPSTNGS_SYSTEM_BOOT_SW_SELECT_KEY "boot_sw_select"	// 0: QCAM, 1: linux

// for PL330 master mode system
#define QPSTNGS_MASTER_SYSTEM_DEVID_KEY "ma_devid"
#define QPSTNGS_MASTER_SYSTEM_SDIO0_KEY "ma_sdio0_en"
#define QPSTNGS_MASTER_SYSTEM_SDIO1_KEY "ma_sdio1_en"
#define QPSTNGS_MASTER_SYSTEM_GPIO_GRP1_KEY "ma_gpiogrp1_en"
#define QPSTNGS_MASTER_SYSTEM_GPIO_GRP2_KEY "ma_gpiogrp2_en"
#define QPSTNGS_MASTER_SYSTEM_SWI2C_KEY "ma_swi2c_en"
#define QPSTNGS_MASTER_SYSTEM_LOCALBUS_MUTEX_KEY "ma_local_bus_mutex_type"
#define QPSTNGS_MASTER_SYSTEM_ETH_KEY "ma_eth_en"
#define QPSTNGS_MASTER_SYSTEM_FRONTEND_KEY "ma_frondend_type"

#define QPSTNGS_MASTER_SYSTEM_USER_DATA0_KEY "ma_userdata0"
#define QPSTNGS_MASTER_SYSTEM_USER_DATA1_KEY "ma_userdata1"
#define QPSTNGS_MASTER_SYSTEM_USER_DATA2_KEY "ma_userdata2"
#define QPSTNGS_MASTER_SYSTEM_USER_DATA3_KEY "ma_userdata3"
#define QPSTNGS_MASTER_SYSTEM_USER_DATA4_KEY "ma_userdata4"
#define QPSTNGS_MASTER_SYSTEM_USER_DATA5_KEY "ma_userdata5"
#define QPSTNGS_MASTER_SYSTEM_USER_DATA6_KEY "ma_userdata6"
#define QPSTNGS_MASTER_SYSTEM_USER_DATA7_KEY "ma_userdata7"
#define QPSTNGS_MASTER_SYSTEM_USER_STRING0_KEY "ma_userstring0"
#define QPSTNGS_MASTER_SYSTEM_USER_STRING1_KEY "ma_userstring1"
#define QPSTNGS_MASTER_SYSTEM_USER_STRING2_KEY "ma_userstring2"
#define QPSTNGS_MASTER_SYSTEM_USER_STRING3_KEY "ma_userstring3"
#define QPSTNGS_MASTER_SYSTEM_USER_STRING4_KEY "ma_userstring4"
#define QPSTNGS_MASTER_SYSTEM_USER_STRING5_KEY "ma_userstring5"
#define QPSTNGS_MASTER_SYSTEM_USER_STRING6_KEY "ma_userstring6"
#define QPSTNGS_MASTER_SYSTEM_USER_STRING7_KEY "ma_userstring7"
#define QPSTNGS_MASTER_SYSTEM_MTD_PARTITION_KEY "ma_mtd_par"


#define ATAG_QTAG	0x51544147

typedef struct tag_qtag {
	TQP_UINT32 ofc_en;
	TQP_UINT32 cmos_reset_en;
	TQP_UINT32 ma_devid;
	TQP_UINT32 ma_sdio0_en;
	TQP_UINT32 ma_sdio1_en;
	TQP_UINT32 ma_gpiogrp1_en;
	TQP_UINT32 ma_gpiogrp2_en;
	TQP_UINT32 ma_swi2c_en;
	TQP_UINT32 ma_local_bus_mutex_type;
	TQP_UINT32 ma_eth_en;
	TQP_UINT32 ma_frondend_type;
	TQP_UINT32 ma_userdata0;
	TQP_UINT32 ma_userdata1;
	TQP_UINT32 ma_userdata2;
	TQP_UINT32 ma_userdata3;
	TQP_UINT32 ma_userdata4;
	TQP_UINT32 ma_userdata5;
	TQP_UINT32 ma_userdata6;
	TQP_UINT32 ma_userdata7;
	TQP_UINT8 ma_userstring0[32];
	TQP_UINT8 ma_userstring1[32];
	TQP_UINT8 ma_userstring2[32];
	TQP_UINT8 ma_userstring3[32];
	TQP_UINT8 ma_userstring4[32];
	TQP_UINT8 ma_userstring5[32];
	TQP_UINT8 ma_userstring6[32];
	TQP_UINT8 ma_userstring7[32];
// --- mtd partition
	TQP_UINT8 ma_mtd_par[128];
// --- GPIO strap value
	TQP_UINT32 GPIO_strap;
} tag_qtag_t;

//
// 'pers'  Persistent settings that will not be changed by flashupdater
//
#define QPSTNGS_PERSISTENT_BLOCKTAG QP_MAKEFOURCC('p','e','r','s')
#define QPSTNGS_PERSISTENT_SERIAL_KEY "serial"


// Type for the block tag, usually a fourCC constant, that identifies
// a block of settings.
typedef TQP_UINT32 TQPStngsBlockTag;

//
// Access functions take a TQPStngsBlockTag parameter and a PTQP_UCHAR
// parameter that points to a NULL-terminated string.
// When a variable-length item is read (string, binary data) a pointer
// to memory allocated by the Settings Manager is returned: QPStngsFree
// must be called to free this memory when the caller is done using it.
// 

//
// Functions to access a 32-bit item
// 
EQP_Errors QPStngsGetUint32 (TQPStngsBlockTag blockTag, PTQP_CHAR keyP, PTQP_UINT32 valueP);
EQP_Errors QPStngsSetUint32 (TQPStngsBlockTag blockTag, PTQP_CHAR keyP, TQP_UINT32 value);

//
// Functions to access a character string
// The returned string is NULL terminated.
//
EQP_Errors QPStngsGetString (TQPStngsBlockTag blockTag, PTQP_CHAR keyP, PTQP_CHAR * stringPP);
EQP_Errors QPStngsSetString (TQPStngsBlockTag blockTag, PTQP_CHAR keyP, PTQP_CHAR stringP);

//
// Functions to access a generic chunk of data
// The size of the chunk is returned in the getter; must be passed in the setter.
// 
EQP_Errors QPStngsGetBinary (TQPStngsBlockTag blockTag, PTQP_CHAR keyP, void ** dataPP, PTQP_UINT32 sizeP);
EQP_Errors QPStngsSetBinary (TQPStngsBlockTag blockTag, PTQP_CHAR keyP, void * dataP, TQP_UINT32 size);

EQP_Errors QPStngsFree (void * dataP);

//
// Functions to delete settings
//
EQP_Errors QPStngsDeleteBlock (TQPStngsBlockTag blockTag);
EQP_Errors QPStngsDeleteKey (TQPStngsBlockTag blockTag, PTQP_CHAR keyP);

#ifdef __cplusplus
}
#endif

#endif //__QPSETTINGS_H__

