///////////////////////////////////////////////////////////////////////////////
//        
//
// API v 0.2
// 
// This is the header definition of the low level
// API for QL201B codec chip
//
//
//
// History:
//	Aug 4, 2006	(EC) Created from David's first draft
//
// Authors: (DT) David Tao
//			(EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __QPXAPI_H__
#define __QPXAPI_H__

#if defined(__cplusplus)
extern "C" {
#endif


#define QPX_OK 0
#define QPX_ERR 1

#define QPX_ENCODE 0
#define QPX_DECODE 1
#define QPX_DISABLE 2
#define QPX_ENABLE 3

/////////////////////////////////////////////////////////////////////////////
//
//


// system API, the API's apply to both video and 
// audio sub-systems.
//


// initialize codec chip 
// this function combines both audio No.1 and video No.1
//
// arguments:
// arg 1 : pVidFW, pointer to the video firmware
// arg 2 : pAudFW, pointer to the audio firmware
//
// return value:
// OK, ERR
//
int QPX_Sys_Init (void * pVidFW, int nVidFWSize,
				 void * pAudFW, int nAudFWSize
				 );


#define QPX_CLOSE 0
#define QPX_OPEN_VES 1
#define QPX_OPEN_AES 2
#define QPX_OPEN_VES_AES 3

// set encode stream type
//
// arguments:
// arg 1 : nType, VES, AES, VES_AES
// arg 2 : pIdxEntries, pointer to an integer between 32 and 128 to indicate the number
//			of index entries the firmware transfers in each block; it is set to the actual
//			value used upon return
//
// return value:
// OK, ERR
//
int QPX_Sys_SetParamEnc (int nType, int * pIdxEntries); 


// callback function for QPX_Sys_TrgComp and
// QPX_Sys_TrgDecomp
//
// arguments:
// arg 1 : reference constant for caller's context recovery
// arg 2 : nParam, VID_PTR_UPDATE, AUD_PTR_UPDATE, INDEX_PTR_UPDATE
// arg 3 : pParam, ring buffer write pointer for encode
//         ring buffer read pointer for decode
//
// return value:
// OK, ERR
//

#define QPX_VIDEOBUF		0x00
#define QPX_AUDIOBUF		0x01
#define QPX_INDEXBUF		0x02
#define QPX_BUFMASK			0x03
#define QPX_LAST			0x04
#define QPX_FRAMEERR		0x08
#define QPX_READY2DISPLAY	0x10
#define QPX_PLAYBACKDONE	0x20

typedef int (*QPX_CALLBACK) (void * refCon, int flags, void * pParam);

#define QPX_INDEX_BYTES	20
#define QPX_TYPE_MASK 0xE0000000
#define QPX_SIZE_MASK 0x03FFFFFF

typedef union _index_entry {
	struct {
		unsigned long		sizeAndType;
		unsigned long		pts;
		unsigned long		dts;
		unsigned long		hiOffset;
		unsigned long		loOffset;
	} fields;
	unsigned char	bytes[QPX_INDEX_BYTES];
} index_entry;


// start encoding based on the stream type
// set in QPX_Sys_SetParamEnc function.
// this function comines both audio No.6 and video No.4
//
// arguments:
// arg 1 : pAudBuf, pointer to the audio ring buffer, only 
//         valid for stream type AES and VES_AES
// arg 2 : nAudBufSize, audio ring buffer size in unit of 
//         bytes. only valid for stream type AES and VES_AES
// arg 3 : pVidBuf, pointer to the video ring buffer, only 
//         valid for stream type VES and VES_AES
// arg 4 : nVidBufSize, video ring buffer size in unit of 
//         bytes. only valid for stream type VES and VES_AES
// arg 5 : pIndexBuf, pointer to the index ring buffer,
//         always valid
// arg 6 : nIndexBufSize, index ring buffer size in unit of 
//         bytes. always valid
// arg 7 : pCallBackFunc, call back function for buffer 
//         notification
// arg 8 : refCon is a reference constant returned to the caller
//		   in the callback
// 
// return value:
// OK, ERR
//
int QPX_Sys_TrgComp (char * pAudBuf, 
					int nAudBufSize, 
					char * pVidBuf, 
					int nVidBufSize,
					char * pIndexBuf, 
					int nIndexBufSize,
					QPX_CALLBACK pCallBackFunc,
					void * refCon
					);

// stop encoding. 
// this function combines both audio No.8 and video No.5
//
// arguments:
// arg 1 : nStopMode, STOP_AT_GOP_BOUNDARY, STOP_IMMEDIATELY
//
// return value:
// OK, ERR
//
#define QPX_ENCSTOP_IMMED 0
#define QPX_ENCSTOP_GOP 1
#define QPX_ENCSTOP_AUTO 2

int QPX_Sys_CanComp (int nStopMode);


// update read pointer for encode
//
// arguments:
// arg 1 : nType, VID, AUD, INDEX
// arg 2 : ptr, new read pointer
//
// return value:
// OK, ERR
//
int QPX_Sys_UpdateReadPtr (int nType, 
						  char * ptr
						  );


// set decode stream type
//
// arguments:
// arg 1 : nType, VES, AES, VES_AES
//
// return value:
// OK, ERR
//
int QPX_Sys_SetParamDec (int nType);

// start decoding based on the stream type
// set in QPX_Sys_SetParamDec function
// this function comines both audio No.7 and video No.13
//
// arguments:
// pAudBuf, pointer to the audio ring buffer, only 
//         valid for stream type AES and VES_AES
// nAudBufSize, audio ring buffer size in unit of 
//         bytes. only valid for stream type AES and VES_AES
// pVidBuf, pointer to the video ring buffer, only 
//         valid for stream type VES and VES_AES
// nVidBufSize, video ring buffer size in unit of 
//         bytes. only valid for stream type VES and VES_AES
// pCallBackFunc, call back function for buffer 
//         notification
// refCon is a reference constant returned to the caller
//		   in the callback
// bManDisplayTurnon is a boolean to enable disable the manual display
//		   turnon feature
// 
// return value:
// OK, ERR

int QPX_Sys_TrgDecomp (char * pAudBuf, 
					  int nAudBufSize, 
					  char * pVidBuf, 
					  int nVidBufSize,
					  QPX_CALLBACK pCallBackFunc,
					  void * refCon,
					  int bExtDisplayTurnon
					  );

// stop decoding
// this function combines both audio No.9 and video No.15
//
// arguments:
// arg 1 : nStopMode, STOP_COMPLETE_ALL, STOP_IMMEDIATELY
// arg 2 : nDisplay, STOP_DISPLAY_LAST_FRAME, STOP_DISPLAY_BLACK
//
// return value:
// OK, ERR
//
#define QPX_DECSTOP_IMMED 0
#define QPX_DECSTOP_COMPL 1
#define QPX_DECDISP_LAST 0
#define QPX_DECDISP_BLACK 1

int QPX_Sys_CanDecomp (int nStopMode, 
					  int nDisplay
					  );

// turn on the display when decoding
// this function is used in conjunction with the manual display turnon feature
// to turnon the display when the API has reported it is ready to display
// (i.e. when the callback is called with the QPX_READY2DISPLAY flag)
//
int QPX_Sys_TurnonDisplay (void);

// turn off the display when decoding
//
int QPX_Sys_TurnoffDisplay (void);


// update write pointer for decode
//
// arguments:
// arg 1 : flags, VID, AUD, INDEX  Last update
// arg 2 : ptr, new write pointer
//
// return value:
// OK, ERR
//

int QPX_Sys_UpdateWritePtr (int flags, 
						   char * ptr
						   );

// this function replace video No.19
//
// arguments:
// arg 1 : nMode, ENCODE, DECODE, DISABLE
//
// return value:
// OK, ERR
//
int QPX_Sys_SetIO (int nMode);


// Control power saving state
// power state has two values:
//	0: power off
//	1: power on

// get power saving state
int QPX_Sys_GetPowerState (int * psP);

// set power saving state
int QPX_Sys_SetPowerState (int ps);


// hardware reset
// this function replace video No.20
//
// return value:
// OK, ERR
//
int QPX_Sys_Hrst (void);


#define QPX_TRISTATE	0
#define QPX_ON			1

// Audio Out control
//
// arguments:
//	state:	QPX_TRISTATE puts the audio output pins in Tri-state mode
//			QPX_ON enables the audio output pins
//
// return value:
// OK, ERR
//
int QPX_Sys_AudioOut (int state);

// Audio Out control
//
// arguments:
//	state:	QPX_TRISTATE puts the video output pins in Tri-state mode
//			QPX_ON enables the video output pins
//
// return value:
// OK, ERR
//
int QPX_Sys_VideoOut (int state);


// Shared memory access for trick mode
//
// arguments:
//	rdAddrP:	pointer to the address of the shared read memory area
//	wrAddrP:	pointer to the address of the shared write memory area
//
// return value:
// OK, ERR
//
int QPX_Sys_GetSharedPtrs (unsigned long * rdAddrP, unsigned long * wrAddrP);


// Pause and resume functions
//
// return value:
// OK, ERR
//
int QPX_Sys_Pause ();
int QPX_Sys_Resume ();

// Encoder freeze frame
//
// return value:
// OK, ERR
//
int QPX_Sys_FreezeComp (int bOn);

// Encoder still frame
//
// return value:
// OK, ERR
//
int QPX_Sys_StillVideoComp (int bOn);

/////////////////////////////////////////////////////////////////////////////
//
//


// audio API, the API's apply to the audio sub-system 
//

// set AAC encode parameters
// audio No.4
//
// arguments:
// arg 1 : pAACTable, audio settings
//
// return value:
// AAC_OK, AAC_ERR
//
typedef struct _audioparam_struct {	
	unsigned long	samplerate;
	unsigned long	k_bitrate;				// in 1000 b/s
	unsigned int	aac_flag : 1;			// ignored
	unsigned int	crc : 1;				// ignored
	unsigned int	channelmode : 2;		// 1 or 2
	unsigned int	cutoff_frequency;		// 18000 Hz default
} audioparam_struct;	

#define QPX_SRATE_44_1K				0
#define QPX_SRATE_48K				1	
#define QPX_SRATE_32K				2
#define QPX_SRATE_22_5K				4
#define QPX_SRATE_24K				5
#define QPX_SRATE_16K				6
#define QPX_SRATE_8K				8
#define QPX_SRATE_UNKNOWN			0xFF

int QPX_AAC_SetParamEnc (audioparam_struct * pAACTable);

// set AAC encoder stereo enhancement parameters
// audio No.11
//
// arguments:
// arg 1 : pEnhanceParam, TBD 
// 
// return value
// AAC_OK, AAC_ERR
//
typedef struct _addsubgain_struct {
	unsigned long	gain1;
	unsigned long	gain2;
	unsigned long	add;
	unsigned long	sub;
	unsigned long	att1;
	unsigned long	att2;
	unsigned long	lgain;
	unsigned long	rgain;
} addsubgain_struct;

int QPX_AAC_SetParamEnhanceEnc (addsubgain_struct * pEnhanceParam);

// get AAC decode parameters
// audio No.5
//
// arguments:
// arg 1 : pAACstr, pointer to an AAC stream header
// arg 1 : nSize, size of AAC stream header
// arg 3 : pAACTable, audio settings
//
// return value:
// AAC_OK, AAC_ERR
//
int QPX_AAC_GetParamDec (void * pAACstr, 
						 unsigned int nSize,
						 audioparam_struct * pAACTable);


/////////////////////////////////////////////////////////////////////////////
//
//


// video API, the API's apply to the video sub-system
//

// profile
//
#define PROFILE_BASELINE			0
#define PROFILE_MAIN				1
#define PROFILE_EXTENDED			2
#define PROFILE_420_10				3
#define PROFILE_422_8				4
#define PROFILE_422_10				5
#define PROFILE_444_12				6

// level
//
#define LEVEL_1						0
#define LEVEL_1_B					1
#define LEVEL_1_1					2
#define LEVEL_1_2					3
#define LEVEL_1_3					4
#define LEVEL_2						5
#define LEVEL_2_1					6
#define LEVEL_2_2					7
#define LEVEL_3						8
#define LEVEL_3_1					9
#define LEVEL_3_2					10
#define LEVEL_4						11
#define LEVEL_4_1					12
#define LEVEL_4_2					13
#define LEVEL_5						14
#define LEVEL_5_1					15

#define FFMODE_AUTO					0
#define FFMODE_FRAME_ONLY			1
#define FFMODE_FIELD_ONLY			2


typedef struct _videoparam_struct
{
	unsigned int	profile;
	unsigned int	level;
	unsigned int	spsr_freq;
	unsigned int	ff_mode;
} videoparam_struct, *pvideoparam_struct;


// set encoder parameters
// video No.2
//
// arguments:
// arg 1 : pVideoParam
//
// return value:
// MP4_OK, MP4_ERR
//
int QPX_MP4_SetParam(videoparam_struct *pVideoParam);


// coding mode
#define CODING_MODE_CAVLC			0
#define CODING_MODE_CABAC			1

// IDR flag
#define IDR_OPEN_GOP				0
#define IDR_CLOSED_GOP				1


typedef struct _qlty_struct
{
	unsigned char	coding_mode;
	unsigned char	cabac_init_idc;
	unsigned int	intra_period;
	unsigned int	lp_filter;
	unsigned int	dis_lf_idc;
	long			alpha_offset;
	long			beta_offset;
	unsigned int	B_num;
	unsigned char	IDR;
	unsigned char	et;
	unsigned int	et_threshold_offset;
	unsigned int	et_threshold_max;
	unsigned int	et_sad_threshold;
	unsigned int	input_filter_strength;
} qlty_struct, *pqlty_struct;


// set encoder quality parameters
// video No.3
//
// arguments:
// arg 1 : pQualtyParam
//
// return value:
// MP4_OK, MP4_ERR
//
int QPX_MP4_SetQltyParam(qlty_struct * pQualtyParam);


typedef struct _bitrate_struct
{
	unsigned char	vbr_flag;
	unsigned long	avg_bit_rate;
	unsigned long	min_bit_rate;
	unsigned long	max_bit_rate;
	unsigned long	qp_update_freq;
	unsigned char	padding_enable;
} bitrate_struct, *pbitrate_struct;


// set encoder bitrate parameters
// video No.6
//
// arguments:
// arg 1 : pBitrateParam
//
// return value:
// MP4_OK, MP4_ERR
//
int QPX_MP4_SetBitrate(bitrate_struct * pBitrateParam);


// input scaling filter 
#define FILTER_MODE_NONE			0
#define FILTER_MODE_1_2				1
#define FILTER_MODE_1_4				2


typedef struct _resolution_struct
{
	unsigned int	width;
	unsigned int	height;
	unsigned int	filter_mode_width;
	unsigned int	filter_mode_height;
} resolution_struct, *presolution_struct;

// set encoder resolution parameters
// this function replaces video No.9
//
// arguments:
// arg 1 : pResolutionParam
//
// return value:
// MP4_OK, MP4_ERR
//
int QPX_MP4_SetResolution (resolution_struct * pResolutionParam); 


#define INCONTROL_MODE_PROG						0
#define INCONTROL_MODE_PROG_SRC_INTRLC_FORMAT	1
#define INCONTROL_MODE_INTERLACE				2
#define INCONTROL_MODE_SPECIAL_1				3
#define INCONTROL_MODE_SPECIAL_2				4

#define INCONTROL_FMT_PROGRESSIVE				0
#define INCONTROL_FMT_INTERLACE					1

#define INCONTROL_CKEDGE_POSITIVE				0
#define INCONTROL_CKEDGE_NEGATIVE				1

typedef struct _incontrol_struct
{
	unsigned int	mode;
	unsigned int	format;
	unsigned int	ck_edge;
	unsigned int	start_line;
} incontrol_struct, *pincontrol_struct;

// set input control parameters
//
// arguments:
// arg 1 : pInControl
//
// return value:
// MP4_OK, MP4_ERR
//
int QPX_MP4_SetInControl (pincontrol_struct pInControl);

typedef struct _fps_struct
{
	unsigned int	fps;
	unsigned int	skip_frame;
} fps_struct, *pfps_struct;

// set encoder frame rate
// video No.16
//
// arguments:
// arg 1 : pFps
//
// return value:
// MP4_OK, MP4_ERR
//
int QPX_MP4_SetFps (pfps_struct pFps);


// get SPS and PPS information for decoder
// video No.12
//
// arguments:
// arg 1 : pSps, pointer to SPS
// arg 2 : nSize, size of the SPS buffer in unit of bytes
// arg 3 : pInfo, pointer to structure of parameters
//
// return value:
// MP4_OK, MP4_ERR
//

#define QPX_AR_UNSPECIFIED	0
#define QPX_AR_SQUARE		1
#define QPX_AR_12_11		2
#define QPX_AR_10_11		3
#define QPX_AR_16_11		4
#define QPX_AR_40_33		5
#define QPX_AR_24_11		6
#define QPX_AR_20_11		7
#define QPX_AR_32_11		8
#define QPX_AR_80_33		9
#define QPX_AR_18_11		10
#define QPX_AR_15_11		11
#define QPX_AR_64_33		12
#define QPX_AR_160_99		13
#define QPX_EXTENDED		255


typedef struct _stminfo_struct {
	/*--- picture size ---*/
	unsigned short	pic_width;
	unsigned short	pic_height;
	/*--- aspect ratio ---*/
	unsigned char	aspect_ratio_info;
	unsigned short	par_width;
	unsigned short	par_height;
} stminfo_struct;

int QPX_MP4_GetSpsPpsInfo (void * pSps, 
						  unsigned int nSize,
						  stminfo_struct * pInfo
						  );


// decoder flush
//
// arguments:
// none
//
// return value:
// MP4_OK, MP4_ERR
//
int QPX_MP4_FlushDecoder (void);


#if defined(__cplusplus)
}
#endif


#endif //__QPXAPI_H__


