///////////////////////////////////////////////////////////////////////////////
//     
//
//		Debug Manager implementation
//
//	Platform-independent debugging utilities
//
// History:
//	Aug 4, 2006	(EC) Created
//
// Authors: (EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////

#include "qptypes.h"

#ifdef QP_LINXKER
#include <linux/string.h>
#include <linux/time.h>
#include <asm/div64.h>
#ifdef QP_DBG_TO_BUFFER
#include <asm/uaccess.h>
#include <linux/fs.h>
#endif
#else
#include <string.h>
#include <stdarg.h>
#ifdef _QP_MACOSXKER
#include "libkern/libkern.h"
#else
#include <stdio.h>
#endif
#endif

#include "qpdbgmgr.h"
#include "qposmgrint.h"
#include "qptimemgrint.h"

#define MAX_COMPS	64
#define MAX_NAMELEN	20

#define UNSPEC_NAME "Unspecified"

typedef struct _TCompBlk {
	TQP_BOOL		registered;
	char			name[MAX_NAMELEN+1];
	unsigned long	level;
#ifdef QP_DBG_TO_BUFFER
	unsigned long	buflevel;
#endif
} TCompBlk, *PTCompBlk;

// This function must be provided in order to output a string to the local
// debugger or output device
//
extern void DBGPrintStr (const char * strP, int level);

#ifdef QP_DEBUG
static TQP_BOOL Inited = QP_FALSE;
static TQP_BOOL PrintName = QP_FALSE;
static TQP_BOOL TimeStamp = QP_FALSE;
static TCompBlk Comps[MAX_COMPS];
static TQP_UINT64 StartTimeMs;


static TQP_BOOL namesMatch (char * n1P, char * n2P);
static void copyName (char * toP, const char * fromP);
#endif

#ifdef DBGPRINT_BUFFER
static void bufDBGPrintStr (const char * strP);

#define PFBUF_SIZE 4<<20
static char PFBuf[PFBUF_SIZE];
static int curIdx = 0;
#endif

#ifdef QP_DBG_TO_BUFFER
static void bufDBGPrintStrNew (const char * strP);
#define PFBUF_SIZE 500000
static char PFBuf[PFBUF_SIZE];
static int curIdx = 0;
static int curbufskipwrite = 0;
#endif

EQP_Errors
QPDbgMInitPriv (void)
{
	EQP_Errors		rv = QPERR_NOTIMPL;

#ifdef QP_DEBUG
	unsigned long	i;

	Comps[0].registered = QP_TRUE;
	Comps[0].level = QPDBGM_LEVEL_WARN;
	copyName(Comps[0].name,UNSPEC_NAME);

	for (i = 1; i < MAX_COMPS; i++) {
		Comps[i].registered = QP_FALSE;
		Comps[i].level = QPDBGM_LEVEL_WARN;
		Comps[i].name[0] = 0;
		Comps[i].name[MAX_NAMELEN] = 0;	// make sure there is always a terminating NULL
	}

	StartTimeMs = TimeGetSysTimeMilliS();

	Inited = QP_TRUE;

	rv = QPERR_SUCCESS;
#endif

	return (rv);
}

EQP_Errors
QPDbgMTermPriv (void)
{
	EQP_Errors		rv = QPERR_NOTIMPL;

#ifdef QP_DEBUG
	Inited = QP_FALSE;
	rv = QPERR_SUCCESS;
#endif

	return (rv);
}


void
QPDbgMGetCaps (PTQP_DbgMCaps capsP)
{
#ifdef QP_DEBUG
	capsP->isDebug = QP_TRUE;
#else
	capsP->isDebug = QP_FALSE;
#endif
	capsP->maxComps = MAX_COMPS;
	capsP->maxNameLen = MAX_NAMELEN;
}


#ifdef QP_DEBUG
#define MAX_DBGSTR 256
#define DBGPF_BUFSIZE (MAX_NAMELEN+MAX_DBGSTR+2)

int
_QPDbgMPrintf (TQP_UINT32 compLevel, const char * format, ...)
{
	int		rv = 0;
	TQP_UINT32	cID = (compLevel >> 4) & 0xFFF;
	TQP_UINT32	dbgLvl = compLevel & 0xF;

	if (!Inited) {
		QPDbgMInitPriv();
	}

#ifdef QP_DBG_TO_BUFFER
	if ((dbgLvl <= Comps[cID].level)||
		(dbgLvl <= Comps[cID].buflevel))
#else
	if (dbgLvl <= Comps[cID].level)
#endif
	{
		char	buff[DBGPF_BUFSIZE];
		int		i = 0;

		va_list args;
		va_start(args,format);

		if (TimeStamp) {
#ifdef QP_DBG_TO_BUFFER
			TQP_UINT64 timeMs = TimeGetSysTimeMilliS();
			i = sprintf(buff,"%010llu: ",timeMs);
#else
			TQP_UINT64 timeMs = TimeGetSysTimeMilliS() - StartTimeMs;
#ifdef QP_LINXKER
			TQP_UINT64 t = timeMs;
			do_div( t, 3600000 );
			TQP_UINT32 hours = t;
#else
			TQP_UINT32 hours = (TQP_UINT32) (timeMs/3600000);
#endif
			TQP_UINT32 msLeft = (TQP_UINT32) (timeMs - hours*3600000);
			TQP_UINT32 minutes = msLeft/60000;
			TQP_UINT32 seconds;
			msLeft -= minutes*60000;
			seconds = msLeft/1000;
			msLeft = msLeft - seconds*1000;
#if defined(_QP_WINKER) || defined(QP_LINXKER)
			i = sprintf(buff,"%10uh%02um%02u.%03us: ",hours,minutes,seconds,msLeft);
#elif defined(_QP_MACOSXKER) || defined(_QP_MACOSX) || defined(_QP_LINUX)
            i = snprintf(buff,DBGPF_BUFSIZE,"%10uh%02um%02u.%03us: ",hours,minutes,seconds,msLeft);
#else
			i = sprintf_s(buff,DBGPF_BUFSIZE,"%10uh%02um%02u.%03us: "
				,hours,minutes,seconds,msLeft);
#endif
#endif
		}

		if (PrintName) {
			int	j = 0;
			char ch;
			while ((ch = Comps[cID].name[j]) != '\0') {
				buff[i++] = ch;
				j++;
			}
			// pad to the max name length
			for (;j < MAX_NAMELEN;j++) {
				buff[i++] = '_';
			}
			buff[i++] = ':';
			buff[i++] = ' ';
		}
#if defined(_QP_WINDOWS)
		rv = _vsnprintf_s(&buff[i],DBGPF_BUFSIZE-i,DBGPF_BUFSIZE-i,format,args);
#elif defined (_QP_MACOSXKER)
		rv = vsnprintf(&buff[i],DBGPF_BUFSIZE-i,format,args);
#else
		rv = vsprintf(&buff[i],format,args);
#endif
#ifdef QP_DBG_TO_BUFFER
	if (dbgLvl <= Comps[cID].level)
	{
#endif
#ifdef DBGPRINT_BUFFER
		bufDBGPrintStr(buff);
#else
#ifdef QP_LINXKER
		printk( buff );
#else
#ifdef _QP_ANDROID
		androidLog(buff);
#else
		DBGPrintStr(buff,dbgLvl);
#endif /* _QP_ANDROID */
#endif /* QP_LINXKER */
#endif /* DBGPRINT_BUFFER */
#ifdef QP_DBG_TO_BUFFER
	}
	if (dbgLvl <= Comps[cID].buflevel)
	{
		bufDBGPrintStrNew(buff);
	}
#endif

		va_end(args);
	}

	return (rv);
}
#endif

//
// Register a component by name
//	returns a component ID in cIDP; nameP is a null terminated string
//	the ID can be zero if there were no more room for a new component
//
EQP_Errors QPDbgMRegComp (TQP_DbgMCompID * cIDP, char * nameP)
{
	EQP_Errors		rv = QPERR_NOTIMPL;

#ifdef QP_DEBUG
	unsigned long	i;
	unsigned long	freeID;

	if (!Inited) {
		QPDbgMInitPriv();
	}

	if (cIDP == QP_NULL) {
		rv = QPERR_PARMS;
		goto regCompExit;
	}
	if (nameP == QP_NULL) {
		rv = QPERR_PARMS;
		goto regCompExit;
	}

	*cIDP = 0;
	// look for the name
	freeID = 0;
	for (i = 1; i < MAX_COMPS; i++) {
		if (!Comps[i].registered) {
			freeID = i;
		}
		if (namesMatch(nameP,Comps[i].name)) {
			*cIDP = (TQP_DbgMCompID) i;
			break;
		}
	}
	if (i == MAX_COMPS) {
		if (freeID == 0) {
			rv = QPERR_NOMEM;
			goto regCompExit;
		}
		*cIDP = (TQP_DbgMCompID) freeID;
		copyName(Comps[freeID].name,nameP);
		Comps[*cIDP].level = QPDBGM_LEVEL_WARN;
	}
	Comps[*cIDP].registered = QP_TRUE;

	rv = QPERR_SUCCESS;

regCompExit:
#endif //QP_DEBUG

	return (rv);
}

//
// Unregister a component
//
EQP_Errors
QPDbgMUnregComp (TQP_DbgMCompID cID)
{
	EQP_Errors		rv = QPERR_NOTIMPL;

#ifdef QP_DEBUG

	if (!Inited) {
		QPDbgMInitPriv();
	}

	if ((cID == 0) || (cID >= MAX_COMPS)) {
		rv = QPERR_PARMS;
		goto unregCompExit;
	}

	Comps[cID].registered = QP_FALSE;
	Comps[cID].name[0] = '\0';
	rv = QPERR_SUCCESS;

unregCompExit:
#endif //QP_DEBUG

	return (rv);
}

///////////////////////////////////////////////////////////////////////////////
// Functions to be called by the environment that outputs the debug
// strings.

void
QPDbgMNamePrinting (TQP_BOOL enable)
{
#ifdef QP_DEBUG
	PrintName = enable;
#endif //QP_DEBUG
}

void
QPDbgMTimeStampPrinting (TQP_BOOL enable)
{
#ifdef QP_DEBUG
	TimeStamp = enable;
#endif //QP_DEBUG
}


//
// Set a component's debug level
//
EQP_Errors
QPDbgMSetCompLevel (TQP_DbgMCompID cID, unsigned long dbgLvl)
{
	EQP_Errors		rv = QPERR_NOTIMPL;

#ifdef QP_DEBUG

	if (!Inited) {
		QPDbgMInitPriv();
	}

	if (cID >= MAX_COMPS) {
		rv = QPERR_PARMS;
		goto setCompLevelExit;
	}
	Comps[cID].level = dbgLvl;

	rv = QPERR_SUCCESS;

setCompLevelExit:
#endif //QP_DEBUG

	return (rv);
}


//
// Get a component's debug level
//
EQP_Errors
QPDbgMGetCompLevel (TQP_DbgMCompID cID, unsigned long * dbgLvlP)
{
	EQP_Errors		rv = QPERR_NOTIMPL;

#ifdef QP_DEBUG

	if (!Inited) {
		QPDbgMInitPriv();
	}

	if (cID >= MAX_COMPS) {
		rv = QPERR_PARMS;
		goto getCompLevelExit;
	}
	if (dbgLvlP == QP_NULL) {
		rv = QPERR_PARMS;
		goto getCompLevelExit;
	}

	*dbgLvlP = Comps[cID].level;

	rv = QPERR_SUCCESS;

getCompLevelExit:
#endif //QP_DEBUG

	return (rv);
}

#ifdef QP_DBG_TO_BUFFER
//
// Set a component's debug level
//
EQP_Errors
QPDbgMSetBufCompLevel (TQP_DbgMCompID cID, unsigned long dbgLvl)
{
	EQP_Errors		rv = QPERR_NOTIMPL;

#ifdef QP_DEBUG

	if (!Inited) {
		QPDbgMInitPriv();
	}

	if (cID >= MAX_COMPS) {
		rv = QPERR_PARMS;
		goto setBufCompLevelExit;
	}
	Comps[cID].buflevel = dbgLvl;

	rv = QPERR_SUCCESS;

setBufCompLevelExit:
#endif //QP_DEBUG

	return (rv);
}


//
// Get a component's debug level
//
EQP_Errors
QPDbgMGetBufCompLevel (TQP_DbgMCompID cID, unsigned long * dbgLvlP)
{
	EQP_Errors		rv = QPERR_NOTIMPL;

#ifdef QP_DEBUG

	if (!Inited) {
		QPDbgMInitPriv();
	}

	if (cID >= MAX_COMPS) {
		rv = QPERR_PARMS;
		goto getBufCompLevelExit;
	}
	if (dbgLvlP == QP_NULL) {
		rv = QPERR_PARMS;
		goto getBufCompLevelExit;
	}

	*dbgLvlP = Comps[cID].buflevel;

	rv = QPERR_SUCCESS;

getBufCompLevelExit:
#endif //QP_DEBUG

	return (rv);
}


EQP_Errors
QPDbgMSetBufInit(void)
{
	curIdx = 0;
	return QPERR_SUCCESS;
}

EQP_Errors QPDbgMSetBufDump(const char * location)
{
	struct file *file = NULL;
	mm_segment_t fs;
	int loop, start, end;

	if (!location)
		return QPERR_SUCCESS;

    if(file == NULL)
        file = filp_open(location, O_RDWR | O_APPEND | O_CREAT, 0644);
    if (IS_ERR(file)) {
        printk("QPDbgMSetBufDump: error occured while opening file %s, for dump buffer exiting...\n", location);
        return QPERR_PARMS;
    }
	curbufskipwrite = 1;

	fs = get_fs();
	set_fs(KERNEL_DS);

	file->f_op->write(file, (char *)&PFBuf[0],
			      curIdx, &file->f_pos);

	set_fs(fs);
	if(file != NULL)
    	filp_close(file, NULL);

	curbufskipwrite = 0;
	return QPERR_SUCCESS;
}
#endif
//
// Get registered components list
//
EQP_Errors
QPDbgMGetCompsList (TQP_DbgMCompID * cIDP, unsigned long * numP)
{
	EQP_Errors		rv = QPERR_NOTIMPL;

#ifdef QP_DEBUG
	unsigned long	arraySize;
	unsigned long	i;
	unsigned long	numReg = 0;

	if (!Inited) {
		QPDbgMInitPriv();
	}

	if (numP == QP_NULL) {
		rv = QPERR_PARMS;
		goto getCompsListExit;
	}

	arraySize = *numP;

	for (i = 0; i < MAX_COMPS; i++) {
		if (Comps[i].registered) {
			if ((cIDP != QP_NULL) && (arraySize > 0)) {
				cIDP[numReg] = (TQP_DbgMCompID) i;
				arraySize--;
			}
			numReg++;
		}
	}
	*numP = numReg;

	rv = QPERR_SUCCESS;

getCompsListExit:
#endif //QP_DEBUG
	return (rv);
}


//
// Get the name of a registered component
//
EQP_Errors QPDbgMGetCompName (TQP_DbgMCompID cID, char * nameP)
{
	EQP_Errors		rv = QPERR_NOTIMPL;

#ifdef QP_DEBUG

	if (!Inited) {
		QPDbgMInitPriv();
	}

	if (nameP == QP_NULL) {
		rv = QPERR_PARMS;
		goto getCompNameExit;
	}
	if (cID >= MAX_COMPS) {
		rv = QPERR_PARMS;
		goto getCompNameExit;
	}

	copyName(nameP,Comps[cID].name);
	rv = QPERR_SUCCESS;

getCompNameExit:
#endif //QP_DEBUG

	return (rv);
}


///////////////////////////////////////////////////////////////////////////////

#ifdef QP_DEBUG

static TQP_BOOL
namesMatch (char * n1P, char * n2P)
{
	TQP_BOOL		rv = QP_TRUE;

	do {
		if (*n1P != *n2P) {
			rv = QP_FALSE;
			break;
		}

	} while ((*n1P++ != '\0') && (*n2P++ != '\0'));

	return (rv);
}


static void
copyName (char * toP, const char * fromP)
{
	unsigned long	i;

	for (i = 0; i < MAX_NAMELEN; i++) {
		toP[i] = fromP[i];
		if (fromP[i] == '\0') {
			break;
		}
	}
}


#ifdef DBGPRINT_BUFFER
static void
bufDBGPrintStr (const char * strP)
{
	int length = strlen(strP);

	if ((curIdx + length) < PFBUF_SIZE) {
		strcpy(&PFBuf[curIdx],strP);
		curIdx += length;
	}
}
#endif

#ifdef QP_DBG_TO_BUFFER
static void
bufDBGPrintStrNew (const char * strP)
{
	int length = strlen(strP);
	if (curbufskipwrite)
		return;
	if ((curIdx + length) < PFBUF_SIZE) {
		strcpy(&PFBuf[curIdx],strP);
		curIdx += length;
	}
}
#endif

#endif //QP_DEBUG

