///////////////////////////////////////////////////////////////////////////////
//        
//
//		File Manager
//
//	Platform-independent File Manager implementation
//
// History:
//	May 29, 2007	(EC) Created
//
// Authors: (EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////


#include "qptypes.h"
#include "qptimemgr.h"
#include "qpmemmgr.h"
#include "qpfilemgr.h"
#include "qpfilemgrint.h"
#include "qpdbgmgr.h"
#include "qposmgr.h"

#define MAX_VOL_NAME_LEN	512
#define MAX_VOLUMES	32


typedef enum _EClientRecStates {
	CLREC_INITED = QP_MAKEFOURCC('c','l','i','n'),
	CLREC_FREED = QP_MAKEFOURCC('c','l','f','r')
} EClientRecStates;

typedef struct _TClientRec {
	EClientRecStates			state;
	struct _TClientRec *		next;
	void *						refCon;
	TQP_FMCallBack				cbP;
	TQP_FMVolID					volIDs[MAX_VOLUMES];
	unsigned long				volIDsNum;		// 0 to MAX_VOLUMES
} TClientRec, *PTClientRec;

#define IsValidClientRec(crP) ((crP != QP_NULL) && (crP->state == CLREC_INITED))

static TQP_OSMSemID ClientQMutex = QP_NULL;
static PTClientRec ClientsQ = QP_NULL;
static TQP_DbgMCompID DBGCompID = 0;
static TQP_OSMThreadID NotifThreadID = QP_NULL;
static TQP_OSMEvtgrpID NotifEvtGroup = QP_NULL;

static void notifThread (void *);
static void cleanupExit (void);


EQP_Errors
QPFileMInitPriv (void)
{
	EQP_Errors			rv = QPERR_SUCCESS;
	TQP_OSMThreadParms	threadParms;

	// register with the Debug Manager first
	QPDbgMRegComp(&DBGCompID,"QP File Manager");

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMInitPriv: entered\n"));

	// Check if already intialized (use the Notification thread ID as indication of
	// being initialized).
	if (NotifThreadID != QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMInitPriv: already initialized\n"));
		goto initExit;
	}

	rv = FileMInit();
	if (!QP_SUCCESS(rv)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMInitPriv: FileMInit failed %i\n",rv));
		goto initExit;
	}

	// Create a semaphore to protect access to the clients records queue
	ClientQMutex = QPOSMCreateMutex();
	if (ClientQMutex == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPFileMInitPriv: could not create the clients queue Mutex %i\n",rv));
		cleanupExit();
		goto initExit;
	}

	// Create an event group to communicate with the notification thread
	rv = QPOSMCreateEvtgrp(&NotifEvtGroup,0);
	if (!QP_SUCCESS(rv)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMInitPriv: could not create the notification Event Group %i\n",rv));
		cleanupExit();
		goto initExit;
	}

	// Create the notification thread
	threadParms.priority = QPOS_THRPRI_DEFAULT;
	threadParms.stackP = QP_NULL;
	threadParms.stackSize = 0;
	threadParms.procP = notifThread;
	threadParms.parm = QP_NULL;
	rv = QPOSMCreateThread(&NotifThreadID,&threadParms);
	if (!QP_SUCCESS(rv)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMInitPriv: could not create the notification thread %i\n",rv));
		cleanupExit();
		goto initExit;
	}

	// start a periodic timer to check for drives mounting/unmounting

	ClientsQ = QP_NULL;

initExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMInitPriv: exiting %i\n",rv));
	return (rv);
}

EQP_Errors
QPFileMTermPriv (void)
{
	EQP_Errors		rv = QPERR_SUCCESS;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMTermPriv: entered\n"));

	// Check if never intialized (use the Notification thread ID as indication of
	// being initialized).
	if (NotifThreadID == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMTermPriv: not initialized\n"));
	}else{
		cleanupExit();
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMTermPriv: exiting %i\n",rv));

	// unregister from the Debug Manager
	QPDbgMUnregComp(DBGCompID);

	return (rv);
}


//
// Returns the list of capabilities for this implementation
//
void QPFileMGetCaps (PTQP_FMCaps capsP)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMGetCaps: entered 0x%X\n",capsP));

	if (capsP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPFileMGetCaps: NULL caps structure pointer\n"));
	}else{
		FileMGetCaps(capsP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMGetCaps: exiting\n"));
}


//
// Registers a client with the File Manager; returns a File Manager ID to be used in other calls
// to identify this client.
// The client can register a callback function (cbP) and the corresponding context recovery
// value (refCon).
//
TQP_FMID QPFileMRegister (void * refCon, TQP_FMCallBack cbP)
{
	PTClientRec	rv = QP_NULL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMRegister: entered 0x%X 0x%X\n"
		,refCon,cbP));

	// allocate memory for the client record
	rv = (PTClientRec) QPMMMalloc(QP_NULL,sizeof(TClientRec));
	if (rv == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMRegister: could not allocate memory for a client record (%i)\n",sizeof(TClientRec)));
		goto registerExit;
	}

	// initialize the new client record
	rv->state = CLREC_INITED;
	rv->cbP = cbP;
	rv->refCon = refCon;

	// insert the client record in the queue
	if (QPOSMLockMutex(ClientQMutex)) {
		rv->next = ClientsQ;
		ClientsQ = rv;
		QPOSMUnlockMutex(ClientQMutex);
	}else{
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPFileMRegister: lock Client Q mutex failed\n"));
		QPMMFree(QP_NULL,rv);
		rv = QP_NULL;
		goto registerExit;
	}

registerExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMRegister: exiting 0x%X\n",rv));
	return (rv);
}

//
// Unregisters a registered client
//
EQP_Errors QPFileMUnregister (TQP_FMID fmID)
{
	EQP_Errors		rv = QPERR_FAIL;
	PTClientRec		crP = (PTClientRec) fmID;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMUnregister: entered 0x%X\n",fmID));

	if (!IsValidClientRec(crP)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMUnregister: invalid client record ID 0x%X\n",fmID));
		goto unregisterExit;
	}

	// Remove the client record from the queue
	if (QPOSMLockMutex(ClientQMutex)) {
		PTClientRec *	prevP = &ClientsQ;
		PTClientRec		curP;

		while ((curP = *prevP) != QP_NULL) {
			if (curP == crP) {
				*prevP = crP->next;
				break;
			}
			prevP = &curP->next;
		}
		// mark this record free
		crP->state = CLREC_FREED;
		// free the memory used by the client record
		QPMMFree(QP_NULL,crP);
		QPOSMUnlockMutex(ClientQMutex);
	}else{
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPFileMUnregister: lock Client Q mutex failed\n"));
		goto unregisterExit;
	}

	rv = QPERR_SUCCESS;

unregisterExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMUnregister: exiting 0x%X\n",rv));
	return (rv);
}


///////////////////////////////////////////////////////////////////////////////
//
// Volume handling
//


//
// Get volume IDs
//
EQP_Errors QPPFGetVolList (TQP_FMID fmID, TQP_FMVolID ** volIDsP, unsigned long * volNumP)
{
	EQP_Errors		rv = QPERR_FAIL;
	PTClientRec		crP = (PTClientRec) fmID;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPPFGetVolList: entered 0x%X 0x%X 0x%X\n",fmID,volIDsP,volNumP));

	if (!IsValidClientRec(crP)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPPFGetVolList: invalid client record ID 0x%X\n",fmID));
		rv = QPERR_INVALID;
		goto getVolListExit;
	}

	if (volNumP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPPFGetVolList: NULL volNumP\n"));
		rv = QPERR_INVALID;
		goto getVolListExit;
	}

	*volNumP = crP->volIDsNum;
	if (volIDsP != QP_NULL) {
		*volIDsP = crP->volIDs;
	}
	rv = QPERR_SUCCESS;

getVolListExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFGetVolList: exiting 0x%X\n",rv));
	return (rv);
}

//
// Get the ID of a Volume from a path name
//
TQP_FMVolID QPFileMGetVolID (TQP_FMID fmID, const PTQP_CHAR nameP)
{
	TQP_FMVolID		rv = QP_NULL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMGetVolID: entered 0x%X\n",nameP));

	if (nameP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMGetVolID: NULL nameP\n"));
	}else{
		rv = FileMGetVolID(nameP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMGetVolID: exiting 0x%X\n",rv));
	return (rv);
}

// Static Volume information

//
// Returns information about an available Volume
//
EQP_Errors QPFileMGetVolInfo (TQP_FMVolID volID, PTQP_FMVolInfo volInfoP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMGetVolInfo: entered 0x%X 0x%X\n",volID,volInfoP));

	if (volInfoP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMGetVolInfo: NULL volInfoP\n"));
		rv = QPERR_INVALID;
	}else{
		rv = FileMGetVolInfo(volID,volInfoP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMGetVolInfo: exiting 0x%X\n",rv));
	return (rv);
}

// Dynamic Volume information

//
// Returns a Volume's name
//
EQP_Errors QPFileMGetVolName (TQP_FMVolID volID, PTQP_CHAR nameP, unsigned long * lenP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMGetVolName: entered 0x%X 0x%X 0x%X\n",volID,nameP,lenP));

	if (lenP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMGetVolName: NULL lenP\n"));
		rv = QPERR_INVALID;
	}else{
		rv = FileMGetVolName(volID,nameP,lenP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMGetVolName: exiting 0x%X\n",rv));
	return (rv);
}

//
// Sets a volume's name
//
EQP_Errors QPFileMSetVolName (TQP_FMVolID volID, const PTQP_CHAR nameP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMSetVolName: entered 0x%X 0x%X\n",volID,nameP));

	if (nameP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMSetVolName: NULL nameP\n"));
		rv = QPERR_INVALID;
	}else{
		rv = FileMSetVolName(volID,nameP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMSetVolName: exiting 0x%X\n",rv));
	return (rv);
}

//
// Returns the size in bytes of the free space on a Volume
//
EQP_Errors QPFileMGetVolFree (TQP_FMVolID volID, PTQP_UINT64 freeP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMGetVolFree: entered 0x%X 0x%X\n",volID,freeP));

	if (freeP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMGetVolFree: NULL freeP\n"));
		rv = QPERR_INVALID;
	}else{
		rv = FileMGetVolFree(volID,freeP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMGetVolFree: exiting 0x%X\n",rv));
	return (rv);
}

//
// Returns TRUE if the volume (volID) is mounted
//
TQP_BOOL QPFileMIsVolAvailable (TQP_FMVolID volID)
{
	TQP_BOOL	rv = QP_FALSE;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMIsVolAvailable: entered 0x%X\n",volID));

	rv = FileMIsVolAvailable(volID);

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMIsVolAvailable: exiting 0x%X\n",rv));
	return (rv);
}


///////////////////////////////////////////////////////////////////////////////
//
// File/Directory management
//

//
// Starts listing a named directory (pnP). The ID of the listing operation is returned at dlIDP.
//
EQP_Errors QPFileMDirListStart (TQP_FMVolID volID, const PTQP_CHAR pnP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMDirListStart: entered 0x%X 0x%X\n",volID,pnP));

	if (pnP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMDirListStart: NULL directory path name\n"));
		rv = QPERR_INVALID;
	}else{
		rv = FileMDirListStart(volID,pnP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMDirListStart: exiting 0x%X\n",rv));
	return (rv);
}

//
// Recursively retrieve an item's name from a directory on which listing was started
// using QPFileMDirListStart.  The item name is returned in (pnP,lenP).  A true/false value,
// indicating whether the item is a file or a sub-directory, is returned through isDirP.
//
EQP_Errors QPFileMDirListNext (TQP_FMVolID volID, PTQP_FMFileInfo fileInfoP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMDirListNext: entered 0x%X 0x%X\n",volID,fileInfoP));

	rv = FileMDirListNext(volID,fileInfoP);

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMDirListNext: exiting 0x%X\n",rv));
	return (rv);
}

EQP_Errors QPFileMGetNameInfo (TQP_FMVolID volID, const PTQP_CHAR pnP, PTQP_FMFileInfo fiP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMGetNameInfo: entered 0x%X 0x%X 0x%X\n",volID,pnP,fiP));

	if ((pnP == QP_NULL) || (fiP == QP_NULL)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMGetNameInfo: NULL pointer parameter\n",pnP,fiP));
		rv = QPERR_INVALID;
	}else{
		rv = FileMGetNameInfo(volID,pnP,fiP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMGetNameInfo: exiting 0x%X\n",rv));
	return (rv);
}

//
// Removes (delete) a named file or directory (pnP).
//
EQP_Errors QPFileMRemove (TQP_FMVolID volID, const PTQP_CHAR pnP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMRemove: entered 0x%X 0x%X\n",volID,pnP));

	if (pnP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMRemove: NULL name pointer\n"));
		rv = QPERR_INVALID;
	}else{
		rv = FileMRemove(volID,pnP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMRemove: exiting 0x%X\n",rv));
	return (rv);
}

//
// Renames a file or a directory.  The old pathname is (pnP); the new pathname is (newPnP).
//
EQP_Errors QPFileMRename (TQP_FMVolID volID, const PTQP_CHAR pnP, const PTQP_CHAR newPnP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMRename: entered 0x%X 0x%X 0x%X\n",volID,pnP,newPnP));

	if ((pnP == QP_NULL) || (newPnP == QP_NULL)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMRename: NULL name pointer 0x%X 0x%X\n",pnP,newPnP));
		rv = QPERR_INVALID;
	}else{
		rv = FileMRename(volID,pnP,newPnP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMRename: exiting 0x%X\n",rv));
	return (rv);
}

//
// Creates a new empty directory.  The new directory pathname is specified by the (pnP).
//
EQP_Errors QPFileMMakeDir (TQP_FMVolID volID, const PTQP_CHAR pnP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMMakeDir: entered 0x%X 0x%X\n",volID,pnP));

	if (pnP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMMakeDir: NULL name pointer\n"));
		rv = QPERR_INVALID;
	}else{
		rv = FileMMakeDir(volID,pnP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMMakeDir: exiting 0x%X\n",rv));
	return (rv);
}




///////////////////////////////////////////////////////////////////////////////
//
// File access
//

//
// Opens a named file (pnP) on a given Volume (volID).  flags describes various
// sharing and access modes; the ID of the open file is returned at (fIDP).
//
EQP_Errors QPFileMOpen (TQP_FMVolID volID, const PTQP_CHAR pnP,
 					TQP_UINT32 flags, TQP_FMFileID * fIDP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMOpen: entered 0x%X 0x%X 0x%X 0x%X\n",volID,pnP,flags,fIDP));

	if ((pnP == QP_NULL) || (fIDP == QP_NULL)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMOpen: NULL parameter pointer 0x%X 0x%X\n",pnP,fIDP));
		rv = QPERR_INVALID;
	}else{
		rv = FileMOpen(volID,pnP,flags,fIDP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMOpen: exiting 0x%X\n",rv));
	return (rv);
}

//
// Returns information about an open file (fID).  File information is returned in the structure
// pointed by fiP.
//
EQP_Errors QPFileMGetFileInfo (TQP_FMVolID fID, PTQP_FMFileInfo fiP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMGetFileInfo: entered 0x%X 0x%X\n",fID,fiP));

	if (fiP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMGetFileInfo: NULL file info pointer\n"));
		rv = QPERR_INVALID;
	}else{
		rv = FileMGetFileInfo(fID,fiP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMGetFileInfo: exiting 0x%X\n",rv));
	return (rv);
}

//
// Reads from an open file (fID) into a local buffer.  bufP points to a caller-allocated
// buffer to read data into.  The number of bytes to read is at lenP on entry.
// The number of bytes actually read is at lenP on exit.  The file pointer is advanced
// by this amount.
//
EQP_Errors QPFileMRead (TQP_FMFileID fID, void * bufP, TQP_UINT64 * lenP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMRead: entered 0x%X 0x%X 0x%X 0x%X\n",fID,bufP,lenP));

	if ((bufP == QP_NULL) || (lenP == QP_NULL)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMRead: NULL parameter pointer 0x%X 0x%X\n",bufP,lenP));
		rv = QPERR_INVALID;
	}else{
		rv = FileMRead(fID,bufP,lenP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMRead: exiting 0x%X\n",rv));
	return (rv);
}

//
// Writes data into an open file (fID).  bufP points to a caller-allocated buffer
// with the data to write.  The number of bytes to write is at lenP on entry.  The
// number of bytes actually written is at lenP on exit.  The file pointer is advanced
// by this amount.
//
EQP_Errors QPFileMWrite (TQP_FMFileID fID, const void * bufP, TQP_UINT64 * lenP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMWrite: entered 0x%X 0x%X 0x%X 0x%X\n",fID,bufP,lenP));

	if ((bufP == QP_NULL) || (lenP == QP_NULL)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMWrite: NULL parameter pointer 0x%X 0x%X\n",bufP,lenP));
		rv = QPERR_INVALID;
	}else{
		rv = FileMWrite(fID,bufP,lenP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMWrite: exiting 0x%X\n",rv));
	return (rv);
}

//
// Returns the current pointer in an open file (fID).  The current pointer position
// is returned at posP.
//
EQP_Errors QPFileMGetPos (TQP_FMFileID fID, TQP_UINT64 * posP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMGetPos: entered 0x%X 0x%X\n",fID,posP));

	if (posP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMGetPos: NULL position pointer\n"));
		rv = QPERR_INVALID;
	}else{
		rv = FileMGetPos(fID,posP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMGetPos: exiting 0x%X\n",rv));
	return (rv);
}

//
// Sets the current pointer in an open file (fID).  The desired pointer position
// is at posP on entry.  The actual pointer position is returned at posP.  It can
// differ from the desired position if attempting to go past the end of file.
//
EQP_Errors QPFileMSetPos (TQP_FMFileID fID, TQP_UINT64 * posP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMSetPos: entered 0x%X 0x%X\n",fID,posP));

	if (posP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMSetPos: NULL position pointer\n"));
		rv = QPERR_INVALID;
	}else{
		rv = FileMSetPos(fID,posP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMSetPos: exiting 0x%X\n",rv));
	return (rv);
}

//
// Returns the length in bytes of an open file (fID).
//
EQP_Errors QPFileMGetLength (TQP_FMFileID fID, TQP_UINT64 * lenP)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMGetLength: entered 0x%X 0x%X\n",fID,lenP));

	if (lenP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPFileMGetLength: NULL length pointer\n"));
		rv = QPERR_INVALID;
	}else{
		rv = FileMGetLength(fID,lenP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMGetLength: exiting 0x%X\n",rv));
	return (rv);
}

//
// Closes an open file (fID).
//
EQP_Errors QPFileMClose (TQP_FMFileID fID)
{
	EQP_Errors		rv = QPERR_FAIL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE)
		,"QPFileMClose: entered 0x%X\n",fID));

	rv = FileMClose(fID);

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPFileMClose: exiting 0x%X\n",rv));
	return (rv);
}




///////////////////////////////////////////////////////////////////////////////
//
// Local static functions
//

static void
notifThread (void * parm)
{
}

static void
cleanupExit (void)
{
	if (NotifThreadID != QP_NULL) {
		// TODO: send a terminate message to the thread
		QPOSMDeleteThread(NotifThreadID);
		NotifThreadID = QP_NULL;
	}
	// Delete all objects
	if (ClientQMutex != QP_NULL) {
		QPOSMDeleteMutex(ClientQMutex);
		ClientQMutex = QP_NULL;
	}
	if (NotifEvtGroup != QP_NULL) {
		QPOSMDeleteEvtgrp(NotifEvtGroup);
		NotifEvtGroup = QP_NULL;
	}
	FileMTerm();
}

	//rv = FileMGetVolsList(crP->volIDs,&volIDsNum);
	//if (!QP_SUCCESS(rv)) {
	//	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPFileMInitPriv: FileMGetVolsList failed %i\n",rv));
	//	cleanupExit();
	//	goto getVolListExit;
	//}

