///////////////////////////////////////////////////////////////////////////////
//        
//
//		Memory Manager
//
// Implements the memory manager abstraction layer.
// Use compile time constants to select an OS or a platform:
//	Windows: _QP_WINDOWS
//	ITRON: _QP_ITRON
//	NuCore: _QP_NUCORE
//
// History:
//	Aug 4, 2006	(EC) Created
//
// Authors: (EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////

#include "qpmemmgr.h"
#include "qpdbgmgr.h"
#include "qpmemmgrint.h"


#ifdef QP_DEBUG
// uncomment this definition to enable the code that shows the
// maximum dynamic memory usage
//#define MAX_MEM_USAGE
#endif

static TQP_DbgMCompID DBGCompID = 0;


EQP_Errors
QPMemMInitPriv (void)
{
	EQP_Errors		rv = QPERR_SUCCESS;

	// register with the Debug Manager first
	QPDbgMRegComp(&DBGCompID,"QP Memory Manager");

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMemMInitPriv: entered\n"));

	rv = MemMInit();

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMemMInitPriv: exiting %i\n",rv));
	return (rv);
}

EQP_Errors
QPMemMTermPriv (void)
{
	EQP_Errors		rv = QPERR_SUCCESS;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMemMTermPriv: entered\n"));

	rv = MemMTerm();

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMemMTermPriv: exiting %i\n",rv));

	// unregister with the Debug Manager
	QPDbgMUnregComp(DBGCompID);

	return (rv);
}

void
QPMMGetCaps (PTQP_MMCaps capsP)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMGetCaps: entered\n"));

	MMGetCaps(capsP);

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMGetCaps: exiting\n"));
}


TQP_MMID
QPMMInit (void * ptr, unsigned long bytes)
{
	TQP_MMID		rv = QP_NULL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMInit: entered 0x%X, %i\n"
		,ptr,bytes));

	rv = MMInit(ptr,bytes);

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMInit: exiting %i\n",rv));

	return (rv);
}


#ifdef MAX_MEM_USAGE
static unsigned long BlocksNum = 0;
static unsigned long MaxBlocksNum = 0;
static unsigned long TotMem = 0;
static unsigned long MaxTotMem = 0;
#endif

void *
QPMMMalloc (TQP_MMID mmID, unsigned long bytes)
{
	void *	rv = QP_NULL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMMalloc: entered 0x%X, %i\n"
		,mmID,bytes));

#ifdef MAX_MEM_USAGE
	bytes += sizeof(unsigned long);
#endif

	rv = MMMalloc(mmID,bytes);
    
#ifdef MAX_MEM_USAGE
	if (rv != QP_NULL) {
		*((unsigned long *) rv) = bytes - sizeof(unsigned long);
		rv = ((char *) rv) + sizeof(unsigned long);
		BlocksNum++;
		TotMem += bytes-sizeof(unsigned long);
		if (BlocksNum > MaxBlocksNum) {
			MaxBlocksNum = BlocksNum;
			QPDbgMPrintf((QPDBGM_LEVEL_WARN,"       new max blocks %i\n",MaxBlocksNum));
		}
		if (TotMem > MaxTotMem) {
			MaxTotMem = TotMem;
			QPDbgMPrintf((QPDBGM_LEVEL_WARN,"       new max mem %i\n",MaxTotMem));
		}
	}
#endif

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMMalloc: exiting %i\n",rv));

	return (rv);
}

void
QPMMFree (TQP_MMID mmID, void * ptr)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMFree: entered 0x%X, 0x%X\n"
		,mmID,ptr));

#ifdef MAX_MEM_USAGE
	unsigned long	size;
	ptr = ((char *) ptr) - sizeof(unsigned long);
	size = *((unsigned long *) ptr);
	BlocksNum--;
	TotMem -= size;
#endif

	MMFree(mmID,ptr);

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMFree: exiting\n"));
}

void
QPMMTerm (TQP_MMID mmID)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMTerm: entered 0x%X\n"
		,mmID));

	MMTerm(mmID);

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMTerm: exiting\n"));
}


void * QPMMVir2Phy (void * ptr)
{
	return (ptr);
}


void *
QPMMMalloc2 (PTQP_UINT32 memTypeP, unsigned long bytes)
{
	void *	rv = QP_NULL;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMMalloc2: entered 0x%X, %i\n"
		,memTypeP,bytes));

	rv = MMMalloc2(memTypeP,bytes);

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMMalloc2: exiting %i\n",rv));

	return (rv);
}

void
QPMMFree2 (void * memP)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMFree2: entered 0x%X\n"
		,memP));

	MMFree2(memP);

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPMMFree2: exiting\n"));
}


EQP_Errors
QPMMStartM2MRequest (PTQP_MMM2MReqParms reqParmsP,
					TQP_MMM2MReqCompl complProc, void * refCon,
					TQP_MMM2MReqID  * reqIDP)
{
#if defined(QP_LINXKER)
	return MMStartM2MRequest(reqParmsP, complProc, refCon, reqIDP);
#else
	return (QPERR_NOTIMPL);
#endif
}


EQP_Errors QPMMM2MRequestState (TQP_MMM2MReqID reqID, EQP_MMM2MReqStates * stateP)
{
	return (QPERR_NOTIMPL);
}

//
// Kill a pending M2M request
//
EQP_Errors QPMMKillM2MRequest (TQP_MMM2MReqID reqID)
{
	return (QPERR_NOTIMPL);
}


