///////////////////////////////////////////////////////////////////////////////
//     
//
//		OS Manager
//
// Implements the OS abstraction layer.
//
// History:
//	Aug 9, 2006	(EC) Changed for the new interrupt handling interface
//	Aug 4, 2006	(EC) Created
//
// Authors: (EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////


#include "qptypes.h"
#include "qposmgr.h"
#include "qpdbgmgr.h"
#include "qpmemmgr.h"
#include "qposmgrint.h"
#include "qposmgrpriv.h"

enum _ISRCheckCodes {
	ISRCHK_INITED	= 0x69727369,	// 'isri'
	ISRCHK_DELETED	= 0x64727369	// 'isrd'
};

typedef struct _TISRDesc {
	TQP_UINT32			check;
	struct _TISRDesc *	next;
	TQP_IntSR			isr;
	void *				refCon;
	TQP_UINT32			isrNumber;
} TISRDesc, *PTISRDesc;

static PTISRDesc ISRQueues[QPOS_ISR_MAXNUMBER] = {QP_NULL,};
static TQP_OSMSemID ISRMutex = QP_NULL;

static TQP_DbgMCompID DBGCompID = 0;
static TQP_DbgMCompID DBGThrCompID = 0;
static TQP_DbgMCompID DBGSemCompID = 0;
TQP_DbgMCompID DBGEvtCompID = 0;

static TQP_BOOL Inited = QP_FALSE;


///////////////////////////////////////////////////////////////////////////////
//
//	PRIVATE CALLS
//
// These calls are meant to be issued by a control module (initialization,
// termination) and should not be called by any other module or manager.
//

//
// Initializes the data structures of the OS manager.
// This call is non reentrant and must be called only once
//
EQP_Errors QPOSMInitPriv (void)
{
	EQP_Errors		rv = QPERR_NOTIMPL;
	int				i;

	// register with the Debug Manager first

	QPDbgMRegComp(&DBGCompID,"QPOSManager General");
	QPDbgMRegComp(&DBGThrCompID,"QPOSManager Threads");
	QPDbgMRegComp(&DBGSemCompID,"QPOSManager Sems");
	QPDbgMRegComp(&DBGEvtCompID,"QPOSManager Events");
	QPDbgMSetCompLevel(DBGEvtCompID,
			QPDBGM_LEVEL_ERROR
	);


	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMInitPriv: entered\n"));

	if (Inited) {
		rv = QPERR_INVALID;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMInitPriv: already initialized\n"));
		goto initExit;
	}

	// platform dependent initializations
	OSMInit();

	// Initialize the ISR queues
	// create a semaphore for the ISR mutex
	rv = QPOSMCreateSem(&ISRMutex,1,1);
	if (!QP_SUCCESS(rv)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMInitPriv: could not create the ISR Mutex %d\n",rv));
		goto initExit;
	}

	for (i = 0; i < QPOS_ISR_MAXNUMBER; i++) {
		ISRQueues[i] = QP_NULL;
	}

	rv = QPERR_SUCCESS;
	Inited = QP_TRUE;

initExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMInitPriv: exiting %d\n",rv));
	return (rv);
}

//
// Terminates the data structures of the OS manager.
// This call is non reentrant and must be called only once
//
EQP_Errors QPOSMTermPriv (void)
{
	EQP_Errors		rv = QPERR_NOTIMPL;
	int				i;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMTermPriv: entered\n"));

	if (!Inited) {
		rv = QPERR_INVALID;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMTermPriv: not initialized\n"));
		goto termExit;
	}

	// kill all the handlers left
	// acquire the ISR mutex
	rv = QPOSMWaitSem(ISRMutex,-1);
	if (QP_SUCCESS(rv)) {
		for (i = 0; i < QPOS_ISR_MAXNUMBER; i++) {
			PTISRDesc	isrP = ISRQueues[i];
			while (isrP != QP_NULL) {
				PTISRDesc	tP = isrP->next;
				isrP->check = ISRCHK_DELETED;
				QPMMFree(0,isrP);
				isrP = tP;
			}
		}
	}
	rv = QPOSMDeleteSem(ISRMutex);
	ISRMutex = QP_NULL;
	if (!QP_SUCCESS(rv)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMTermPriv: cannot delete IH semaphore %d\n",rv));
	}

	OSMTerm();

	Inited = QP_FALSE;

termExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMTermPriv: exiting %d\n",rv));

	// unregister with the Debug Manager
	QPDbgMUnregComp(DBGCompID);

	return (rv);
}


///////////////////////////////////////////////////////////////////////////////
//
// GETCAPS
//
//

void
QPOSMGetCaps (PTQP_OSMCaps capsP)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMGetCaps: entered 0x%X\n",capsP));

	if (capsP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMGetCaps: NULL caps structure pointer\n"));
	}else{
		// platform-dependent capabilities
		OSMGetCaps(capsP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMGetCaps: exiting\n"));
}


///////////////////////////////////////////////////////////////////////////////
//
//	THREADS MANAGEMENT
//

EQP_Errors QPOSMCreateThread (TQP_OSMThreadID * tIDP,
 				PTQP_OSMThreadParms parmsP)
{
	EQP_Errors		rv = QPERR_NOTIMPL;
	PTThreadBlock	twsP;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGThrCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMCreateThread: entered 0x%X 0x%X\n"
		,tIDP,parmsP));

	if ((tIDP == QP_NULL) || (parmsP == QP_NULL)) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGThrCompID,QPDBGM_LEVEL_ERROR),"QPOSMCreateThread: NULL parameter pointer 0x%X 0x%X\n"
			,tIDP,parmsP));
		goto createThreadExit;
	}
	if (parmsP->procP == QP_NULL) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGThrCompID,QPDBGM_LEVEL_ERROR),"QPOSMCreateThread: NULL procedure pointer\n"));
		goto createThreadExit;
	}
	if ((parmsP->stackP != QP_NULL) && (parmsP->stackSize == 0)) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGThrCompID,QPDBGM_LEVEL_ERROR),"QPOSMCreateThread: zero stack size (0x%X)\n",parmsP->stackP));
		goto createThreadExit;
	}

	if (parmsP->priority > QPOS_THRPRI_HIGHEST) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGThrCompID,QPDBGM_LEVEL_ERROR),"QPOSMCreateThread: priority too high %d, reset to default\n"
			,parmsP->priority));
		parmsP->priority = QPOS_THRPRI_DEFAULT;
	}

	twsP = (PTThreadBlock) QPMMMalloc(0,sizeof(TThreadBlock));
	if (twsP != QP_NULL) {
		twsP->parm = parmsP->parm;
		twsP->procP = parmsP->procP;
		rv = OSMCreateThread(twsP,parmsP);
		if (QP_SUCCESS(rv)) {
			*tIDP = twsP;
		}else{
			QPMMFree(0,twsP);
		}
	}else{
		rv = QPERR_NOMEM;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGThrCompID,QPDBGM_LEVEL_ERROR),"QPOSMCreateThread: cannot allocate memory "
			"for thread wrapper structure (%d)\n",sizeof(TThreadBlock)));
	}

createThreadExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGThrCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMCreateThread: exiting %d\n",rv));
	return (rv);
}


EQP_Errors QPOSMDeleteThread (TQP_OSMThreadID tID)
{
	EQP_Errors		rv = QPERR_NOTIMPL;
	PTThreadBlock	twsP;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGThrCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMDeleteThread: entered 0x%X\n",tID));

	twsP = (PTThreadBlock) tID;

	if (twsP != QP_NULL) {
		rv = OSMDeleteThread(twsP);
		if (rv == QPERR_SUCCESS) {
			QPMMFree(0,twsP);
		}
	}else{
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGThrCompID,QPDBGM_LEVEL_ERROR),"QPOSMDeleteThread: NULL thread ID"));
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGThrCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMDeleteThread: exiting %d\n",rv));
	return (rv);
}


///////////////////////////////////////////////////////////////////////////////
//
//	SEMAPHORES
//


EQP_Errors QPOSMCreateSem (TQP_OSMSemID * sIDP,
 				unsigned long	maxCount,
				unsigned long	initCount)
{
	EQP_Errors			rv = QPERR_NOTIMPL;
	PTSemaphoreBlock	semBlkP;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMCreateSem: entered 0x%X, %d, %d\n"
		,sIDP,maxCount,initCount));

	if (sIDP == 0) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_ERROR),"QPOSMCreateSem: NULL ID pointer\n"));
		goto createSemExit;
	}
	if (maxCount == 0) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_ERROR),"QPOSMCreateSem: max count is zero\n"));
		goto createSemExit;
	}
	if (maxCount < initCount) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_ERROR),"QPOSMCreateSem: init count is off range %d (max %d)\n"
			,initCount,maxCount));
		goto createSemExit;
	}

	semBlkP = (PTSemaphoreBlock) QPMMMalloc(0,sizeof(TSemaphoreBlock));
	if (semBlkP == QP_NULL) {
		rv = QPERR_NOMEM;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_ERROR),"QPOSMCreateSem: cannot allocate memory "
			"for semaphore block (%d)\n",sizeof(TSemaphoreBlock)));
	}else{
		rv = OSMCreateSem(semBlkP,maxCount,initCount);
		if (QP_SUCCESS(rv)) {
			*sIDP = semBlkP;
			semBlkP->check = SEMCHK_INITED;
		}else{
			semBlkP->check = SEMCHK_DELETED;
			QPMMFree(0,semBlkP);
		}
	}

createSemExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMCreateSem: exiting %d\n",rv));
	return (rv);
}


//
// Delete an existing semaphore.
//
EQP_Errors QPOSMDeleteSem (TQP_OSMSemID sID)
{
	EQP_Errors			rv = QPERR_NOTIMPL;
	PTSemaphoreBlock	semBlkP = (PTSemaphoreBlock) sID;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMDeleteSem: entered 0x%X\n",sID));

	if (semBlkP == QP_NULL) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_ERROR),"QPOSMDeleteSem: NULL semaphore ID\n"));
		goto deleteSemExit;
	}
	if (semBlkP->check != SEMCHK_INITED) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_ERROR),"QPOSMDeleteSem: invalid semaphore ID (0x%X)\n"
			,semBlkP->check));
		goto deleteSemExit;
	}
	rv = OSMDeleteSem(semBlkP);
	semBlkP->check = SEMCHK_DELETED;
	QPMMFree(0,semBlkP);

deleteSemExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMDeleteSem: exiting %d\n",rv));
	return (rv);
}

//
// Signal a semaphore.
//
EQP_Errors QPOSMSignalSem (TQP_OSMSemID sID)
{
	EQP_Errors			rv = QPERR_NOTIMPL;
	PTSemaphoreBlock	semBlkP = (PTSemaphoreBlock) sID;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMSignalSem: entered 0x%X\n",sID));

	if (semBlkP == QP_NULL) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_ERROR),"QPOSMSignalSem: NULL semaphore ID\n"));
		goto signalSemExit;
	}
	if (semBlkP->check != SEMCHK_INITED) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_ERROR),"QPOSMSignalSem: invalid semaphore ID (0x%X)\n"
			,semBlkP->check));
		goto signalSemExit;
	}

	rv = OSMSignalSem(semBlkP);

signalSemExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMSignalSem: exiting %d\n",rv));
	return (rv);
}

//
// Wait on a semaphore.
//
EQP_Errors QPOSMWaitSem (TQP_OSMSemID sID, long timeoutMS)
{
	EQP_Errors			rv = QPERR_NOTIMPL;
	PTSemaphoreBlock	semBlkP = (PTSemaphoreBlock) sID;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMWaitSem: entered 0x%X, %d\n",sID,timeoutMS));

	if (semBlkP == QP_NULL) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_ERROR),"QPOSMWaitSem: NULL semaphore ID\n"));
		goto waitSemExit;
	}
	if (semBlkP->check != SEMCHK_INITED) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_ERROR),"QPOSMWaitSem: invalid semaphore ID (0x%X)\n"
			,semBlkP->check));
		goto waitSemExit;
	}

	rv = OSMWaitSem(semBlkP,timeoutMS);

waitSemExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGSemCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMWaitSem: exiting %d\n",rv));
	return (rv);
}

//
// Helper functions
//

TQP_OSMSemID
QPOSMCreateMutex (void)
{
	TQP_OSMSemID		rv = QP_NULL;

	QPOSMCreateSem(&rv,1,1);

	return (rv);
}

TQP_BOOL
QPOSMLockMutex (TQP_OSMSemID mID)
{
	return (QPOSMWaitSem(mID,-1) == QPERR_SUCCESS);
}

void
QPOSMUnlockMutex (TQP_OSMSemID mID)
{
	QPOSMSignalSem(mID);
}

void
QPOSMDeleteMutex (TQP_OSMSemID mID)
{
	QPOSMDeleteSem(mID);
}


///////////////////////////////////////////////////////////////////////////////
//
//	EVENT GROUPS
//


EQP_Errors QPOSMCreateEvtgrp (TQP_OSMEvtgrpID * eIDP,
 				TQP_UINT32 initState)
{
	EQP_Errors		rv = QPERR_FAIL;
	PTEventBlock	evtBlkP;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMCreateEvtgrp: entered 0x%X 0x%X\n"
		,eIDP,initState));

	evtBlkP = (PTEventBlock) QPMMMalloc(0,sizeof(TEventBlock));
	if (evtBlkP == QP_NULL) {
		rv = QPERR_NOMEM;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMCreateEvtgrp: cannot allocate memory "
			"for event block (%d)\n",sizeof(TEventBlock)));
	}else{
		evtBlkP->mutexID = QPOSMCreateMutex();
		if (evtBlkP->mutexID != QP_NULL) {
			rv = OSMCreateEvtgrp(evtBlkP,initState);
			if (QP_SUCCESS(rv)) {
				*eIDP = evtBlkP;
				evtBlkP->check = EVGCHK_INITED;
			}else{
				QPOSMDeleteMutex(evtBlkP->mutexID);
				QPMMFree(0,evtBlkP);
			}
		}else{
			QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMCreateEvtgrp: could not create mutex\n"));
		}
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMCreateEvtgrp: exiting %d\n",rv));
	return (rv);
}

//
// Delete an existing Event Group object
//
EQP_Errors QPOSMDeleteEvtgrp (TQP_OSMEvtgrpID eID)
{
	EQP_Errors		rv = QPERR_NOTIMPL;
	PTEventBlock	evtBlkP = (PTEventBlock) eID;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMDeleteEvtgrp: entered 0x%X\n",eID));

	if (evtBlkP == QP_NULL) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMDeleteEvtgrp: NULL event group ID\n"));
		goto deleteEvtgrpExit;
	}
	if (QPOSMLockMutex(evtBlkP->mutexID)) {
		if (evtBlkP->check != EVGCHK_INITED) {
			rv = QPERR_PARMS;
			QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMDeleteEvtgrp: invalid event group ID (0x%X)\n"
				,evtBlkP->check));
			goto deleteEvtgrpExit;
		}

		rv = OSMDeleteEvtgrp(evtBlkP);
		evtBlkP->check = EVGCHK_DELETED;
		QPOSMDeleteMutex(evtBlkP->mutexID);
		QPMMFree(0,evtBlkP);
	}else{
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMDeleteEvtgrp: could not lock mutex\n"));
	}

deleteEvtgrpExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMDeleteEvtgrp: exiting %d\n",rv));
	return (rv);
}

//
// Wait for a given set of events to occur
//
//TQP_DbgMCompID	g_QPDBGEvtCompID = 0;
EQP_Errors QPOSMWaitEvtgrp (TQP_OSMEvtgrpID	eID,
 							TQP_UINT32		mask,
 							TQP_UINT32 *	evtStatusP,
							long			timeoutMS)
{
	EQP_Errors		rv = QPERR_SUCCESS;
	PTEventBlock	evtBlkP = (PTEventBlock) eID;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMWaitEvtgrp: entered 0x%X, 0x%X, 0x%X, %d\n"
		,eID,mask,evtStatusP,timeoutMS));

	if (evtBlkP == QP_NULL) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMWaitEvtgrp: NULL event group ID\n"));
		goto waitEvtgrpExit;
	}

	if (QPOSMLockMutex(evtBlkP->mutexID)) {
		switch (evtBlkP->check) {
		case EVGCHK_INITED:
			evtBlkP->check = EVGCHK_WAITING;
			break;
		case EVGCHK_WAITING:
			rv = QPERR_FAIL;
			QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMWaitEvtgrp: already waiting on this event group\n"));
			break;
		case EVGCHK_DELETED:
			rv = QPERR_PARMS;
			QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMWaitEvtgrp: event deleted\n"));
			break;
		default:
			rv = QPERR_PARMS;
			QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMWaitEvtgrp: unknown event state 0x%X\n",evtBlkP->check));
			break;
		}
		QPOSMUnlockMutex(evtBlkP->mutexID);
	}else{
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMWaitEvtgrp: could not lock mutex(1)\n"));
	}

	if (rv != QPERR_SUCCESS) {
		goto waitEvtgrpExit;
	}

	rv = OSMWaitEvtgrp(evtBlkP,mask,evtStatusP,timeoutMS);
	if ((rv != QPERR_SUCCESS) && (rv != QPERR_TIMEOUT)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMWaitEvtgrp: OSMWaitEvtgrp failed 0x%Xn",rv));
		goto waitEvtgrpExit;
	}
	if (QPOSMLockMutex(evtBlkP->mutexID)) {
		evtBlkP->check = EVGCHK_INITED;
		QPOSMUnlockMutex(evtBlkP->mutexID);
	}else{
		rv = QPERR_FAIL;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMWaitEvtgrp: could not lock mutex(2)\n"));
		goto waitEvtgrpExit;
	}

waitEvtgrpExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMWaitEvtgrp: exiting (0x%X) %d\n",eID,rv));
	return (rv);
}

//
// Set a set of events
//
EQP_Errors QPOSMSetEvtgrp (TQP_OSMEvtgrpID eID,
 							TQP_UINT32 mask)
{
	EQP_Errors		rv = QPERR_FAIL;
	PTEventBlock	evtBlkP = (PTEventBlock) eID;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMSetEvtgrp: entered 0x%X, 0x%X\n",eID,mask));

	if (evtBlkP == QP_NULL) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMSetEvtgrp: NULL event group ID\n"));
		goto setEvtgrpExit;
	}

	switch (evtBlkP->check) {
	case EVGCHK_INITED:
	case EVGCHK_WAITING:
		rv = OSMSetEvtgrp(evtBlkP,mask);
		break;
	case EVGCHK_DELETED:
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMSetEvtgrp: event deleted\n"));
		break;
	default:
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMSetEvtgrp: unknown event state 0x%X\n",evtBlkP->check));
		break;
	}

setEvtgrpExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMSetEvtgrp: exiting %d\n",rv));
	return (rv);
}

//
// Clear a set of events
//
EQP_Errors QPOSMClearEvtgrp (TQP_OSMEvtgrpID eID,
 							TQP_UINT32 mask)
{
	EQP_Errors		rv = QPERR_FAIL;
	PTEventBlock	evtBlkP = (PTEventBlock) eID;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMClearEvtgrp: entered 0x%X, 0x%X\n",eID,mask));

	if (evtBlkP == QP_NULL) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMClearEvtgrp: NULL event group ID\n"));
		goto clearEvtgrpExit;
	}
	switch (evtBlkP->check) {
	case EVGCHK_INITED:
	case EVGCHK_WAITING:
		rv = OSMClearEvtgrp(evtBlkP,mask);
		break;
	case EVGCHK_DELETED:
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMClearEvtgrp: event deleted\n"));
		break;
	default:
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_ERROR),"QPOSMClearEvtgrp: unknown event state 0x%X\n",evtBlkP->check));
		break;
	}

clearEvtgrpExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGEvtCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMClearEvtgrp: exiting %d\n",rv));
	return (rv);
}



///////////////////////////////////////////////////////////////////////////////
//
//	INTERRUPT HANDLING
//

//
// Register an interrupt handler
//
EQP_Errors QPOSMRegisterISR (TQP_OSMIntSRID *	isrIDP,
							 TQP_IntSR			isr,
							 void *				refCon,
							 TQP_UINT32			isrNumber)
{
	EQP_Errors		rv = QPERR_NOTIMPL;
	PTISRDesc		newISRP;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMRegisterISR: entered 0x%X 0x%X 0x%X %d\n"
		,isrIDP,isr,refCon,isrNumber));

	if (isrIDP == QP_NULL) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMRegisterISR: NULL ID pointer\n"));
		goto regIRSExit;
	}
	if (isr == QP_NULL) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMRegisterISR: NULL interrupt routine\n"));
		goto regIRSExit;
	}
	if (isrNumber >= QPOS_ISR_MAXNUMBER) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMRegisterISR: ISR number out of range %d\n"
			,isrNumber));
		goto regIRSExit;
	}

	rv = OSMRegisterISR(isrNumber);
	if (QP_FAIL(rv)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPOSMRegisterISR: OSMRegisterISR failed %d\n",rv));
		goto regIRSExit;
	}

	// insert the new ISR in the corresponding queue
	// protect the ISRQueues array with a mutex
	if (QPOSMLockMutex(ISRMutex)) {
		newISRP = QPMMMalloc(QP_NULL,sizeof(TISRDesc));
		if (newISRP != QP_NULL) {
			newISRP->check = ISRCHK_INITED;
			newISRP->isr = isr;
			newISRP->refCon = refCon;
			newISRP->isrNumber = isrNumber;
			newISRP->next = ISRQueues[isrNumber];
			ISRQueues[isrNumber] = newISRP;
			rv = QPERR_SUCCESS;
		}else{
			QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMRegisterISR: could not allocate"
				" memory for ISR descriptor (%d)\n",sizeof(TISRDesc)));
			rv = QPERR_NOMEM;
		}

		*isrIDP = newISRP;
		QPOSMUnlockMutex(ISRMutex);
	}else{
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMRegisterISR: lock ISR mutex failed\n"));
		goto regIRSExit;
	}

regIRSExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMRegisterISR: exiting %d\n",rv));
	return (rv);
}


//
// Unregister an interrupt handler
//
EQP_Errors QPOSMUnregisterISR (TQP_OSMIntSRID isrID)
{
	EQP_Errors		rv = QPERR_NOTIMPL;
	PTISRDesc		isrP = (PTISRDesc) isrID;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMUnregisterISR: entered 0x%X\n"
		,isrID));

	if (isrP == QP_NULL) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMUnregisterISR: NULL ISR ID\n"));
		goto unregisterISRExit;
	}
	if (isrP->check != ISRCHK_INITED) {
		rv = QPERR_PARMS;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMUnregisterISR: Invalid ISR ID\n"));
		goto unregisterISRExit;
	}

	if (QPOSMLockMutex(ISRMutex)) {
		PTISRDesc		curISRP;
		PTISRDesc *		prevISRPP;
		curISRP = ISRQueues[isrP->isrNumber];
		prevISRPP = &ISRQueues[isrP->isrNumber];
		while (curISRP != QP_NULL) {
			if (((TQP_OSMIntSRID) curISRP) == isrID) {
				break;
			}
			prevISRPP = &curISRP->next;
			curISRP = curISRP->next;
		}
		if (curISRP != QP_NULL) {
			// found it
			*prevISRPP = curISRP->next;
			curISRP->check = ISRCHK_DELETED;
			QPMMFree(0,curISRP);
			rv = QPERR_SUCCESS;
		}else{
			rv = QPERR_FAIL;
			QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMUnregisterISR: could not find ISR in its queue\n"));
		}
		QPOSMUnlockMutex(ISRMutex);
	}else{
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPOSMUnregisterISR: lock ISR mutex failed\n"));
		goto unregisterISRExit;
	}

unregisterISRExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPOSMUnregisterISR: exiting %d\n",rv));
	return (rv);
}


void QPOSMISREntry (TQP_UINT32 isrNumber)
{
	if (isrNumber < QPOS_ISR_MAXNUMBER) {
		if (QPOSMLockMutex(ISRMutex)) {
			PTISRDesc	isrDescP = ISRQueues[isrNumber];
			while (isrDescP != QP_NULL) {
				if ((isrDescP->isr != QP_NULL) && (isrDescP->check == ISRCHK_INITED)) {
					isrDescP->isr(isrDescP->refCon);
				}
				isrDescP = isrDescP->next;
			}
			QPOSMUnlockMutex(ISRMutex);
		}
	}
}


void TaskWrapper (PTThreadBlock	twsP)
{
	if (twsP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGThrCompID,QPDBGM_LEVEL_ERROR),"TaskWrapper: NULL thread block pointer"));
	}else{
		// call the thread's entry point
		(*twsP->procP)(twsP->parm);
	}

	return;
}
