///////////////////////////////////////////////////////////////////////////////
//        
//
//		Internal OS Manager interface
//
// Use compile time constants to select an OS:
//	Windows: _QP_WINDOWS
//	ITRON: _QP_ITRON
//	NuCore: _QP_NUCORE
//
// History:
//	Aug 9, 2006	(EC) Removed interrupt handling definitions (new interface)
//	Aug 4, 2006	(EC) Created
//
// Authors: (EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////



#ifndef __QPOSMGRINT_H__
#define __QPOSMGRINT_H__

#ifdef QP_LINXKER
#include "../common/qptypes.h"
#else
#include "qptypes.h"
#endif

#if defined(_QP_WINDOWS)
#include <windows.h>
#elif defined(_QP_WINKER)
#include <wdm.h>
#elif defined(_QP_ITRON)
#elif defined(_QP_LINUX) || defined(_QP_MACOSX)
#include <pthread.h>
#elif defined(QP_LINXKER)
//#include <linux/videodev.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/ctype.h>
#include <linux/types.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
#include <linux/semaphore.h>
#else
#include <asm/semaphore.h>
#endif
#if LINUX_VERSION_CODE > KERNEL_VERSION(3,3,0)
#include <asm/switch_to.h>
#else
#include <asm/system.h>
#endif
#endif

#if defined(_QP_MACOSXKER)
#include <IOKit/IOLib.h>
#include <mach/semaphore.h>
#endif

#ifdef QP_LINXKER
#include "../common/qposmgr.h"
#else
#include "qposmgr.h"
#endif

#define DEF_STACK_SIZE (32*1024)

enum _ThreadCheckCodes {
	THRCHK_INITED	= 0x69726874,	// 'thri'
	THRCHK_STOPPED	= 0x73726874,	// 'thrs'
	THRCHK_ENDED	= 0x65726874,	// 'thre'
	THRCHK_DELETED	= 0x64726874	// 'thrd'
};
enum _SemCheckCodes {
	SEMCHK_INITED	= 0x696D6573,	// 'semi'
	SEMCHK_DELETED	= 0x646D6573	// 'semd'
};
enum _EvtflgCheckCodes {
	EVGCHK_INITED	= 0x69677665,	// 'evgi'
	EVGCHK_WAITING	= 0x77677665,	// 'evgw'
	EVGCHK_DELETED	= 0x64677665	// 'evgd'
};

typedef struct _TThreadBlock {
	TQP_ThrProc		procP;
	void *			parm;
#if defined(_QP_WINDOWS)
	HANDLE			threadH;
#elif defined(_QP_WINKER)
	PVOID			threadObj;
#elif defined(_QP_ITRON)
	ID				taskID;
#elif defined(_QP_LINUX) || defined(_QP_MACOSX)
	pthread_t		threadId;
#elif defined(QP_LINXKER)
	pid_t			threadT;
#elif defined(_QP_MACOSXKER)
	IOThread		thrID;
#endif
} TThreadBlock, *PTThreadBlock;

#if defined(_QP_LINUX) || defined(_QP_MACOSX)

#define QP_EVENT_SET QP_TRUE
#define QP_EVENT_CLEAR QP_FALSE

/* Type of the mutex_cond object */
typedef struct mutex_cond{
	pthread_mutex_t i_mutex;
	pthread_cond_t i_cv;
}mutex_cond_t;

/* Type of the List element object */

typedef struct List_element{
	struct List_element *next;
	struct List_element *prev;
	struct mutex_cond *i_mutex_cond;
}List_element_t;

/* Type of the List object, Any other implementation
 * of List can be used
 */

typedef struct List{
	struct List_element *start_list;
	struct List_element *end_list;
	pthread_mutex_t i_mutex;	 /* This mutex is used to protect the List critical section */
}List_t;

/* Type of the Event Handle object */

typedef struct event_object_s {
	TQP_UINT32	i_signalled;	/* 0:non-signalled(reset), 1:signalled(set) */
	struct List	*i_list;
}event_object_t;

typedef struct
{
	pthread_mutex_t mutex;
	pthread_cond_t condition;
	int semCount;
	int maxCount;
} sem_private_struct, *sem_private;

#endif

typedef struct _TEventBlock {
	TQP_INT32		check;
	TQP_INT32		bits;
	TQP_OSMSemID	mutexID;
#if defined(_QP_WINDOWS)
	HANDLE			events[32];
	HANDLE			statusMutex;
#elif defined(_QP_WINKER)
	KEVENT			events[32];
	KSPIN_LOCK		spinlock;
#elif defined(_QP_ITRON)
	ID				evtID;
#elif defined(_QP_LINUX) || defined(_QP_MACOSX)
	pthread_mutex_t	status_mutex;
	pthread_cond_t	status_cv;
#elif defined( QP_LINXKER )
	spinlock_t 		lock;
	struct semaphore	sem;
	struct timer_list	timer;
#endif
} TEventBlock, *PTEventBlock;

typedef struct _TSemaphoreBlock {
	TQP_INT32		check;
#if defined(_QP_WINDOWS)
	HANDLE			handle;
#elif defined(_QP_WINKER)
	KSEMAPHORE		sem;
#elif defined(_QP_ITRON)
	ID				semID;
#elif defined(_QP_LINUX) || defined(_QP_MACOSX)
	sem_private 	token;
#elif defined(QP_LINXKER)
	struct semaphore sem;
#elif defined(_QP_MACOSXKER)
	unsigned long	count;
	unsigned long	maxCount;
	semaphore_t		sem;
#endif
} TSemaphoreBlock, *PTSemaphoreBlock;

void TaskWrapper (PTThreadBlock	twsP);

extern EQP_Errors OSMInit (void);
extern void OSMTerm (void);
extern void OSMGetCaps (PTQP_OSMCaps);
extern EQP_Errors OSMCreateThread (PTThreadBlock, PTQP_OSMThreadParms);
extern EQP_Errors OSMDeleteThread (PTThreadBlock);
extern EQP_Errors OSMCreateSem (PTSemaphoreBlock, unsigned long	maxCount, unsigned long	initCount);
extern EQP_Errors OSMDeleteSem (PTSemaphoreBlock);
extern EQP_Errors OSMSignalSem (PTSemaphoreBlock);
extern EQP_Errors OSMWaitSem (PTSemaphoreBlock, long timeoutMS);
extern EQP_Errors OSMCreateEvtgrp (PTEventBlock, TQP_UINT32 initState);
extern EQP_Errors OSMDeleteEvtgrp (PTEventBlock);
extern EQP_Errors OSMWaitEvtgrp (PTEventBlock, TQP_UINT32 mask,TQP_UINT32 * evtStatusP, long timeoutMS);
extern EQP_Errors OSMSetEvtgrp (PTEventBlock, TQP_UINT32 mask);
extern EQP_Errors OSMClearEvtgrp (PTEventBlock, TQP_UINT32 mask);
extern EQP_Errors OSMRegisterISR (TQP_UINT32 isrNumber);
//extern EQP_Errors OSMDelayMilliS (TQP_UINT32 milliS);
//extern EQP_Errors OSMDelayMicroS (TQP_UINT32 microS);
//extern TQP_UINT64 OSMGetSysTime (void);


#endif //__QPOSMGRINT_H__
