///////////////////////////////////////////////////////////////////////////////
//        
//
//		Platform Manager
//
// Handles platform-specific information.
//
// History:
//	Sep 13,2006	(EC) Created
//
// Authors: (EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////


#include "../common/qptypes.h"
#include "../common/qpdbgmgr.h"
#include "../common/qppfmgr.h"
#include "../common/qposmgr.h"

extern const TQP_UINT8 QL201AudioFW[];
extern const TQP_UINT32 QL201AudioFWSize;
extern const TQP_UINT8 QL201VideoFW[];
extern const TQP_UINT32 QL201VideoFWSize;
extern const TQP_UINT8 QL300AudioFW[];
extern const TQP_UINT32 QL300AudioFWSize;
extern const TQP_UINT8 QL300VideoFW[];
extern const TQP_UINT32 QL300VideoFWSize;
extern const TQP_UINT8 QL331AudioFW[];
extern const TQP_UINT32 QL331AudioFWSize;
extern const TQP_UINT8 QL331VideoFW[];
extern const TQP_UINT32 QL331VideoFWSize;

static TQP_DbgMCompID DBGCompID = 0;

static EAttrStates AttrStates[QPPF_ACODE_NUM];

static TAccessMode AccessMode;
static TBusType BusType;
static TSlaveIF SlaveIF;
static TMemory Memory;
static TInterrupt Interrupt;
static TVideoIn VideoIn;
static TAudioIn AudioIn;
static TVideoOut VideoOut;
static TAudioOut AudioOut;
static TVideoDec VideoDec;
static TVideoEnc VideoEnc;
static TTuner Tuner;
static TTVAudio TVAudio;
static TAudCodec AudCodec;
static TChip Chip;
static TFWAPI FWAPI;
static TCodecLib CodecLib;
static TI2C I2C;
static TGPIO GPIO;

static TQP_BOOL FWDFlag = QP_FALSE;
static TQP_BOOL HWIFlag = QP_FALSE;

static void getAccessMode (TQP_BOOL reset, PTAccessMode attrP);
static void getBusType (TQP_BOOL reset, PTBusType attrP);
static void getSlaveIF (TQP_BOOL reset, PTSlaveIF attrP);
static void getMemory (TQP_BOOL reset, PTMemory attrP);
static void getInterrupt (TQP_BOOL reset, PTInterrupt attrP);
static void getVideoIn (TQP_BOOL reset, PTVideoIn attrP);
static void getAudioIn (TQP_BOOL reset, PTAudioIn attrP);
static void getVideoOut (TQP_BOOL reset, PTVideoOut attrP);
static void getAudioOut (TQP_BOOL reset, PTAudioOut attrP);
static void getVideoDec (TQP_BOOL reset, PTVideoDec attrP);
static void getVideoEnc (TQP_BOOL reset, PTVideoEnc attrP);
static void getTuner (TQP_BOOL reset, PTTuner attrP);
static void getTVAudio (TQP_BOOL reset, PTTVAudio attrP);
static void getAudCodec (TQP_BOOL reset, PTAudCodec attrP);
static void getChip (TQP_BOOL reset, PTChip attrP);
static void getFWAPI (TQP_BOOL reset, PTFWAPI attrP);
static void getCodecLib (TQP_BOOL reset, PTCodecLib attrP);
static void getI2C (TQP_BOOL reset, PTI2C attrP);
static void getGPIO (TQP_BOOL reset, PTGPIO attrP);

static EQP_Errors setAccessMode (PTAccessMode attrP);
static EQP_Errors setBusType (PTBusType attrP);
static EQP_Errors setSlaveIF (PTSlaveIF attrP);
static EQP_Errors setMemory (PTMemory attrP);
static EQP_Errors setInterrupt (PTInterrupt attrP);
static EQP_Errors setVideoIn (PTVideoIn attrP);
static EQP_Errors setAudioIn (PTAudioIn attrP);
static EQP_Errors setVideoOut (PTVideoOut attrP);
static EQP_Errors setAudioOut (PTAudioOut attrP);
static EQP_Errors setVideoDec (PTVideoDec attrP);
static EQP_Errors setVideoEnc (PTVideoEnc attrP);
static EQP_Errors setTuner (PTTuner attrP);
static EQP_Errors setTVAudio (PTTVAudio attrP);
static EQP_Errors setAudCodec (PTAudCodec attrP);
static EQP_Errors setChip (PTChip attrP);
static EQP_Errors setFWAPI (PTFWAPI attrP);
static EQP_Errors setCodecLib (PTCodecLib attrP);
static EQP_Errors setI2C (PTI2C attrP);
static EQP_Errors setGPIO (PTGPIO attrP);

static void resetAccessMode (void);
static void resetBusType (void);
static void resetSlaveIF (void);
static void resetMemory (void);
static void resetInterrupt (void);
static void resetVideoIn (void);
static void resetAudioIn (void);
static void resetVideoOut (void);
static void resetAudioOut (void);
static void resetVideoDec (void);
static void resetVideoEnc (void);
static void resetTuner (void);
static void resetTVAudio (void);
static void resetAudCodec (void);
static void resetChip (void);
static void resetFWAPI (void);
static void resetCodecLib (void);
static void resetI2C (void);
static void resetGPIO (void);


static EQP_Errors getFW (TQP_UINT8 ** ptrP, PTQP_UINT32 sizeP, const TQP_UINT8 * arrayP, TQP_UINT32 size);


EQP_Errors
QPPFMInitPriv (void)
{
	EQP_Errors		rv = QPERR_SUCCESS;

	// register with the Debug Manager first
	QPDbgMRegComp(&DBGCompID,"QP Platform Manager");

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMInitPriv: entered\n"));

	resetAccessMode();
	resetBusType();
	resetSlaveIF();
	resetMemory();
	resetInterrupt();
	resetVideoIn();
	resetAudioIn();
	resetVideoOut();
	resetAudioOut();
	resetVideoDec();
	resetVideoEnc();
	resetTuner();
	resetTVAudio();
	resetAudCodec();
	resetChip();
	resetFWAPI();
	resetCodecLib();
	resetI2C();
	resetGPIO();

	FWDFlag = QP_TRUE;
	HWIFlag = QP_FALSE;
	
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMInitPriv: exiting %i\n",rv));
	return (rv);
}

EQP_Errors
QPPFMTermPriv (void)
{
	EQP_Errors		rv = QPERR_SUCCESS;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMTermPriv: entered\n"));

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMTermPriv: exiting %i\n",rv));

	// unregister with the Debug Manager
	QPDbgMUnregComp(DBGCompID);

	return (rv);
}

//
// Returns the list of capabilities for this implementation
//
void QPPFMGetCaps (PTQP_PFMCaps capsP)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMGetCaps: entered 0x%X\n"
		,capsP));

	if (capsP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPPFMGetCaps: NULL caps structure pointer\n"));
	}else{
		capsP->version.major = QPPFM_MAJORVERSION;
		capsP->version.minor = QPPFM_MINORVERSION;
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMGetCaps: exiting\n"));
}


//
// Returns an attribute's current value and status information
//
EAttrStates
QPPFMGetAttr (TQP_BOOL reset, EAttrCodes code, PTQP_PFMAttrStruct attrP)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMGetAttr: entered %i,%i,0x%X\n"
		,reset,code,attrP));

	if (attrP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPPFMGetAttr: NULL structure pointer\n"));
		goto getAttrExit;
	}
	switch (code) {
	case QPPF_ACODE_ACCESS:
		getAccessMode(reset,&attrP->accessMode);
		break;
	case QPPF_ACODE_BUS:
		getBusType(reset,&attrP->busType);
		break;
	case QPPF_ACODE_SLAVEIF:
		getSlaveIF(reset,&attrP->slaveIF);
		break;
	case QPPF_ACODE_MEMORY:
		getMemory(reset,&attrP->memory);
		break;
	case QPPF_ACODE_INTERRUPT:
		getInterrupt(reset,&attrP->interrupt);
		break;
	case QPPF_ACODE_VIDEOIN:
		getVideoIn(reset,&attrP->videoIn);
		break;
	case QPPF_ACODE_AUDIOIN:
		getAudioIn(reset,&attrP->audioIn);
		break;
	case QPPF_ACODE_VIDEOOUT:
		getVideoOut(reset,&attrP->videoOut);
		break;
	case QPPF_ACODE_AUDIOOUT:
		getAudioOut(reset,&attrP->audioOut);
		break;
	case QPPF_ACODE_VIDEODEC:
		getVideoDec(reset,&attrP->videoDec);
		break;
	case QPPF_ACODE_VIDEOENC:
		getVideoEnc(reset,&attrP->videoEnc);
		break;
	case QPPF_ACODE_TUNER:
		getTuner(reset,&attrP->tuner);
		break;
	case QPPF_ACODE_TVAUDIO:
		getTVAudio(reset,&attrP->tvAudio);
		break;
	case QPPF_ACODE_AUDIOCODEC:
		getAudCodec(reset,&attrP->audCodec);
		break;
	case QPPF_ACODE_CHIP:
		getChip(reset,&attrP->chip);
		break;
	case QPPF_ACODE_FWAPI:
		getFWAPI(reset,&attrP->fwAPI);
		break;
	case QPPF_ACODE_CODECLIB:
		getCodecLib(reset,&attrP->codecLib);
		break;
	case QPPF_ACODE_I2C:
		getI2C(reset,&attrP->i2c);
		break;
	case QPPF_ACODE_GPIO:
		getGPIO(reset,&attrP->gpio);
		break;
	default:
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPPFMGetAttr: unsupported attribute code %i\n",code));
		goto getAttrExit;
	}

getAttrExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMGetAttr: exiting %i\n",AttrStates[code]));

	return (AttrStates[code]);
}


//
// Sets an attribute to a new value (overriding current value)
//
EQP_Errors QPPFMSetAttr (EAttrCodes code, PTQP_PFMAttrStruct attrP)
{
	EQP_Errors	rv = QPERR_PARMS;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMSetAttr: entered %i,0x%X\n"
		,code,attrP));

	if (attrP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPPFMSetAttr: NULL structure pointer\n"));
		goto setAttrExit;
	}
	switch (code) {
	case QPPF_ACODE_ACCESS:
		rv = setAccessMode(&attrP->accessMode);
		break;
	case QPPF_ACODE_BUS:
		rv = setBusType(&attrP->busType);
		break;
	case QPPF_ACODE_SLAVEIF:
		rv = setSlaveIF(&attrP->slaveIF);
		break;
	case QPPF_ACODE_MEMORY:
		rv = setMemory(&attrP->memory);
		break;
	case QPPF_ACODE_INTERRUPT:
		rv = setInterrupt(&attrP->interrupt);
		break;
	case QPPF_ACODE_VIDEOIN:
		rv = setVideoIn(&attrP->videoIn);
		break;
	case QPPF_ACODE_AUDIOIN:
		rv = setAudioIn(&attrP->audioIn);
		break;
	case QPPF_ACODE_VIDEOOUT:
		rv = setVideoOut(&attrP->videoOut);
		break;
	case QPPF_ACODE_AUDIOOUT:
		rv = setAudioOut(&attrP->audioOut);
		break;
	case QPPF_ACODE_VIDEODEC:
		rv = setVideoDec(&attrP->videoDec);
		break;
	case QPPF_ACODE_VIDEOENC:
		rv = setVideoEnc(&attrP->videoEnc);
		break;
	case QPPF_ACODE_TUNER:
		rv = setTuner(&attrP->tuner);
		break;
	case QPPF_ACODE_TVAUDIO:
		rv = setTVAudio(&attrP->tvAudio);
		break;
	case QPPF_ACODE_AUDIOCODEC:
		rv = setAudCodec(&attrP->audCodec);
		break;
	case QPPF_ACODE_CHIP:
		rv = setChip(&attrP->chip);
		break;
	case QPPF_ACODE_FWAPI:
		rv = setFWAPI(&attrP->fwAPI);
		break;
	case QPPF_ACODE_CODECLIB:
		rv = setCodecLib(&attrP->codecLib);
		break;
	case QPPF_ACODE_I2C:
		rv = setI2C(&attrP->i2c);
		break;
	case QPPF_ACODE_GPIO:
		rv = setGPIO(&attrP->gpio);
		break;
	default:
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"QPPFMSetAttr: unsupported attribute code %i\n",code));
		goto setAttrExit;
	}

	AttrStates[code] = QPPF_ATYPE_OVERRIDDEN;

	rv = QPERR_SUCCESS;

setAttrExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMSetAttr: exiting 0x%X\n",rv));

	return (rv);
}


//
// Gets the current status of the "automatic firmware download" flag
//
TQP_BOOL QPPFMGetFWDwnldFlag (void)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMGetFWDwnldFlag: entered\n"));

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMGetFWDwnldFlag: exiting %i\n",FWDFlag));
	return (FWDFlag);
}


//
// Sets the status of the "automatic firmware download" flag
//
void QPPFMSetFWDwnldFlag (TQP_BOOL fwDF)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMSetFWDwnldFlag: entered %i\n",fwDF));

	FWDFlag = fwDF;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMSetFWDwnldFlag: exiting\n"));
}

//
// Gets the current status of the "do not init hardware" flag
//
TQP_BOOL QPPFMGetHWInitFlag (void)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMGetHWInitFlag: entered\n"));

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMGetHWInitFlag: exiting %i\n",HWIFlag));
	return (HWIFlag);
}


//
// Sets the status of the "do not init hardware" flag
//
void QPPFMSetHWInitFlag (TQP_BOOL hwIF)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMSetHWInitdFlag: entered %i\n",hwIF));

	HWIFlag = hwIF;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMSetHWInitdFlag: exiting\n"));
}


//
// Returns pointer and size of the built-in video firmware
// If ptrP is NULL, only the size is returned, if sizeP is also NULL,
// the FW buffer is freed.
//
EQP_Errors
QPPFMGetVideoFW (TQP_UINT8 ** ptrP, PTQP_UINT32 sizeP)
{
	EQP_Errors		rv = QPERR_SUCCESS;
	PTQP_UINT8		fwArray = 0;
	TQP_UINT32		fwSize = 0;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMGetVideoFW: entered 0x%X, 0x%X\n"
		,ptrP,sizeP));

	switch (Chip.type) {
	case QPPF_CHIPTYPE_BERINGER:
		fwArray = (PTQP_UINT8) QL201VideoFW;
		fwSize = QL201VideoFWSize;
		break;
	case QPPF_CHIPTYPE_ARTESA:
	case QPPF_CHIPTYPE_ARTESA_B0:
	case QPPF_CHIPTYPE_ARTESA_VIRTEX_4:
	case QPPF_CHIPTYPE_ARTESA_VIRTEX_5:
		fwArray = (PTQP_UINT8) QL300VideoFW;
		fwSize = QL300VideoFWSize;
		break;
	case QPPF_CHIPTYPE_STONEGATE:
	case QPPF_CHIPTYPE_STONEGATE_VIRTEX_4:
	case QPPF_CHIPTYPE_STONEGATE_VIRTEX_5:
		fwArray = (PTQP_UINT8) QL331VideoFW;
		fwSize = QL331VideoFWSize;
		break;
	default:
		rv = QPERR_NOTFOUND;
		break;
	}

	if (QP_SUCCESS(rv)) {
		rv = getFW(ptrP,sizeP,fwArray,fwSize);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMGetVideoFW: exiting %i\n",rv));
	return (rv);
}


//
// Returns pointer and size of the built-in audio firmware
// If ptrP is NULL, only the size is returned, if sizeP is also NULL,
// the FW buffer is freed
//
EQP_Errors
QPPFMGetAudioFW (TQP_UINT8 ** ptrP, PTQP_UINT32 sizeP)
{
	EQP_Errors		rv = QPERR_SUCCESS;
	PTQP_UINT8		fwArray = 0;
	TQP_UINT32		fwSize = 0;

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMGetAudioFW: entered 0x%X, 0x%X\n"
		,ptrP,sizeP));

	switch (Chip.type) {
	case QPPF_CHIPTYPE_BERINGER:
		fwArray = (PTQP_UINT8) QL201AudioFW;
		fwSize = QL201AudioFWSize;
		break;
	case QPPF_CHIPTYPE_ARTESA:
	case QPPF_CHIPTYPE_ARTESA_B0:
	case QPPF_CHIPTYPE_ARTESA_VIRTEX_4:
	case QPPF_CHIPTYPE_ARTESA_VIRTEX_5:
		fwArray = (PTQP_UINT8) QL300AudioFW;
		fwSize = QL300AudioFWSize;
		break;
	case QPPF_CHIPTYPE_STONEGATE:
	case QPPF_CHIPTYPE_STONEGATE_VIRTEX_4:
	case QPPF_CHIPTYPE_STONEGATE_VIRTEX_5:
		fwArray = (PTQP_UINT8) QL331AudioFW;
		fwSize = QL331AudioFWSize;
		break;
	default:
		rv = QPERR_NOTFOUND;
		break;
	}

	if (QP_SUCCESS(rv)) {
		rv = getFW(ptrP,sizeP,fwArray,fwSize);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPPFMGetAudioFW: exiting %i\n",rv));
	return (rv);
}


///////////////////////////////////////////////////////////////////////////////
//
// GET functions
//

static void
getAccessMode (TQP_BOOL reset, PTAccessMode attrP)
{
	if (reset) {
		resetAccessMode();
	}
	*attrP = AccessMode;
}

static void
getBusType (TQP_BOOL reset, PTBusType attrP)
{
	if (reset) {
		resetBusType();
	}
	*attrP = BusType;
}

static void
getSlaveIF (TQP_BOOL reset, PTSlaveIF attrP)
{
	if (reset) {
		resetSlaveIF();
	}
	*attrP = SlaveIF;
}

static void
getMemory (TQP_BOOL reset, PTMemory attrP)
{
	if (reset) {
		resetMemory();
	}
	*attrP = Memory;
}


static void
getInterrupt (TQP_BOOL reset, PTInterrupt attrP)
{
	if (reset) {
		resetInterrupt();
	}
	*attrP = Interrupt;
}

static void
getVideoIn (TQP_BOOL reset, PTVideoIn attrP)
{
	if (reset) {
		resetVideoIn();
	}
	*attrP = VideoIn;
}

static void
getAudioIn (TQP_BOOL reset, PTAudioIn attrP)
{
	if (reset) {
		resetAudioIn();
	}
	*attrP = AudioIn;
}

static void
getVideoOut (TQP_BOOL reset, PTVideoOut attrP)
{
	if (reset) {
		resetVideoOut();
	}
	*attrP = VideoOut;
}

static void
getAudioOut (TQP_BOOL reset, PTAudioOut attrP)
{
	if (reset) {
		resetAudioOut();
	}
	*attrP = AudioOut;
}

static void
getVideoDec (TQP_BOOL reset, PTVideoDec attrP)
{
	if (reset) {
		resetVideoDec();
	}
	*attrP = VideoDec;
}

static void
getVideoEnc (TQP_BOOL reset, PTVideoEnc attrP)
{
	if (reset) {
		resetVideoEnc();
	}
	*attrP = VideoEnc;
}

static void
getTuner (TQP_BOOL reset, PTTuner attrP)
{
	if (reset) {
		resetTuner();
	}
	*attrP = Tuner;
}

static void
getTVAudio (TQP_BOOL reset, PTTVAudio attrP)
{
	if (reset) {
		resetTVAudio();
	}
	*attrP = TVAudio;
}

static void
getAudCodec (TQP_BOOL reset, PTAudCodec attrP)
{
	if (reset) {
		resetAudCodec();
	}
	*attrP = AudCodec;
}

static void
getChip (TQP_BOOL reset, PTChip attrP)
{
	if (reset) {
		resetChip();
	}
	*attrP = Chip;
}

static void
getFWAPI (TQP_BOOL reset, PTFWAPI attrP)
{
	if (reset) {
		resetFWAPI();
	}
	*attrP = FWAPI;
}

static void
getCodecLib (TQP_BOOL reset, PTCodecLib attrP)
{
	if (reset) {
		resetCodecLib();
	}
	*attrP = CodecLib;
}

static void
getI2C (TQP_BOOL reset, PTI2C attrP)
{
	if (reset) {
		resetI2C();
	}
	*attrP = I2C;
}

static void
getGPIO (TQP_BOOL reset, PTGPIO attrP)
{
	if (reset) {
		resetGPIO();
	}
	*attrP = GPIO;
}


///////////////////////////////////////////////////////////////////////////////
//
// SET functions
//

static EQP_Errors
setAccessMode (PTAccessMode attrP)
{
	AccessMode = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setBusType (PTBusType attrP)
{
	BusType = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setSlaveIF (PTSlaveIF attrP)
{
	SlaveIF = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setMemory (PTMemory attrP)
{
	Memory = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setInterrupt (PTInterrupt attrP)
{
	Interrupt = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setVideoIn (PTVideoIn attrP)
{
	VideoIn = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setAudioIn (PTAudioIn attrP)
{
	AudioIn = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setVideoOut (PTVideoOut attrP)
{
	VideoOut = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setAudioOut (PTAudioOut attrP)
{
	AudioOut = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setVideoDec (PTVideoDec attrP)
{
	VideoDec = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setVideoEnc (PTVideoEnc attrP)
{
	VideoEnc = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setTuner (PTTuner attrP)
{
	Tuner = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setTVAudio (PTTVAudio attrP)
{
	TVAudio = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setAudCodec (PTAudCodec attrP)
{
	AudCodec = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setChip (PTChip attrP)
{
	Chip = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setFWAPI (PTFWAPI attrP)
{
	FWAPI = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setCodecLib (PTCodecLib attrP)
{
	CodecLib = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setI2C (PTI2C attrP)
{
	I2C = *attrP;

	return (QPERR_SUCCESS);
}

static EQP_Errors
setGPIO (PTGPIO attrP)
{
	GPIO = *attrP;

	return (QPERR_SUCCESS);
}


///////////////////////////////////////////////////////////////////////////////
//
// RESET functions
//

static void
resetAccessMode (void)
{
	AccessMode.mode = QPPF_DEF_ACCESSMODE;
	AttrStates[QPPF_ACODE_ACCESS] = QPPF_ATYPE_HARDCODED;
}

static void
resetBusType (void)
{
	BusType.type = QPPF_DEF_BUSTYPE;
	AttrStates[QPPF_ACODE_BUS] = QPPF_ATYPE_HARDCODED;
}

static void
resetSlaveIF (void)
{
	SlaveIF.regBase = QPPF_DEF_SLAVEIF_REG;
	SlaveIF.memBase = QPPF_DEF_SLAVEIF_MEM;
	SlaveIF.pageSize = QPPF_DEF_SLAVEIF_PAGE;
	AttrStates[QPPF_ACODE_SLAVEIF] = QPPF_ATYPE_HARDCODED;
}

static void
resetMemory (void)
{
	Memory.megaBits = QPPF_DEF_MEMORY_MBITS;
	Memory.type = QPPF_DEF_MEMORY_TYPE;
	AttrStates[QPPF_ACODE_MEMORY] = QPPF_ATYPE_HARDCODED;
}

static void
resetInterrupt (void)
{
	Interrupt.number = QPPF_DEF_INTNUM;
	AttrStates[QPPF_ACODE_INTERRUPT] = QPPF_ATYPE_HARDCODED;
}

static void
resetVideoIn (void)
{
	VideoIn.edge = QPPF_DEF_VIUCLKEDGE;
	VideoIn.format = QPPF_DEF_VIUFORMAT;
	VideoIn.mode = QPPF_DEF_VIUMODE;
	VideoIn.syncCode1 = QPPF_DEF_VIUSYNCCODE1;
	VideoIn.syncCode2 = QPPF_DEF_VIUSYNCCODE2;
	VideoIn.startLine = QPPF_DEF_VIUSTARTLINE;
	VideoIn.startPixel = QPPF_DEF_VIUSTARTPEL;
	AttrStates[QPPF_ACODE_VIDEOIN] = QPPF_ATYPE_HARDCODED;
}

static void
resetAudioIn (void)
{
	AudioIn.bclk = QPPF_DEF_AIBCLK;
	AudioIn.i2s = QPPF_DEF_AII2S;
	AudioIn.lrclk = QPPF_DEF_AILRCLK;
	AudioIn.msb = QPPF_DEF_AIMSB;
	AudioIn.rj = QPPF_DEF_AIRJ;
	AttrStates[QPPF_ACODE_AUDIOIN] = QPPF_ATYPE_HARDCODED;
}

static void
resetVideoOut (void)
{
	VideoOut.enable = QPPF_DEF_VOENABLE;
	VideoOut.mode = QPPF_DEF_VOUMODE;
	VideoOut.startLine = QPPF_DEF_VOUSTARTLINE;
	VideoOut.startPixel = QPPF_DEF_VOUSTARTPEL;
	AttrStates[QPPF_ACODE_VIDEOOUT] = QPPF_ATYPE_HARDCODED;
}

static void
resetAudioOut (void)
{
	AudioOut.enable = QPPF_DEF_AOENABLE;
	AudioOut.bclk = QPPF_DEF_AOBCLK;
	AudioOut.i2s = QPPF_DEF_AOI2S;
	AudioOut.lrclk = QPPF_DEF_AOLRCLK;
	AudioOut.msb = QPPF_DEF_AOMSB;
	AudioOut.rj = QPPF_DEF_AORJ;
	AudioOut.mode = QPPF_DEF_AOMODE;
	AttrStates[QPPF_ACODE_AUDIOOUT] = QPPF_ATYPE_HARDCODED;
}

static void
resetVideoDec (void)
{
	VideoDec.type = QPPF_DEF_VIDDEC_TYPE;
	AttrStates[QPPF_ACODE_VIDEODEC] = QPPF_ATYPE_HARDCODED;
}

static void
resetVideoEnc (void)
{
	VideoEnc.type = QPPF_DEF_VIDENC_TYPE;
	VideoEnc.std = QPPF_DEF_VIDSTD;
	AttrStates[QPPF_ACODE_VIDEOENC] = QPPF_ATYPE_HARDCODED;
}

static void
resetTuner (void)
{
	Tuner.type = QPPF_DEF_TUNER_TYPE;
	AttrStates[QPPF_ACODE_TUNER] = QPPF_ATYPE_HARDCODED;
}

static void
resetTVAudio (void)
{
	TVAudio.type = QPPF_DEF_TVAUDIO_TYPE;
	AttrStates[QPPF_ACODE_TVAUDIO] = QPPF_ATYPE_HARDCODED;
}

static void
resetAudCodec (void)
{
	AudCodec.type = QPPF_DEF_AUDCODEC_TYPE;
	AttrStates[QPPF_ACODE_AUDIOCODEC] = QPPF_ATYPE_HARDCODED;
}

static void
resetChip (void)
{
	Chip.type = QPPF_DEF_CHIP_TYPE;
	AttrStates[QPPF_ACODE_CHIP] = QPPF_ATYPE_HARDCODED;
}

#ifdef QP_LINXKER
static void
resetFWAPI (void)
{

}

static void
resetCodecLib (void)
{

}

static void
resetI2C (void)
{

}

static void
resetGPIO (void)
{

}
#else
static void
resetFWAPI (void)
{
	FWAPI.version = QPPF_DEF_FWAPI_VERSION;
	AttrStates[QPPF_ACODE_FWAPI] = QPPF_ATYPE_HARDCODED;
}

static void
resetCodecLib (void)
{
	CodecLib.intMode = QPPF_DEF_CODECLIB_INTMODE;
	CodecLib.memMode = QPPF_DEF_CODECLIB_MEMMODE;
	AttrStates[QPPF_ACODE_CODECLIB] = QPPF_ATYPE_HARDCODED;
}

static void
resetI2C (void)
{
	I2C.type = QPPF_DEF_I2C_TYPE;
	I2C.clkBit = QPPF_DEF_I2C_CLKBIT;
	I2C.dataBit = QPPF_DEF_I2C_DATABIT;
	AttrStates[QPPF_ACODE_I2C] = QPPF_ATYPE_HARDCODED;
}

static void
resetGPIO (void)
{
	GPIO.directions = QPPF_DEF_GPIO_DIRECTIONS;
	GPIO.values = QPPF_DEF_GPIO_VALUES;
	AttrStates[QPPF_ACODE_GPIO] = QPPF_ATYPE_HARDCODED;
}
#endif

static EQP_Errors
getFW (TQP_UINT8 ** ptrP, PTQP_UINT32 sizeP, const TQP_UINT8 * arrayP, TQP_UINT32 size)
{
	EQP_Errors	rv = QPERR_PARMS;

	if ((ptrP != QP_NULL) && (sizeP == QP_NULL)) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR)
			,"getFW: NULL size pointer but not-NULL pointer pointer\n"));
		goto getFWExit;
	}

	if (ptrP != QP_NULL) {
		*ptrP = (TQP_UINT8 *) arrayP;
	}
	if (sizeP != QP_NULL) {
		*sizeP = size;
	}
	rv = QPERR_SUCCESS;

getFWExit:
	return (rv);
}
