///////////////////////////////////////////////////////////////////////////////
//        
//
//		Time Manager
//
// Implements platform-independent timing services.
//
// History:
//	Aug 23, 2007	(EC) Created
//
// Authors: (EC) Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////


#ifdef QP_LINXKER
#include "../common/qptypes.h"
#include "../common/qpdbgmgr.h"
#include "../common/qptimemgr.h"
#else
#include "qptypes.h"
#include "qpdbgmgr.h"
#include "qptimemgr.h"
#endif
#include "qptimemgrint.h"
#include "qptimemgrpriv.h"

static TQP_DbgMCompID DBGCompID = 0;

static TQP_BOOL Inited = QP_FALSE;


///////////////////////////////////////////////////////////////////////////////
//
//	PRIVATE CALLS
//
// These calls are meant to be issued by a control module (initialization,
// termination) and should not be called by any other module or manager.
//

//
// Initializes the data structures of the time manager.
// This call is non reentrant and must be called only once
//
EQP_Errors QPTMInitPriv (void)
{
	EQP_Errors		rv = QPERR_NOTIMPL;

	// register with the Debug Manager first

	QPDbgMRegComp(&DBGCompID,"QPTimeManager");

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMInitPriv: entered\n"));

	if (Inited) {
		rv = QPERR_INVALID;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPTMInitPriv: already initialized\n"));
		goto initExit;
	}

	// platform dependent initializations
	TimeInit();

	rv = QPERR_SUCCESS;
	Inited = QP_TRUE;

initExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMInitPriv: exiting %i\n",rv));
	return (rv);
}

//
// Terminates the data structures of the Time manager.
// This call is non reentrant and must be called only once
//
EQP_Errors QPTMTermPriv (void)
{
	EQP_Errors		rv = QPERR_NOTIMPL;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMTermPriv: entered\n"));

	if (!Inited) {
		rv = QPERR_INVALID;
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPTMTermPriv: not initialized\n"));
		goto termExit;
	}

	TimeTerm();

	Inited = QP_FALSE;

termExit:
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMTermPriv: exiting %i\n",rv));

	// unregister with the Debug Manager
	QPDbgMUnregComp(DBGCompID);

	return (rv);
}


///////////////////////////////////////////////////////////////////////////////
//
// GETCAPS
//
//

void
QPTimeGetCaps (PTQP_TMCaps capsP)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTimeGetCaps: entered 0x%X\n",capsP));

	if (capsP == QP_NULL) {
		QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_ERROR),"QPTimeGetCaps: NULL caps structure pointer\n"));
	}else{
		// platform-dependent capabilities
		TimeGetCaps(capsP);
	}

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTimeGetCaps: exiting\n"));
}


///////////////////////////////////////////////////////////////////////////////
//
// Getting/setting time
//

TQPTM_Time
QPTMGetSysTimeMilliS (void)
{
	TQPTM_Time		rv = 0;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMGetSysTimeMilliS: entered\n"));

	rv = TimeGetSysTimeMilliS();

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMGetSysTimeMilliS: exiting %X%08X\n"
		,(unsigned long) (rv>>32),(unsigned long) (rv&0xFFFFFFFF)));

	return (rv);
}

TQPTM_Time
QPTMGetSysTimeMicroS (void)
{
	TQPTM_Time		rv = 0;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"TimeGetSysTimeMicroS: entered\n"));

	rv = TimeGetSysTimeMicroS();

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"TimeGetSysTimeMicroS: exiting %X%08X\n"
		,(unsigned long) (rv>>32),(unsigned long) (rv&0xFFFFFFFF)));

	return (rv);
}

TQPTM_Time
QPTMGetSysTimeNanoS (void)
{
	TQPTM_Time		rv = 0;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMGetSysTimeNanoS: entered\n"));

	rv = TimeGetSysTimeNanoS();

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMGetSysTimeNanoS: exiting %X%08X\n"
		,(unsigned long) (rv>>32),(unsigned long) (rv&0xFFFFFFFF)));

	return (rv);
}

TQPTM_Time
QPTMGetSysTicks (void)
{
	TQPTM_Time		rv = 0;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMGetSysTicks: entered\n"));

	rv = TimeGetSysTicks();

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMGetSysTicks: exiting %X%08X\n"
		,(unsigned long) (rv>>32),(unsigned long) (rv&0xFFFFFFFF)));

	return (rv);
}


//
// Delay thread execution.
//
EQP_Errors
QPTMDelayMilliS (TQP_UINT32 milliS)
{
	EQP_Errors		rv = QPERR_NOTIMPL;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMDelayMilliS: entered %i\n",milliS));

	rv = TimeDelayMilliS(milliS);

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMDelayMilliS: exiting %i\n",rv));
	return (rv);
}

EQP_Errors
QPTMDelayMicroS (TQP_UINT32 microS)
{
	EQP_Errors		rv = QPERR_NOTIMPL;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMDelayMicroS: entered %i\n",microS));

	rv = TimeDelayMicroS(microS);

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMDelayMicroS: exiting %i\n",rv));
	return (rv);
}


///////////////////////////////////////////////////////////////////////////////
//
// Timers
//

EQP_Errors
QPTMStartTimer (TQP_TMTimerID *	tIDP, TQP_TMCallBack cbP, void * refCon, TQP_UINT32 microS)
{
	EQP_Errors		rv = QPERR_NOTIMPL;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMStartTimer: entered\n"));

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMStartTimer: exiting %i\n",rv));
	return (rv);
}

EQP_Errors
QPTMStartPeriodicTimer (TQP_TMTimerID *	tIDP, TQP_TMCallBack cbP, void * refCon, TQP_UINT32 microS)
{
	EQP_Errors		rv = QPERR_NOTIMPL;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMStartPeriodicTimer: entered\n"));

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMStartPeriodicTimer: exiting %i\n",rv));
	return (rv);
}

EQP_Errors
QPTMKillTimer (TQP_TMTimerID tID)
{
	EQP_Errors		rv = QPERR_NOTIMPL;
	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMKillTimer: entered\n"));

	QPDbgMPrintf((QPDBGMCOMPLVL(DBGCompID,QPDBGM_LEVEL_EXTRACE),"QPTMKillTimer: exiting %i\n",rv));
	return (rv);
}

