///////////////////////////////////////////////////////////////////////////////
//        
//
// Cadi7401.c
//
// Description: 
//
//  ADI7401 implementation
// 
// History:
//	June 18, 2008 created
//
// Authors: Andy Yu
//
///////////////////////////////////////////////////////////////////////////////


#include "cadi7401.h"
#include "qptimemgr.h"
#include "qpdbgmgr.h"
#include "util.h"


/////////////////////////////////////////////////////////////////////////////
//
//

#define	ADI7401_BRIGHTNESS_MIN				0x00
#define	ADI7401_BRIGHTNESS_MAX				0x1E
#define	ADI7401_BRIGHTNESS_DEFAULT		0x10

#define	ADI7401_CONTRAST_MIN					0x00   // not supported
#define	ADI7401_CONTRAST_MAX					0x7F   // not supported
#define	ADI7401_CONTRAST_DEFAULT			0x44   // not supported

#define	ADI7401_SATURATION_MIN				0x00   // not supported
#define	ADI7401_SATURATION_MAX				0x7F   // not supported
#define	ADI7401_SATURATION_DEFAULT		0x40   // not supported

#define	ADI7401_HUE_MIN								0x00
#define	ADI7401_HUE_MAX								0xFF
#define	ADI7401_HUE_DEFAULT						0x80

#define	ADI7401_SHARPNESS_MIN					0      // not supported
#define	ADI7401_SHARPNESS_MAX					15     // not supported
#define	ADI7401_SHARPNESS_DEFAULT			8      // not supported


extern TQP_DbgMCompID	g_QPPeripheralDBGCompID;


/////////////////////////////////////////////////////////////////////////////
//
//


static EQP_Errors CADI7401_readRegister(CADI7401 *This,
									   TQP_UINT8 regAddr, 
									   PTQP_UINT8 pValue
									   )
{
    return (II2C_WriteThenRead(This->m_pI2C,
							   This->m_cI2CDevAddr,	//device address
							   1,					//write size
							   &regAddr,			//write buffer
							   1,					//read size
							   pValue				//read buffer
							   ));
}



static EQP_Errors CADI7401_writeRegister(CADI7401 *This,
										TQP_UINT8 regAddr, 
										TQP_UINT8 value
										)
{
    TQP_UINT8 buffer[2];

	buffer[0] = regAddr;
	buffer[1] = value;

    return (II2C_Write(This->m_pI2C, 
					   This->m_cI2CDevAddr,	//device address
					   2,					//write size
					   buffer				//write buffer
					   ));
}


static EQP_Errors CADI7401_WriteBlock(CADI7401 *This,
									 TQP_UINT32 *pRegTable
									 )
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;
	TQP_UINT32	dwTotalReg = 0;

	// get total number of registers to write
	//
	while (-1 != pRegTable[dwTotalReg]) 
	{
		dwTotalReg++;
	}

	dwTotalReg >>= 1;

	// do multiple writes
	//
	while (dwTotalReg)
	{
		qpStatus = CADI7401_writeRegister(This,
										 (TQP_UINT8)pRegTable[0], 
										 (TQP_UINT8)pRegTable[1]
										 );
		if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CADI7401_WriteBlock() ****** writeRegister failed! status(%d) ******\n",
						 qpStatus
						 ));
			break;
		}

		pRegTable += 2;
		dwTotalReg--;
	}

	return (qpStatus);
}



/////////////////////////////////////////////////////////////////////////////
//
//


// IGenericDevice
//
static EQP_Errors CADI7401_InitDevice(IVideoDecoder *This_p, 
									 QP_DEVICE_CALLBACK pFuncCallback,
									 TQP_PVOID context,
									 TQP_PVOID pInitData
									 )
{
    TQP_UINT8     buffer[2];   // Andy 2008.10.06: Monterey, control ADI7401
    TQP_UINT8     ucRegAddr;
    TQP_UINT8     ucValue;

    CVideoDecoder *pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
    CADI7401      *This = GET_INHERITED_OBJECT(CADI7401, m_VideoDecoder, pVideoDecoder);

    static TQP_UINT32 dwADI7401RegValsNtsc[] = 
    {
        0x05, 0x01,
        0x06, 0x06,
        0x1D, 0x47,
        0x3A, 0x11,
        0x3B, 0x80,
        0x3C, 0x5C,
        0x6B, 0x83,
        0x85, 0x19,
        -1,-1
    };

    static TQP_UINT32 dwADI7401RegValsPal[] = 
    {
        0x05, 0x01,
        0x06, 0x07,
        0x1D, 0x47,
        0x3A, 0x11,
        0x3B, 0x80,
        0x3C, 0x5C,
        0x6B, 0x83,
        0x85, 0x19,
        -1,-1
    };

    static TQP_UINT32 dwADI7401RegVals720p[] = 
    {
        0x05, 0x01,
        0x06, 0x0A,
        0x1D, 0x47,
        0x3A, 0x21,
        0x3B, 0x80,
        0x3C, 0x5C,
        0x6B, 0x83,
        0x85, 0x19,
        -1,-1
    };

    static TQP_UINT32 dwADI7401RegVals1080i[] = 
    {
        0x05, 0x01,
        0x06, 0x0C,
        0x1D, 0x47,
        0x3A, 0x21,
        0x3B, 0x80,
        0x3C, 0x5D,
        0x6B, 0x83,
        0x85, 0x19,
        -1,-1
    };

    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
                 "CADI7401_InitDevice()\n"
                ));

    if (CObject_IsInitialized(&pVideoDecoder->m_Object))
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
                     "CADI7401_InitDevice() Already Initialized. Skipping...\n" 
                    ));
      return (QPERR_SUCCESS);
    }
	
    // save callback address and context
    //
    pVideoDecoder->m_pDeviceCallback = pFuncCallback;
    pVideoDecoder->m_callbackContext = context;

    // Andy 2008.10.06: enable ADI7401 start
    ucRegAddr = 0x03;
    II2C_WriteThenRead(This->m_pI2C,
                                  ADI7401_DEVICE_ADDR,      //device address: ADI7401
                                  1,         //write size
                                  &ucRegAddr,  //write buffer
                                  1,         //read size
                                  &ucValue     //read buffer
                                 );
    buffer[0] = 0x03;
    buffer[1] = (ucValue & 0xBF);   // clear register 0x03 bit 6
    II2C_Write(This->m_pI2C,
                          ADI7401_DEVICE_ADDR,           //device address: ADI7401
                          2,              //write size
                          buffer          //write buffer
                         );

    ucRegAddr = 0x04;
    II2C_WriteThenRead(This->m_pI2C,
                                  ADI7401_DEVICE_ADDR,      //device address: ADI7401
                                  1,         //write size
                                  &ucRegAddr,  //write buffer
                                  1,         //read size
                                  &ucValue     //read buffer
                                 );
    buffer[0] = 0x04;
    buffer[1] = (ucValue | 0x08);   // set register 0x04 bit 3
    II2C_Write(This->m_pI2C,
                          ADI7401_DEVICE_ADDR,           //device address: ADI7401
                          2,              //write size
                          buffer          //write buffer
                         );

    ucRegAddr = 0x1D;
    II2C_WriteThenRead(This->m_pI2C,
                                  ADI7401_DEVICE_ADDR,      //device address: ADI7401
                                  1,         //write size
                                  &ucRegAddr,  //write buffer
                                  1,         //read size
                                  &ucValue     //read buffer
                                 );
    buffer[0] = 0x1D;
    buffer[1] = (ucValue & 0x7F);   // clear register 0x1D bit 8
    II2C_Write(This->m_pI2C,
                          ADI7401_DEVICE_ADDR,           //device address: ADI7401
                          2,              //write size
                          buffer          //write buffer
                         );

    QPTMDelayMilliS(10);   // delay 10ms
    // Andy 2008.10.06: enable ADI7401 end

    // Initialize ADI7401 with the default value
    //
    CADI7401_WriteBlock(This,
                        dwADI7401RegVals720p
                       );

    // program the video decoder with settings read from the registry
    //
    //QPVideoDecoder_SetVideoSystem(This_p, 
    //                              pVideoDecoder->m_VidStandard
    //                             );        
    
    //QPVideoDecoder_SetVideoRes(This_p, 1920, 1080);
    
    QPVideoDecoder_SelectVideoSource(This_p, 
                                  pVideoDecoder->m_VidSource
                                 );

    QPVideoDecoder_SetBrightness(This_p, 
                                 pVideoDecoder->m_nBrightness
                                );
    QPVideoDecoder_SetContrast(This_p, 
                               pVideoDecoder->m_nContrast
                              );
    QPVideoDecoder_SetSaturation(This_p, 
                                 pVideoDecoder->m_nSaturation
                                );
    QPVideoDecoder_SetHue(This_p, 
                          pVideoDecoder->m_nHue
                         );
    QPVideoDecoder_SetSharpness(This_p, 
                                pVideoDecoder->m_nSharpness
                               );

    // CObject Init
    //
    CObject_Init(&pVideoDecoder->m_Object);

    return (QPERR_SUCCESS);
}



// IVideoDecoder
//
static EQP_Errors CADI7401_SetVideoRes(IVideoDecoder *This_p,
									   TQP_UINT32 dwWid,
				  					   TQP_UINT32 dwHgt,
				  					   TQP_UINT32 dwScanFmt,
				  					   TQP_UINT32 dwFps
									   )
{
    static TQP_UINT32 dwADI7401RegValsNtsc[] = 
    {
        0x05, 0x01,
        0x06, 0x06,
        0x1D, 0x47,
        0x3A, 0x11,
        0x3B, 0x80,
        0x3C, 0x5C,
        0x6B, 0x83,
        0x85, 0x19,
        -1,-1
    };

    static TQP_UINT32 dwADI7401RegValsPal[] = 
    {
        0x05, 0x01,
        0x06, 0x07,
        0x1D, 0x47,
        0x3A, 0x11,
        0x3B, 0x80,
        0x3C, 0x5C,
        0x6B, 0x83,
        0x85, 0x19,
        -1,-1
    };

    static TQP_UINT32 dwADI7401RegVals720p[] = 
    {
        0x05, 0x01,
        0x06, 0x0A,
        0x1D, 0x47,
        0x3A, 0x21,
        0x3B, 0x80,
        0x3C, 0x5C,
        0x6B, 0x83,
        0x85, 0x19,
        -1,-1
    };

    static TQP_UINT32 dwADI7401RegVals1080i[] = 
    {
        0x05, 0x01,
        0x06, 0x0C,
        0x1D, 0x47,
        0x3A, 0x21,
        0x3B, 0x80,
        0x3C, 0x5D,
        0x6B, 0x83,
        0x85, 0x19,
        -1,-1
    };

	TQP_UINT32		*pVals;
	EQP_Errors		qpStatus;
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CADI7401		*This = GET_INHERITED_OBJECT(CADI7401, m_VideoDecoder, pVideoDecoder);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
				 "CADI7401_SetVideoRes() width(%d) height(%d)\n",
				 dwWid, dwHgt
				 ));

	if (((720 == dwWid) && (480 == dwHgt)) || ((640 == dwWid) && (480 == dwHgt)))
	{
		pVals = dwADI7401RegValsNtsc;
	}
	else if ((720 == dwWid) && (576 == dwHgt))
	{
		pVals = dwADI7401RegValsPal;
	}
	else if ((1280 == dwWid) && (720 == dwHgt))
	{
		pVals = dwADI7401RegVals720p;
	}
	else if ((1920 == dwWid) && (1080 == dwHgt))
	{
		pVals = dwADI7401RegVals1080i;
  }
  else
  {
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CADI7401_SetVideoRes() Unsupported video resolution: %dx%d \n", 
					 dwWid, dwHgt
					 ));
		return (QPERR_PARMS);
	}

	qpStatus = CADI7401_WriteBlock(This,
								   pVals
								   );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CADI7401_SetVideoRes() ****** WriteBlock() failed to write ADI7401 registers! ******\n"
					 ));
		return (QPERR_FAIL);
	}

	//pVideoDecoder->m_VidStandard = standard;

	return (QPERR_SUCCESS);
}



// constructor
//
CADI7401 * CADI7401_Constructor(CADI7401 *pADI7401, 
								CObject *pParent,
								TQP_UINT32 dwAttributes,
								ITuner *pTuner,
								ITVAudio *pTvAudio,
								II2C *pI2C,
								TQP_UINT8 cI2CDevAddr
								)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CADI7401_Constructor()\n"
				 ));

	if (pADI7401)
	{
		CVideoDecoder	*pVideoDecoder;

		pVideoDecoder = CVideoDecoder_Constructor(&pADI7401->m_VideoDecoder, 
												  pParent,
												  dwAttributes,
												  pTuner,
												  pTvAudio
												  );

		if (!pVideoDecoder)
		{
			pADI7401 = QP_NULL;
		}
		else
		{
			// fill in the function table
			//

			// IGenericDevice
			pADI7401->m_VideoDecoder.m_iVideoDecoder.InitDevice = CADI7401_InitDevice;

			// IVideoDecoder
			pADI7401->m_VideoDecoder.m_iVideoDecoder.SetVideoRes = CADI7401_SetVideoRes;

			// initialize members
			//
			pADI7401->m_pI2C = pI2C;
			pADI7401->m_cI2CDevAddr = cI2CDevAddr;

			// inherited members
			//
			pADI7401->m_VideoDecoder.m_nBrightnessDef = ADI7401_BRIGHTNESS_DEFAULT;
			pADI7401->m_VideoDecoder.m_nBrightnessMin = ADI7401_BRIGHTNESS_MIN;
			pADI7401->m_VideoDecoder.m_nBrightnessMax = ADI7401_BRIGHTNESS_MAX;
			pADI7401->m_VideoDecoder.m_nBrightness = ADI7401_BRIGHTNESS_DEFAULT;
			pADI7401->m_VideoDecoder.m_nContrastDef = ADI7401_CONTRAST_DEFAULT; 
			pADI7401->m_VideoDecoder.m_nContrastMin = ADI7401_CONTRAST_MIN;
			pADI7401->m_VideoDecoder.m_nContrastMax = ADI7401_CONTRAST_MAX;
			pADI7401->m_VideoDecoder.m_nContrast = ADI7401_CONTRAST_DEFAULT;
			pADI7401->m_VideoDecoder.m_nSaturationDef = ADI7401_SATURATION_DEFAULT;
			pADI7401->m_VideoDecoder.m_nSaturationMin = ADI7401_SATURATION_MIN;
			pADI7401->m_VideoDecoder.m_nSaturationMax = ADI7401_SATURATION_MAX;
			pADI7401->m_VideoDecoder.m_nSaturation = ADI7401_SATURATION_DEFAULT;
			pADI7401->m_VideoDecoder.m_nHueDef = ADI7401_HUE_DEFAULT;
			pADI7401->m_VideoDecoder.m_nHueMin = ADI7401_HUE_MIN;
			pADI7401->m_VideoDecoder.m_nHueMax = ADI7401_HUE_MAX;
			pADI7401->m_VideoDecoder.m_nHue = ADI7401_HUE_DEFAULT;
			pADI7401->m_VideoDecoder.m_nSharpnessDef = ADI7401_SHARPNESS_DEFAULT;
			pADI7401->m_VideoDecoder.m_nSharpnessMin = ADI7401_SHARPNESS_MIN;
			pADI7401->m_VideoDecoder.m_nSharpnessMax = ADI7401_SHARPNESS_MAX;
			pADI7401->m_VideoDecoder.m_nSharpness = ADI7401_SHARPNESS_DEFAULT;
		}
	}

	return (pADI7401);
}



// destructor
//
void CADI7401_Destructor(CADI7401 *This)
{
    TQP_UINT8     buffer[2];
    TQP_UINT8     ucRegAddr;
    TQP_UINT8     ucValue;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CADI7401_Destructor()\n"
				 ));

    // Andy 2008.10.06: disable ADV7401
    ucRegAddr = 0x03;
    II2C_WriteThenRead(This->m_pI2C,
                                  ADI7401_DEVICE_ADDR,      //device address: ADI7401
                                  1,         //write size
                                  &ucRegAddr,  //write buffer
                                  1,         //read size
                                  &ucValue     //read buffer
                                 );
    buffer[0] = 0x03;
    buffer[1] = (ucValue | 0x40);   // set register 0x03 bit 6
    II2C_Write(This->m_pI2C,
                          ADI7401_DEVICE_ADDR,           //device address: ADI7401
                          2,              //write size
                          buffer          //write buffer
                         );

    ucRegAddr = 0x04;
    II2C_WriteThenRead(This->m_pI2C,
                                  ADI7401_DEVICE_ADDR,      //device address: ADI7401
                                  1,         //write size
                                  &ucRegAddr,  //write buffer
                                  1,         //read size
                                  &ucValue     //read buffer
                                 );
    buffer[0] = 0x04;
    buffer[1] = (ucValue & 0xF7);   // clear register 0x04 bit 3
    II2C_Write(This->m_pI2C,
                          ADI7401_DEVICE_ADDR,           //device address: ADI7401
                          2,              //write size
                          buffer          //write buffer
                         );

    ucRegAddr = 0x1D;
    II2C_WriteThenRead(This->m_pI2C,
                                  ADI7401_DEVICE_ADDR,      //device address: ADI7401
                                  1,         //write size
                                  &ucRegAddr,  //write buffer
                                  1,         //read size
                                  &ucValue     //read buffer
                                 );
    buffer[0] = 0x1D;
    buffer[1] = (ucValue | 0x80);   // set register 0x1D bit 8
    II2C_Write(This->m_pI2C,
                          ADI7401_DEVICE_ADDR,           //device address: ADI7401
                          2,              //write size
                          buffer          //write buffer
                         );

    QPTMDelayMilliS(10);   // delay 10ms

    CVideoDecoder_Destructor(&This->m_VideoDecoder);
}

