///////////////////////////////////////////////////////////////////////////////
//        
//
// cadv7614.c
//
// Description: 
//   ADV7614 implementation
// 
// History:
//	 Sep 2, 2009 created
//
// Authors: Jack Lee
//
///////////////////////////////////////////////////////////////////////////////


#include "cadv7614.h"
#include "qptimemgr.h"
#include "qpdbgmgr.h"
#include "util.h"


/////////////////////////////////////////////////////////////////////////////
//
//

#define	ADV7614_BRIGHTNESS_MIN				0x00   // not supported
#define	ADV7614_BRIGHTNESS_MAX				0xFF   // not supported
#define	ADV7614_BRIGHTNESS_DEFAULT			0x80   // not supported

#define	ADV7614_CONTRAST_MIN				0x00   // not supported
#define	ADV7614_CONTRAST_MAX				0x7F   // not supported
#define	ADV7614_CONTRAST_DEFAULT			0x44   // not supported

#define	ADV7614_SATURATION_MIN				0x00   // not supported
#define	ADV7614_SATURATION_MAX				0x7F   // not supported
#define	ADV7614_SATURATION_DEFAULT			0x40   // not supported

#define	ADV7614_HUE_MIN						(-128)	// 0x80   // not supported
#define	ADV7614_HUE_MAX						(+127)	// 0x7F   // not supported
#define	ADV7614_HUE_DEFAULT					0x00   // not supported

#define	ADV7614_SHARPNESS_MIN				0      // not supported
#define	ADV7614_SHARPNESS_MAX				15     // not supported
#define	ADV7614_SHARPNESS_DEFAULT			8      // not supported


extern TQP_DbgMCompID	g_QPPeripheralDBGCompID;


/////////////////////////////////////////////////////////////////////////////
//
//


static EQP_Errors CADV7614_readRegister(CADV7614 *This,
									   TQP_UINT8 regAddr, 
									   PTQP_UINT8 pValue
									   )
{
    return (II2C_WriteThenRead(This->m_pI2C,
							   This->m_cI2CDevAddr,	//device address
							   1,					//write size
							   &regAddr,			//write buffer
							   1,					//read size
							   pValue				//read buffer
							   ));
}



static EQP_Errors CADV7614_writeRegister(CADV7614 *This,
										TQP_UINT8 regAddr, 
										TQP_UINT8 value
										)
{
    TQP_UINT8 buffer[2];

	buffer[0] = regAddr;
	buffer[1] = value;

    return (II2C_Write(This->m_pI2C, 
					   This->m_cI2CDevAddr,	//device address
					   2,					//write size
					   buffer				//write buffer
					   ));
}


static EQP_Errors CADV7614_WriteBlock(CADV7614 *This, TQP_UINT8 *pRegTable)
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;

	// get total number of registers to write
	//
	while (0xff != pRegTable[0])
	{
		qpStatus = II2C_Write(This->m_pI2C, 
					   pRegTable[0],	//device address
					   2,				//write size
					   pRegTable+1		//write buffer
					   );										 
		if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CADV7614_WriteBlock() ****** writeRegister failed! status(%d) ******\n",
						 qpStatus
						 ));
			break;
		}

		pRegTable += 3;
	}

	return (qpStatus);
}



/////////////////////////////////////////////////////////////////////////////
//
//


// IGenericDevice
//
EQP_Errors CADV7614_InitDevice(IVideoDecoder *This_p, 
							   QP_DEVICE_CALLBACK pFuncCallback,
							   TQP_PVOID context,
							   TQP_PVOID pInitData
							   )
{
    TQP_UINT8     buffer[2];
    TQP_UINT8     ucRegAddr;
    TQP_UINT8     ucValue;

    CVideoDecoder *pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
    CADV7614      *This = GET_INHERITED_OBJECT(CADV7614, m_VideoDecoder, pVideoDecoder);

    static TQP_UINT8 dwADV7614RegVals[] = 
    {
		ADV7614_DEVICE_ADDR_IO,       0x15, 0xBE, // Tristate all outputs from video core
		ADV7614_DEVICE_ADDR_HDMI,     0x1A, 0x1A, // Mute audio
		ADV7614_DEVICE_ADDR_HDMI,     0x01, 0x78, // Open termination
		ADV7614_DEVICE_ADDR_IO,       0x0C, 0x42, // Power up part and power down VDP
		ADV7614_DEVICE_ADDR_IO,       0xF7, 0x78, // DPP Address to 78
		ADV7614_DEVICE_ADDR_IO,       0x00, 0x1F, // Vid std 1920 x 1080
		ADV7614_DEVICE_ADDR_IO,       0x01, 0x05, // Prim_Mode HDMI-COMP
		ADV7614_DEVICE_ADDR_IO,       0x02, 0xF5, // Auto CSC, YPrPb out, Set ALT_DATA_SAT 
		ADV7614_DEVICE_ADDR_IO,       0x03, 0x80, // 24 bit SDR 422 Mode 0
		ADV7614_DEVICE_ADDR_IO,       0x04, 0xA2, // 
		ADV7614_DEVICE_ADDR_IO,       0x05, 0x28, // AV Codes Off
		ADV7614_DEVICE_ADDR_IO,       0x06, 0xA6, // VS and HS Inverted
		ADV7614_DEVICE_ADDR_IO,       0x0B, 0x44, // Power down ESDP block
		ADV7614_DEVICE_ADDR_IO,       0x14, 0x6A, // Drive strength adjustment
		ADV7614_DEVICE_ADDR_IO,       0x15, 0x80, // Disable Tristate of Pins
		ADV7614_DEVICE_ADDR_IO,       0x33, 0x60, // LLC DLL Mux Enable
		ADV7614_DEVICE_ADDR_CP,       0x69, 0x30, // Enable CP CSC
		ADV7614_DEVICE_ADDR_CP,       0xBA, 0x01, // Set HDMI FreeRun
		ADV7614_DEVICE_ADDR_CP,	      0xCF, 0x01, // Power off macrovision
		ADV7614_DEVICE_ADDR_AFE_DPLL, 0x00, 0xFF, // Power Down ADC's and there associated clocks
		ADV7614_DEVICE_ADDR_AFE_DPLL, 0x01, 0xFE, // Power down ref buffer/bandgap/clamps/sync strippers/input mux/output buffer
		ADV7614_DEVICE_ADDR_AFE_DPLL, 0x13, 0x80, // Set LLC DLL Phase
		ADV7614_DEVICE_ADDR_AFE_DPLL, 0xB5, 0x01, // Setting MCLK to 256Fs
		ADV7614_DEVICE_ADDR_AFE_DPLL, 0xC8, 0x40, // ADI recommended setting
		ADV7614_DEVICE_ADDR_HDMI,     0x01, 0x00, // Enable clock terminators
		ADV7614_DEVICE_ADDR_HDMI,     0x0D, 0x84, // ADI recommended write
		ADV7614_DEVICE_ADDR_HDMI,     0x1A, 0x0A, // Set Unmute Delay to 1_5 sec
		ADV7614_DEVICE_ADDR_HDMI,     0x3D, 0x10, // ADI Recommended Setting
		ADV7614_DEVICE_ADDR_HDMI,     0x3E, 0x39, // ADI recommended setting
		ADV7614_DEVICE_ADDR_HDMI,     0x48, 0x40, // 
		ADV7614_DEVICE_ADDR_HDMI,     0x57, 0x90, // Enable dynamic PLL control
		ADV7614_DEVICE_ADDR_HDMI,     0x58, 0x01, // Recommended PLL setting
		ADV7614_DEVICE_ADDR_HDMI,     0x59, 0xA3, // ADI recommended setting
		ADV7614_DEVICE_ADDR_HDMI,     0x93, 0x8B, // Equaliser ADI recommended setting
		ADV7614_DEVICE_ADDR_HDMI,     0x94, 0x2D, // Equaliser ADI recommended setting
		ADV7614_DEVICE_ADDR_HDMI,     0x96, 0x01, // Enable automatic EQ changing
		ADV7614_DEVICE_ADDR_HDMI,     0x1A, 0x0A, // Unmute audio

        0xff, 0xff, 0xff
    };

    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
                 "CADV7614_InitDevice()\n"
                ));

    if (CObject_IsInitialized(&pVideoDecoder->m_Object))
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
                     "CADV7614_InitDevice() Already Initialized. Skipping...\n" 
                    ));
      return (QPERR_SUCCESS);
    }
	
    // save callback address and context
    //
    pVideoDecoder->m_pDeviceCallback = pFuncCallback;
    pVideoDecoder->m_callbackContext = context;

    // Initialize ADV7614 with the default value
    //
    CADV7614_WriteBlock(This, dwADV7614RegVals);

    // program the video decoder with settings read from the registry
    //
    QPVideoDecoder_SetVideoSystem(This_p, pVideoDecoder->m_VidStandard);        
    QPVideoDecoder_SelectVideoSource(This_p, pVideoDecoder->m_VidSource);

    QPVideoDecoder_SetBrightness(This_p, pVideoDecoder->m_nBrightness);
    QPVideoDecoder_SetContrast(This_p, pVideoDecoder->m_nContrast);
    QPVideoDecoder_SetSaturation(This_p, pVideoDecoder->m_nSaturation);
    QPVideoDecoder_SetHue(This_p, pVideoDecoder->m_nHue);
    QPVideoDecoder_SetSharpness(This_p, pVideoDecoder->m_nSharpness);

    // CObject Init
    //
    CObject_Init(&pVideoDecoder->m_Object);

    return (QPERR_SUCCESS);
}



// IVideoDecoder
//
EQP_Errors CADV7614_SetVideoRes(IVideoDecoder *This_p,
								TQP_UINT32 dwWid,
				  				TQP_UINT32 dwHgt,
				  				TQP_UINT32 dwScanFmt,
				  				TQP_UINT32 dwFps
				  				)
{
    CVideoDecoder *pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
    CADV7614      *This = GET_INHERITED_OBJECT(CADV7614, m_VideoDecoder, pVideoDecoder);
    
    static TQP_UINT8 dwRes1[] = 
    {
    	ADV7614_DEVICE_ADDR_IO,       0x00, 0x1F, 
        0xff, 0xff, 0xff
    };

    static TQP_UINT8 dwRes2[] = 
    {
		ADV7614_DEVICE_ADDR_IO,       0x00, 0x0A, 
        0xff, 0xff, 0xff
    };
  
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CADV7614_SetVideoRes() : Video Res Set to %dx %d\n",
				 dwWid,
				 dwHgt
				 ));    

	if (dwHgt >= 720) {
		CADV7614_WriteBlock(This, dwRes1);
	} else {
		CADV7614_WriteBlock(This, dwRes2);
	}

	return (QPERR_SUCCESS);
}



// constructor
//
CADV7614 * CADV7614_Constructor(CADV7614 *pADV7614, 
								CObject *pParent,
								TQP_UINT32 dwAttributes,
								ITuner *pTuner,
								ITVAudio *pTvAudio,
								II2C *pI2C,
								TQP_UINT8 cI2CDevAddr
								)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CADV7614_Constructor()\n"
				 ));

	if (pADV7614)
	{
		CVideoDecoder	*pVideoDecoder;

		pVideoDecoder = CVideoDecoder_Constructor(&pADV7614->m_VideoDecoder, 
												  pParent,
												  dwAttributes,
												  pTuner,
												  pTvAudio
												  );

		if (!pVideoDecoder)
		{
			pADV7614 = QP_NULL;
		}
		else
		{
			// fill in the function table
			//

			// IGenericDevice
			pADV7614->m_VideoDecoder.m_iVideoDecoder.InitDevice = CADV7614_InitDevice;

			// IVideoDecoder
			pADV7614->m_VideoDecoder.m_iVideoDecoder.SetVideoRes = CADV7614_SetVideoRes;

			// initialize members
			//
			pADV7614->m_pI2C = pI2C;
			pADV7614->m_cI2CDevAddr = cI2CDevAddr;

			// inherited members
			//
			pADV7614->m_VideoDecoder.m_nBrightnessDef = ADV7614_BRIGHTNESS_DEFAULT;
			pADV7614->m_VideoDecoder.m_nBrightnessMin = ADV7614_BRIGHTNESS_MIN;
			pADV7614->m_VideoDecoder.m_nBrightnessMax = ADV7614_BRIGHTNESS_MAX;
			pADV7614->m_VideoDecoder.m_nBrightness = ADV7614_BRIGHTNESS_DEFAULT;
			pADV7614->m_VideoDecoder.m_nContrastDef = ADV7614_CONTRAST_DEFAULT; 
			pADV7614->m_VideoDecoder.m_nContrastMin = ADV7614_CONTRAST_MIN;
			pADV7614->m_VideoDecoder.m_nContrastMax = ADV7614_CONTRAST_MAX;
			pADV7614->m_VideoDecoder.m_nContrast = ADV7614_CONTRAST_DEFAULT;
			pADV7614->m_VideoDecoder.m_nSaturationDef = ADV7614_SATURATION_DEFAULT;
			pADV7614->m_VideoDecoder.m_nSaturationMin = ADV7614_SATURATION_MIN;
			pADV7614->m_VideoDecoder.m_nSaturationMax = ADV7614_SATURATION_MAX;
			pADV7614->m_VideoDecoder.m_nSaturation = ADV7614_SATURATION_DEFAULT;
			pADV7614->m_VideoDecoder.m_nHueDef = ADV7614_HUE_DEFAULT;
			pADV7614->m_VideoDecoder.m_nHueMin = ADV7614_HUE_MIN;
			pADV7614->m_VideoDecoder.m_nHueMax = ADV7614_HUE_MAX;
			pADV7614->m_VideoDecoder.m_nHue = ADV7614_HUE_DEFAULT;
			pADV7614->m_VideoDecoder.m_nSharpnessDef = ADV7614_SHARPNESS_DEFAULT;
			pADV7614->m_VideoDecoder.m_nSharpnessMin = ADV7614_SHARPNESS_MIN;
			pADV7614->m_VideoDecoder.m_nSharpnessMax = ADV7614_SHARPNESS_MAX;
			pADV7614->m_VideoDecoder.m_nSharpness = ADV7614_SHARPNESS_DEFAULT;
		}
	}

	return (pADV7614);
}



// destructor
//
void CADV7614_Destructor(CADV7614 *This)
{
    TQP_UINT8     buffer[2];
    TQP_UINT8     ucRegAddr;
    TQP_UINT8     ucValue;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CADV7614_Destructor()\n"
				 ));

	// Tristate all outputs from video core
    buffer[0] = 0x15;
    buffer[1] = 0xbe;
    II2C_Write(This->m_pI2C,
               ADV7614_DEVICE_ADDR_IO,           //device address
               2,              //write size
               buffer          //write buffer
               );

    QPTMDelayMilliS(10);   // delay 10ms

    CVideoDecoder_Destructor(&This->m_VideoDecoder);
}

