///////////////////////////////////////////////////////////////////////////////
//        
//
// Ci2c.c
//
// Description: 
//
//  I2C provider implementation
// 
// History:
//	Aug 28, 2006	(dt) created
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////

#if defined(QP_LINXKER)
#include "qlautoconf.h"
#endif

#include "ci2c.h"
#include "cusbcntl.h"
#include "util.h"
#include "qpdbgmgr.h"
#include "hci.h"
#include "cqlcodec.h"
#if defined(_WIN32) && !defined(_QP_WINKER)
#include "qldev_interface.h"
#endif //WIN32 &!_QP_WINKER
#if (!defined(QP_LINXKER) && !defined(_QP_LINUX) && !defined(_QP_MACOSX)) || defined(CONFIG_PL330SOC)
#include "cgpio_i2c.h"
#endif

extern TQP_DbgMCompID	g_QPPeripheralDBGCompID;


/////////////////////////////////////////////////////////////////////////////
//
//

// Dummy I2C function
//

static EQP_Errors Ci2cDummyRead(II2C *This_p,
							  TQP_UINT8 cDeviceAddr,
							  int nSize,
							  PTQP_UINT8 pBuf
							  )
{
	return	QPERR_FAIL;
}



static EQP_Errors Ci2cDummyWrite(II2C *This_p,
							   TQP_UINT8 cDeviceAddr,
							   int nSize,
							   PTQP_UINT8 pBuf
							   )
{
	return	QPERR_FAIL;
}



static EQP_Errors Ci2cDummyWriteThenRead(II2C *This_p,
									   TQP_UINT8 cDeviceAddr,
									   int nWriteSize,
									   const PTQP_UINT8 pWriteBuf,
									   int nReadSize,
									   PTQP_UINT8 pReadBuf
									   )
{
	return	QPERR_FAIL;
}

/////////////////////////////////////////////////////////////////////////////
//
//

// use USB controller I2C function
//

static EQP_Errors Ci2cUsbRead(II2C *This_p, 
							  TQP_UINT8 cDeviceAddr, 
							  int nSize, 
							  PTQP_UINT8 pBuf
							  )
{
	CI2C	*This = GET_INHERITED_OBJECT(CI2C, m_ii2c, This_p);

	return (CUsbCntl_I2CRead(This->m_pUsbCntl, 
							 cDeviceAddr, 
							 nSize, 
							 pBuf
							 ));
}



static EQP_Errors Ci2cUsbWrite(II2C *This_p, 
							   TQP_UINT8 cDeviceAddr, 
							   int nSize, 
							   PTQP_UINT8 pBuf
							   )
{
	CI2C	*This = GET_INHERITED_OBJECT(CI2C, m_ii2c, This_p);

	return (CUsbCntl_I2CWrite(This->m_pUsbCntl, 
							  cDeviceAddr, 
							  nSize, 
							  pBuf
							  ));
}



static EQP_Errors Ci2cUsbWriteThenRead(II2C *This_p, 
									   TQP_UINT8 cDeviceAddr, 
									   int nWriteSize, 
									   const PTQP_UINT8 pWriteBuf,
									   int nReadSize,
									   PTQP_UINT8 pReadBuf
									   )
{
	CI2C	*This = GET_INHERITED_OBJECT(CI2C, m_ii2c, This_p);

	return (CUsbCntl_I2CWriteThenRead(This->m_pUsbCntl, 
									  cDeviceAddr, 
									  nWriteSize, 
									  pWriteBuf,
									  nReadSize,
									  pReadBuf
									  ));
}

/////////////////////////////////////////////////////////////////////////////
//
//

// use USB controller Software I2C function
//

static EQP_Errors Ci2cUsbSWI2CRead(II2C *This_p,
							  TQP_UINT8 cDeviceAddr,
							  int nSize,
							  PTQP_UINT8 pBuf
							  )
{
	CI2C	*This = GET_INHERITED_OBJECT(CI2C, m_ii2c, This_p);

	return (CUsbCntl_SWI2CRead(This->m_pUsbCntl,
							 cDeviceAddr,
							 nSize,
							 pBuf
							 ));
}



static EQP_Errors Ci2cUsbSWI2CWrite(II2C *This_p,
							   TQP_UINT8 cDeviceAddr,
							   int nSize,
							   PTQP_UINT8 pBuf
							   )
{
	CI2C	*This = GET_INHERITED_OBJECT(CI2C, m_ii2c, This_p);

	return (CUsbCntl_SWI2CWrite(This->m_pUsbCntl,
							  cDeviceAddr,
							  nSize,
							  pBuf
							  ));
}



static EQP_Errors Ci2cUsbSWI2CWriteThenRead(II2C *This_p,
									   TQP_UINT8 cDeviceAddr,
									   int nWriteSize,
									   const PTQP_UINT8 pWriteBuf,
									   int nReadSize,
									   PTQP_UINT8 pReadBuf
									   )
{
	CI2C	*This = GET_INHERITED_OBJECT(CI2C, m_ii2c, This_p);

	return (CUsbCntl_SWI2CWriteThenRead(This->m_pUsbCntl,
									  cDeviceAddr,
									  nWriteSize,
									  pWriteBuf,
									  nReadSize,
									  pReadBuf
									  ));
}


/////////////////////////////////////////////////////////////////////////////
//
//

// use QL201 I2C function
//

#define QL201_I2C_ID					0x0055

#define REG_QL201_I2C_CONTROL			0x0500
#define REG_QL201_I2C_WRITE_BUF1		0x0504
#define REG_QL201_I2C_WRITE_BUF2		0x0508
#define REG_QL201_I2C_READ_BUF1			0x050C
#define REG_QL201_I2C_READ_BUF2			0x0510
#define REG_QL201_I2C_DEVICE_ID			0x0514
#define REG_QL201_I2C_DEBUG				0x0518


#pragma pack(4)


typedef union _QL201_I2C_CONTROL
{
	TQP_UINT32	Read; 
	struct
	{
		TQP_UINT32	master_wnum		: 3;	// bit 0 - 2
		TQP_UINT32	master_rnum		: 3;	// bit 3 - 5
		TQP_UINT32	_10b			: 1;	// bit 6
		TQP_UINT32	target_addr_id	: 10;	// bit 7 - 16
		TQP_UINT32	reserved		: 14;	// bit 17 - 30
		TQP_UINT32	t_en			: 1;	// bit 31
	} Reg;
} QL201_I2C_CONTROL, *PQL201_I2C_CONTROL;


typedef union _QL201_I2C_DATA_BUF
{
	TQP_UINT32	Read; 
	struct
	{
		TQP_UINT8	data[4];
	} Reg;
} QL201_I2C_DATA_BUF, *PQL201_I2C_DATA_BUF;


typedef union _QL201_I2C_DEVICE_ID
{
	TQP_UINT32	Read; 
	struct
	{
		TQP_UINT32	sa				: 10;	// bit 0 - 9
		TQP_UINT32	reserved		: 22;	// bit 10 - 31
	} Reg;
} QL201_I2C_DEVICE_ID, *PQL201_I2C_DEVICE_ID;

#pragma pack()


static void CI2CQl201_InitDeviceID(CI2C *This)
{
	QL201_I2C_DEVICE_ID	id;

	id.Read = 0;
	id.Reg.sa = QL201_I2C_ID;

	QPHCI_RegisterWrite(&This->m_pMpegCodec->m_hci,
						REG_QL201_I2C_DEVICE_ID, 
						id.Read
						);
}



static TQP_BOOL CI2CQl201_TxReady(CI2C *This)
{
	QL201_I2C_CONTROL	cntl;
	EQP_Errors			qpStatus;

	qpStatus = QPHCI_RegisterRead(&This->m_pMpegCodec->m_hci,
								  REG_QL201_I2C_CONTROL, 
								  &cntl.Read
								  );
	if (!QP_SUCCESS(qpStatus))
		return (QP_FALSE);

	return (0 == cntl.Reg.t_en);
}



EQP_Errors _CI2CQl201_Read(CI2C *This, 
						   TQP_UINT8 cDeviceAddr, 
						   int nRdSize, 
						   PTQP_UINT8 pRdBuf,
						   int nWrSize, 
						   PTQP_UINT8 pWrBuf,
						   TQP_BOOL fCritical
						   )
{
	EQP_Errors			qpStatus = QPERR_SUCCESS;
#if !defined(CONFIG_HOST_PLATFORM_PL330SOC)
	TQP_INT64			llTimeTo = 500;	// 500 msec timeout
#else
	TQP_INT64			llTimeTo = 2000;	// try 3000 times
#endif
	QL201_I2C_CONTROL	cntl;
	QL201_I2C_DATA_BUF	data[2];
	int					i;

	if (!pRdBuf || 
		(nRdSize >= 8) ||
		(nWrSize >= 8) ||
		((nWrSize != 0) && !pWrBuf)
		)
	{
		return (QPERR_PARMS);
	}

	if (fCritical)
	{
		ENTER_CRITICAL(&This->m_pMpegCodec->m_Object)
	}
#if defined(CONFIG_HOST_PLATFORM_PL330SOC)
	else
	{
		if (This->m_hw_i2c_in_use) 
		{
			return QPERR_FAIL;
		}
	}

	This->m_hw_i2c_in_use = 1;
#endif


#if 0
	if (!CI2CQl201_TxReady(This))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "_CI2CQl201_Read(0x%x) not ready!\n",
					 cDeviceAddr
					 ));
		CI2CQl201_InitDeviceID(This);
		qpStatus = QPERR_FAIL;
	}
	else
	{
#endif
		// write I2C data
		//
		data[0].Read = 0;
		data[1].Read = 0;

		for (i = 0; i < nWrSize; i++)
		{
			data[i / 4].Reg.data[i % 4] = pWrBuf[i];
		}

		if (nWrSize > 0)
		{
			qpStatus = QPHCI_RegisterWrite(&This->m_pMpegCodec->m_hci,
										   REG_QL201_I2C_WRITE_BUF1, 
										   data[0].Read
										   );
			if (QP_SUCCESS(qpStatus) && 
				(nWrSize > 4)
				)
			{
				qpStatus = QPHCI_RegisterWrite(&This->m_pMpegCodec->m_hci,
											   REG_QL201_I2C_WRITE_BUF2, 
											   data[1].Read
											   );
			}
		}

		// start I2C read
		//
		if (QP_SUCCESS(qpStatus))
		{
			cntl.Read = 0;

			cntl.Reg.target_addr_id = cDeviceAddr;
			cntl.Reg.master_wnum = nWrSize;
			cntl.Reg.master_rnum = nRdSize;
			cntl.Reg.t_en = 1;

			qpStatus = QPHCI_RegisterWrite(&This->m_pMpegCodec->m_hci,
										   REG_QL201_I2C_CONTROL, 
										   cntl.Read
										   );

			// wait for t_en bit auto clear
			//
			if (QP_SUCCESS(qpStatus))
			{
#if !defined(CONFIG_HOST_PLATFORM_PL330SOC)
				BeginTimer(&llTimeTo);
#endif

				while (0 != cntl.Reg.t_en) 
				{
#if !defined(CONFIG_HOST_PLATFORM_PL330SOC)
					if (Timeout(llTimeTo, 1))
#else
					if (0 == --llTimeTo)
#endif
					{
						QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
									 "_CI2CQl201_Read(0x%x) wait for I2C read timeout\n",
									 cDeviceAddr
									 ));
						qpStatus = QPERR_FAIL;
						break;
					}

					qpStatus = QPHCI_RegisterRead(&This->m_pMpegCodec->m_hci, 
												  REG_QL201_I2C_CONTROL, 
												  &cntl.Read
												  );
					if (!QP_SUCCESS(qpStatus))
						break;
				}
			}

			// read back I2C data
			//
			if (QP_SUCCESS(qpStatus))
			{
				qpStatus = QPHCI_RegisterRead(&This->m_pMpegCodec->m_hci, 
											  REG_QL201_I2C_READ_BUF1, 
											  &data[0].Read
											  );
				if (QP_SUCCESS(qpStatus) &&
					(nRdSize > 4)
					)
				{
					qpStatus = QPHCI_RegisterRead(&This->m_pMpegCodec->m_hci,
												  REG_QL201_I2C_READ_BUF2,
												  &data[1].Read
												  );
				}

				for (i = 0; i < nRdSize; i++)
				{						
					pRdBuf[i] = data[i / 4].Reg.data[i % 4];
				}
			}
		}
#if 0
	}
#endif

#if defined(CONFIG_HOST_PLATFORM_PL330SOC)
	This->m_hw_i2c_in_use = 0;
#endif

	if (fCritical)
	{
		LEAVE_CRITICAL(&This->m_pMpegCodec->m_Object)
	}

	return (qpStatus);
}



EQP_Errors _CI2CQl201_Write(CI2C *This, 
						    TQP_UINT8 cDeviceAddr, 
						    int nSize, 
						    PTQP_UINT8 pBuf,
						    TQP_BOOL fCritical
						    )
{
	EQP_Errors			qpStatus;
#if !defined(CONFIG_HOST_PLATFORM_PL330SOC)
	TQP_INT64			llTimeTo = 500;	// 500 msec timeout
#else
	TQP_INT64			llTimeTo = 2000;	// try 3000 times
#endif
	QL201_I2C_CONTROL	cntl;
	QL201_I2C_DATA_BUF	data[2];
	int					i;

	if (!pBuf ||
		nSize >= 8
		)
	{
		return (QPERR_PARMS);
	}


	if (fCritical)
	{
		ENTER_CRITICAL(&This->m_pMpegCodec->m_Object)
	} 
#if defined(CONFIG_HOST_PLATFORM_PL330SOC)
	else
	{
		if (This->m_hw_i2c_in_use) 
		{
			return QPERR_FAIL;
		}
	}

	This->m_hw_i2c_in_use = 1;
#endif

#if 0
	if (!CI2CQl201_TxReady(This))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "_CI2CQl201_Write(0x%x) not ready!\n",
					 cDeviceAddr
					 ));
		CI2CQl201_InitDeviceID(This);
		qpStatus = QPERR_FAIL;
	}
	else
	{
#endif
		// write I2C data
		//
		data[0].Read = 0;
		data[1].Read = 0;

		for (i = 0; i < nSize; i++)
		{
			data[i / 4].Reg.data[i % 4] = pBuf[i];
		}

		qpStatus = QPHCI_RegisterWrite(&This->m_pMpegCodec->m_hci,
									   REG_QL201_I2C_WRITE_BUF1, 
									   data[0].Read
									   );
		if (QP_SUCCESS(qpStatus) && 
			(nSize > 4)
			)
		{
			qpStatus = QPHCI_RegisterWrite(&This->m_pMpegCodec->m_hci,
										   REG_QL201_I2C_WRITE_BUF2, 
										   data[1].Read
										   );
		}

		// start I2C write
		//
		if (QP_SUCCESS(qpStatus))
		{
			cntl.Read = 0;

			cntl.Reg.target_addr_id = cDeviceAddr;
			cntl.Reg.master_wnum = nSize;
			cntl.Reg.t_en = 1;

			qpStatus = QPHCI_RegisterWrite(&This->m_pMpegCodec->m_hci, 
										   REG_QL201_I2C_CONTROL, 
										   cntl.Read
										   );

			// wait for t_en bit auto clear
			//
			if (QP_SUCCESS(qpStatus))
			{
#if !defined(CONFIG_HOST_PLATFORM_PL330SOC)
				BeginTimer(&llTimeTo);
#endif

				while (0 != cntl.Reg.t_en) 
				{
#if !defined(CONFIG_HOST_PLATFORM_PL330SOC)
					if (Timeout(llTimeTo, 1))
#else
					if (0 == --llTimeTo)
#endif
					{
						QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
									 "_CI2CQl201_Write(0x%x) len(%d) (0x%x) wait for I2C write timeout\n",
									 cDeviceAddr,
									 nSize,
									 pBuf[0]
									 ));
						qpStatus = QPERR_FAIL;
						break;
					}

					qpStatus = QPHCI_RegisterRead(&This->m_pMpegCodec->m_hci,
												  REG_QL201_I2C_CONTROL, 
												  &cntl.Read
												  );
					if (!QP_SUCCESS(qpStatus))
						break;
				}
			}
		}
#if 0
	}
#endif

#if defined(CONFIG_HOST_PLATFORM_PL330SOC)
	This->m_hw_i2c_in_use = 0;
#endif

	if (fCritical)
	{
		LEAVE_CRITICAL(&This->m_pMpegCodec->m_Object)
	}

	return (qpStatus);
}




static EQP_Errors CI2CQl201_Read(II2C *This_p,
								 TQP_UINT8 cDeviceAddr, 
								 int nSize, 
								 PTQP_UINT8 pBuf
								 )
{
	CI2C	*This =	GET_INHERITED_OBJECT(CI2C, m_ii2c, This_p);

	return (_CI2CQl201_Read(This,
							cDeviceAddr, 
							nSize, 
							pBuf,
							0,
							QP_NULL,
							QP_TRUE
							));
}


static EQP_Errors CI2CQl201_Write(II2C *This_p,
								  TQP_UINT8 cDeviceAddr, 
								  int nSize, 
								  PTQP_UINT8 pBuf
								  )
{
	CI2C	*This = GET_INHERITED_OBJECT(CI2C, m_ii2c, This_p);

	return (_CI2CQl201_Write(This,
							 cDeviceAddr, 
							 nSize, 
							 pBuf,
							 QP_TRUE
							 ));
}


static EQP_Errors CI2CQl201_WriteThenRead(II2C *This_p,
										  TQP_UINT8 cDeviceAddr, 
										  int nWriteSize, 
										  const PTQP_UINT8 pWriteBuf,
										  int nReadSize,
										  PTQP_UINT8 pReadBuf
										  )
{
	CI2C	*This =	GET_INHERITED_OBJECT(CI2C, m_ii2c, This_p);

	return (_CI2CQl201_Read(This,
							cDeviceAddr, 
							nReadSize, 
							pReadBuf,
							nWriteSize,
							(PTQP_UINT8)pWriteBuf,
							QP_TRUE
							));
}


/////////////////////////////////////////////////////////////////////////////
//
//

// use 201 emulation mode I2C function
//

#if defined(_WIN32) && !defined(_QP_WINKER)

static EQP_Errors Ci2cEmuRead(II2C *This_p, 
							  TQP_UINT8 cDeviceAddr, 
							  int nSize, 
							  PTQP_UINT8 pBuf
							  )
{
	TQP_BOOL	ret;

	ret = ReadI2C(cDeviceAddr, 
				  (TQP_UINT8)nSize, 
				  pBuf
				  );
	return (ret ? QPERR_SUCCESS : QPERR_FAIL);
}



static EQP_Errors Ci2cEmuWrite(II2C *This_p, 
							   TQP_UINT8 cDeviceAddr, 
							   int nSize, 
							   PTQP_UINT8 pBuf
							   )
{
	TQP_BOOL	ret;

	ret = WriteI2C(cDeviceAddr, 
				   (TQP_UINT8)nSize, 
				   pBuf
				   );
	return (ret ? QPERR_SUCCESS : QPERR_FAIL);
}



static EQP_Errors Ci2cEmuWriteThenRead(II2C *This_p, 
									   TQP_UINT8 cDeviceAddr, 
									   int nWriteSize, 
									   const PTQP_UINT8 pWriteBuf,
									   int nReadSize,
									   PTQP_UINT8 pReadBuf
									   )
{
	TQP_BOOL	ret;

	ret = WriteThenReadI2C(cDeviceAddr, 
						   (TQP_UINT8)nWriteSize, 
						   pWriteBuf,
						   (TQP_UINT8)nReadSize,
						   pReadBuf
						   );
	return (ret ? QPERR_SUCCESS : QPERR_FAIL);
}

#else //!(WIN32 &!_QP_WINKER)

static EQP_Errors Ci2cEmuRead(II2C *This_p, 
							  TQP_UINT8 cDeviceAddr, 
							  int nSize, 
							  PTQP_UINT8 pBuf
							  )
{
	return (QPERR_SUCCESS);
}



static EQP_Errors Ci2cEmuWrite(II2C *This_p, 
							   TQP_UINT8 cDeviceAddr, 
							   int nSize, 
							   PTQP_UINT8 pBuf
							   )
{
	return (QPERR_SUCCESS);
}



static EQP_Errors Ci2cEmuWriteThenRead(II2C *This_p, 
									   TQP_UINT8 cDeviceAddr, 
									   int nWriteSize, 
									   const PTQP_UINT8 pWriteBuf,
									   int nReadSize,
									   PTQP_UINT8 pReadBuf
									   )
{
	return (QPERR_SUCCESS);
}

#endif //WIN32 &!_QP_WINKER

#if (!defined(QP_LINXKER) && !defined(_QP_LINUX) && !defined(_QP_MACOSX)) || defined(CONFIG_PL330SOC)
static EQP_Errors CI2CGPIORead(II2C *This_p,
								 TQP_UINT8 cDeviceAddr, 
								 int nSize, 
								 PTQP_UINT8 pBuf
								 )
{
	EQP_Errors err = QPERR_SUCCESS;
	struct i2c_msg i2cMsg[2];
	
	i2cMsg[0].addr = cDeviceAddr;
	i2cMsg[0].buf = pBuf;// buffer will overlapp
	i2cMsg[0].flags = I2C_M_RD;
	i2cMsg[0].len = (TQP_UINT16)nSize;

	if (i2c_transfer(
#ifndef _QP_QPSOS
			This_p,
#endif
			&i2cMsg[0],
			1
			)<0) {
		err = QPERR_FAIL;
	}
	
	return err;
}

static EQP_Errors CI2CGPIOWrite(II2C *This_p,
								  TQP_UINT8 cDeviceAddr, 
								  int nSize, 
								  PTQP_UINT8 pBuf
								  )
{
	EQP_Errors err = QPERR_SUCCESS;
	struct i2c_msg i2cMsg[2];
	
	i2cMsg[0].addr = cDeviceAddr;
	i2cMsg[0].len = (TQP_UINT16)nSize;
	i2cMsg[0].buf = pBuf;
	i2cMsg[0].flags = I2C_M_WRITE;
	
	if (i2c_transfer(
#ifndef _QP_QPSOS
			This_p,
#endif		
			&i2cMsg[0],
			1
			) < 0) {
		err = QPERR_FAIL;
	}
	return err;
}

static EQP_Errors CI2CGPIOWriteThenRead(II2C *This_p,
										  TQP_UINT8 cDeviceAddr, 
										  int nWriteSize, 
										  const PTQP_UINT8 pWriteBuf,
										  int nReadSize,
										  PTQP_UINT8 pReadBuf
										  )
{
	EQP_Errors err = QPERR_SUCCESS;
	struct i2c_msg i2cMsg[2];

	i2cMsg[0].addr = cDeviceAddr;
	i2cMsg[0].buf = pWriteBuf;
	i2cMsg[0].flags = I2C_M_WRITE;	// for write
	i2cMsg[0].len = (TQP_UINT16)nWriteSize;// usually '1' for subaddress

	i2cMsg[1].addr = cDeviceAddr;
	i2cMsg[1].buf = pReadBuf;// buffer to rcv data
	i2cMsg[1].flags = I2C_M_RD;
	i2cMsg[1].len = (TQP_UINT16)nReadSize;

	if (i2c_transfer(
#ifndef _QP_QPSOS
			This_p,
#endif		
			&i2cMsg[0],
			2
			)<0) {
		err = QPERR_FAIL;
	}
	
	return err;
}
#endif //!QP_LINXKER

/////////////////////////////////////////////////////////////////////////////
//
//

// constructor
//
CI2C * CI2C_Constructor(CI2C *pI2c, 
						CObject *pParent,
						TQP_UINT32 dwAttributes,
						QPI2C_TYPE type,
						TQP_UINT32 clk,
						TQP_UINT32 data,
						CUsbCntl *pUsbCntl,
						CQLCodec *pMpegCodec
						)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CI2C_Constructor() type(%d)\n", 
				 type
				 ));

	if (pI2c)
	{
		CObject_Constructor(&pI2c->m_Object, 
							pParent, 
							dwAttributes
							);
		pI2c->m_type = type;
		pI2c->m_clk = clk;
		pI2c->m_data = data;
		pI2c->m_pUsbCntl = pUsbCntl;
		pI2c->m_pMpegCodec = pMpegCodec;
		pI2c->m_hw_i2c_in_use = 0;

		// fill in the function table
		//
		if (QPI2C_TYPE_DUMMY == pI2c->m_type)
		{
			// II2C
			pI2c->m_ii2c.Read = Ci2cDummyRead;
			pI2c->m_ii2c.Write = Ci2cDummyWrite;
			pI2c->m_ii2c.WriteThenRead = Ci2cDummyWriteThenRead;
		}
		else if (QPI2C_TYPE_USE_USB == pI2c->m_type)
		{
			// II2C
			pI2c->m_ii2c.Read = Ci2cUsbRead;
			pI2c->m_ii2c.Write = Ci2cUsbWrite;
			pI2c->m_ii2c.WriteThenRead = Ci2cUsbWriteThenRead;
		}
		else if (QPI2C_TYPE_USE_QLCODEC == pI2c->m_type)
		{
			// II2C
			pI2c->m_ii2c.Read = CI2CQl201_Read;
			pI2c->m_ii2c.Write = CI2CQl201_Write;
			pI2c->m_ii2c.WriteThenRead = CI2CQl201_WriteThenRead;
		}
		else if (QPI2C_TYPE_USE_201_EMULATION == pI2c->m_type)
		{
			// II2C
			pI2c->m_ii2c.Read = Ci2cEmuRead;
			pI2c->m_ii2c.Write = Ci2cEmuWrite;
			pI2c->m_ii2c.WriteThenRead = Ci2cEmuWriteThenRead;
		}
		else if (QPI2C_TYPE_SOC_INTERNAL_HW == pI2c->m_type)
		{
			// II2C
			pI2c->m_ii2c.Read = CI2CQl201_Read;
			pI2c->m_ii2c.Write = CI2CQl201_Write;
			pI2c->m_ii2c.WriteThenRead = CI2CQl201_WriteThenRead;
		}
#if (!defined(QP_LINXKER) && !defined(_QP_LINUX) && !defined(_QP_MACOSX)) || defined(CONFIG_PL330SOC)
		else if (QPI2C_TYPE_SOC_INTERNAL_GPIO == pI2c->m_type)
		{
			i2c_init(
#ifndef _QP_QPSOS
					 &pI2c->m_ii2c, &pI2c->m_pMpegCodec->m_hci,
#endif //!_QP_QPSOS
					 pI2c->m_clk, 
					 pI2c->m_data, 
					 1, 
					 450
					 );

			pI2c->m_ii2c.Read = CI2CGPIORead;
			pI2c->m_ii2c.Write = CI2CGPIOWrite;
			pI2c->m_ii2c.WriteThenRead = CI2CGPIOWriteThenRead;
		}
#endif
		else if (QPI2C_TYPE_USE_USB_SWI2C == pI2c->m_type)
		{
			// II2C
			pI2c->m_ii2c.Read = Ci2cUsbSWI2CRead;
			pI2c->m_ii2c.Write = Ci2cUsbSWI2CWrite;
			pI2c->m_ii2c.WriteThenRead = Ci2cUsbSWI2CWriteThenRead;
		}
		else
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CI2C_Constructor() unknown type(%d)\n", 
						 type
						 ));
			CI2C_Destructor(pI2c);
			return (QP_NULL);
		}
	}

	return (pI2c);
}


// destructor
//
void CI2C_Destructor(CI2C *This)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CI2C_Destructor()\n"
				 ));
#if (!defined(QP_LINXKER) && !defined(_QP_LINUX) && !defined(_QP_MACOSX)) || defined(CONFIG_PL330SOC)
	if (QPI2C_TYPE_SOC_INTERNAL_GPIO == This->m_type)
		i2c_term(
#ifndef _QP_QPSOS
			&This->m_ii2c		
#endif		
			);
#endif

	CObject_Destructor(&This->m_Object);
}




