///////////////////////////////////////////////////////////////////////////////
//        
//
// Cmi2010.c
//
// Description: 
//
//  Micron MI2010(MT9D111) implementation
// 
// History:
//	July 17, 2009	(dt) created
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


#include "cmi2010.h"
#include "qpdbgmgr.h"
#include "util.h"


/////////////////////////////////////////////////////////////////////////////
//
//

#define	MI2010_BRIGHTNESS_MIN					(-64)
#define	MI2010_BRIGHTNESS_MAX					64
#define	MI2010_BRIGHTNESS_DEFAULT				0x00

#define	MI2010_CONTRAST_MIN						0x00
#define	MI2010_CONTRAST_MAX						64
#define	MI2010_CONTRAST_DEFAULT					32

#define	MI2010_SATURATION_MIN					0x00
#define	MI2010_SATURATION_MAX					128
#define	MI2010_SATURATION_DEFAULT				64

#define	MI2010_HUE_MIN							(-128)	// 0x80
#define	MI2010_HUE_MAX							(+127)	// 0x7F
#define	MI2010_HUE_DEFAULT						0x00

#define	MI2010_SHARPNESS_MIN					0
#define	MI2010_SHARPNESS_MAX					128
#define	MI2010_SHARPNESS_DEFAULT				64

#define	MI2010_EDGE_ENHANCEMENT_MIN				0x00
#define	MI2010_EDGE_ENHANCEMENT_MAX				0xFF
#define	MI2010_EDGE_ENHANCEMENT_DEFAULT			0x80

#define	MI2010_GAMMA_MIN						0x00
#define	MI2010_GAMMA_MAX						64
#define	MI2010_GAMMA_DEFAULT					32

#define	MI2010_LOW_LIGHT_BOOST_MIN				1
#define	MI2010_LOW_LIGHT_BOOST_MAX				3
#define	MI2010_LOW_LIGHT_BOOST_DEFAULT			2

#define	MI2010_EXPOSURE_MIN						0x00
#define	MI2010_EXPOSURE_MAX						0xFF
#define	MI2010_EXPOSURE_DEFAULT					0x80

#define	MI2010_WHITE_BALANCE_MIN				2800
#define	MI2010_WHITE_BALANCE_MAX				6500
#define	MI2010_WHITE_BALANCE_DEFAULT			2500

#define	MI2010_FOCUS_MIN						0x00
#define	MI2010_FOCUS_MAX						0xFF
#define	MI2010_FOCUS_DEFAULT					0x80


extern TQP_DbgMCompID	g_QPPeripheralDBGCompID;


/////////////////////////////////////////////////////////////////////////////
//
//


const TQP_UINT8 Sensor_INIT_COMMON_table[] = 
{
// [COMMON] //////////////////////////////////////////////////
0xF0, 0x00, 0x00,
0x65, 0xA0, 0x00,		// bypassed PLL (prepare for soft reset)

0xF0, 0x00, 0x01,		
0xC3, 0x05, 0x01,		// MCU_BOOT_MODE (MCU reset)

0xF0, 0x00, 0x00,		
0x0D, 0x00, 0x21,		// RESET_REG (enable soft reset)
0x00, 0x00, 0x00, 0x00, 0x00, 0x01,  //Delay
                 		
0x0D, 0x00, 0x00,		// RESET_REG (disable soft reset)
0x00, 0x00, 0x00, 0x00, 0x00, 0x64,  //Delay

0x2B, 0x00, 0x20,		// Doesn't do anything, used as a placeholder
0x25, 0x40, 0x00,		// Line valid Control

// [Core]
0x01, 0x00, 0x1C, 		// ROW_WINDOW_START_REG
0x02, 0x00, 0x3C, 		// COL_WINDOW_START_REG
0x03, 0x04, 0xB4,		// ROW_WINDOW_SIZE_REG 
0x04, 0x06, 0x42, 		// COL_WINDOW_SIZE_REG 

0xF0, 0x00, 0x01,
0xC6, 0xA1, 0x02,
0xC8, 0x00, 0x00,

// Black Level and Digital Gain


// Color Matrix/Color Correction
0xC6, 0x21, 0x60,
0xC8, 0x39, 0x23, 		// COLOR_CORR_MATRIX_SCALE_14
0xC6, 0x21, 0x61,
0xC8, 0x04, 0xE4, 		// COLOR_CORR_MATRIX_SCALE_11
0xC6, 0x21, 0x62,
0xC8, 0xE2, 0xFB,		// COLOR_CORR_MATRIX_1_2
0xC6, 0x21, 0x63,
0xC8, 0x90, 0x0B, 		// COLOR_CORR_MATRIX_3_4
0xC6, 0x21, 0x64,
0xC8, 0x3B, 0xEA, 		// COLOR_CORR_MATRIX_5_6
0xC6, 0x21, 0x65,
0xC8, 0xCD, 0x6B, 		// COLOR_CORR_MATRIX_7_8
0xC6, 0x21, 0x66,
0xC8, 0x3F, 0xC3, 		// COLOR_CORR_MATRIX_9

// Lens correction
0xF0, 0x00, 0x02,
0x80, 0x01, 0xA8,		//lens_correction_control
0x81, 0x6A, 0x35,		//zone_bounds_x1_x2
0x82, 0x2F, 0x9F,		//zone_bounds_x0_x3
0x83, 0x8D, 0x5E,		//zone_bounds_x4_x5
0x84, 0x51, 0x28,		//zone_bounds_y1_y2
0x85, 0x23, 0x79,		//zone_bounds_y0_y3
0x86, 0x68, 0x45,		//zone_bounds_y4_y5
0x87, 0x0C, 0x0C,		//center_offset
0x88, 0x01, 0xC7,		//fx_red
0x8B, 0x01, 0x46,		//fy_red
0x8E, 0x0A, 0x3E,		//df_dx_red
0x91, 0x08, 0xDE,		//df_dy_red
0x94, 0x27, 0xD8,		//second_deriv_zone_0_red
0x97, 0x11, 0x46,		//second_deriv_zone_1_red
0x9A, 0x56, 0x48,		//second_deriv_zone_2_red
0x9D, 0x14, 0x6A,		//second_deriv_zone_3_red
0xA0, 0x3B, 0x46,		//second_deriv_zone_4_red
0xA3, 0x1B, 0x6C,		//second_deriv_zone_5_red
0xA6, 0x5B, 0x45,		//second_deriv_zone_6_red
0xA9, 0x0D, 0xCC,		//second_deriv_zone_7_red
0x89, 0x01, 0x84,		//fx_green
0x8C, 0x01, 0x22,		//fy_green
0x8F, 0x0C, 0x6B,		//df_dx_green
0x92, 0x0A, 0x67,		//df_dy_green
0x95, 0x0C, 0xBD,		//second_deriv_zone_0_green
0x98, 0x10, 0x41,		//second_deriv_zone_1_green
0x9B, 0x50, 0x3C,		//second_deriv_zone_2_green
0x9E, 0x0F, 0x5B,		//second_deriv_zone_3_green
0xA1, 0x38, 0x41,		//second_deriv_zone_4_green
0xA4, 0x16, 0x5B,		//second_deriv_zone_5_green
0xA7, 0x51, 0x49,		//second_deriv_zone_6_green
0xAA, 0xD9, 0xA6,		//second_deriv_zone_7_green
0x8A, 0x01, 0x62,		//fx_blue
0x8D, 0x00, 0xFD,		//fy_blue
0x90, 0x0B, 0xF9,		//df_dx_blue
0x93, 0x0A, 0x44,		//df_dy_blue
0x96, 0x21, 0xD5,		//second_deriv_zone_0_blue
0x99, 0x0F, 0x3B,		//second_deriv_zone_1_blue
0x9C, 0x47, 0x32,		//second_deriv_zone_2_blue
0x9F, 0x0F, 0x51,		//second_deriv_zone_3_blue
0xA2, 0x2E, 0x3C,		//second_deriv_zone_4_blue
0xA5, 0x14, 0x48,		//second_deriv_zone_5_blue
0xA8, 0x45, 0x36,		//second_deriv_zone_6_blue
0xAB, 0xFD, 0xD4,		//second_deriv_zone_7_blue
0xAC, 0x00, 0x83,		//x2_factors
0xAD, 0x00, 0x00,		//global_offset_fxy_function
0xAE, 0x00, 0x00,		//k_factor_in_k_fx_fy

0xF0, 0x00, 0x01,
0x08, 0x01, 0xFC,

// Defect Correction

// Auto Focus Statistics

//AE & AWB Setting
0xC6, 0xA1, 0x37, 		// MCU_ADDRESS [SEQ_PREVIEW_3_AE]
0xC8, 0x00, 0x03, 		// MCU_DATA_0
0xC6, 0xA1, 0x39, 		// MCU_ADDRESS [SEQ_PREVIEW_3_AWB]
0xC8, 0x00, 0x03, 		// MCU_DATA_0
0xC6, 0xA1, 0x38, 		// MCU_ADDRESS [SEQ_PREVIEW_3_FD]
0xC8, 0x00, 0x01, 		// MCU_DATA_0
0xC6, 0xA1, 0x3B,		// MCU_ADDRESS [SEQ_PREVIEW_3_HG]
0xC8, 0x00, 0x03, 		// MCU_DATA_0

// Defect Correction

// Auto Focus Statistics

// AE
0xC6, 0xA2, 0x06, 		// MCU_ADDRESS [AE_TARGET]
0xC8, 0x00, 0x48, 		// MCU_DATA_0
0xC6, 0xA2, 0x07, 		// MCU_ADDRESS [AE_GATE]
0xC8, 0x00, 0x04, 		// MCU_DATA_0
0xC6, 0x22, 0x14,		// MCU_ADDRESS [AE_MAX_DGAIN_AE1]
0xC8, 0x00, 0x96,		// MCU_DATA_0
0xC6, 0xA2, 0x16,		// MCU_ADDRESS [AE_MAX_DGAIN_AE2]
0xC8, 0x00, 0x32,		// MCU_DATA_0

// AWB
0xC6, 0xA2, 0x06, 		// MCU_ADDRESS [AE_TARGET]
0xC8, 0x00, 0x48, 		// MCU_DATA_0
0xC6, 0xA2, 0x07, 		// MCU_ADDRESS [AE_GATE]
0xC8, 0x00, 0x04, 		// MCU_DATA_0
0xC6, 0x23, 0x06, 		// MCU_ADDRESS [AWB_CCM_L_0]
0xC8, 0x02, 0x34, 		// MCU_DATA_0
0xC6, 0x23, 0x08, 		// MCU_ADDRESS [AWB_CCM_L_1]
0xC8, 0xFE, 0xBC, 		// MCU_DATA_0
0xC6, 0x23, 0x0A, 		// MCU_ADDRESS [AWB_CCM_L_2]
0xC8, 0x00, 0x14, 		// MCU_DATA_0
0xC6, 0x23, 0x0C, 		// MCU_ADDRESS [AWB_CCM_L_3]
0xC8, 0xFF, 0xCB, 		// MCU_DATA_0
0xC6, 0x23, 0x0E, 		// MCU_ADDRESS [AWB_CCM_L_4]
0xC8, 0x01, 0x7C, 		// MCU_DATA_0
0xC6, 0x23, 0x10, 		// MCU_ADDRESS [AWB_CCM_L_5]
0xC8, 0xFF, 0xC1, 		// MCU_DATA_0
0xC6, 0x23, 0x12, 		// MCU_ADDRESS [AWB_CCM_L_6]
0xC8, 0xFF, 0x65, 		// MCU_DATA_0
0xC6, 0x23, 0x14, 		// MCU_ADDRESS [AWB_CCM_L_7]
0xC8, 0xFD, 0x81,		// MCU_DATA_0
0xC6, 0x23, 0x16, 		// MCU_ADDRESS [AWB_CCM_L_8]
0xC8, 0x04, 0x4D, 		// MCU_DATA_0
0xC6, 0x23, 0x18, 		// MCU_ADDRESS [AWB_CCM_L_9]
0xC8, 0x00, 0x20, 		// MCU_DATA_0
0xC6, 0x23, 0x1A, 		// MCU_ADDRESS [AWB_CCM_L_10]
0xC8, 0x00, 0x35, 		// MCU_DATA_0
0xC6, 0x23, 0x1C, 		// MCU_ADDRESS [AWB_CCM_RL_0]
0xC8, 0xFF, 0x52, 		// MCU_DATA_0
0xC6, 0x23, 0x1E, 		// MCU_ADDRESS [AWB_CCM_RL_1]
0xC8, 0x00, 0xAA, 		// MCU_DATA_0
0xC6, 0x23, 0x20, 		// MCU_ADDRESS [AWB_CCM_RL_2]
0xC8, 0x00, 0x10, 		// MCU_DATA_0
0xC6, 0x23, 0x22, 		// MCU_ADDRESS [AWB_CCM_RL_3]
0xC8, 0xFF, 0xD5, 		// MCU_DATA_0
0xC6, 0x23, 0x24, 		// MCU_ADDRESS [AWB_CCM_RL_4]
0xC8, 0x00, 0x39, 		// MCU_DATA_0
0xC6, 0x23, 0x26, 		// MCU_ADDRESS [AWB_CCM_RL_5]
0xC8, 0xFF, 0xF2, 		// MCU_DATA_0
0xC6, 0x23, 0x28, 		// MCU_ADDRESS [AWB_CCM_RL_6]
0xC8, 0x00, 0x66, 		// MCU_DATA_0
0xC6, 0x23, 0x2A, 		// MCU_ADDRESS [AWB_CCM_RL_7]
0xC8, 0x01, 0xCB, 		// MCU_DATA_0
0xC6, 0x23, 0x2C, 		// MCU_ADDRESS [AWB_CCM_RL_8]
0xC8, 0xFD, 0xA1, 		// MCU_DATA_0
0xC6, 0x23, 0x2E, 		// MCU_ADDRESS [AWB_CCM_RL_9]
0xC8, 0x00, 0x0F, 		// MCU_DATA_0
0xC6, 0x23, 0x30, 		// MCU_ADDRESS [AWB_CCM_RL_10]
0xC8, 0xFF, 0xFD, 		// MCU_DATA_0
0xC6, 0xA3, 0x48, 		// MCU_ADDRESS [AWB_GAIN_BUFFER_SPEED]
0xC8, 0x00, 0x08, 		// MCU_DATA_0
0xC6, 0xA3, 0x49, 		// MCU_ADDRESS [AWB_JUMP_DIVISOR]
0xC8, 0x00, 0x02, 		// MCU_DATA_0
0xC6, 0xA3, 0x4A, 		// MCU_ADDRESS [AWB_GAIN_MIN]
0xC8, 0x00, 0x53, 		// MCU_DATA_0
0xC6, 0xA3, 0x4B, 		// MCU_ADDRESS [AWB_GAIN_MAX]
0xC8, 0x00, 0xC0, 		// MCU_DATA_0
0xC6, 0xA3, 0x4F, 		// MCU_ADDRESS [AWB_CCM_POSITION_MIN]
0xC8, 0x00, 0x04, 		// MCU_DATA_0
0xC6, 0xA3, 0x50, 		// MCU_ADDRESS [AWB_CCM_POSITION_MAX]
0xC8, 0x00, 0x7F, 		// MCU_DATA_0
0xC6, 0xA3, 0x52, 		// MCU_ADDRESS [AWB_SATURATION]
0xC8, 0x00, 0x83, 		// MCU_DATA_0
0xC6, 0xA3, 0x53, 		// MCU_ADDRESS [AWB_MODE]
0xC8, 0x00, 0x00, 		// MCU_DATA_0
0xC6, 0xA3, 0x5B, 		// MCU_ADDRESS [AWB_STEADY_BGAIN_OUT_MIN]
0xC8, 0x00, 0x73, 		// MCU_DATA_0
0xC6, 0xA3, 0x5C, 		// MCU_ADDRESS [AWB_STEADY_BGAIN_OUT_MAX]
0xC8, 0x00, 0x8C, 		// MCU_DATA_0
0xC6, 0xA3, 0x5D, 		// MCU_ADDRESS [AWB_STEADY_BGAIN_IN_MIN]
0xC8, 0x00, 0x7C, 		// MCU_DATA_0
0xC6, 0xA3, 0x5E, 		// MCU_ADDRESS [AWB_STEADY_BGAIN_IN_MAX]
0xC8, 0x00, 0x83, 		// MCU_DATA_0
0xC6, 0x23, 0x5F, 		// MCU_ADDRESS [AWB_CNT_PXL_TH]
0xC8, 0x00, 0x64, 		// MCU_DATA_0
0xC6, 0xA3, 0x61, 		// MCU_ADDRESS [AWB_TG_MIN0]
0xC8, 0x00, 0xC0, 		// MCU_DATA_0
0xC6, 0xA3, 0x62, 		// MCU_ADDRESS [AWB_TG_MAX0]
0xC8, 0x00, 0xF4, 		// MCU_DATA_0
0xC6, 0xA3, 0x02, 		// MCU_ADDRESS [AWB_WINDOW_POS]
0xC8, 0x00, 0x00, 		// MCU_DATA_0
0xC6, 0xA3, 0x03, 		// MCU_ADDRESS [AWB_WINDOW_SIZE]
0xC8, 0x00, 0xEF, 		// MCU_DATA_0
0xC6, 0xA3, 0x63, 		// MCU_ADDRESS [AWB_X0]
0xC8, 0x00, 0x00, 		// MCU_DATA_0
0xC6, 0xA3, 0x64, 		// MCU_ADDRESS [AWB_KR_L]
0xC8, 0x00, 0x82, 		// MCU_DATA_0
0xC6, 0xA3, 0x65, 		// MCU_ADDRESS [AWB_KG_L]
0xC8, 0x00, 0x80, 		// MCU_DATA_0
0xC6, 0xA3, 0x66, 		// MCU_ADDRESS [AWB_KB_L]
0xC8, 0x00, 0x80, 		// MCU_DATA_0

// Histogram
0xC6, 0x22, 0xD3,
0xC8, 0x00, 0x00,		// HISTOGRAM_LOWER_BOUNDS
0xC6, 0x22, 0xD4,
0xC8, 0x54, 0xA0,		// HISTOGRAM_UPPER_BOUNDS
0xC6, 0x22, 0xD5,
0xC8, 0x02, 0x00,		// BIN_DEFINITIONS_1
0xC6, 0x22, 0xD6,
0xC8, 0x04, 0xC0,		// BIN_DEFINITIONS_2
0xC6, 0x22, 0xD7,
0xC8, 0x2D, 0x5D,		// HISTOGRAM_WINDOW_SIZE

//Gamma
0xC6, 0xA7, 0x45, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_0]
0xC8, 0x00, 0x00, 		// MCU_DATA_0
0xC6, 0xA7, 0x46, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_1]
0xC8, 0x00, 0x17, 		// MCU_DATA_0
0xC6, 0xA7, 0x47, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_2]
0xC8, 0x00, 0x28, 		// MCU_DATA_0
0xC6, 0xA7, 0x48, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_3]
0xC8, 0x00, 0x3D, 	// MCU_DATA_0
0xC6, 0xA7, 0x49, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_4]
0xC8, 0x00, 0x5B, 		// MCU_DATA_0
0xC6, 0xA7, 0x4A, 	// MCU_ADDRESS [MODE_GAM_TABLE_A_5]
0xC8, 0x00, 0x74, 		// MCU_DATA_0
0xC6, 0xA7, 0x4B, 	// MCU_ADDRESS [MODE_GAM_TABLE_A_6]
0xC8, 0x00, 0x89, 		// MCU_DATA_0
0xC6, 0xA7, 0x4C, 	// MCU_ADDRESS [MODE_GAM_TABLE_A_7]
0xC8, 0x00, 0x9B, 		// MCU_DATA_0
0xC6, 0xA7, 0x4D, 	// MCU_ADDRESS [MODE_GAM_TABLE_A_8]
0xC8, 0x00, 0xAA, 	// MCU_DATA_0
0xC6, 0xA7, 0x4E, 	// MCU_ADDRESS [MODE_GAM_TABLE_A_9]
0xC8, 0x00, 0xB7, 		// MCU_DATA_0
0xC6, 0xA7, 0x4F, 	// MCU_ADDRESS [MODE_GAM_TABLE_A_10]
0xC8, 0x00, 0xC3, 	// MCU_DATA_0
0xC6, 0xA7, 0x50, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_11]
0xC8, 0x00, 0xCD, 	// MCU_DATA_0
0xC6, 0xA7, 0x51, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_12]
0xC8, 0x00, 0xD6, 	// MCU_DATA_0
0xC6, 0xA7, 0x52, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_13]
0xC8, 0x00, 0xDE, 	// MCU_DATA_0
0xC6, 0xA7, 0x53, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_14]
0xC8, 0x00, 0xE6, 		// MCU_DATA_0
0xC6, 0xA7, 0x54, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_15]
0xC8, 0x00, 0xED, 	// MCU_DATA_0
0xC6, 0xA7, 0x55, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_16]
0xC8, 0x00, 0xF3, 		// MCU_DATA_0
0xC6, 0xA7, 0x56, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_17]
0xC8, 0x00, 0xF9, 		// MCU_DATA_0
0xC6, 0xA7, 0x57, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_18]
0xC8, 0x00, 0xFF, 		// MCU_DATA_0
0xC6, 0xA7, 0x43, 		// MCU_ADDRESS [MODE_GAM_CONT_A]
0xC8, 0x00, 0x03, 		// MCU_DATA_0

0xC6, 0x21, 0xB2, 
0xC8, 0x14, 0x00,		// GAMMA_CURVE_KNEES_0_1
0xC6, 0x21, 0xB3, 
0xC8, 0x3A, 0x22, 		// GAMMA_CURVE_KNEES_2_3
0xC6, 0x21, 0xB4, 
0xC8, 0x76, 0x5D, 	// GAMMA_CURVE_KNEES_4_5
0xC6, 0x21, 0xB5, 
0xC8, 0x96, 0x88, 		// GAMMA_CURVE_KNEES_6_7
0xC6, 0x21, 0xB6, 
0xC8, 0xAF, 0xA3, 		// GAMMA_CURVE_KNEES_8_9
0xC6, 0x21, 0xB7, 
0xC8, 0xC4, 0xBA, 	// GAMMA_CURVE_KNEES_10_11
0xC6, 0x21, 0xB8, 
0xC8, 0xD7, 0xCE, 	// GAMMA_CURVE_KNEES_12_13
0xC6, 0x21, 0xB9, 
0xC8, 0xE8, 0xE0, 	// GAMMA_CURVE_KNEES_14_15
0xC6, 0x21, 0xBA, 
0xC8, 0xF8, 0xF0, 		// GAMMA_CURVE_KNEES_16_17
0xC6, 0xA1, 0xBB, 
0xC8, 0x00, 0xFF,	 	// GAMMA_CURVE_KNEE_18

0xC6, 0x27, 0x72,		// fifo_conf0(b)
0xC8, 0x00, 0x67,		//      = 103
0xC6, 0x27, 0x74,		// fifo_conf1 (b)
0xC8, 0xE0, 0xE1,		//      = 57569
0xC6, 0xA7, 0x76,		// fifo_conf2 (b)
0xC8, 0x00, 0xE1,		//      = 225

0xC6, 0x27, 0x6D,		// fifo_conf1 (a)
0xC8, 0xE0, 0xE1,		//      = 57569
0xC6, 0xA7, 0x6F,		// fifo_conf2 (a)
0xC8, 0x00, 0xE1,		//      = 225

0xC6, 0xA7, 0x7D,		// mcu_address             
0xC8, 0x00, 0x02,		// (6) mode_output_format_a
0xC6, 0xA7, 0x7E,		// mcu_address               
0xC8, 0x00, 0x00,		// (6) mode_output_format_b  

0xFF, 0xFF, 0xFF, 0xFF , 0xFF, 0xFF
};

const TQP_UINT8 Sensor_YUY2_HD_init[] =
{
// Fmt : YUY2 (1280*720)
//------ HD ------------------------------------------------------------------
//1280x720 start
0xF0, 0x00, 0x00,
0x66, 0x1E, 0x05, 		// PLL_REG, 24M
0x67, 0x05, 0x00, 		// PLL2_REG
0x65, 0xA0, 0x00, 		// CLOCK_ENABLING
0x00, 0x00, 0x00,		0x00, 0x00, 0x01,     //Delay      
0x65, 0x20, 0x00, 		// CLOCK_ENABLING
//0x00, 0x00, 0x00,		0x00, 0x00, 0x42,     //Delay

0x05, 0x01, 0x1E, 		// HORZ_BLANK_B
0x06, 0x00, 0x0D, 		// VERT_BLANK_B
//0x20, 0x03, 0x03,		// Flip

0xF0, 0x00, 0x01,          
0xC6, 0x27, 0x0B, 		// MCU_ADDRESS [MODE_CONFIG]
0xC8, 0x00, 0x30, 		// MCU_DATA_0
0xC6, 0x27, 0x74, 		// MCU_ADDRESS [MODE_FIFO_CONF1_B]
0xC8, 0x05, 0x01, 		// MCU_DATA_0

0xC6, 0x27, 0x09, 		// MCU_ADDRESS [MODE_OUTPUT_HEIGHT_B]
0xC8, 0x02, 0xD8,		// MCU_DATA_0,2d0
0xC6, 0x27, 0x07, 		// MCU_ADDRESS [MODE_OUTPUT_WIDTH_B]
0xC8, 0x05, 0x02,		// MCU_DATA_0,504

0xC6, 0x27, 0x35, 		// MCU_ADDRESS [MODE_CROP_X0_B]
0xC8, 0x00, 0x00, 		// MCU_DATA_0
0xC6, 0x27, 0x37, 		// MCU_ADDRESS [MODE_CROP_X1_B]
0xC8, 0x05, 0x02,		// MCU_DATA_0
0xC6, 0x27, 0x39, 		// MCU_ADDRESS [MODE_CROP_Y0_B]
0xC8, 0x00, 0x00, 		// MCU_DATA_0
0xC6, 0x27, 0x3B, 		// MCU_ADDRESS [MODE_CROP_Y1_B]
0xC8, 0x02, 0xD4,		// MCU_DATA_0,2d4

0xC6, 0x27, 0x1B, 		// MCU_ADDRESS [MODE_SENSOR_ROW_START_B]
0xC8, 0x00, 0x1C, 		// MCU_DATA_0
0xC6, 0x27, 0x1D, 		// MCU_ADDRESS [MODE_SENSOR_COL_START_B]
0xC8, 0x00, 0x3C, 		// MCU_DATA_0
0xC6, 0x27, 0x1F, 		// MCU_ADDRESS [MODE_SENSOR_ROW_HEIGHT_B]
0xC8, 0x02, 0xD4,		// MCU_DATA_0,2d4
0xC6, 0x27, 0x21, 		// MCU_ADDRESS [MODE_SENSOR_COL_WIDTH_B]
0xC8, 0x05, 0x02,		// MCU_DATA_0,504
0xC6, 0x27, 0x1B, 		// MCU_ADDRESS [MODE_SENSOR_ROW_START_B]
0xC8, 0x00, 0xFE, 		// MCU_DATA_0
0xC6, 0x27, 0x1D, 		// MCU_ADDRESS [MODE_SENSOR_COL_START_B]
0xC8, 0x00, 0xBE, 		// MCU_DATA_0

0xC6, 0xA1, 0x20, 		// MCU_ADDRESS [SEQ_CAP_MODE]
0xC8, 0x00, 0x72, 		// MCU_DATA_0
0x00, 0x00, 0x00,		0x00, 0x02, 0x00,//Delay
0xC6, 0xA1, 0x03, 		// MCU_ADDRESS [SEQ_CMD]
0xC8, 0x00, 0x02, 		// MCU_DATA_0

0xC6, 0xA1, 0x20, 		// MCU_ADDRESS [SEQ_CAP_MODE]
0xC8, 0x00, 0x72, 		// MCU_DATA_0
0xC6, 0xA1, 0x03, 		// MCU_ADDRESS [SEQ_CMD]
0xC8, 0x00, 0x05, 		// MCU_DATA_0

0xC6, 0xA2, 0x0E, 	    // MCU_ADDRESS
0xC8, 0x00, 0x10, 		// MCU_DATA_0 7.5fps 16/120

0xC6, 0x22, 0x0B,		// Max R12 (B)(Shutter Delay)
0xC8, 0x04, 0x5B,		//      = 1115
0xC6, 0xA2, 0x17,       // IndexTH23
0xC8, 0x00, 0x08,		//      = 8
0xC6, 0x22, 0x28,		//RowTime (msclk per)/4
0xC8, 0x01, 0x88,		//      = 392
0xC6, 0x22, 0x2F,		//R9 Step
0xC8, 0x00, 0xD3,		//      = 211
0xC6, 0xA4, 0x08,		//search_f1_50
0xC8, 0x00, 0x31,		//      = 49
0xC6, 0xA4, 0x09,		//search_f2_50
0xC8, 0x00, 0x33,		//      = 51
0xC6, 0xA4, 0x0A,		//search_f1_60
0xC8, 0x00, 0x29,		//      = 41
0xC6, 0xA4, 0x0B,		//search_f2_60
0xC8, 0x00, 0x2B,		//      = 43
0xC6, 0x24, 0x11,		//R9_Step_60
0xC8, 0x00, 0xD3,		//      = 211
0xC6, 0x24, 0x13,		//R9_Step_50
0xC8, 0x00, 0xFD,		//      = 253

0xC6, 0xA1, 0x03, 		// MCU_ADDRESS
0xC8, 0x00, 0x05, 		// MCU_DATA_0

0xC6, 0xA1, 0x20, 		// MCU_ADDRESS [SEQ_CAP_MODE]
0xC8, 0x00, 0x72, 		// MCU_DATA_0
0x00, 0x00, 0x00,		0x00, 0x00, 0x44,	//Delay
0xC6, 0xA1, 0x03, 		// MCU_ADDRESS [SEQ_CMD]
0xC8, 0x00, 0x02, 		// MCU_DATA_0
//1280x720 end
0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};



const TQP_UINT8 Sensor_INIT_COMMON_table_30fps[] = 
{
// [COMMON] //////////////////////////////////////////////////
0xF0, 0x00, 0x00,
0x65, 0xA0, 0x00,		// bypassed PLL (prepare for soft reset)

0xF0, 0x00, 0x01,		
0xC3, 0x05, 0x01,		// MCU_BOOT_MODE (MCU reset)

0xF0, 0x00, 0x00,		
0x0D, 0x00, 0x21,		// RESET_REG (enable soft reset)
0x00, 0x00, 0x00, 0x00, 0x00, 0x01,  //Delay
                 		
0x0D, 0x00, 0x00,		// RESET_REG (disable soft reset)
0x00, 0x00, 0x00, 0x00, 0x00, 0x64,  //Delay

0x2B, 0x00, 0x20,		// Doesn't do anything, used as a placeholder
0x25, 0x40, 0x00,		// Line valid Control

// [Core]
0x01, 0x00, 0x1C, 		// ROW_WINDOW_START_REG
0x02, 0x00, 0x3C, 		// COL_WINDOW_START_REG
0x03, 0x04, 0xB4,		// ROW_WINDOW_SIZE_REG 
0x04, 0x06, 0x42, 		// COL_WINDOW_SIZE_REG 

0xF0, 0x00, 0x01,
0xC6, 0xA1, 0x02,
0xC8, 0x00, 0x00,

// Black Level and Digital Gain


// Color Matrix/Color Correction
0xC6, 0x21, 0x60,
0xC8, 0x39, 0x23, 		// COLOR_CORR_MATRIX_SCALE_14
0xC6, 0x21, 0x61,
0xC8, 0x04, 0xE4, 		// COLOR_CORR_MATRIX_SCALE_11
0xC6, 0x21, 0x62,
0xC8, 0xE2, 0xFB,		// COLOR_CORR_MATRIX_1_2
0xC6, 0x21, 0x63,
0xC8, 0x90, 0x0B, 		// COLOR_CORR_MATRIX_3_4
0xC6, 0x21, 0x64,
0xC8, 0x3B, 0xEA, 		// COLOR_CORR_MATRIX_5_6
0xC6, 0x21, 0x65,
0xC8, 0xCD, 0x6B, 		// COLOR_CORR_MATRIX_7_8
0xC6, 0x21, 0x66,
0xC8, 0x3F, 0xC3, 		// COLOR_CORR_MATRIX_9

// Lens correction
0xF0, 0x00, 0x02,
0x80, 0x01, 0xA8,		//lens_correction_control
0x81, 0x6A, 0x35,		//zone_bounds_x1_x2
0x82, 0x2F, 0x9F,		//zone_bounds_x0_x3
0x83, 0x8D, 0x5E,		//zone_bounds_x4_x5
0x84, 0x51, 0x28,		//zone_bounds_y1_y2
0x85, 0x23, 0x79,		//zone_bounds_y0_y3
0x86, 0x68, 0x45,		//zone_bounds_y4_y5
0x87, 0x0C, 0x0C,		//center_offset
0x88, 0x01, 0xC7,		//fx_red
0x8B, 0x01, 0x46,		//fy_red
0x8E, 0x0A, 0x3E,		//df_dx_red
0x91, 0x08, 0xDE,		//df_dy_red
0x94, 0x27, 0xD8,		//second_deriv_zone_0_red
0x97, 0x11, 0x46,		//second_deriv_zone_1_red
0x9A, 0x56, 0x48,		//second_deriv_zone_2_red
0x9D, 0x14, 0x6A,		//second_deriv_zone_3_red
0xA0, 0x3B, 0x46,		//second_deriv_zone_4_red
0xA3, 0x1B, 0x6C,		//second_deriv_zone_5_red
0xA6, 0x5B, 0x45,		//second_deriv_zone_6_red
0xA9, 0x0D, 0xCC,		//second_deriv_zone_7_red
0x89, 0x01, 0x84,		//fx_green
0x8C, 0x01, 0x22,		//fy_green
0x8F, 0x0C, 0x6B,		//df_dx_green
0x92, 0x0A, 0x67,		//df_dy_green
0x95, 0x0C, 0xBD,		//second_deriv_zone_0_green
0x98, 0x10, 0x41,		//second_deriv_zone_1_green
0x9B, 0x50, 0x3C,		//second_deriv_zone_2_green
0x9E, 0x0F, 0x5B,		//second_deriv_zone_3_green
0xA1, 0x38, 0x41,		//second_deriv_zone_4_green
0xA4, 0x16, 0x5B,		//second_deriv_zone_5_green
0xA7, 0x51, 0x49,		//second_deriv_zone_6_green
0xAA, 0xD9, 0xA6,		//second_deriv_zone_7_green
0x8A, 0x01, 0x62,		//fx_blue
0x8D, 0x00, 0xFD,		//fy_blue
0x90, 0x0B, 0xF9,		//df_dx_blue
0x93, 0x0A, 0x44,		//df_dy_blue
0x96, 0x21, 0xD5,		//second_deriv_zone_0_blue
0x99, 0x0F, 0x3B,		//second_deriv_zone_1_blue
0x9C, 0x47, 0x32,		//second_deriv_zone_2_blue
0x9F, 0x0F, 0x51,		//second_deriv_zone_3_blue
0xA2, 0x2E, 0x3C,		//second_deriv_zone_4_blue
0xA5, 0x14, 0x48,		//second_deriv_zone_5_blue
0xA8, 0x45, 0x36,		//second_deriv_zone_6_blue
0xAB, 0xFD, 0xD4,		//second_deriv_zone_7_blue
0xAC, 0x00, 0x83,		//x2_factors
0xAD, 0x00, 0x00,		//global_offset_fxy_function
0xAE, 0x00, 0x00,		//k_factor_in_k_fx_fy

0xF0, 0x00, 0x01,
0x08, 0x01, 0xFC,

// Defect Correction

// Auto Focus Statistics

//AE & AWB Setting
0xC6, 0xA1, 0x37, 		// MCU_ADDRESS [SEQ_PREVIEW_3_AE]
0xC8, 0x00, 0x03, 		// MCU_DATA_0
0xC6, 0xA1, 0x39, 		// MCU_ADDRESS [SEQ_PREVIEW_3_AWB]
0xC8, 0x00, 0x03, 		// MCU_DATA_0
0xC6, 0xA1, 0x38, 		// MCU_ADDRESS [SEQ_PREVIEW_3_FD]
0xC8, 0x00, 0x01, 		// MCU_DATA_0
0xC6, 0xA1, 0x3B,		// MCU_ADDRESS [SEQ_PREVIEW_3_HG]
0xC8, 0x00, 0x03, 		// MCU_DATA_0

// Defect Correction

// Auto Focus Statistics

// AE
0xC6, 0xA2, 0x06, 		// MCU_ADDRESS [AE_TARGET]
0xC8, 0x00, 0x48, 		// MCU_DATA_0
0xC6, 0xA2, 0x07, 		// MCU_ADDRESS [AE_GATE]
0xC8, 0x00, 0x04, 		// MCU_DATA_0
0xC6, 0x22, 0x14,		// MCU_ADDRESS [AE_MAX_DGAIN_AE1]
0xC8, 0x00, 0x96,		// MCU_DATA_0
0xC6, 0xA2, 0x16,		// MCU_ADDRESS [AE_MAX_DGAIN_AE2]
0xC8, 0x00, 0x32,		// MCU_DATA_0

// AWB
0xC6, 0xA2, 0x06, 		// MCU_ADDRESS [AE_TARGET]
0xC8, 0x00, 0x48, 		// MCU_DATA_0
0xC6, 0xA2, 0x07, 		// MCU_ADDRESS [AE_GATE]
0xC8, 0x00, 0x04, 		// MCU_DATA_0
0xC6, 0x23, 0x06, 		// MCU_ADDRESS [AWB_CCM_L_0]
0xC8, 0x02, 0x34, 		// MCU_DATA_0
0xC6, 0x23, 0x08, 		// MCU_ADDRESS [AWB_CCM_L_1]
0xC8, 0xFE, 0xBC, 		// MCU_DATA_0
0xC6, 0x23, 0x0A, 		// MCU_ADDRESS [AWB_CCM_L_2]
0xC8, 0x00, 0x14, 		// MCU_DATA_0
0xC6, 0x23, 0x0C, 		// MCU_ADDRESS [AWB_CCM_L_3]
0xC8, 0xFF, 0xCB, 		// MCU_DATA_0
0xC6, 0x23, 0x0E, 		// MCU_ADDRESS [AWB_CCM_L_4]
0xC8, 0x01, 0x7C, 		// MCU_DATA_0
0xC6, 0x23, 0x10, 		// MCU_ADDRESS [AWB_CCM_L_5]
0xC8, 0xFF, 0xC1, 		// MCU_DATA_0
0xC6, 0x23, 0x12, 		// MCU_ADDRESS [AWB_CCM_L_6]
0xC8, 0xFF, 0x65, 		// MCU_DATA_0
0xC6, 0x23, 0x14, 		// MCU_ADDRESS [AWB_CCM_L_7]
0xC8, 0xFD, 0x81, 		// MCU_DATA_0
0xC6, 0x23, 0x16, 		// MCU_ADDRESS [AWB_CCM_L_8]
0xC8, 0x04, 0x4D, 		// MCU_DATA_0
0xC6, 0x23, 0x18, 		// MCU_ADDRESS [AWB_CCM_L_9]
0xC8, 0x00, 0x20, 		// MCU_DATA_0
0xC6, 0x23, 0x1A, 		// MCU_ADDRESS [AWB_CCM_L_10]
0xC8, 0x00, 0x35, 		// MCU_DATA_0
0xC6, 0x23, 0x1C, 		// MCU_ADDRESS [AWB_CCM_RL_0]
0xC8, 0xFF, 0x52, 		// MCU_DATA_0
0xC6, 0x23, 0x1E, 		// MCU_ADDRESS [AWB_CCM_RL_1]
0xC8, 0x00, 0xAA, 		// MCU_DATA_0
0xC6, 0x23, 0x20, 		// MCU_ADDRESS [AWB_CCM_RL_2]
0xC8, 0x00, 0x10, 		// MCU_DATA_0
0xC6, 0x23, 0x22, 		// MCU_ADDRESS [AWB_CCM_RL_3]
0xC8, 0xFF, 0xD5, 		// MCU_DATA_0
0xC6, 0x23, 0x24, 		// MCU_ADDRESS [AWB_CCM_RL_4]
0xC8, 0x00, 0x39, 		// MCU_DATA_0
0xC6, 0x23, 0x26, 		// MCU_ADDRESS [AWB_CCM_RL_5]
0xC8, 0xFF, 0xF2, 		// MCU_DATA_0
0xC6, 0x23, 0x28, 		// MCU_ADDRESS [AWB_CCM_RL_6]
0xC8, 0x00, 0x66, 		// MCU_DATA_0
0xC6, 0x23, 0x2A, 		// MCU_ADDRESS [AWB_CCM_RL_7]
0xC8, 0x01, 0xCB, 		// MCU_DATA_0
0xC6, 0x23, 0x2C, 		// MCU_ADDRESS [AWB_CCM_RL_8]
0xC8, 0xFD, 0xA1, 		// MCU_DATA_0
0xC6, 0x23, 0x2E, 		// MCU_ADDRESS [AWB_CCM_RL_9]
0xC8, 0x00, 0x0F, 		// MCU_DATA_0
0xC6, 0x23, 0x30, 		// MCU_ADDRESS [AWB_CCM_RL_10]
0xC8, 0xFF, 0xFD, 		// MCU_DATA_0
0xC6, 0xA3, 0x48, 		// MCU_ADDRESS [AWB_GAIN_BUFFER_SPEED]
0xC8, 0x00, 0x08, 		// MCU_DATA_0
0xC6, 0xA3, 0x49, 		// MCU_ADDRESS [AWB_JUMP_DIVISOR]
0xC8, 0x00, 0x02, 		// MCU_DATA_0
0xC6, 0xA3, 0x4A, 		// MCU_ADDRESS [AWB_GAIN_MIN]
0xC8, 0x00, 0x53, 		// MCU_DATA_0
0xC6, 0xA3, 0x4B, 		// MCU_ADDRESS [AWB_GAIN_MAX]
0xC8, 0x00, 0xC0, 		// MCU_DATA_0
0xC6, 0xA3, 0x4F, 		// MCU_ADDRESS [AWB_CCM_POSITION_MIN]
0xC8, 0x00, 0x04, 		// MCU_DATA_0
0xC6, 0xA3, 0x50, 		// MCU_ADDRESS [AWB_CCM_POSITION_MAX]
0xC8, 0x00, 0x7F, 		// MCU_DATA_0
0xC6, 0xA3, 0x52, 		// MCU_ADDRESS [AWB_SATURATION]
0xC8, 0x00, 0x83, 		// MCU_DATA_0
0xC6, 0xA3, 0x53, 		// MCU_ADDRESS [AWB_MODE]
0xC8, 0x00, 0x00, 		// MCU_DATA_0
0xC6, 0xA3, 0x5B, 		// MCU_ADDRESS [AWB_STEADY_BGAIN_OUT_MIN]
0xC8, 0x00, 0x73, 		// MCU_DATA_0
0xC6, 0xA3, 0x5C, 		// MCU_ADDRESS [AWB_STEADY_BGAIN_OUT_MAX]
0xC8, 0x00, 0x8C, 		// MCU_DATA_0
0xC6, 0xA3, 0x5D, 		// MCU_ADDRESS [AWB_STEADY_BGAIN_IN_MIN]
0xC8, 0x00, 0x7C, 		// MCU_DATA_0
0xC6, 0xA3, 0x5E, 		// MCU_ADDRESS [AWB_STEADY_BGAIN_IN_MAX]
0xC8, 0x00, 0x83, 		// MCU_DATA_0
0xC6, 0x23, 0x5F, 		// MCU_ADDRESS [AWB_CNT_PXL_TH]
0xC8, 0x00, 0x64, 		// MCU_DATA_0
0xC6, 0xA3, 0x61, 		// MCU_ADDRESS [AWB_TG_MIN0]
0xC8, 0x00, 0xC0, 		// MCU_DATA_0
0xC6, 0xA3, 0x62, 		// MCU_ADDRESS [AWB_TG_MAX0]
0xC8, 0x00, 0xF4, 		// MCU_DATA_0
0xC6, 0xA3, 0x02, 		// MCU_ADDRESS [AWB_WINDOW_POS]
0xC8, 0x00, 0x00, 		// MCU_DATA_0
0xC6, 0xA3, 0x03, 		// MCU_ADDRESS [AWB_WINDOW_SIZE]
0xC8, 0x00, 0xEF, 		// MCU_DATA_0
0xC6, 0xA3, 0x63, 		// MCU_ADDRESS [AWB_X0]
0xC8, 0x00, 0x00, 		// MCU_DATA_0
0xC6, 0xA3, 0x64, 		// MCU_ADDRESS [AWB_KR_L]
0xC8, 0x00, 0x82, 		// MCU_DATA_0
0xC6, 0xA3, 0x65, 		// MCU_ADDRESS [AWB_KG_L]
0xC8, 0x00, 0x80, 		// MCU_DATA_0
0xC6, 0xA3, 0x66, 		// MCU_ADDRESS [AWB_KB_L]
0xC8, 0x00, 0x80, 		// MCU_DATA_0

// Histogram
0xC6, 0x22, 0xD3,
0xC8, 0x00, 0x00,		// HISTOGRAM_LOWER_BOUNDS
0xC6, 0x22, 0xD4,
0xC8, 0x54, 0xA0,		// HISTOGRAM_UPPER_BOUNDS
0xC6, 0x22, 0xD5,
0xC8, 0x02, 0x00,		// BIN_DEFINITIONS_1
0xC6, 0x22, 0xD6,
0xC8, 0x04, 0xC0,		// BIN_DEFINITIONS_2
0xC6, 0x22, 0xD7,
0xC8, 0x2D, 0x5D,		// HISTOGRAM_WINDOW_SIZE

//Gamma
0xC6, 0xA7, 0x45, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_0]
0xC8, 0x00, 0x00, 		// MCU_DATA_0
0xC6, 0xA7, 0x46, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_1]
0xC8, 0x00, 0x17, 		// MCU_DATA_0
0xC6, 0xA7, 0x47, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_2]
0xC8, 0x00, 0x28, 		// MCU_DATA_0
0xC6, 0xA7, 0x48, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_3]
0xC8, 0x00, 0x3D, 		// MCU_DATA_0
0xC6, 0xA7, 0x49, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_4]
0xC8, 0x00, 0x5B, 		// MCU_DATA_0
0xC6, 0xA7, 0x4A, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_5]
0xC8, 0x00, 0x74, 		// MCU_DATA_0
0xC6, 0xA7, 0x4B, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_6]
0xC8, 0x00, 0x89, 		// MCU_DATA_0
0xC6, 0xA7, 0x4C, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_7]
0xC8, 0x00, 0x9B, 		// MCU_DATA_0
0xC6, 0xA7, 0x4D, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_8]
0xC8, 0x00, 0xAA, 		// MCU_DATA_0
0xC6, 0xA7, 0x4E, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_9]
0xC8, 0x00, 0xB7, 		// MCU_DATA_0
0xC6, 0xA7, 0x4F, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_10]
0xC8, 0x00, 0xC3, 		// MCU_DATA_0
0xC6, 0xA7, 0x50, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_11]
0xC8, 0x00, 0xCD, 		// MCU_DATA_0
0xC6, 0xA7, 0x51, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_12]
0xC8, 0x00, 0xD6, 		// MCU_DATA_0
0xC6, 0xA7, 0x52, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_13]
0xC8, 0x00, 0xDE, 		// MCU_DATA_0
0xC6, 0xA7, 0x53, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_14]
0xC8, 0x00, 0xE6, 		// MCU_DATA_0
0xC6, 0xA7, 0x54, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_15]
0xC8, 0x00, 0xED, 		// MCU_DATA_0
0xC6, 0xA7, 0x55, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_16]
0xC8, 0x00, 0xF3, 		// MCU_DATA_0
0xC6, 0xA7, 0x56, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_17]
0xC8, 0x00, 0xF9, 		// MCU_DATA_0
0xC6, 0xA7, 0x57, 		// MCU_ADDRESS [MODE_GAM_TABLE_A_18]
0xC8, 0x00, 0xFF, 		// MCU_DATA_0
0xC6, 0xA7, 0x43, 		// MCU_ADDRESS [MODE_GAM_CONT_A]
0xC8, 0x00, 0x03, 		// MCU_DATA_0

0xC6, 0x21, 0xB2, 
0xC8, 0x14, 0x00,		// GAMMA_CURVE_KNEES_0_1
0xC6, 0x21, 0xB3, 
0xC8, 0x3A, 0x22, 		// GAMMA_CURVE_KNEES_2_3
0xC6, 0x21, 0xB4, 
0xC8, 0x76, 0x5D, 		// GAMMA_CURVE_KNEES_4_5
0xC6, 0x21, 0xB5, 
0xC8, 0x96, 0x88, 		// GAMMA_CURVE_KNEES_6_7
0xC6, 0x21, 0xB6, 
0xC8, 0xAF, 0xA3, 		// GAMMA_CURVE_KNEES_8_9
0xC6, 0x21, 0xB7, 
0xC8, 0xC4, 0xBA, 		// GAMMA_CURVE_KNEES_10_11
0xC6, 0x21, 0xB8, 
0xC8, 0xD7, 0xCE, 		// GAMMA_CURVE_KNEES_12_13
0xC6, 0x21, 0xB9, 
0xC8, 0xE8, 0xE0, 		// GAMMA_CURVE_KNEES_14_15
0xC6, 0x21, 0xBA, 
0xC8, 0xF8, 0xF0, 		// GAMMA_CURVE_KNEES_16_17
0xC6, 0xA1, 0xBB, 
0xC8, 0x00, 0xFF,	 	// GAMMA_CURVE_KNEE_18

0xC6, 0x27, 0x72,		// fifo_conf0(b)
0xC8, 0x00, 0x67,		//      = 103
0xC6, 0x27, 0x74,		// fifo_conf1 (b)
0xC8, 0xE0, 0xE1,		//      = 57569
0xC6, 0xA7, 0x76,		// fifo_conf2 (b)
0xC8, 0x00, 0xE1,		//      = 225

0xC6, 0x27, 0x6D,		// fifo_conf1 (a)
0xC8, 0xE0, 0xE1,		//      = 57569
0xC6, 0xA7, 0x6F,		// fifo_conf2 (a)
0xC8, 0x00, 0xE1,		//      = 225

0xC6, 0xA7, 0x7D,		// mcu_address             
0xC8, 0x00, 0x02,		// (6) mode_output_format_a
0xC6, 0xA7, 0x7E,		// mcu_address               
0xC8, 0x00, 0x00,		// (6) mode_output_format_b  

0xFF, 0xFF, 0xFF, 0xFF , 0xFF, 0xFF
};

const TQP_UINT8 Sensor_YUY2_HD_init_30fps[] =
{
// Fmt : MJPEG (1280*720)
//------ HD ------------------------------------------------------------------
//1280x720 start
0xF0, 0x00, 0x00,
0x66, 0x23, 0x05, 		// PLL_REG, 24M
0x67, 0x05, 0x00, 		// PLL2_REG
0x65, 0xA0, 0x00, 		// CLOCK_ENABLING
0x00, 0x00, 0x00,		0x00, 0x00, 0x01,           
0x65, 0x20, 0x00, 		// CLOCK_ENABLING
//0x00, 0x00, 0x00,		0x00, 0x00, 0x42, 

0x05, 0x01, 0x1E, 		// HORZ_BLANK_B
0x06, 0x00, 0x0D, 		// VERT_BLANK_B
//0x20, 0x03, 0x03,		// Flip

0xF0, 0x00, 0x01,          
0xC6, 0x27, 0x0B, 		// MCU_ADDRESS [MODE_CONFIG]
0xC8, 0x00, 0x30, 		// MCU_DATA_0
0xC6, 0x27, 0x74, 		// MCU_ADDRESS [MODE_FIFO_CONF1_B]
0xC8, 0x05, 0x01, 		// MCU_DATA_0

0xC6, 0x27, 0x09, 		// MCU_ADDRESS [MODE_OUTPUT_HEIGHT_B]
0xC8, 0x02, 0xD8,		// MCU_DATA_0,2d0
0xC6, 0x27, 0x07, 		// MCU_ADDRESS [MODE_OUTPUT_WIDTH_B]
0xC8, 0x05, 0x02,		// MCU_DATA_0,504

0xC6, 0x27, 0x35, 		// MCU_ADDRESS [MODE_CROP_X0_B]
0xC8, 0x00, 0x00, 		// MCU_DATA_0
0xC6, 0x27, 0x37, 		// MCU_ADDRESS [MODE_CROP_X1_B]
0xC8, 0x05, 0x02,		// MCU_DATA_0
0xC6, 0x27, 0x39, 		// MCU_ADDRESS [MODE_CROP_Y0_B]
0xC8, 0x00, 0x00, 		// MCU_DATA_0
0xC6, 0x27, 0x3B, 		// MCU_ADDRESS [MODE_CROP_Y1_B]
0xC8, 0x02, 0xD4,		// MCU_DATA_0,2d4

0xC6, 0x27, 0x1B, 		// MCU_ADDRESS [MODE_SENSOR_ROW_START_B]
0xC8, 0x00, 0x1C, 		// MCU_DATA_0
0xC6, 0x27, 0x1D, 		// MCU_ADDRESS [MODE_SENSOR_COL_START_B]
0xC8, 0x00, 0x3C, 		// MCU_DATA_0
0xC6, 0x27, 0x1F, 		// MCU_ADDRESS [MODE_SENSOR_ROW_HEIGHT_B]
0xC8, 0x02, 0xD4,		// MCU_DATA_0,2d4
0xC6, 0x27, 0x21, 		// MCU_ADDRESS [MODE_SENSOR_COL_WIDTH_B]
0xC8, 0x05, 0x02,		// MCU_DATA_0,504
0xC6, 0x27, 0x1B, 		// MCU_ADDRESS [MODE_SENSOR_ROW_START_B]
0xC8, 0x00, 0xFE, 		// MCU_DATA_0
0xC6, 0x27, 0x1D, 		// MCU_ADDRESS [MODE_SENSOR_COL_START_B]
0xC8, 0x00, 0xBE, 		// MCU_DATA_0

0xC6, 0xA1, 0x20, 		// MCU_ADDRESS [SEQ_CAP_MODE]
0xC8, 0x00, 0x72, 		// MCU_DATA_0
0x00, 0x00, 0x00,		0x00, 0x02, 0x00,//Da
0xC6, 0xA1, 0x03, 		// MCU_ADDRESS [SEQ_CMD]
0xC8, 0x00, 0x02, 		// MCU_DATA_0

0xC6, 0xA1, 0x20, 		// MCU_ADDRESS [SEQ_CAP_MODE]
0xC8, 0x00, 0x72, 		// MCU_DATA_0
0xC6, 0xA1, 0x03, 		// MCU_ADDRESS [SEQ_CMD]
0xC8, 0x00, 0x05, 		// MCU_DATA_0

//0xC6, 0xA2, 0x17, 		// MCU_ADDRESS (HD)
//0xC8, 0x00, 0x04, 		// MCU_DATA_0
0xC6, 0xA2, 0x0E, 		// MCU_ADDRESS
0xC8, 0x00, 0x10, 		// MCU_DATA_0 7.5fps 16/120

0xC6, 0x22, 0x0B,		// Max R12 (B)(Shutter Delay)
0xC8, 0x04, 0x5B,		//      = 1115
0xC6, 0xA2, 0x17,		// IndexTH23
0xC8, 0x00, 0x08,		//      = 8
0xC6, 0x22, 0x28,		//RowTime (msclk per)/4
0xC8, 0x01, 0x88,		//      = 392
0xC6, 0x22, 0x2F,		//R9 Step
0xC8, 0x00, 0xD3,		//      = 211
0xC6, 0xA4, 0x08,		//search_f1_50
0xC8, 0x00, 0x31,		//      = 49
0xC6, 0xA4, 0x09,		//search_f2_50
0xC8, 0x00, 0x33,		//      = 51
0xC6, 0xA4, 0x0A,		//search_f1_60
0xC8, 0x00, 0x29,		//      = 41
0xC6, 0xA4, 0x0B,		//search_f2_60
0xC8, 0x00, 0x2B,		//      = 43
0xC6, 0x24, 0x11,		//R9_Step_60
0xC8, 0x00, 0xD3,		//      = 211
0xC6, 0x24, 0x13,		//R9_Step_50
0xC8, 0x00, 0xFD,		//      = 253

0xF0, 0x00, 0x01,
0xC6, 0xA2, 0x17, 		// MCU_ADDRESS [IndexTH23]
0xC8, 0x00, 0x03, 		// MCU_DATA_0 40fps  3/120
0xC6, 0xA2, 0x0E, 		// MCU_ADDRESS [maxIndex]
0xC8, 0x00, 0x04,		// MCU_DATA_0 30fps = 4/120
0xC6, 0xA2, 0x10,		// MCU_ADDRESS [AE_MAX_VIRT_GAIN]
0xC8, 0x00, 0x80,		// MCU_DATA_0
0xC6, 0xA2, 0x18,		// MCU_ADDRESS [AE_MAX_GAIN23]
0xC8, 0x00, 0x28,		// MCU_DATA_0

0xC6, 0xA1, 0x02,
0xC8, 0x00, 0x0E,
0xC6, 0x22, 0x25,
0xC8, 0x00, 0x10,

//0x00, 0x00, 0x00,		0x00, 0x00, 0x22,
0xC6, 0xA1, 0x03, 		// MCU_ADDRESS
0xC8, 0x00, 0x05, 		// MCU_DATA_0

0xC6, 0xA1, 0x20, 		// MCU_ADDRESS [SEQ_CAP_MODE]
0xC8, 0x00, 0x72, 		// MCU_DATA_0
0x00, 0x00, 0x00,		0x00, 0x00, 0x22,	//Delay
0xC6, 0xA1, 0x03, 		// MCU_ADDRESS [SEQ_CMD]
0xC8, 0x00, 0x02, 		// MCU_DATA_0
//1280x720 end

0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};




/////////////////////////////////////////////////////////////////////////////
//
//


static EQP_Errors CMi2010_readRegister(CMi2010 *This,
									   TQP_UINT8 regAddr, 
									   PTQP_UINT16 pValue
									   )
{
	EQP_Errors	qpStatus;
    TQP_UINT8	buffer[2];
	TQP_UINT16	value;

    qpStatus = II2C_WriteThenRead(This->m_pI2C,
								  This->m_cI2CDevAddr,	//device address
								  1,					//write size
								  &regAddr,				//write buffer
								  2,					//read size
								  buffer				//read buffer
								  );
	if (QP_SUCCESS(qpStatus))
	{
		value = ((TQP_UINT16)buffer[0] & 0xFF);
		value = (value << 8) + buffer[1];
		*pValue = value;
	}
	return (qpStatus);
}



static EQP_Errors CMi2010_writeRegister(CMi2010 *This,
										TQP_UINT8 regAddr, 
										TQP_UINT16 value
										)
{
    TQP_UINT8	buffer[3];

	buffer[0] = regAddr;
	buffer[1] = (TQP_UINT8)((value >> 8) & 0xFF);
	buffer[2] = (TQP_UINT8)(value & 0xFF);

    return (II2C_Write(This->m_pI2C, 
					   This->m_cI2CDevAddr,	//device address
					   3,					//write size
					   buffer				//write buffer
					   ));
}


static EQP_Errors CMi2010_WriteBlock(CMi2010 *This,
									 TQP_UINT8 *pbAray
									 )
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;
	TQP_UINT32	uwAdr = 0;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_WARN),
				 "CMi2010_WriteBlock() (0x%x)\n",
				 pbAray
				 ));


	uwAdr = 0x0000;

	while (QP_TRUE)
	{
		if ((pbAray[uwAdr] & 
			pbAray[uwAdr + 1] & 
			pbAray[uwAdr + 2] & 
			pbAray[uwAdr + 3] &
			pbAray[uwAdr + 4] & 
			pbAray[uwAdr + 5]) == 0xFF
			)
		{
			// end of the array block
			//
			break;
		}
		else if ((pbAray[uwAdr] |
				 pbAray[uwAdr + 1] |
				 pbAray[uwAdr + 2] |
				 pbAray[uwAdr + 3]) == 0x00
				 )
		{
			// Delay
			//
			DelayMilliSeconds((((TQP_UINT16)pbAray[uwAdr + 4] << 8) & 0xFF00) + pbAray[uwAdr + 5]);
			uwAdr += 6;
			continue;
		}

		qpStatus = CMi2010_writeRegister(This, 
										 pbAray[uwAdr], 
										 (((TQP_UINT16)pbAray[uwAdr + 1] << 8) & 0xFF00) + pbAray[uwAdr + 2]
										 );
		if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CMi2010_WriteBlock() ****** writeRegister failed! status(%d) ******\n",
						 qpStatus
						 ));
			break;
		}
		uwAdr += 3;
	};
	return (qpStatus);
}



/////////////////////////////////////////////////////////////////////////////
//
//


// IGenericDevice
//
static EQP_Errors CMi2010_InitDevice(IVideoDecoder *This_p, 
									 QP_DEVICE_CALLBACK pFuncCallback,
									 TQP_PVOID context,
									 TQP_PVOID pInitData
									 )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMi2010_InitDevice()\n"
				 ));

	if (CObject_IsInitialized(&pVideoDecoder->m_Object))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
					 "CMi2010_InitDevice() Already Initialized. Skipping...\n" 
					 ));
		return (QPERR_SUCCESS);
	}

	// save callback address and context
	//
	pVideoDecoder->m_pDeviceCallback = pFuncCallback;
	pVideoDecoder->m_callbackContext = context;


	// Initialize MI2010 with the default value
	//
	CMi2010_WriteBlock(This,
					   (TQP_UINT8 *)Sensor_INIT_COMMON_table_30fps
					   );
	CMi2010_WriteBlock(This,
					   (TQP_UINT8 *)Sensor_YUY2_HD_init_30fps
					   );

	// program the sensor with settings read from the registry
	//
	QPVideoDecoder_SetVideoSystem(This_p, 
								  pVideoDecoder->m_VidStandard
								  );
	QPVideoDecoder_SelectVideoSource(This_p, 
									 pVideoDecoder->m_VidSource
									 );

	QPVideoDecoder_SetVcrTiming(This_p, 
								pVideoDecoder->m_VCRTiming
								);

	QPVideoDecoder_SetBrightness(This_p, 
								 pVideoDecoder->m_nBrightness
								 );
	QPVideoDecoder_SetContrast(This_p, 
							   pVideoDecoder->m_nContrast
							   );
	QPVideoDecoder_SetSaturation(This_p, 
								 pVideoDecoder->m_nSaturation
								 );
	QPVideoDecoder_SetHue(This_p, 
						  pVideoDecoder->m_nHue
						  );
	QPVideoDecoder_SetSharpness(This_p, 
								pVideoDecoder->m_nSharpness
								);

	// CObject Init
	//
	CObject_Init(&pVideoDecoder->m_Object);

	return (QPERR_SUCCESS);
}



// IVideoDecoder
//
static EQP_Errors CMi2010_SetVideoRes(IVideoDecoder *This_p,
									  TQP_UINT32 dwWid,
				  					  TQP_UINT32 dwHgt,
				  					  TQP_UINT32 dwScanFmt,
				  					  TQP_UINT32 dwFps
				  					  )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);
	EQP_Errors		qpStatus;

	static TQP_UINT8	ucRegValsGen[] = 
	{
		0xFF, 0xFF, 0xFF, 0xFF , 0xFF, 0xFF
	};

	qpStatus = CMi2010_WriteBlock(This,
								  ucRegValsGen
								  );

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMi2010_SetVideoRes() : Video Res Set to %dx %d, status(%d)\n",
				 dwWid,
				 dwHgt,
				 qpStatus
				 ));

	return (qpStatus);
}


// Set Decoder Brightness
static EQP_Errors CMi2010_SetBrightness(IVideoDecoder *This_p,
										int nParamValue
										) 
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_nBrightness = nParamValue; 
	return (QPERR_SUCCESS);
}


// Set Decoder Contrast
static EQP_Errors CMi2010_SetContrast(IVideoDecoder *This_p,
									  int nParamValue
									  ) 
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_nContrast = nParamValue; 
	return (QPERR_SUCCESS);
}


// Set Decoder Saturation
static EQP_Errors CMi2010_SetSaturation(IVideoDecoder *This_p,
										int nParamValue
										) 
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_nSaturation = nParamValue; 
	return (QPERR_SUCCESS);
}



// Set Decoder Hue
static EQP_Errors CMi2010_SetHue(IVideoDecoder *This_p,
								 int nParamValue
								 ) 
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_nHue = nParamValue; 
	return (QPERR_SUCCESS);
}



// Set Decoder Sharpness
static EQP_Errors CMi2010_SetSharpness(IVideoDecoder *This_p,
									   int nParamValue
									   ) 
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_nSharpness = nParamValue; 
	return (QPERR_SUCCESS);
}



static EQP_Errors CMi2010_SetEdgeEnhancement(IVideoDecoder *This_p, 
											 int nParamValue
											 )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_nEdgeEnhancement = nParamValue; 
	return (QPERR_SUCCESS);
}



static EQP_Errors CMi2010_SetGamma(IVideoDecoder *This_p, 
								   int nParamValue
								   )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_nGamma = nParamValue; 
	return (QPERR_SUCCESS);
}



static EQP_Errors CMi2010_SetLowLightBoost(IVideoDecoder *This_p, 
										   int nParamValue
										   )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_nLowLightBoost = nParamValue; 
	return (QPERR_SUCCESS);
}



static EQP_Errors CMi2010_SetExposure(IVideoDecoder *This_p, 
									  int nParamValue
									  )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_nExposure = nParamValue; 
	return (QPERR_SUCCESS);
}



static EQP_Errors CMi2010_SetAutoExposureEnabled(IVideoDecoder *This_p, 
												 TQP_UINT32 enable
												 )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_AutoExposureEnable = enable; 
	return (QPERR_SUCCESS);
}



static EQP_Errors CMi2010_SetWhiteBalance(IVideoDecoder *This_p, 
										  int nParamValue
										  )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_nWhiteBalance = nParamValue; 
	return (QPERR_SUCCESS);
}



static EQP_Errors CMi2010_SetAutoWhiteBalanceEnabled(IVideoDecoder *This_p, 
													 TQP_UINT32 enable
													 )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_AutoWhiteBalanceEnable = enable; 
	return (QPERR_SUCCESS);
}



static EQP_Errors CMi2010_SetFocus(IVideoDecoder *This_p, 
								   int nParamValue
								   )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_nFocus = nParamValue; 
	return (QPERR_SUCCESS);
}



static EQP_Errors CMi2010_SetAutoFocusEnabled(IVideoDecoder *This_p, 
											  TQP_UINT32 enable
											  )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMi2010			*This = GET_INHERITED_OBJECT(CMi2010, m_VideoDecoder, pVideoDecoder);

	pVideoDecoder->m_AutoFocusEnable = enable; 
	return (QPERR_SUCCESS);
}


// constructor
//
CMi2010 * CMi2010_Constructor(CMi2010 *pMi2010, 
							  CObject *pParent,
							  TQP_UINT32 dwAttributes,
							  ITuner *pTuner,
							  ITVAudio *pTvAudio,
							  II2C *pI2C,
							  TQP_UINT8 cI2CDevAddr
							  )
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMi2010_Constructor()\n"
				 ));

	if (pMi2010)
	{
		CVideoDecoder	*pVideoDecoder;

		pVideoDecoder = CVideoDecoder_Constructor(&pMi2010->m_VideoDecoder, 
												  pParent,
												  dwAttributes,
												  pTuner,
												  pTvAudio
												  );

		if (!pVideoDecoder)
		{
			pMi2010 = QP_NULL;
		}
		else
		{
			// fill in the function table
			//

			// IGenericDevice
			pMi2010->m_VideoDecoder.m_iVideoDecoder.InitDevice = CMi2010_InitDevice;

			// IVideoDecoder
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetVideoRes = CMi2010_SetVideoRes;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetBrightness = CMi2010_SetBrightness;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetContrast = CMi2010_SetContrast;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetSaturation = CMi2010_SetSaturation;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetHue = CMi2010_SetHue;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetSharpness = CMi2010_SetSharpness;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetEdgeEnhancement = CMi2010_SetEdgeEnhancement;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetGamma = CMi2010_SetGamma;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetLowLightBoost = CMi2010_SetLowLightBoost;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetExposure = CMi2010_SetExposure;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetAutoExposureEnabled = CMi2010_SetAutoExposureEnabled;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetWhiteBalance = CMi2010_SetWhiteBalance;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetAutoWhiteBalanceEnabled = CMi2010_SetAutoWhiteBalanceEnabled;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetFocus = CMi2010_SetFocus;
			pMi2010->m_VideoDecoder.m_iVideoDecoder.SetAutoFocusEnabled = CMi2010_SetAutoFocusEnabled;

			// initialize members
			//
			pMi2010->m_pI2C = pI2C;
			pMi2010->m_cI2CDevAddr = cI2CDevAddr;

			// inherited members
			//
			pMi2010->m_VideoDecoder.m_nBrightnessDef = MI2010_BRIGHTNESS_DEFAULT;
			pMi2010->m_VideoDecoder.m_nBrightnessMin = MI2010_BRIGHTNESS_MIN;
			pMi2010->m_VideoDecoder.m_nBrightnessMax = MI2010_BRIGHTNESS_MAX;
			pMi2010->m_VideoDecoder.m_nBrightness = MI2010_BRIGHTNESS_DEFAULT;
			pMi2010->m_VideoDecoder.m_nContrastDef = MI2010_CONTRAST_DEFAULT; 
			pMi2010->m_VideoDecoder.m_nContrastMin = MI2010_CONTRAST_MIN;
			pMi2010->m_VideoDecoder.m_nContrastMax = MI2010_CONTRAST_MAX;
			pMi2010->m_VideoDecoder.m_nContrast = MI2010_CONTRAST_DEFAULT;
			pMi2010->m_VideoDecoder.m_nSaturationDef = MI2010_SATURATION_DEFAULT;
			pMi2010->m_VideoDecoder.m_nSaturationMin = MI2010_SATURATION_MIN;
			pMi2010->m_VideoDecoder.m_nSaturationMax = MI2010_SATURATION_MAX;
			pMi2010->m_VideoDecoder.m_nSaturation = MI2010_SATURATION_DEFAULT;
			pMi2010->m_VideoDecoder.m_nHueDef = MI2010_HUE_DEFAULT;
			pMi2010->m_VideoDecoder.m_nHueMin = MI2010_HUE_MIN;
			pMi2010->m_VideoDecoder.m_nHueMax = MI2010_HUE_MAX;
			pMi2010->m_VideoDecoder.m_nHue = MI2010_HUE_DEFAULT;
			pMi2010->m_VideoDecoder.m_nSharpnessDef = MI2010_SHARPNESS_DEFAULT;
			pMi2010->m_VideoDecoder.m_nSharpnessMin = MI2010_SHARPNESS_MIN;
			pMi2010->m_VideoDecoder.m_nSharpnessMax = MI2010_SHARPNESS_MAX;
			pMi2010->m_VideoDecoder.m_nSharpness = MI2010_SHARPNESS_DEFAULT;
			pMi2010->m_VideoDecoder.m_nEdgeEnhancementDef = MI2010_EDGE_ENHANCEMENT_DEFAULT;
			pMi2010->m_VideoDecoder.m_nEdgeEnhancementMin = MI2010_EDGE_ENHANCEMENT_MIN;
			pMi2010->m_VideoDecoder.m_nEdgeEnhancementMax = MI2010_EDGE_ENHANCEMENT_MAX;
			pMi2010->m_VideoDecoder.m_nEdgeEnhancement = MI2010_EDGE_ENHANCEMENT_DEFAULT;
			pMi2010->m_VideoDecoder.m_nGammaDef = MI2010_GAMMA_DEFAULT;
			pMi2010->m_VideoDecoder.m_nGammaMin = MI2010_GAMMA_MIN;
			pMi2010->m_VideoDecoder.m_nGammaMax = MI2010_GAMMA_MAX;
			pMi2010->m_VideoDecoder.m_nGamma = MI2010_GAMMA_DEFAULT;
			pMi2010->m_VideoDecoder.m_nLowLightBoostDef = MI2010_LOW_LIGHT_BOOST_DEFAULT;
			pMi2010->m_VideoDecoder.m_nLowLightBoostMin = MI2010_LOW_LIGHT_BOOST_MIN;
			pMi2010->m_VideoDecoder.m_nLowLightBoostMax = MI2010_LOW_LIGHT_BOOST_MAX;
			pMi2010->m_VideoDecoder.m_nLowLightBoost = MI2010_LOW_LIGHT_BOOST_DEFAULT;
			pMi2010->m_VideoDecoder.m_nExposureDef = MI2010_EXPOSURE_DEFAULT;
			pMi2010->m_VideoDecoder.m_nExposureMin = MI2010_EXPOSURE_MIN;
			pMi2010->m_VideoDecoder.m_nExposureMax = MI2010_EXPOSURE_MAX;
			pMi2010->m_VideoDecoder.m_nExposure = MI2010_EXPOSURE_DEFAULT;
			pMi2010->m_VideoDecoder.m_AutoExposureEnable = QP_TRUE;
			pMi2010->m_VideoDecoder.m_nWhiteBalanceDef = MI2010_WHITE_BALANCE_DEFAULT;
			pMi2010->m_VideoDecoder.m_nWhiteBalanceMin = MI2010_WHITE_BALANCE_MIN;
			pMi2010->m_VideoDecoder.m_nWhiteBalanceMax = MI2010_WHITE_BALANCE_MAX;
			pMi2010->m_VideoDecoder.m_nWhiteBalance = MI2010_WHITE_BALANCE_DEFAULT;
			pMi2010->m_VideoDecoder.m_AutoWhiteBalanceEnable = QP_TRUE;
			pMi2010->m_VideoDecoder.m_nFocusDef = MI2010_FOCUS_DEFAULT;
			pMi2010->m_VideoDecoder.m_nFocusMin = MI2010_FOCUS_MIN;
			pMi2010->m_VideoDecoder.m_nFocusMax = MI2010_FOCUS_MAX;
			pMi2010->m_VideoDecoder.m_nFocus = MI2010_FOCUS_DEFAULT;
			pMi2010->m_VideoDecoder.m_AutoFocusEnable = QP_FALSE;
		}
	}

	return (pMi2010);
}



// destructor
//
void CMi2010_Destructor(CMi2010 *This)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMi2010_Destructor()\n"
				 ));

	CVideoDecoder_Destructor(&This->m_VideoDecoder);
}

