///////////////////////////////////////////////////////////////////////////////
//        
//
// Cmt9v112.c
//
// Description: 
//
//  Micron MT9V112 implementation
// 
// History:
//	May 11, 2007	(dt) created
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


#include "cmt9v112.h"
#include "qpdbgmgr.h"
#include "util.h"


/////////////////////////////////////////////////////////////////////////////
//
//

#define	MT9V112_BRIGHTNESS_MIN					0x00
#define	MT9V112_BRIGHTNESS_MAX					0xFF
#define	MT9V112_BRIGHTNESS_DEFAULT				0x80

#define	MT9V112_CONTRAST_MIN					0x00
#define	MT9V112_CONTRAST_MAX					0x7F
#define	MT9V112_CONTRAST_DEFAULT				0x44

#define	MT9V112_SATURATION_MIN					0x00
#define	MT9V112_SATURATION_MAX					0x7F
#define	MT9V112_SATURATION_DEFAULT				0x40

#define	MT9V112_HUE_MIN							(-128)	// 0x80
#define	MT9V112_HUE_MAX							(+127)	// 0x7F
#define	MT9V112_HUE_DEFAULT						0x00

#define	MT9V112_SHARPNESS_MIN					0
#define	MT9V112_SHARPNESS_MAX					15
#define	MT9V112_SHARPNESS_DEFAULT				8


extern TQP_DbgMCompID	g_QPPeripheralDBGCompID;


/////////////////////////////////////////////////////////////////////////////
//
//


static EQP_Errors CMt9V112_readRegister(CMt9V112 *This,
										TQP_UINT8 regAddr, 
										PTQP_UINT8 pValue
										)
{
    return (II2C_WriteThenRead(This->m_pI2C,
							   This->m_cI2CDevAddr,	//device address
							   1,					//write size
							   &regAddr,			//write buffer
							   1,					//read size
							   pValue				//read buffer
							   ));
}



static EQP_Errors CMt9V112_writeRegister(CMt9V112 *This,
										 TQP_UINT8 regAddr, 
										 TQP_UINT8 value
										 )
{
    TQP_UINT8 buffer[2];

	buffer[0] = regAddr;
	buffer[1] = value;

    return (II2C_Write(This->m_pI2C, 
					   This->m_cI2CDevAddr,	//device address
					   2,					//write size
					   buffer				//write buffer
					   ));
}


static EQP_Errors CMt9V112_WriteBlock(CMt9V112 *This,
									  TQP_UINT32 *pRegTable
									  )
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;
	TQP_UINT32	dwTotalReg = 0;

	// get total number of registers to write
	//
	while (-1 != pRegTable[dwTotalReg]) 
	{
		dwTotalReg++;
	}

	dwTotalReg >>= 1;

	// do multiple writes
	//
	while (dwTotalReg)
	{
		qpStatus = CMt9V112_writeRegister(This,
										  (TQP_UINT8)pRegTable[0], 
										  (TQP_UINT8)pRegTable[1]
										  );
		if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CMt9V112_WriteBlock() ****** writeRegister failed! status(%d) ******\n",
						 qpStatus
						 ));
			break;
		}

		pRegTable += 2;
		dwTotalReg--;
	}

	return (qpStatus);
}



/////////////////////////////////////////////////////////////////////////////
//
//


// IGenericDevice
//
static EQP_Errors CMt9V112_InitDevice(IVideoDecoder *This_p, 
									  QP_DEVICE_CALLBACK pFuncCallback,
									  TQP_PVOID context,
									  TQP_PVOID pInitData
									  )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMt9V112		*This = GET_INHERITED_OBJECT(CMt9V112, m_VideoDecoder, pVideoDecoder);

	static TQP_UINT32 dwMt9V112RegVals[] = 
	{
		0xF0,0x00,
		0xF1,0x00,
		0x0D,0x00,
		0xF1,0x09,
		0x0D,0x00,
		0xF1,0x08,
		0x00,0x12,
		0xF1,0x29,
		0x01,0x00,
		0xF1,0x12,
		0x02,0x00,
		0xF1,0x26,
		0x03,0x01,
		0xF1,0xE2,
		0x04,0x02,
		0xF1,0x82,
		0x05,0x00,
		0xF1,0xCF,
		0x06,0x00,
		0xF1,0x0D,
		0x07,0x04,
		0xF1,0x1A,
		0x08,0x00,
		0xF1,0x0D,
		0x09,0x02,
		0xF1,0x0E,
		0x0A,0x00,
		0xF1,0x11,
		0x0B,0x00,
		0xF1,0x00,
		0x0C,0x00,
		0xF1,0x00,
		0x0D,0x00,
		0xF1,0x08,
		0x20,0x07,
		0xF1,0x00,
		0x21,0x04,
		0xF1,0x0C,
		0x22,0x01,
		0xF1,0x2B,
		0x23,0x06,
		0xF1,0x08,
		0x24,0x40,
		0xF1,0x00,
		0x2B,0x00,
		0xF1,0x42,
		0x2C,0x00,
		0xF1,0x5C,
		0x2D,0x00,
		0xF1,0x58,
		0x2E,0x00,
		0xF1,0x42,
		0x2F,0x00,
		0xF1,0x42,
		0x30,0x04,
		0xF1,0x3E,
		0x35,0x20,
		0xF1,0x29,
		0x41,0x00,
		0xF1,0xD7,
		0x58,0x00,
		0xF1,0x00,
		0x59,0x00,
		0xF1,0x0C,
		0x5F,0x36,
		0xF1,0x30,
		0x60,0x00,
		0xF1,0x80,
		0x61,0x00,
		0xF1,0x2C,
		0x62,0x00,
		0xF1,0x28,
		0x63,0x00,
		0xF1,0x2D,
		0x64,0x00,
		0xF1,0x2C,
		0xC8,0x00,
		0xF1,0x0B,
		0xF0,0x00,
		0xF1,0x00,
		0xF1,0x00,
		0xF1,0x00,
		0xFF,0x12,
		0xF1,0x29,

		0xf0,0x00,
		0xf1,0x01,
		0x05,0x00,
		0xF1,0x0B,
		0x06,0x60,
		0xF1,0x0E,
		0x08,0x00,
		0xF1,0x80,
		0x25,0x00,
		0xF1,0x2D,
		0x34,0x00,
		0xF1,0x00,
		0x35,0xFF,
		0xF1,0x00,
		0x3A,0x0a,
		0xF1,0x00,
		0x3B,0x04,
		0xF1,0x3E,
		0x3C,0x04,
		0xF1,0x00,
		0x47,0x00,
		0xF1,0x18,
		0x48,0x00,
		0xF1,0x00,
		0x53,0x1A,
		0xF1,0x08,
		0x54,0x60,
		0xF1,0x3D,
		0x55,0xAB,
		0xF1,0x8C,
		0x56,0xDA,
		0xF1,0xC4,
		0x57,0xFF,
		0xF1,0xED,
		0x58,0x00,
		0xF1,0x00,
		0x80,0x00,
		0xF1,0x03,
		0x81,0x00,
		0xF1,0x00,
		0x82,0x00,
		0xF1,0x00,
		0x83,0x00,
		0xF1,0x00,
		0x84,0x00,
		0xF1,0x00,
		0x85,0x00,
		0xF1,0x00,
		0x86,0x00,
		0xF1,0x00,
		0x87,0x00,
		0xF1,0x00,
		0x88,0x00,
		0xF1,0x00,
		0x89,0x00,
		0xF1,0x00,
		0x8A,0x00,
		0xF1,0x00,
		0x8B,0x00,
		0xF1,0x00,
		0x8C,0x00,
		0xF1,0x00,
		0x8D,0x00,
		0xF1,0x00,
		0x8E,0x00,
		0xF1,0x00,
		0x8F,0x00,
		0xF1,0x00,
		0x90,0x00,
		0xF1,0x00,
		0x91,0x00,
		0xF1,0x00,
		0x92,0x00,
		0xF1,0x00,
		0x93,0x00,
		0xF1,0x00,
		0x94,0x00,
		0xF1,0x00,
		0x95,0x00,
		0xF1,0x00,
		0x9B,0x0a,
		0xF1,0x00,
		0x9D,0x24,
		0xF1,0xAE,
		0x9F,0x00,
		0xF1,0x00,
		0xA0,0x02,
		0xF1,0x80,
		0xA1,0x02,
		0xF1,0x80,
		0xA2,0x00,
		0xF1,0x00,
		0xA3,0x01,
		0xF1,0xE0,
		0xA4,0x01,
		0xF1,0xE0,
		0xA5,0x00,
		0xF1,0x00,
		0xA6,0x02,
		0xF1,0x80,
		0xA7,0x01,
		0xF1,0x40,
		0xA8,0x00,
		0xF1,0x00,
		0xA9,0x01,
		0xF1,0xE0,
		0xAA,0x00,
		0xF1,0xF0,
		0xAB,0x02,
		0xF1,0x80,
		0xAC,0x01,
		0xF1,0xE0,
		0xAE,0x0C,
		0xF1,0x09,
		0xAF,0x00,
		0xF1,0x00,
		0xB6,0x00,
		0xF1,0x00,
		0xB7,0x00,
		0xF1,0x00,
		0xB8,0x00,
		0xF1,0x00,
		0xB9,0x00,
		0xF1,0x00,
		0xBA,0x00,
		0xF1,0x00,
		0xBB,0x00,
		0xF1,0x00,
		0xBC,0x00,
		0xF1,0x00,
		0xBD,0x00,
		0xF1,0x00,
		0xBE,0x00,
		0xF1,0x00,
		0xBF,0x00,
		0xF1,0x00,
		0xC0,0x00,
		0xF1,0x00,
		0xC1,0x00,
		0xF1,0x00,
		0xC2,0x00,
		0xF1,0x00,
		0xC3,0x00,
		0xF1,0x00,
		0xC4,0x00,
		0xF1,0x00,
		0xC8,0x1F,
		0xF1,0x0B,
		0xDC,0x10,
		0xF1,0x05,
		0xDD,0x56,
		0xF1,0x2E,
		0xDE,0xB3,
		0xF1,0x8E,
		0xDF,0xE1,
		0xF1,0xCD,
		0xE0,0xFF,
		0xF1,0xF1,
		0xE1,0x00,
		0xF1,0x00,
		0xE2,0x70,
		0xF1,0x00,
		0xE3,0xB0,
		0xF1,0x23,

		0xf0,0x00,
		0xF1,0x02,
		0x02,0x00,
		0xF1,0xEA,
		0x03,0x29,
		0xF1,0x1a,
		0x04,0x04,
		0xF1,0xa4,
		0x09,0x00,
		0xF1,0x97,
		0x0A,0x00,
		0xF1,0x72,
		0x0B,0x00,
		0xF1,0x1e,
		0x0C,0x00,
		0xF1,0x1d,
		0x0D,0x00,
		0xF1,0x7e,
		0x0E,0x00,
		0xF1,0x72,
		0x0F,0x00,
		0xF1,0x11,
		0x10,0x00,
		0xF1,0x34,
		0x11,0x00,
		0xF1,0x82,
		0x15,0x00,
		0xF1,0x81,
		0x16,0x00,
		0xF1,0x1A,
		0x17,0x00,
		0xF1,0x31,
		0x18,0x00,
		0xF1,0x00,
		0x19,0x00,
		0xF1,0x1D,
		0x1A,0x00,
		0xF1,0x06,
		0x1B,0x00,
		0xF1,0x07,
		0x1C,0x00,
		0xF1,0x3E,
		0x1D,0x00,
		0xF1,0x6C,
		0x1E,0x00,
		0xF1,0x3E,
		0x1F,0x00,
		0xF1,0x90,
		0x21,0x80,
		0xF1,0x80,
		0x22,0x90,
		0xF1,0x80,
		0x23,0x88,
		0xF1,0x78,
		0x24,0x7F,
		0xF1,0x00,
		0x26,0x80,
		0xF1,0x00,
		0x27,0x80,
		0xF1,0x08,
		0x28,0xEF,
		0xF1,0x02,
		0x29,0x84,
		0xF1,0x7C,
		0x2A,0x00,
		0xF1,0xD0,
		0x2B,0x60,
		0xF1,0x20,
		0x2C,0x60,
		0xF1,0x20,
		0x2D,0xF0,
		0xF1,0xA0,
		0x2E,0x0C,
		0xF1,0x44,
		0x2F,0xDF,
		0xF1,0x20,
		0x33,0x14,
		0xF1,0x6E,
		0x36,0x78,
		0xF1,0x10,
		0x37,0x03,
		0xF1,0x00,
		0x38,0x04,
		0xF1,0x40,
		0x39,0x05,
		0xF1,0x62,
		0x3A,0x03,
		0xF1,0x57,
		0x3B,0x05,
		0xF1,0x5B,
		0x3C,0x05,
		0xF1,0x5B,
		0x3D,0x17,
		0xF1,0xDD,
		0x3E,0x1C,
		0xF1,0xFF,
		0x46,0xD9,
		0xF1,0x00,
		0x57,0x01,
		0xF1,0x46,
		0x58,0x01,
		0xF1,0x87,
		0x59,0x02,
		0xF1,0x0E,
		0x5A,0x02,
		0xF1,0x77,
		0x5B,0x80,
		0xF1,0x02,
		0x5C,0x0B,
		0xF1,0x07,
		0x5D,0x0F,
		0xF1,0x0B,
		0x5E,0x67,
		0xF1,0x4C,
		0x5F,0x3D,
		0xF1,0x2C,
		0x60,0x00,
		0xF1,0x02,
		0x61,0x59,
		0xF1,0x55,
		0x62,0x10,
		0xF1,0x10,
		0x65,0x00,
		0xF1,0x00,
		0x67,0x40,
		0xF1,0x10,
		0x82,0x03,
		0xF1,0xFC,
		0x83,0x03,
		0xF1,0x01,
		0x84,0x00,
		0xF1,0xC1,
		0x85,0x03,
		0xF1,0xA1,
		0x86,0x03,
		0xF1,0xD4,
		0x87,0x03,
		0xF1,0xD7,
		0x88,0x03,
		0xF1,0x99,
		0x89,0x03,
		0xF1,0xF8,
		0x8A,0x00,
		0xF1,0x1C,
		0x8B,0x03,
		0xF1,0xBD,
		0x8C,0x03,
		0xF1,0xDB,
		0x8D,0x03,
		0xF1,0xBD,
		0x8E,0x03,
		0xF1,0xFC,
		0x8F,0x03,
		0xF1,0xDE,
		0x90,0x03,
		0xF1,0xDE,
		0x91,0x03,
		0xF1,0xDE,
		0x92,0x03,
		0xF1,0xDE,
		0x93,0x00,
		0xF1,0x1F,
		0x94,0x00,
		0xF1,0x41,
		0x95,0x03,
		0xF1,0x63,
		0xC8,0x1F,
		0xF1,0x0B,
		0xCB,0x00,
		0xF1,0x01,
		0xD2,0x00,
		0xF1,0x7F,
		0xD3,0x00,
		0xF1,0x00,
		0xD4,0x02,
		0xF1,0x08,
		0xD5,0x00,
		0xF1,0x00,
		0xEF,0x00,
		0xF1,0x08,
		0xF2,0x00,
		0xF1,0x00,
		0xF5,0x00,
		0xF1,0x40,
		0xF6,0x00,
		0xF1,0x7F,
		0xFF,0xA8,
		0xF1,0x80,
		-1,-1
	};

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMt9V112_InitDevice()\n"
				 ));

	if (CObject_IsInitialized(&pVideoDecoder->m_Object))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
					 "CMt9V112_InitDevice() Already Initialized. Skipping...\n" 
					 ));
		return (QPERR_SUCCESS);
	}

	// save callback address and context
	//
	pVideoDecoder->m_pDeviceCallback = pFuncCallback;
	pVideoDecoder->m_callbackContext = context;


	// Initialize MT9V112 with the default value
	//
	CMt9V112_WriteBlock(This,
						dwMt9V112RegVals
						);

	// program the video decoder with settings read from the registry
	//
	QPVideoDecoder_SetVideoSystem(This_p, 
								  pVideoDecoder->m_VidStandard
								  );
	QPVideoDecoder_SelectVideoSource(This_p, 
									 pVideoDecoder->m_VidSource
									 );

	QPVideoDecoder_SetVcrTiming(This_p, 
								pVideoDecoder->m_VCRTiming
								);

	QPVideoDecoder_SetBrightness(This_p, 
								 pVideoDecoder->m_nBrightness
								 );
	QPVideoDecoder_SetContrast(This_p, 
							   pVideoDecoder->m_nContrast
							   );
	QPVideoDecoder_SetSaturation(This_p, 
								 pVideoDecoder->m_nSaturation
								 );
	QPVideoDecoder_SetHue(This_p, 
						  pVideoDecoder->m_nHue
						  );
	QPVideoDecoder_SetSharpness(This_p, 
								pVideoDecoder->m_nSharpness
								);

	// CObject Init
	//
	CObject_Init(&pVideoDecoder->m_Object);

	return (QPERR_SUCCESS);
}



// IVideoDecoder
//
static EQP_Errors CMt9V112_SetVideoRes(IVideoDecoder *This_p,
									   TQP_UINT32 dwWid,
				  					   TQP_UINT32 dwHgt,
				  					   TQP_UINT32 dwScanFmt,
				  					   TQP_UINT32 dwFps
				  					   )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CMt9V112		*This = GET_INHERITED_OBJECT(CMt9V112, m_VideoDecoder, pVideoDecoder);
	EQP_Errors		qpStatus;

	static TQP_UINT32	dwRegValsGen[14] = 
	{
		0xF0,0x00,
		0xF1,0x01,
		0xA1,0x02,
		0xF1,0x80,
		0xA4,0x01,
		0xF1,0xE0,
		-1,-1
	};

	dwRegValsGen[5] = ((dwWid >> 8) & 0x000000FF);
	dwRegValsGen[7] = (dwWid & 0x000000FF);
	dwRegValsGen[9] = ((dwHgt >> 8) & 0x000000FF);
	dwRegValsGen[11] = (dwHgt & 0x000000FF);

	qpStatus = CMt9V112_WriteBlock(This,
								   dwRegValsGen
								   );

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMt9V112_SetVideoRes() : Video Res Set to %dx %d, status(%d)\n",
				 dwWid,
				 dwHgt,
				 qpStatus
				 ));

	return (qpStatus);
}



// constructor
//
CMt9V112 * CMt9V112_Constructor(CMt9V112 *pMt9V112, 
								CObject *pParent,
								TQP_UINT32 dwAttributes,
								ITuner *pTuner,
								ITVAudio *pTvAudio,
								II2C *pI2C,
								TQP_UINT8 cI2CDevAddr
								)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMt9V112_Constructor()\n"
				 ));

	if (pMt9V112)
	{
		CVideoDecoder	*pVideoDecoder;

		pVideoDecoder = CVideoDecoder_Constructor(&pMt9V112->m_VideoDecoder, 
												  pParent,
												  dwAttributes,
												  pTuner,
												  pTvAudio
												  );

		if (!pVideoDecoder)
		{
			pMt9V112 = QP_NULL;
		}
		else
		{
			// fill in the function table
			//

			// IGenericDevice
			pMt9V112->m_VideoDecoder.m_iVideoDecoder.InitDevice = CMt9V112_InitDevice;

			// IVideoDecoder
			pMt9V112->m_VideoDecoder.m_iVideoDecoder.SetVideoRes = CMt9V112_SetVideoRes;

			// initialize members
			//
			pMt9V112->m_pI2C = pI2C;
			pMt9V112->m_cI2CDevAddr = cI2CDevAddr;

			// inherited members
			//
			pMt9V112->m_VideoDecoder.m_nBrightnessDef = MT9V112_BRIGHTNESS_DEFAULT;
			pMt9V112->m_VideoDecoder.m_nBrightnessMin = MT9V112_BRIGHTNESS_MIN;
			pMt9V112->m_VideoDecoder.m_nBrightnessMax = MT9V112_BRIGHTNESS_MAX;
			pMt9V112->m_VideoDecoder.m_nBrightness = MT9V112_BRIGHTNESS_DEFAULT;
			pMt9V112->m_VideoDecoder.m_nContrastDef = MT9V112_CONTRAST_DEFAULT; 
			pMt9V112->m_VideoDecoder.m_nContrastMin = MT9V112_CONTRAST_MIN;
			pMt9V112->m_VideoDecoder.m_nContrastMax = MT9V112_CONTRAST_MAX;
			pMt9V112->m_VideoDecoder.m_nContrast = MT9V112_CONTRAST_DEFAULT;
			pMt9V112->m_VideoDecoder.m_nSaturationDef = MT9V112_SATURATION_DEFAULT;
			pMt9V112->m_VideoDecoder.m_nSaturationMin = MT9V112_SATURATION_MIN;
			pMt9V112->m_VideoDecoder.m_nSaturationMax = MT9V112_SATURATION_MAX;
			pMt9V112->m_VideoDecoder.m_nSaturation = MT9V112_SATURATION_DEFAULT;
			pMt9V112->m_VideoDecoder.m_nHueDef = MT9V112_HUE_DEFAULT;
			pMt9V112->m_VideoDecoder.m_nHueMin = MT9V112_HUE_MIN;
			pMt9V112->m_VideoDecoder.m_nHueMax = MT9V112_HUE_MAX;
			pMt9V112->m_VideoDecoder.m_nHue = MT9V112_HUE_DEFAULT;
			pMt9V112->m_VideoDecoder.m_nSharpnessDef = MT9V112_SHARPNESS_DEFAULT;
			pMt9V112->m_VideoDecoder.m_nSharpnessMin = MT9V112_SHARPNESS_MIN;
			pMt9V112->m_VideoDecoder.m_nSharpnessMax = MT9V112_SHARPNESS_MAX;
			pMt9V112->m_VideoDecoder.m_nSharpness = MT9V112_SHARPNESS_DEFAULT;
		}
	}

	return (pMt9V112);
}



// destructor
//
void CMt9V112_Destructor(CMt9V112 *This)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CMt9V112_Destructor()\n"
				 ));

	CVideoDecoder_Destructor(&This->m_VideoDecoder);
}

