///////////////////////////////////////////////////////////////////////////////
//        
//
// Cov7720.c
//
// Description: 
//
//  OV7720 implementation
// 
// History:
//	April 28, 2008	(dt) created
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


#include "cov7720.h"
#include "qpdbgmgr.h"
#include "util.h"


/////////////////////////////////////////////////////////////////////////////
//
//

#define	OV7720_BRIGHTNESS_MIN				0x00
#define	OV7720_BRIGHTNESS_MAX				0xFF
#define	OV7720_BRIGHTNESS_DEFAULT			0x80

#define	OV7720_CONTRAST_MIN					0x00
#define	OV7720_CONTRAST_MAX					0x7F
#define	OV7720_CONTRAST_DEFAULT				0x44

#define	OV7720_SATURATION_MIN				0x00
#define	OV7720_SATURATION_MAX				0x7F
#define	OV7720_SATURATION_DEFAULT			0x40

#define	OV7720_HUE_MIN						(-128)	// 0x80
#define	OV7720_HUE_MAX						(+127)	// 0x7F
#define	OV7720_HUE_DEFAULT					0x00

#define	OV7720_SHARPNESS_MIN				0
#define	OV7720_SHARPNESS_MAX				15
#define	OV7720_SHARPNESS_DEFAULT			8


extern TQP_DbgMCompID	g_QPPeripheralDBGCompID;


/////////////////////////////////////////////////////////////////////////////
//
//


static EQP_Errors COV7720_readRegister(COV7720 *This,
									   TQP_UINT8 regAddr, 
									   PTQP_UINT8 pValue
									   )
{
    return (II2C_WriteThenRead(This->m_pI2C,
							   This->m_cI2CDevAddr,	//device address
							   1,					//write size
							   &regAddr,			//write buffer
							   1,					//read size
							   pValue				//read buffer
							   ));
}



static EQP_Errors COV7720_writeRegister(COV7720 *This,
										TQP_UINT8 regAddr, 
										TQP_UINT8 value
										)
{
    TQP_UINT8 buffer[2];

	buffer[0] = regAddr;
	buffer[1] = value;

    return (II2C_Write(This->m_pI2C, 
					   This->m_cI2CDevAddr,	//device address
					   2,					//write size
					   buffer				//write buffer
					   ));
}


static EQP_Errors COV7720_WriteBlock(COV7720 *This,
									 TQP_UINT32 *pRegTable
									 )
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;
	TQP_UINT32	dwTotalReg = 0;

	// get total number of registers to write
	//
	while (-1 != pRegTable[dwTotalReg]) 
	{
		dwTotalReg++;
	}

	dwTotalReg >>= 1;

	// do multiple writes
	//
	while (dwTotalReg)
	{
		qpStatus = COV7720_writeRegister(This,
										 (TQP_UINT8)pRegTable[0], 
										 (TQP_UINT8)pRegTable[1]
										 );
		if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "COV7720_WriteBlock() ****** writeRegister failed! status(%d) ******\n",
						 qpStatus
						 ));
			break;
		}

		pRegTable += 2;
		dwTotalReg--;
	}

	return (qpStatus);
}



/////////////////////////////////////////////////////////////////////////////
//
//


// IGenericDevice
//
static EQP_Errors COV7720_InitDevice(IVideoDecoder *This_p, 
									 QP_DEVICE_CALLBACK pFuncCallback,
									 TQP_PVOID context,
									 TQP_PVOID pInitData
									 )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	COV7720			*This = GET_INHERITED_OBJECT(COV7720, m_VideoDecoder, pVideoDecoder);

	static TQP_UINT32 dwOV7720RegVals[] = 
	{
		0x12,0x80,
  
		0x3D,0x00,
		0x1A,0xF1,
		0x0c,0x80,
		0x11,0x01,
		0x32,0x00,

		0x12,0x00,

		0x0d,0x41,
		0x20,0x10,
		0x4e,0x0f,
		0x3e,0xf3,
		0x41,0x20,
		0x63,0xe0,
		0x64,0xff,
		0x9e,0x41,

		0x13,0xf0,
		0x22,0x7f,
		0x23,0x03,
		0x24,0x40,
		0x25,0x30,
		0x26,0xa1,
		0x13,0xf7,

		0x90,0x04,
		0x91,0x01,
		0x92,0x04,

		0x94,0xb0,
		0x95,0x9d,
		0x96,0x13,
		0x97,0x1c,
		0x98,0x98,
		0x99,0xb4,
		0x9a,0x1e,

		0x7e,0x07,
		0x7f,0x12,
		0x80,0x28,
		0x81,0x4e,
		0x82,0x5b,
		0x83,0x65,
		0x84,0x6e,
		0x85,0x77,
		0x86,0x7f,
		0x87,0x87,
		0x88,0x95,
		0x89,0xa2,
		0x8a,0xba,
		0x8b,0xd0,
		0x8c,0xe4,
		0x8d,0x25,

		-1,-1
	};

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "COV7720_InitDevice()\n"
				 ));

	if (CObject_IsInitialized(&pVideoDecoder->m_Object))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
					 "COV7720_InitDevice() Already Initialized. Skipping...\n" 
					 ));
		return (QPERR_SUCCESS);
	}

	// save callback address and context
	//
	pVideoDecoder->m_pDeviceCallback = pFuncCallback;
	pVideoDecoder->m_callbackContext = context;


	// Initialize OV7720 with the default value
	//
	COV7720_WriteBlock(This,
					   dwOV7720RegVals
					   );

	// program the video decoder with settings read from the registry
	//
	QPVideoDecoder_SetVideoSystem(This_p, 
								  pVideoDecoder->m_VidStandard
								  );
	QPVideoDecoder_SelectVideoSource(This_p, 
									 pVideoDecoder->m_VidSource
									 );

	QPVideoDecoder_SetBrightness(This_p, 
								 pVideoDecoder->m_nBrightness
								 );
	QPVideoDecoder_SetContrast(This_p, 
							   pVideoDecoder->m_nContrast
							   );
	QPVideoDecoder_SetSaturation(This_p, 
								 pVideoDecoder->m_nSaturation
								 );
	QPVideoDecoder_SetHue(This_p, 
						  pVideoDecoder->m_nHue
						  );
	QPVideoDecoder_SetSharpness(This_p, 
								pVideoDecoder->m_nSharpness
								);

	// CObject Init
	//
	CObject_Init(&pVideoDecoder->m_Object);

	return (QPERR_SUCCESS);
}



// IVideoDecoder
//
static EQP_Errors COV7720_SetVideoRes(IVideoDecoder *This_p,
									  TQP_UINT32 dwWid,
				  					  TQP_UINT32 dwHgt,
				  					  TQP_UINT32 dwScanFmt,
				  					  TQP_UINT32 dwFps
				  					  )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	COV7720			*This = GET_INHERITED_OBJECT(COV7720, m_VideoDecoder, pVideoDecoder);
	TQP_UINT32		dwVal;
  
	dwVal = dwWid / 4;  
	COV7720_writeRegister(This, 
						  0x29, 
						  (TQP_UINT8)dwVal	//0xA0 (VGA)
						  ); 
	COV7720_writeRegister(This, 
						  0x18, 
						  (TQP_UINT8)dwVal	//0xA0(VGA)
						  ); 
  
	dwVal = dwHgt / 2;
	COV7720_writeRegister(This, 
						  0x2c, 
						  (TQP_UINT8)dwVal	//0xF0 (VGA)
						  ); 
	COV7720_writeRegister(This, 
						  0x1a, 
						  (TQP_UINT8)dwVal	//0xF0 (VGA)
						  ); 

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "COV7720_SetVideoRes() : Video Res Set to %dx %d\n",
				 dwWid,
				 dwHgt
				 ));

	return (QPERR_SUCCESS);
}



// constructor
//
COV7720 * COV7720_Constructor(COV7720 *pOV7720, 
								CObject *pParent,
								TQP_UINT32 dwAttributes,
								ITuner *pTuner,
								ITVAudio *pTvAudio,
								II2C *pI2C,
								TQP_UINT8 cI2CDevAddr
								)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "COV7720_Constructor()\n"
				 ));

	if (pOV7720)
	{
		CVideoDecoder	*pVideoDecoder;

		pVideoDecoder = CVideoDecoder_Constructor(&pOV7720->m_VideoDecoder, 
												  pParent,
												  dwAttributes,
												  pTuner,
												  pTvAudio
												  );

		if (!pVideoDecoder)
		{
			pOV7720 = QP_NULL;
		}
		else
		{
			// fill in the function table
			//

			// IGenericDevice
			pOV7720->m_VideoDecoder.m_iVideoDecoder.InitDevice = COV7720_InitDevice;

			// IVideoDecoder
			pOV7720->m_VideoDecoder.m_iVideoDecoder.SetVideoRes = COV7720_SetVideoRes;

			// initialize members
			//
			pOV7720->m_pI2C = pI2C;
			pOV7720->m_cI2CDevAddr = cI2CDevAddr;

			// inherited members
			//
			pOV7720->m_VideoDecoder.m_nBrightnessDef = OV7720_BRIGHTNESS_DEFAULT;
			pOV7720->m_VideoDecoder.m_nBrightnessMin = OV7720_BRIGHTNESS_MIN;
			pOV7720->m_VideoDecoder.m_nBrightnessMax = OV7720_BRIGHTNESS_MAX;
			pOV7720->m_VideoDecoder.m_nBrightness = OV7720_BRIGHTNESS_DEFAULT;
			pOV7720->m_VideoDecoder.m_nContrastDef = OV7720_CONTRAST_DEFAULT; 
			pOV7720->m_VideoDecoder.m_nContrastMin = OV7720_CONTRAST_MIN;
			pOV7720->m_VideoDecoder.m_nContrastMax = OV7720_CONTRAST_MAX;
			pOV7720->m_VideoDecoder.m_nContrast = OV7720_CONTRAST_DEFAULT;
			pOV7720->m_VideoDecoder.m_nSaturationDef = OV7720_SATURATION_DEFAULT;
			pOV7720->m_VideoDecoder.m_nSaturationMin = OV7720_SATURATION_MIN;
			pOV7720->m_VideoDecoder.m_nSaturationMax = OV7720_SATURATION_MAX;
			pOV7720->m_VideoDecoder.m_nSaturation = OV7720_SATURATION_DEFAULT;
			pOV7720->m_VideoDecoder.m_nHueDef = OV7720_HUE_DEFAULT;
			pOV7720->m_VideoDecoder.m_nHueMin = OV7720_HUE_MIN;
			pOV7720->m_VideoDecoder.m_nHueMax = OV7720_HUE_MAX;
			pOV7720->m_VideoDecoder.m_nHue = OV7720_HUE_DEFAULT;
			pOV7720->m_VideoDecoder.m_nSharpnessDef = OV7720_SHARPNESS_DEFAULT;
			pOV7720->m_VideoDecoder.m_nSharpnessMin = OV7720_SHARPNESS_MIN;
			pOV7720->m_VideoDecoder.m_nSharpnessMax = OV7720_SHARPNESS_MAX;
			pOV7720->m_VideoDecoder.m_nSharpness = OV7720_SHARPNESS_DEFAULT;
		}
	}

	return (pOV7720);
}



// destructor
//
void COV7720_Destructor(COV7720 *This)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "COV7720_Destructor()\n"
				 ));

	CVideoDecoder_Destructor(&This->m_VideoDecoder);
}

