///////////////////////////////////////////////////////////////////////////////
//        
//
// Cpe1005.h
//
// Description: 
//
//  AcuteLogic camera module PE1005 implementation
// 
// History:
//	April 30, 2008 created
//
// Authors: Andy Yu
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __CPE1005_H__
#define __CPE1005_H__

#include "cviddecoder.h"
#include "qpi2c.h"

#ifdef __cplusplus
extern "C" {
#endif


/////////////////////////////////////////////////////////////////////////////
//
//

// default PE1005 I2C address
//
#define PE1005_DEVICE_ADDR   (0x39)   // 0x72 for write, 0x73 for read

// PRS definitions
//
#define PRS_ACK              (0)
#define PRS_PCS_ERROR        (1)

// RES definitions
#define RET_CMDEXEC_MASK            (0x7FFF)
#define RET_CMDEXEC_OK              (0)

#define RET_CMDEXEC_REQFULL         (-1)
#define RET_CMDEXEC_FAIL            (-2)
#define RET_CMDEXEC_FATAL           (-3)

#define RET_CMDEXEC_UNKNOWN         (1)
#define RET_CMDEXEC_BUSY            (2)
#define RET_CMDEXEC_NOTSUPPORT      (3)
#define RET_CMDEXEC_PRMERROR        (4)

#define RET_CMDEXEC_LENS_OUTOFSPEED (100)
#define RET_CMDEXEC_LENS_OUTOFPOS   (101)
#define RET_CMDEXEC_LENS_OUTOFSTEP  (102)

// PE1005 IO mapping (PCA9535 GPIO definition)
// Output pins
#define PE1005_PON       0x01   // G0.0
#define PE1005_INT_IN    0x01   // G1.0
#define PE1005_SHUTDOWN  0x08   // G1.3
#define PE1005_HRSTX     0x10   // G1.4
#define PE1005_POWER     0x80   // G1.7

// Input pins
#define PE1005_INT_OUT   0x02   // G0.1


/////////////////////////////////////////////////////////////////////////////
//
//

// packet format
//
typedef union _PE1005_PACKET_HEADER
{
    TQP_UINT16    Read;
    struct
    {
        TQP_UINT16 Reserved2:  8;   // Bits 7 - 0
        TQP_UINT16 PRS:        4;   // Bits 11 - 8
        TQP_UINT16 Reserved1:  2;   // Bits 13 - 12
        TQP_UINT16 SR:         1;   // Bits 14
        TQP_UINT16 ALT:        1;   // Bits 15
    } stPakHead;
} PE1005_PACKET_HEADER, *PPE1005_PACKET_HEADER;


typedef union _PE1005_COMMAND_HEADER
{
    TQP_UINT16     Read;
    struct 
    {
        TQP_UINT16 TPY:        1;   // Bits 0
        TQP_UINT16 ID:         8;   // Bits 8 - 1
        TQP_UINT16 CAT:        4;   // Bits 12 - 9
        TQP_UINT16 CLS:        1;   // Bits 14 - 13
        TQP_UINT16 RES:        1;   // Bits 15
    } stCmdHeadRes0;
    struct 
    {
        TQP_UINT16 ERRORCODE: 15;   // Bits 14 - 0
        TQP_UINT16 RES:        1;   // Bits 15
    } stCmdHeadRes1;
} PE1005_COMMAND_HEADER, *PPE1005_COMMAND_HEADER;


typedef union _PE1005_COMMAND_PARAMETER
{
    TQP_UINT8      Read[32];
    struct   // DVC mode (only different on P0 - P3)
    {
        // P0 - P3
        TQP_UINT32 RSV1:       13;   // Bits 12 - 0, reserved
        TQP_UINT32 ASS:         1;   // Bits 13
        TQP_UINT32 ISE:         1;   // Bits 14
        TQP_UINT32 ZTG:         1;   // Bits 15
        TQP_UINT32 RSV2:        2;   // Bits 17 - 16, reserved
        TQP_UINT32 FLR:         1;   // Bits 18
        TQP_UINT32 ISS:         1;   // Bits 19
        TQP_UINT32 FRMALN:      2;   // Bits 21 - 20
        TQP_UINT32 HS:          1;   // Bits 22
        TQP_UINT32 HR:          1;   // Bits 23
        TQP_UINT32 FRMRT:       3;   // Bits 26 - 24
        TQP_UINT32 FRMSZ:       3;   // Bits 29 - 27
        TQP_UINT32 CAMMD:       2;   // Bits 31 - 30

        // P4 - P7
        TQP_UINT32 EVIDX:       3;   // Bits 3 - 0
        TQP_UINT32 AESP:        1;   // Bits 4
        TQP_UINT32 AELCK:       1;   // Bits 5
        TQP_UINT32 BLC:         2;   // Bits 7 - 6
        TQP_UINT32 METMD:       2;   // Bits 9 - 8
        TQP_UINT32 ISOIDX:      3;   // Bits 12 -10
        TQP_UINT32 ISOMD:       1;   // Bits 13
        TQP_UINT32 AEMD:        2;   // Bits 15 - 14
        TQP_UINT32 TVIDX:       5;   // Bits 20 - 16
        TQP_UINT32 AVIDX:       3;   // Bits 23 - 21
        TQP_UINT32 HMR:         1;   // Bits 24
        TQP_UINT32 VMR:         1;   // Bits 25
        TQP_UINT32 RSV3:        6;   // Bits 31 - 26, reserved

        // P8 - P11
        TQP_UINT32 ZOOMPOS:    10;   // Bits 9 - 0
        TQP_UINT32 WBCS:        2;   // Bits 11 - 10
        TQP_UINT32 FCSPOS:      4;   // Bits 15 - 12
        TQP_UINT32 AFWDIDX:     4;   // Bits 19 - 16
        TQP_UINT32 AFWD:        1;   // Bits 20
        TQP_UINT32 FMD:         1;   // Bits 21
        TQP_UINT32 AFRNG:       2;   // Bits 23 - 22
        TQP_UINT32 RSV4:        2;   // Bits 25 - 24, reserved
        TQP_UINT32 WBMDIDX:     4;   // Bits 29 - 26
        TQP_UINT32 WBMD:        2;   // Bits 31 - 30

        // P12 - P15
        TQP_UINT32 NR:          2;   // Bits 1 - 0
        TQP_UINT32 RSV5:        1;   // Bits 2, reserved
        TQP_UINT32 SCNIDX:      4;   // Bits 6 - 3
        TQP_UINT32 SCNMD:       1;   // Bits 7
        TQP_UINT32 FOCALLEN:   24;   // Bits 31 - 8

        // P16 - P19
        TQP_UINT32 RSV6:       16;   // Bits 15 - 0, reserved
        TQP_UINT32 EFFCT:       4;   // Bits 19 - 16
        TQP_UINT32 CRMCB:       2;   // Bits 21 - 20
        TQP_UINT32 CRMCR:       2;   // Bits 23 - 22
        TQP_UINT32 HUECB:       3;   // Bits 26 - 24
        TQP_UINT32 HUECR:       3;   // Bits 29 - 27
        TQP_UINT32 SHARP:       2;   // Bits 31 - 30

        // P20 - P23
        TQP_UINT8  EV_i;   // P20
        TQP_UINT8  EV_s;   // P21
        TQP_UINT8  TV_i;   // P22
        TQP_UINT8  TV_s;   // P23

        // P24 - P27
        TQP_UINT8  AV_i;   // P24
        TQP_UINT8  AV_s;   // P25
        TQP_UINT8  SV_i;   // P26
        TQP_UINT8  SV_s;   // P27

        // P28 - P31
        TQP_UINT32 RSV7;
    } stDVCMode;
    struct   // DSC mode (only different on P0 - P3)
    {
        // P0 - P3
        TQP_UINT32 AFS:         1;   // Bits 0
        TQP_UINT32 FLCK:        1;   // Bits 1
        TQP_UINT32 RSV:         5;   // Bits 6 - 2, reserved
        TQP_UINT32 AFRSLT:      9;   // Bits 15 - 7
        TQP_UINT32 ISWRN:       1;   // Bits 16
        TQP_UINT32 CNTSPD:      2;   // Bits 18 - 17
        TQP_UINT32 STLDRV:      2;   // Bits 20 - 19
        TQP_UINT32 AEBC:        1;   // Bits 21
        TQP_UINT32 HS:          1;   // Bits 22
        TQP_UINT32 PRT:         1;   // Bits 23
        TQP_UINT32 AEBST:       3;   // Bits 26 - 24
        TQP_UINT32 STLSZ:       3;   // Bits 29 - 27
        TQP_UINT32 CAMMD:       2;   // Bits 31 - 30

        // P4 - P7
        TQP_UINT32 EVIDX:       3;   // Bits 3 - 0
        TQP_UINT32 AESP:        1;   // Bits 4
        TQP_UINT32 AELCK:       1;   // Bits 5
        TQP_UINT32 BLC:         2;   // Bits 7 - 6
        TQP_UINT32 METMD:       2;   // Bits 9 - 8
        TQP_UINT32 ISOIDX:      3;   // Bits 12 -10
        TQP_UINT32 ISOMD:       1;   // Bits 13
        TQP_UINT32 AEMD:        2;   // Bits 15 - 14
        TQP_UINT32 TVIDX:       5;   // Bits 20 - 16
        TQP_UINT32 AVIDX:       3;   // Bits 23 - 21
        TQP_UINT32 HMR:         1;   // Bits 24
        TQP_UINT32 VMR:         1;   // Bits 25
        TQP_UINT32 RSV3:        6;   // Bits 31 - 26, reserved

        // P8 - P11
        TQP_UINT32 ZOOMPOS:    10;   // Bits 9 - 0
        TQP_UINT32 WBCS:        2;   // Bits 11 - 10
        TQP_UINT32 FCSPOS:      4;   // Bits 15 - 12
        TQP_UINT32 AFWDIDX:     4;   // Bits 19 - 16
        TQP_UINT32 AFWD:        1;   // Bits 20
        TQP_UINT32 FMD:         1;   // Bits 21
        TQP_UINT32 AFRNG:       2;   // Bits 23 - 22
        TQP_UINT32 RSV4:        2;   // Bits 25 - 24, reserved
        TQP_UINT32 WBMDIDX:     4;   // Bits 29 - 26
        TQP_UINT32 WBMD:        2;   // Bits 31 - 30

        // P12 - P15
        TQP_UINT32 NR:          2;   // Bits 1 - 0
        TQP_UINT32 RSV5:        1;   // Bits 2, reserved
        TQP_UINT32 SCNIDX:      4;   // Bits 6 - 3
        TQP_UINT32 SCNMD:       1;   // Bits 7
        TQP_UINT32 FOCALLEN:   24;   // Bits 31 - 8

        // P16 - P19
        TQP_UINT32 RSV6:       16;   // Bits 15 - 0, reserved
        TQP_UINT32 EFFCT:       4;   // Bits 19 - 16
        TQP_UINT32 CRMCB:       2;   // Bits 21 - 20
        TQP_UINT32 CRMCR:       2;   // Bits 23 - 22
        TQP_UINT32 HUECB:       3;   // Bits 26 - 24
        TQP_UINT32 HUECR:       3;   // Bits 29 - 27
        TQP_UINT32 SHARP:       2;   // Bits 31 - 30

        // P20 - P23
        TQP_UINT8  EV_i;   // P20
        TQP_UINT8  EV_s;   // P21
        TQP_UINT8  TV_i;   // P22
        TQP_UINT8  TV_s;   // P23

        // P24 - P27
        TQP_UINT8  AV_i;   // P24
        TQP_UINT8  AV_s;   // P25
        TQP_UINT8  SV_i;   // P26
        TQP_UINT8  SV_s;   // P27

        // P28 - P31
        TQP_UINT32 RSV7;
    } stDSCMode;
} PE1005_COMMAND_PARAMETER, *PPE1005_COMMAND_PARAMETER;


// command packet
typedef union _PE1005_PACKET
{
    TQP_UINT8                     Read[44];
    struct
    {
        PE1005_PACKET_HEADER      stPakHeader;         // 2 bytes
        TQP_UINT8                 Reserved1;           // 1 bytes, byte swap
        TQP_UINT8                 ucRetry;             // 1 bytes, byte swap

        PE1005_COMMAND_HEADER     stCmdHeader;         // 2 bytes
        TQP_UINT16                usCmdPrmSize;        // 2 bytes
        PE1005_COMMAND_PARAMETER  stCmdPrm;            // 32 bytes

        TQP_UINT16                Reserved2;           // 2 bytes
        TQP_UINT16                usPakChkSum;         // 2 bytes
    } stPakAll;
} PE1005_PACKET, *PPE1005_PACKET;


/////////////////////////////////////////////////////////////////////////////
//
//

typedef struct CPE1005 
{
    // public CVideoDecoder
    //
    CVideoDecoder  m_VideoDecoder;

    // CPE1005
    II2C           *m_pI2C;         // I2C interface
    TQP_UINT8      m_cI2CDevAddr;   // i2c device address

    // public PE1005_PACKET
    PE1005_PACKET  m_CmdPacket;
    PE1005_PACKET  m_ResPacket;

    // public PCA9535_GPO
    TQP_UINT8      m_PCA9535Gout0;
    TQP_UINT8      m_PCA9535Gout1;
} CPE1005;


CPE1005 * CPE1005_Constructor(CPE1005 *pPE1005, 
							  CObject *pParent,
							  TQP_UINT32 dwAttributes,
							  ITuner *pTuner,
							  ITVAudio *pTvAudio,
							  II2C *pI2C,
							  TQP_UINT8 cI2CDevAddr
							  );
void CPE1005_Destructor(CPE1005 *This);

#ifdef __cplusplus
}
#endif

#endif //__CPE1005_H__


