///////////////////////////////////////////////////////////////////////////////
//        
//
// Csaa7115.c
//
// Description: 
//
//  SAA 7115 implementation
// 
// History:
//	Sep 5, 2006	(dt) created
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


#include "csaa7115.h"
#include "qpdbgmgr.h"
#include "util.h"


extern TQP_DbgMCompID	g_QPPeripheralDBGCompID;


/////////////////////////////////////////////////////////////////////////////
//
//

#define SAA7115_STATUS_MASK_ERROR		0x80000000


/////////////////////////////////////////////////////////////////////////////
//
//


static EQP_Errors CSaa7115_readRegister(CSaa7115 *This,
										TQP_UINT8 regAddr, 
										PTQP_UINT8 pValue
										)
{
    return (II2C_WriteThenRead(This->m_pI2C,
							   This->m_cI2CDevAddr,	//device address
							   1,					//write size
							   &regAddr,			//write buffer
							   1,					//read size
							   pValue				//read buffer
							   ));
}



static EQP_Errors CSaa7115_writeRegister(CSaa7115 *This,
										 TQP_UINT8 regAddr, 
										 TQP_UINT8 value
										 )
{
    TQP_UINT8 buffer[2];

	buffer[0] = regAddr;
	buffer[1] = value;

    return (II2C_Write(This->m_pI2C, 
					   This->m_cI2CDevAddr,	//device address
					   2,					//write size
					   buffer				//write buffer
					   ));
}


static EQP_Errors CSaa7115_WriteBlock(CSaa7115 *This,
									  TQP_UINT32 *pRegTable
									  )
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;
	TQP_UINT32	dwTotalReg = 0;

	// get total number of registers to write
	//
	while (-1 != pRegTable[dwTotalReg]) 
	{
		dwTotalReg++;
	}

	dwTotalReg >>= 1;

	// do multiple writes
	//
	while (dwTotalReg)
	{
		qpStatus = CSaa7115_writeRegister(This,
										  (TQP_UINT8)pRegTable[0], 
										  (TQP_UINT8)pRegTable[1]
										  );
		if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CSaa7115_WriteBlock() ****** writeRegister failed! status(%d) ******\n",
						 qpStatus
						 ));
			break;
		}

		pRegTable += 2;
		dwTotalReg--;
	}

	return (qpStatus);
}



static EQP_Errors CSaa7115_GetStatus(CSaa7115 *This, 
									 TQP_UINT32 *pStatus
									 )
{
	TQP_UINT8	cRegData[2];
	TQP_UINT32	dwTemp;
	EQP_Errors	qpStatus;

	qpStatus = CSaa7115_readRegister(This,
									 SAA7115_REG_STATUS_BYTE_DECODER_1,
									 &cRegData[0]
									 );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CSaa7115_GetStatus() ****** readRegister failed to read SAA7115_STATUS_BYTE registers! ******\n" 
					 ));
		*pStatus = SAA7115_STATUS_MASK_ERROR;
		return (QPERR_FAIL);
	}

	qpStatus = CSaa7115_readRegister(This,
									 SAA7115_REG_STATUS_BYTE_DECODER_2,
									 &cRegData[1]
									 );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CSaa7115_GetStatus() ****** readRegister failed to read SAA7115_STATUS_BYTE registers! ******\n"
					 ));
		*pStatus = SAA7115_STATUS_MASK_ERROR;
		return (QPERR_FAIL);
	}

	dwTemp = (TQP_UINT32)cRegData[1] & 0xFF;
	dwTemp = dwTemp << 8;

	*pStatus = ((TQP_UINT32)cRegData[0] & 0xFF) | dwTemp;

#if 0
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7115_GetStatus() 0x%x\n", 
				 *pStatus
				 ));
#endif //0

	return (QPERR_SUCCESS);
}





/////////////////////////////////////////////////////////////////////////////
//
//


// IGenericDevice
//
static EQP_Errors CSaa7115_InitDevice(IVideoDecoder *This_p, 
									  QP_DEVICE_CALLBACK pFuncCallback,
									  TQP_PVOID context,
									  TQP_PVOID pInitData
									  )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CSaa7115		*This = GET_INHERITED_OBJECT(CSaa7115, m_VideoDecoder, pVideoDecoder);

	static TQP_UINT32 dw7115RegVals[] = 
	{
		0x01,0x48,
		0x02,0xc3,
		0x03,0x20,
		0x04,0x90,
		0x05,0x90,
		0x06,0xEB,
		0x07,0xE0,
#if 1
		0x08,0x68,	// no auto detect, VCR timing
#else
		0x08,0xb0,
#endif
		0x09,0x40,
		0x0A,0x80,
		0x0B,0x44,
		0x0C,0x40,
		0x0D,0x00,
		0x0e,0x07,
		0x0F,0x2a,
		0x10,0x06,
		0x11,0x00,
		0x12,0x00,
		0x13,0x00,
		0x14,0x01,
		0x15,0x01,
		0x16,0x12,
		0x17,0xd8,
		0x18,0x40,
		0x19,0x80,
		0x1A,0x77,
		0x1B,0x42,
		0x1C,0xA9,
		0x1D,0x01,
		0x40,0x40,
		0x41,0x00,
		0x42,0x00,
		0x43,0x00,
		0x44,0x00,
		0x45,0x00,
		0x46,0x00,
		0x47,0x00,
		0x48,0x00,
		0x49,0x00,
		0x4a,0x00,
		0x4b,0x00,
		0x4c,0x00,
		0x4d,0x00,
		0x4e,0x00,
		0x4f,0x00,
		0x50,0x00,
		0x51,0x00,
		0x52,0x00,
		0x53,0x00,
		0x54,0x4f,
		0x55,0x00,
		0x56,0x00,
		0x57,0x00,
		0x58,0x00,
		0x59,0x00,
		0x5a,0x06,
		0x5b,0x83,
		0x5c,0x00,
		0x5d,0x0c,
		0x5e,0x00,
		0x5f,0x00,
		0x83,0x31,
		0x88,0x8a,
		0x30,0xbc,
		0x31,0xdf,
		0x32,0x02,
		0x33,0x00,
		0x34,0xcd,
		0x35,0xcc,
		0x36,0x3a,
		0x37,0x00,
		0x38,0x03,
		0x39,0x10,
		0x3a,0x00,
		0x3b,0x00,
		0x3c,0x00,
		0x3d,0x00,
		0x3e,0x00,
		0x3f,0x00,
		0xF0,0xAD,
		0xF1,0x05,
		0xF5,0xAD,
		0xF6,0x01,
		0xF2,0x50,
		0xF3,0x46,
		0xF4,0x00,
		0xF7,0x4B,
		0xF8,0x00,
		0xF9,0x4B,
		0xFA,0x00,
		0xFB,0x4B,
		0xFC,0x00,
		0xFD,0x00,
		0xFE,0x00,
		0xFF,0x88,
		0xCC,0xD0,
		0xCD,0x02,
		0xD0,0x01,
		0xD1,0x00,
		0xD2,0x00,
		0xD4,0x80,
		0xD5,0x40,
		0xD6,0x40,
		0xD8,0x00,
		0xD9,0x04,
		0xDA,0x00,
		0xDC,0x00,
		0xDD,0x02,
		0xDE,0x00,
		0xCE,0xFD,
		0xCF,0x00,
		0xE0,0x00,
		0xE1,0x04,
		0xE2,0x00,
		0xE3,0x04,
		0xE4,0x01,
		0xE8,0x00,
		0xE9,0x00,
		0xEA,0x00,
		0xEB,0x00,
		0xEC,0x00,
		0xED,0x00,
		0xEE,0x00,
		0x80,0x00,
		0x88,0x0B,
		0xC0,0x00,
		0xC1,0x08,
		0xC2,0x00,
		0xC3,0x80,
		0xC4,0x02,
		0xC5,0x00,
		0xC6,0xD0,
		0xC7,0x02,
		0xC8,0x14,
		0xC9,0x00,
		0xCA,0xFD,
		0xCB,0x00,
		0x81,0x01,
		0x82,0x00,
		0x83,0x31,
		0x84,0x20,
		0x85,0x21,
		0x86,0xC5,
		0x87,0x00,
		0x88,0x0B,
		0x80,0x20,
		0x88,0xf0,
		0x87,0x00,
		-1,-1
	};

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7115_InitDevice()\n"
				 ));

	if (CObject_IsInitialized(&pVideoDecoder->m_Object))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
					 "CSaa7115_InitDevice() Already Initialized. Skipping...\n" 
					 ));
		return (QPERR_SUCCESS);
	}

	// save callback address and context
	//
	pVideoDecoder->m_pDeviceCallback = pFuncCallback;
	pVideoDecoder->m_callbackContext = context;


	// Initialize 7115 with the default value
	//
	CSaa7115_WriteBlock(This,
						dw7115RegVals
						);

	// program the video decoder with settings read from the registry
	//
	QPVideoDecoder_SetVideoSystem(This_p, 
								  pVideoDecoder->m_VidStandard
								  );
	QPVideoDecoder_SelectVideoSource(This_p, 
									 pVideoDecoder->m_VidSource
									 );

	QPVideoDecoder_SetVcrTiming(This_p, 
								pVideoDecoder->m_VCRTiming
								);

	QPVideoDecoder_SetBrightness(This_p, 
								 pVideoDecoder->m_nBrightness
								 );
	QPVideoDecoder_SetContrast(This_p, 
							   pVideoDecoder->m_nContrast
							   );
	QPVideoDecoder_SetSaturation(This_p, 
								 pVideoDecoder->m_nSaturation
								 );
	QPVideoDecoder_SetHue(This_p, 
						  pVideoDecoder->m_nHue
						  );
	QPVideoDecoder_SetSharpness(This_p, 
								pVideoDecoder->m_nSharpness
								);

	// CObject Init
	//
	CObject_Init(&pVideoDecoder->m_Object);

	return (QPERR_SUCCESS);
}



// IVideoDecoder
//
static EQP_Errors CSaa7115_SelectVideoSource(IVideoDecoder *This_p,
											 VIDEO_SOURCE source
											 )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CSaa7115		*This = GET_INHERITED_OBJECT(CSaa7115, m_VideoDecoder, pVideoDecoder);

	SAA7115_ANALOG_INPUT_CONTROL_1	inputControl;

	// do not reproram default contrast, saturation and hue.
	//
    TQP_UINT32	dwRegValsSVideo[] = 
	{
		0x02,0xC6,
		0x09,0x80,
		-1,-1
	};

    TQP_UINT32	dwRegValsComposite[] = 
	{
		0x02,0xC3,
		0x09,0x40,
		-1,-1
	};

    TQP_UINT32 dwRegValsTuner[] = 
	{
		0x02,0xC3,
		0x09,0x40,
		-1,-1
	};

    TQP_UINT32 dwRegValsInvalid[] = 
	{
		0x09,0x53,	
		-1,-1
    };
	TQP_UINT32	*pVals;
	EQP_Errors	qpStatus;


	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7115_SelectVideoSource()\n"
				 ));

	inputControl.Read = 0;

	inputControl.Reg.FUSE = 3;

	// Map source to decoder mode
	//
	switch (source)
	{
		case VIDEO_SOURCE_CVIDEO1:
			inputControl.Reg.MODE = This->m_CVideoMode;
			pVals = dwRegValsComposite;
			break;

		case VIDEO_SOURCE_SVIDEO1:
			inputControl.Reg.MODE = This->m_SVideoMode;
			pVals = dwRegValsSVideo;
			break;

		case VIDEO_SOURCE_TVTUNER:
			inputControl.Reg.MODE = This->m_TVVideoMode;
			pVals = dwRegValsTuner;
			break;

		case VIDEO_SOURCE_INVALID:
			// set to a non-existent source
			inputControl.Reg.MODE = This->m_InvalidVideoMode;
			pVals = dwRegValsInvalid;
			break;

		default:
			return (QPERR_NOTIMPL);
	}


	qpStatus = CSaa7115_writeRegister(This,
									  SAA7115_REG_ANALOG_INPUT_CONTROL_1, 
									  inputControl.Read
									  );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CSaa7115_SelectVideoSource() ****** writeRegister() failed to write Saa7115 register! ******\n"
					 ));
		return (QPERR_FAIL);
	}


	qpStatus = !QP_SUCCESS(qpStatus) ? qpStatus : CSaa7115_WriteBlock(This, 
																	  pVals
																	  );

	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CSaa7115_SelectVideoSource() ****** WriteBlock() failed to write Saa7115 registers! ******\n" 
					 ));
		return (QPERR_FAIL);
	}

	pVideoDecoder->m_VidSource = source;

	return (QPERR_SUCCESS);
}



static EQP_Errors CSaa7115_SetVideoSystem(IVideoDecoder *This_p,
										  VIDEO_STANDARD standard
										  )
{
	TQP_UINT32	dwRegValsPAL[] = 
	{
		0x08,0x28,
		-1, -1
	};
	TQP_UINT32	dwRegValsSECAM[] = 
	{
		0x08,0x28,
		-1, -1
	};
	TQP_UINT32	dwRegValsNTSC[] = 
	{
		0x08,0x68,
		-1, -1
    };

	TQP_UINT32		*pVals;
	EQP_Errors		qpStatus;
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CSaa7115		*This = GET_INHERITED_OBJECT(CSaa7115, m_VideoDecoder, pVideoDecoder);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7115_SetVideoSystem() standard(%d)\n",
				 standard
				 ));

	if (VIDEO_STANDARD_NTSC == standard)
		pVals = dwRegValsNTSC;
	else if (VIDEO_STANDARD_PAL == standard)
		pVals = dwRegValsPAL;
	else
		pVals = dwRegValsSECAM;

	qpStatus = CSaa7115_WriteBlock(This,
								   pVals
								   );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CSaa7115_SetVideoSystem() ****** WriteBlock() failed to write Saa7115 registers! ******\n"
					 ));
		return (QPERR_FAIL);
	}

	pVideoDecoder->m_VidStandard = standard;

	return (QPERR_SUCCESS);
}



static EQP_Errors CSaa7115_SetVideoRes(IVideoDecoder *This_p,
									   TQP_UINT32 dwWid,
				  					   TQP_UINT32 dwHgt,
				  					   TQP_UINT32 dwScanFmt,
				  					   TQP_UINT32 dwFps
				  					   )
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7115_SetVideoRes() : Video Res Set to %dx %d\n",
				 dwWid,
				 dwHgt
				 ));

	return (QPERR_SUCCESS);
}



static EQP_Errors CSaa7115_SetVcrTiming(IVideoDecoder *This_p,
										TQP_UINT32 enable
										)
{
	SAA7115_SYNC_CONTROL	syncControl;
	EQP_Errors				qpStatus;

	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CSaa7115		*This = GET_INHERITED_OBJECT(CSaa7115, m_VideoDecoder, pVideoDecoder);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7115_SetVcrTiming() %d\n", 
				 enable
				 ));

	// Get Current Settings
	//
	qpStatus = CSaa7115_readRegister(This,
									 SAA7115_REG_SYNC_CONTROL, 
									 &syncControl.Read
									 );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CSaa7115_SetVcrTiming() ****** readRegister failed! ******\n" 
					 ));
		return (QPERR_FAIL);
	}

	// Set Horizontal time constant
	//
	syncControl.Reg.HTC = enable ? 1 : 3;


	qpStatus = CSaa7115_writeRegister(This,
									  SAA7115_REG_SYNC_CONTROL, 
									  syncControl.Read
									  );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CSaa7115_SetVcrTiming() ****** writeRegister() failed to write Saa7115 registers! ******\n" 
					 ));
		return (QPERR_FAIL);
	}

	pVideoDecoder->m_VCRTiming = enable;

	return (QPERR_SUCCESS);
}



static EQP_Errors CSaa7115_SetOutputEnabled(IVideoDecoder *This_p,
											TQP_UINT32 enable
											)
{
	SAA7115_GLOBAL_XPORT_DELAY_ENABLE	xportEnable;
	EQP_Errors							qpStatus;
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CSaa7115		*This = GET_INHERITED_OBJECT(CSaa7115, m_VideoDecoder, pVideoDecoder);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7115_SetOutputEnabled() %d\n", 
				 enable
				 ));

	// Get Current Settings
	//
	qpStatus = CSaa7115_readRegister(This,
									 SAA7115_REG_GLOBAL_XPORT_DELAY_ENABLE, 
									 &xportEnable.Read
									 );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CSaa7115_SetOutputEnabled() ****** readRegister failed! ******\n"
					 ));
		return (QPERR_FAIL);
	}

	// Set x-port enable
	//
	xportEnable.Reg.XPE = enable ? 1 : 0;

	qpStatus = CSaa7115_writeRegister(This,
									  SAA7115_REG_GLOBAL_XPORT_DELAY_ENABLE, 
									  xportEnable.Read
									  );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CSaa7115_SetOutputEnabled() ****** writeRegister() failed to write Saa7115 registers! ******\n"
					 ));
		return (QPERR_FAIL);
	}

	pVideoDecoder->m_OutputEnable = enable; 
	return (QPERR_SUCCESS);
}


// Set Decoder Brightness
static EQP_Errors CSaa7115_SetBrightness(IVideoDecoder *This_p,
										 int nParamValue
										 )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CSaa7115		*This = GET_INHERITED_OBJECT(CSaa7115, m_VideoDecoder, pVideoDecoder);

	if (!pVideoDecoder->m_bMuted)
	{
		EQP_Errors	qpStatus;

		qpStatus = CSaa7115_writeRegister(This,
										  SAA7115_REG_LUMINANCE_BRIGHTNESS, 
										  (TQP_UINT8)nParamValue
										  );
		if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CSaa7115_SetBrightness() ****** writeRegister() failed to write Saa7115 registers! ******\n"
						 ));
			return (QPERR_FAIL);
		}

		pVideoDecoder->m_nBrightness = nParamValue;
	}

	return (QPERR_SUCCESS);
}



// Set Decoder Contrast
static EQP_Errors CSaa7115_SetContrast(IVideoDecoder *This_p,
									   int nParamValue
									   )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CSaa7115		*This = GET_INHERITED_OBJECT(CSaa7115, m_VideoDecoder, pVideoDecoder);

	if (!pVideoDecoder->m_bMuted)
	{
		EQP_Errors	qpStatus;

		qpStatus = CSaa7115_writeRegister(This,
										  SAA7115_REG_LUMINANCE_CONTRAST, 
										  (TQP_UINT8)nParamValue
										  );
		if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CSaa7115_SetContrast() ****** writeRegister() failed to write Saa7115 registers! ******\n"
						 ));
			return (QPERR_FAIL);
		}

		pVideoDecoder->m_nContrast = nParamValue;
	}

	return (QPERR_SUCCESS);
}



// Set Decoder Saturation
static EQP_Errors CSaa7115_SetSaturation(IVideoDecoder *This_p,
										 int nParamValue
										 )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CSaa7115		*This = GET_INHERITED_OBJECT(CSaa7115, m_VideoDecoder, pVideoDecoder);

	if (!pVideoDecoder->m_bMuted)
	{
		EQP_Errors	qpStatus;

		qpStatus = CSaa7115_writeRegister(This,
										  SAA7115_REG_CHROMA_SATURATION, 
										  (TQP_UINT8)nParamValue
										  );
		if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CSaa7115_SetSaturation() ****** writeRegister() failed to write Saa7115 registers! ******\n"
						 ));
			return (QPERR_FAIL);
		}

		pVideoDecoder->m_nSaturation = nParamValue;
	}

	return (QPERR_SUCCESS);
}



// Set Hue Saturation
static EQP_Errors CSaa7115_SetHue(IVideoDecoder *This_p,
								  int nParamValue
								  )
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CSaa7115		*This = GET_INHERITED_OBJECT(CSaa7115, m_VideoDecoder, pVideoDecoder);

	if (!pVideoDecoder->m_bMuted)
	{
		EQP_Errors	qpStatus;

		qpStatus = CSaa7115_writeRegister(This,
										  SAA7115_REG_CHROMA_HUE, 
										  (TQP_UINT8)nParamValue
										  );
		if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CSaa7115_SetHue() ****** writeRegister() failed to write Saa7115 registers! ******\n"
						 ));
			return (QPERR_FAIL);
		}

		pVideoDecoder->m_nHue = nParamValue;
	}

	return (QPERR_SUCCESS);
}



int sharpnessToKsTable[16] = 
{ //0  1   2   3   4   5   6   7  8  9  10 11 12 13 14 15   
	8, 15, 14, 13, 12, 11, 10, 9, 7, 6, 5, 4, 3, 2, 1, 0  
};


int sharpnessTable[16] = 
{ // 0   1   2   3   4   5  6  7  8  9 10 11 12 13 14 15   
	15, 14, 13, 12, 11, 10, 9, 8, 0, 7, 6, 5, 4, 3, 2, 1
};



// Set Sharpness
static EQP_Errors CSaa7115_SetSharpness(IVideoDecoder *This_p,
										int nParamValue
										)
{
	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CSaa7115		*This = GET_INHERITED_OBJECT(CSaa7115, m_VideoDecoder, pVideoDecoder);

	if (!pVideoDecoder->m_bMuted)
	{
		SAA7115_LUMINANCE_CONTROL	sharpness;
		EQP_Errors					qpStatus;


		// Get Current Settings
		//
		qpStatus = CSaa7115_readRegister(This,
										 SAA7115_REG_LUMINANCE_CONTROL, 
										 &sharpness.Read
										 );
		if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CSaa7115_SetSharpness() ****** readRegister failed! ******\n"
						 ));
			return (QPERR_FAIL);
		}

		sharpness.Reg.LUFI = (TQP_UINT8)sharpnessTable[nParamValue];


		qpStatus = CSaa7115_writeRegister(This,
										  SAA7115_REG_LUMINANCE_CONTROL, 
										  sharpness.Read
										  );
		if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CSaa7115_SetSharpness() ****** writeRegister() failed to write Saa7115 registers! ******\n"
						 ));
			return (QPERR_FAIL);
		}

		pVideoDecoder->m_nSharpness = nParamValue;
	}

	return (QPERR_SUCCESS);
}




static EQP_Errors CSaa7115_DetectCopyProtection(IVideoDecoder *This_p,
												TQP_BOOL *pCopyProtected,
												PTQP_UINT8 pStatus
												)
{
	EQP_Errors					qpStatus;
	TQP_UINT32					dwStatus;
	SAA7115_STATUS_BYTE_DECODER	decoderStatus;

	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CSaa7115		*This = GET_INHERITED_OBJECT(CSaa7115, m_VideoDecoder, pVideoDecoder);


	if (!pCopyProtected)
	{
		return (QPERR_PARMS);
	}

	qpStatus = CSaa7115_GetStatus(This,
								  &dwStatus
								  );

	if (!QP_SUCCESS(qpStatus))
		return (qpStatus);

	if (dwStatus != SAA7115_STATUS_MASK_ERROR)
	{
		decoderStatus.Read = (TQP_UINT16)dwStatus;

		if (!decoderStatus.Reg.HLVLN && 
			decoderStatus.Reg.INTL && 
			decoderStatus.Reg.COPRO
			)
		{
			*pCopyProtected = QP_TRUE;
		}
		else
		{
			*pCopyProtected = QP_FALSE;
		}

		return (QPERR_SUCCESS);
	}

	return (QPERR_FAIL);
}



static EQP_Errors CSaa7115_DetectVideoStandard(IVideoDecoder *This_p,
											   PVIDEO_STANDARD pStandard,
											   PTQP_UINT32 pulSignalLocked,
											   PTQP_UINT32 pulNumberOfLines
											   )
{
	EQP_Errors					qpStatus;
	TQP_UINT32					dwStatus;
	SAA7115_STATUS_BYTE_DECODER	decoderStatus;

	CVideoDecoder	*pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
	CSaa7115		*This = GET_INHERITED_OBJECT(CSaa7115, m_VideoDecoder, pVideoDecoder);

	*pulSignalLocked = 0;
	*pulNumberOfLines = 0;
	*pStandard = pVideoDecoder->m_VidStandard;

	qpStatus = CSaa7115_GetStatus(This,
								  &dwStatus
								  );
	if (!QP_SUCCESS(qpStatus))
		return (qpStatus);

	if (dwStatus != SAA7115_STATUS_MASK_ERROR)
	{
		decoderStatus.Read = (TQP_UINT16)dwStatus;

		if (!decoderStatus.Reg.HLVLN)
		{
			*pulSignalLocked = 1;

			if (decoderStatus.Reg.INTL)
			{
				if (decoderStatus.Reg.FIDT)
				{
					*pulNumberOfLines = 525;
				}
				else
				{
					*pulNumberOfLines = 625;
				}
			}
		}

		switch (decoderStatus.Reg.DCSTD)
		{
			case 1:
				*pStandard = VIDEO_STANDARD_NTSC;
				break;

			case 2:
				*pStandard = VIDEO_STANDARD_PAL;
				break;

			case 3:
				*pStandard = VIDEO_STANDARD_SECAM;
				break;

			case 0:
				*pStandard = VIDEO_STANDARD_INVALID;
				break;

			default:
				break;
		}
	}

	return (QPERR_SUCCESS);
}



// constructor
//
CSaa7115 * CSaa7115_Constructor(CSaa7115 *pSaa7115, 
								CObject *pParent,
								TQP_UINT32 dwAttributes,
								ITuner *pTuner,
								ITVAudio *pTvAudio,
								II2C *pI2C,
								TQP_UINT8 cI2CDevAddr
								)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7115_Constructor()\n"
				 ));

	if (pSaa7115)
	{
		CVideoDecoder	*pVideoDecoder;

		pVideoDecoder = CVideoDecoder_Constructor(&pSaa7115->m_VideoDecoder, 
												  pParent,
												  dwAttributes,
												  pTuner,
												  pTvAudio
												  );

		if (!pVideoDecoder)
		{
			pSaa7115 = QP_NULL;
		}
		else
		{
			// fill in the function table
			//

			// IGenericDevice
			pSaa7115->m_VideoDecoder.m_iVideoDecoder.InitDevice = CSaa7115_InitDevice;


			// IVideoDecoder
			pSaa7115->m_VideoDecoder.m_iVideoDecoder.SelectVideoSource = CSaa7115_SelectVideoSource;
			pSaa7115->m_VideoDecoder.m_iVideoDecoder.SetVideoSystem = CSaa7115_SetVideoSystem;
			pSaa7115->m_VideoDecoder.m_iVideoDecoder.SetVideoRes = CSaa7115_SetVideoRes;
			pSaa7115->m_VideoDecoder.m_iVideoDecoder.SetVcrTiming = CSaa7115_SetVcrTiming;
			pSaa7115->m_VideoDecoder.m_iVideoDecoder.SetOutputEnabled = CSaa7115_SetOutputEnabled;
			pSaa7115->m_VideoDecoder.m_iVideoDecoder.SetBrightness = CSaa7115_SetBrightness;
			pSaa7115->m_VideoDecoder.m_iVideoDecoder.SetContrast = CSaa7115_SetContrast;
			pSaa7115->m_VideoDecoder.m_iVideoDecoder.SetSaturation = CSaa7115_SetSaturation;
			pSaa7115->m_VideoDecoder.m_iVideoDecoder.SetHue = CSaa7115_SetHue;
			pSaa7115->m_VideoDecoder.m_iVideoDecoder.SetSharpness = CSaa7115_SetSharpness;
			pSaa7115->m_VideoDecoder.m_iVideoDecoder.DetectCopyProtection = CSaa7115_DetectCopyProtection;
			pSaa7115->m_VideoDecoder.m_iVideoDecoder.DetectVideoStandard = CSaa7115_DetectVideoStandard;

			// initialize members
			//
			pSaa7115->m_pI2C = pI2C;
			pSaa7115->m_cI2CDevAddr = cI2CDevAddr;
			pSaa7115->m_SVideoMode = SAA7115_INPUT_MODE_8;
			pSaa7115->m_CVideoMode = SAA7115_INPUT_MODE_3;
			pSaa7115->m_TVVideoMode = SAA7115_INPUT_MODE_4;
			pSaa7115->m_InvalidVideoMode = SAA7115_INPUT_MODE_1;

			// inherited members
			//
			pSaa7115->m_VideoDecoder.m_nBrightnessDef = SAA7115_BRIGHTNESS_DEFAULT;
			pSaa7115->m_VideoDecoder.m_nBrightnessMin = SAA7115_BRIGHTNESS_MIN;
			pSaa7115->m_VideoDecoder.m_nBrightnessMax = SAA7115_BRIGHTNESS_MAX;
			pSaa7115->m_VideoDecoder.m_nBrightness = SAA7115_BRIGHTNESS_DEFAULT;
			pSaa7115->m_VideoDecoder.m_nContrastDef = SAA7115_CONTRAST_DEFAULT; 
			pSaa7115->m_VideoDecoder.m_nContrastMin = SAA7115_CONTRAST_MIN;
			pSaa7115->m_VideoDecoder.m_nContrastMax = SAA7115_CONTRAST_MAX;
			pSaa7115->m_VideoDecoder.m_nContrast = SAA7115_CONTRAST_DEFAULT;
			pSaa7115->m_VideoDecoder.m_nSaturationDef = SAA7115_SATURATION_DEFAULT;
			pSaa7115->m_VideoDecoder.m_nSaturationMin = SAA7115_SATURATION_MIN;
			pSaa7115->m_VideoDecoder.m_nSaturationMax = SAA7115_SATURATION_MAX;
			pSaa7115->m_VideoDecoder.m_nSaturation = SAA7115_SATURATION_DEFAULT;
			pSaa7115->m_VideoDecoder.m_nHueDef = SAA7115_HUE_DEFAULT;
			pSaa7115->m_VideoDecoder.m_nHueMin = SAA7115_HUE_MIN;
			pSaa7115->m_VideoDecoder.m_nHueMax = SAA7115_HUE_MAX;
			pSaa7115->m_VideoDecoder.m_nHue = SAA7115_HUE_DEFAULT;
			pSaa7115->m_VideoDecoder.m_nSharpnessDef = SAA7115_SHARPNESS_DEFAULT;
			pSaa7115->m_VideoDecoder.m_nSharpnessMin = SAA7115_SHARPNESS_MIN;
			pSaa7115->m_VideoDecoder.m_nSharpnessMax = SAA7115_SHARPNESS_MAX;
			pSaa7115->m_VideoDecoder.m_nSharpness = SAA7115_SHARPNESS_DEFAULT;
		}
	}

	return (pSaa7115);
}



// destructor
//
void CSaa7115_Destructor(CSaa7115 *This)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7115_Destructor()\n"
				 ));

	CVideoDecoder_Destructor(&This->m_VideoDecoder);
}

