///////////////////////////////////////////////////////////////////////////////
//        
//
// Csaa7128.c
//
// Description: 
//
//  SAA 7128 implementation
// 
// History:
//	Sep 5, 2006	(dt) created
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


#include "csaa7128.h"
#include "util.h"
#include "qpdbgmgr.h"

extern TQP_DbgMCompID	g_QPPeripheralDBGCompID;


/////////////////////////////////////////////////////////////////////////////
//
//

static EQP_Errors CSaa7128_readRegister(CSaa7128 *This, 
										TQP_UINT8 regAddr, 
										PTQP_UINT8 pValue
										)
{
    return (II2C_WriteThenRead(This->m_pI2C, 
							   This->m_cI2CDevAddr,	//device address
							   1,					//write size
							   &regAddr,			//write buffer
							   1,					//read size
							   pValue				//read buffer
							   ));
}



static EQP_Errors CSaa7128_writeRegister(CSaa7128 *This, 
										 TQP_UINT8 regAddr, 
										 TQP_UINT8 value
										 )
{
    TQP_UINT8 buffer[2];

	buffer[0] = regAddr;
	buffer[1] = value;

    return (II2C_Write(This->m_pI2C, 
					   This->m_cI2CDevAddr,	//device address
					   2,					//write size
					   buffer				//write buffer
					   ));
}



static EQP_Errors CSaa7128_WriteBlock(CSaa7128 *This, 
									  TQP_UINT32 *pRegTable
									  )
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;
	TQP_UINT32	dwTotalReg = 0;

	// get total number of registers to write
	//
	while (-1 != pRegTable[dwTotalReg]) 
	{
		dwTotalReg++;
	}

	dwTotalReg >>= 1;

	// do multiple writes
	//
	while (dwTotalReg)
	{
		qpStatus = CSaa7128_writeRegister(This,
										  (TQP_UINT8)pRegTable[0], 
										  (TQP_UINT8)pRegTable[1]
										  );
		if (!QP_SUCCESS(qpStatus))
			return (qpStatus);

		pRegTable += 2;
		dwTotalReg--;
	}

	return (qpStatus);
}



/////////////////////////////////////////////////////////////////////////////
//
//


// IGenericDevice
//
static EQP_Errors CSaa7128_InitDevice(IVideoEncoder *This_p, 
									  QP_DEVICE_CALLBACK pFuncCallback,
									  TQP_PVOID context,
									  TQP_PVOID pInitData
									  )
{
	PQPVIDENCODER_INITDATA	pVidEncInitData = (PQPVIDENCODER_INITDATA)pInitData;
	VIDEO_STANDARD			standard;
	CVideoEncoder			*pVideoEncoder = GET_INHERITED_OBJECT(CVideoEncoder, m_iVideoEncoder, This_p);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7128_InitDevice()\n"
				 ));

	if (CObject_IsInitialized(&pVideoEncoder->m_Object))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
					 "CSaa7128_InitDevice() Already Initialized. Skipping...\n"
					 ));
		return (QPERR_SUCCESS);
	}

	// save callback address and context
	//
	pVideoEncoder->m_pDeviceCallback = pFuncCallback;
	pVideoEncoder->m_callbackContext = context;

	// default standard
	//
	if (pVidEncInitData)
	{
		standard = pVidEncInitData->standard;

		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
					 "CSaa7128_InitDevice() input standard(%d), remembered standard(%d)\n",
					 standard, 
					 pVideoEncoder->m_VidStandard
					 ));
	}
	else
	{
		standard = pVideoEncoder->m_VidStandard;
	}

	IVideoEncoder_SetVideoStandard(This_p,
								   standard
								   );

	pVideoEncoder->m_fVideoEncoderEnabled = QP_TRUE;

	// CObject Init
	//
	CObject_Init(&pVideoEncoder->m_Object);

	return (QPERR_SUCCESS);
}



// IVideoEncoder
//

// Enables encoder device.
static EQP_Errors CSaa7128_Enable(IVideoEncoder *This_p)
{
	TQP_UINT8		ucReg;
	EQP_Errors		qpStatus;
	CVideoEncoder	*pVideoEncoder = GET_INHERITED_OBJECT(CVideoEncoder, m_iVideoEncoder, This_p);
	CSaa7128		*This = GET_INHERITED_OBJECT(CSaa7128, m_VideoEncoder, pVideoEncoder);


	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7128_Enable()\n"
				 ));

	if (VIDEO_STANDARD_NTSC == pVideoEncoder->m_VidStandard)
		ucReg = 0x05;
	else
		ucReg = 0x02;

	qpStatus = CSaa7128_writeRegister(This,
									  0x61, 
									  ucReg
									  );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CSaa7128_Enable() ****** writeRegister() failed to write Saa7128 registers! ******\n" 
					 ));
		return (QPERR_FAIL);
	}

	pVideoEncoder->m_fVideoEncoderEnabled = QP_TRUE;
	return (QPERR_SUCCESS);
}



// Disables encoder device.
static EQP_Errors CSaa7128_Disable(IVideoEncoder *This_p)
{
	TQP_UINT8		ucReg;
	EQP_Errors		qpStatus;
	CVideoEncoder	*pVideoEncoder = GET_INHERITED_OBJECT(CVideoEncoder, m_iVideoEncoder, This_p);
	CSaa7128		*This = GET_INHERITED_OBJECT(CSaa7128, m_VideoEncoder, pVideoEncoder);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7128_Disable()\n"
				 ));

	if (VIDEO_STANDARD_NTSC == pVideoEncoder->m_VidStandard)
		ucReg = 0x45;
	else
		ucReg = 0x42;

	qpStatus = CSaa7128_writeRegister(This,
									  0x61, 
									  ucReg
									  );
	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CSaa7128_Disable() ****** writeRegister failed to write Saa7128 registers! ******\n"
					 ));
		return (QPERR_FAIL);
	}

	pVideoEncoder->m_fVideoEncoderEnabled = QP_FALSE;
	return (QPERR_SUCCESS);
}



// Sets video system of the encoder.
//
static EQP_Errors CSaa7128_SetVideoStandard(IVideoEncoder *This_p, 
											VIDEO_STANDARD standard
											)
{
	static TQP_UINT32	dw7128RegValsNtsc[] = 
	{
		0x26,0x0d,
		0x27,0x00,
		0x28,0x19,
		0x29,0x1d,
		0x2a,0x77,
		0x2b,0x41,
		0x2c,0x88,
		0x2d,0x3f,
		0x2e,0x00,
		0x2f,0x00,
		0x30,0x00,
		0x31,0x00,
		0x32,0x00,
		0x33,0x00,
		0x34,0x00,
		0x35,0x00,
		0x36,0x00,
		0x37,0x00,
		0x38,0x00,
		0x39,0x00,
		0x3a,0x13,
		0x54,0x8c,
		0x5a,0x27,
		0x5b,0x88,
		0x5c,0xc0,
		0x5d,0x3f,
		0x5e,0x36,
		0x5f,0x36,
		0x60,0x00,
		0x61,0x05,
		0x62,0x4a,
		0x63,0x1f,
		0x64,0x7c,
		0x65,0xf0,
		0x66,0x21,
		0x67,0x77,
		0x68,0x41,
		0x69,0x88,
		0x6a,0x41,
		0x6b,0x12,
		0x6c,0xf2,
		0x6d,0x00,
		0x6e,0x90,
		0x6f,0x14,
		0x70,0x41,
		0x71,0xc3,
		0x72,0x00,
		0x73,0x3e,
		0x74,0xb8,
		0x75,0x03,
		0x76,0x15,
		0x77,0x16,
		0x78,0x15,
		0x79,0x16,
		0x7a,0x1a,
		0x7b,0x01,
		0x7c,0xc0,
		0x7d,0x00,
		0x7e,0x00,
		0x7f,0x00,
		-1,   -1
	};

	static TQP_UINT32	dw7128RegValsPal[] = 
	{
		0x26,0x0d,
		0x27,0x00,
		0x28,0x21,
		0x29,0x1d,
		0x2a,0x77,
		0x2b,0x41,
		0x2c,0x88,
		0x2d,0x3f,
		0x2e,0x00,
		0x2f,0x00,
		0x30,0x00,
		0x31,0x00,
		0x32,0x00,
		0x33,0x00,
		0x34,0x00,
		0x35,0x00,
		0x36,0x00,
		0x37,0x00,
		0x38,0x00,
		0x39,0x00,
		0x3a,0x13,
		0x54,0x8c,
		0x5a,0x3f,
		0x5b,0x7d,
		0x5c,0xaf,
		0x5d,0x23,
		0x5e,0x35,
		0x5f,0x35,
		0x60,0x00,
		0x61,0x02,
		0x62,0x2f,
		0x63,0xcb,
		0x64,0x8a,
		0x65,0x09,
		0x66,0x2a,
		0x67,0x77,
		0x68,0x41,
		0x69,0x88,
		0x6a,0x41,
		0x6b,0x12,
		0x6c,0xf2,
		0x6d,0x00,
		0x6e,0xa0,
		0x6f,0x00,
		0x70,0x41,
		0x71,0xc3,
		0x72,0x00,
		0x73,0x3e,
		0x74,0xb8,
		0x75,0x03,
		0x76,0x15,
		0x77,0x16,
		0x78,0x15,
		0x79,0x16,
		0x7a,0x1a,
		0x7b,0x01,
		0x7c,0xc0,
		0x7d,0x00,
		0x7e,0x00,
		0x7f,0x00,
		-1,   -1
	};

	TQP_UINT32		*pVals;
	EQP_Errors		qpStatus;
	CVideoEncoder	*pVideoEncoder = GET_INHERITED_OBJECT(CVideoEncoder, m_iVideoEncoder, This_p);
	CSaa7128		*This = GET_INHERITED_OBJECT(CSaa7128, m_VideoEncoder, pVideoEncoder);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7128_SetVideoStandard() standard(%d)\n",
				 standard
				 ));

	if (VIDEO_STANDARD_NTSC == standard)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
					 "CSaa7128_SetVideoStandard() NTSC\n"
					 ));
		pVals = dw7128RegValsNtsc;
	}
	else if (VIDEO_STANDARD_PAL == standard)
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
					 "CSaa7128_SetVideoStandard() PAL\n"
					 ));
		pVals = dw7128RegValsPal;
	}
	else
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CSaa7128_SetVideoStandard() Unknown Video Standard: %d \n", 
					 standard
					 ));
		return (QPERR_PARMS);
	}

	// save the video standard
	//
	pVideoEncoder->m_VidStandard = standard;

	// Initialize 7128 with the default value
	//
	qpStatus = CSaa7128_WriteBlock(This,
								   pVals
								   );

	if (!QP_SUCCESS(qpStatus))
	{
		QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
					 "CSaa7128_SetVideoStandard() WriteBlock Failed!\n"
					 ));
		return (QPERR_FAIL);
	}

	return (QPERR_SUCCESS);
}



// constructor
//
CSaa7128 * CSaa7128_Constructor(CSaa7128 *pSaa7128,
								CObject *pParent,
								TQP_UINT32 dwAttributes,
								II2C *pI2C,
								TQP_UINT8 cI2CDevAddr
								)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7128_Constructor()\n"
				 ));

	if (pSaa7128)
	{
		CVideoEncoder	*pVideoEncoder;

		pVideoEncoder = CVideoEncoder_Constructor(&pSaa7128->m_VideoEncoder, 
												  pParent,
												  dwAttributes
												  );
		if (!pVideoEncoder)
		{
			pSaa7128 = QP_NULL;
		}
		else
		{
			// fill in the function table
			//

			// IGenericDevice
			pSaa7128->m_VideoEncoder.m_iVideoEncoder.InitDevice = CSaa7128_InitDevice;

			// IVideoEncoder
			pSaa7128->m_VideoEncoder.m_iVideoEncoder.Enable = CSaa7128_Enable;
			pSaa7128->m_VideoEncoder.m_iVideoEncoder.Disable = CSaa7128_Disable;
			pSaa7128->m_VideoEncoder.m_iVideoEncoder.SetVideoStandard = CSaa7128_SetVideoStandard;

			// initialize members
			//
			pSaa7128->m_pI2C = pI2C;
			pSaa7128->m_cI2CDevAddr = cI2CDevAddr;
		}
	}

	return (pSaa7128);
}



// destructor
//
void CSaa7128_Destructor(CSaa7128 *This)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSaa7128_Destructor()\n"
				 ));

	CVideoEncoder_Destructor(&This->m_VideoEncoder);
}


