///////////////////////////////////////////////////////////////////////////////
//        
//
// Csil9011.c
//
// Description: 
//
//  SIL9011 implementation
// 
// History:
//	May 12, 2008 created
//
// Authors: Andy Yu
//
///////////////////////////////////////////////////////////////////////////////


#include "csil9011.h"
#include "qptimemgr.h"
#include "qpdbgmgr.h"
#include "util.h"


/////////////////////////////////////////////////////////////////////////////
//
//

#define	SIL9011_BRIGHTNESS_MIN				0x00   // not supported
#define	SIL9011_BRIGHTNESS_MAX				0xFF   // not supported
#define	SIL9011_BRIGHTNESS_DEFAULT		0x80   // not supported

#define	SIL9011_CONTRAST_MIN					0x00   // not supported
#define	SIL9011_CONTRAST_MAX					0x7F   // not supported
#define	SIL9011_CONTRAST_DEFAULT			0x44   // not supported

#define	SIL9011_SATURATION_MIN				0x00   // not supported
#define	SIL9011_SATURATION_MAX				0x7F   // not supported
#define	SIL9011_SATURATION_DEFAULT		0x40   // not supported

#define	SIL9011_HUE_MIN						(-128)	// 0x80   // not supported
#define	SIL9011_HUE_MAX						(+127)	// 0x7F   // not supported
#define	SIL9011_HUE_DEFAULT					0x00   // not supported

#define	SIL9011_SHARPNESS_MIN					0      // not supported
#define	SIL9011_SHARPNESS_MAX					15     // not supported
#define	SIL9011_SHARPNESS_DEFAULT			8      // not supported


extern TQP_DbgMCompID	g_QPPeripheralDBGCompID;


/////////////////////////////////////////////////////////////////////////////
//
//


static EQP_Errors CSIL9011_readRegister(CSIL9011 *This,
									   TQP_UINT8 regAddr, 
									   PTQP_UINT8 pValue
									   )
{
    return (II2C_WriteThenRead(This->m_pI2C,
							   This->m_cI2CDevAddr,	//device address
							   1,					//write size
							   &regAddr,			//write buffer
							   1,					//read size
							   pValue				//read buffer
							   ));
}



static EQP_Errors CSIL9011_writeRegister(CSIL9011 *This,
										TQP_UINT8 regAddr, 
										TQP_UINT8 value
										)
{
    TQP_UINT8 buffer[2];

	buffer[0] = regAddr;
	buffer[1] = value;

    return (II2C_Write(This->m_pI2C, 
					   This->m_cI2CDevAddr,	//device address
					   2,					//write size
					   buffer				//write buffer
					   ));
}


static EQP_Errors CSIL9011_WriteBlock(CSIL9011 *This,
									 TQP_UINT32 *pRegTable
									 )
{
	EQP_Errors	qpStatus = QPERR_SUCCESS;
	TQP_UINT32	dwTotalReg = 0;

	// get total number of registers to write
	//
	while (-1 != pRegTable[dwTotalReg]) 
	{
		dwTotalReg++;
	}

	dwTotalReg >>= 1;

	// do multiple writes
	//
	while (dwTotalReg)
	{
		qpStatus = CSIL9011_writeRegister(This,
										 (TQP_UINT8)pRegTable[0], 
										 (TQP_UINT8)pRegTable[1]
										 );
		if (!QP_SUCCESS(qpStatus))
		{
			QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
						 "CSIL9011_WriteBlock() ****** writeRegister failed! status(%d) ******\n",
						 qpStatus
						 ));
			break;
		}

		pRegTable += 2;
		dwTotalReg--;
	}

	return (qpStatus);
}



/////////////////////////////////////////////////////////////////////////////
//
//


// IGenericDevice
//
static EQP_Errors CSIL9011_InitDevice(IVideoDecoder *This_p, 
									 QP_DEVICE_CALLBACK pFuncCallback,
									 TQP_PVOID context,
									 TQP_PVOID pInitData
									 )
{
    TQP_UINT8     buffer[2];   // Andy 2008.10.06: Monterey, control ADI7401
    TQP_UINT8     ucRegAddr;
    TQP_UINT8     ucValue;

    CVideoDecoder *pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
    CSIL9011      *This = GET_INHERITED_OBJECT(CSIL9011, m_VideoDecoder, pVideoDecoder);

    static TQP_UINT32 dwSIL9011RegVals[] = 
    {
        0x08,0x05,
        0x08,0x05,
        0x09,0x91,
        0x4A,0x90,
        0xb5,0x04,
        //0x5f,0xD0,
        0x5f,0xC0,

        -1,-1
    };

    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
                 "CSIL9011_InitDevice()\n"
                ));

    if (CObject_IsInitialized(&pVideoDecoder->m_Object))
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
                     "CSIL9011_InitDevice() Already Initialized. Skipping...\n" 
                    ));
      return (QPERR_SUCCESS);
    }
	
    // save callback address and context
    //
    pVideoDecoder->m_pDeviceCallback = pFuncCallback;
    pVideoDecoder->m_callbackContext = context;

    // Andy 2008.10.06: enable SIL9011
    ucRegAddr = 0x00;
    II2C_WriteThenRead(This->m_pI2C,
                                  SIL9011_DEVICE_ADDR_2,        //device address: SIL9011 through device address 0x68
                                  1,           //write size
                                  &ucRegAddr,  //write buffer
                                  1,           //read size
                                  &ucValue     //read buffer
                                 );
    buffer[0] = 0x00;
    buffer[1] = (ucValue | 0xC0);   // set register 0x00 bit 6 & 7
    II2C_Write(This->m_pI2C,
                          SIL9011_DEVICE_ADDR_2,           //device address: SIL9011 through device address 0x68
                          2,              //write size
                          buffer          //write buffer
                         );

    // Initialize SIL9011 with the default value
    //
    CSIL9011_WriteBlock(This,
                        dwSIL9011RegVals
                       );

    // program the video decoder with settings read from the registry
    //
    QPVideoDecoder_SetVideoSystem(This_p, 
                                  pVideoDecoder->m_VidStandard
                                 );        
    QPVideoDecoder_SelectVideoSource(This_p, 
                                  pVideoDecoder->m_VidSource
                                 );

    QPVideoDecoder_SetBrightness(This_p, 
                                 pVideoDecoder->m_nBrightness
                                );
    QPVideoDecoder_SetContrast(This_p, 
                               pVideoDecoder->m_nContrast
                              );
    QPVideoDecoder_SetSaturation(This_p, 
                                 pVideoDecoder->m_nSaturation
                                );
    QPVideoDecoder_SetHue(This_p, 
                          pVideoDecoder->m_nHue
                         );
    QPVideoDecoder_SetSharpness(This_p, 
                                pVideoDecoder->m_nSharpness
                               );

    // CObject Init
    //
    CObject_Init(&pVideoDecoder->m_Object);

    return (QPERR_SUCCESS);
}



// IVideoDecoder
//
static EQP_Errors CSIL9011_SetVideoRes(IVideoDecoder *This_p,
									   TQP_UINT32 dwWid,
				  					   TQP_UINT32 dwHgt,
				  					   TQP_UINT32 dwScanFmt,
				  					   TQP_UINT32 dwFps
				  					   )
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSIL9011_SetVideoRes() : Video Res Set to %dx %d\n",
				 dwWid,
				 dwHgt
				 ));

	return (QPERR_SUCCESS);
}



// constructor
//
CSIL9011 * CSIL9011_Constructor(CSIL9011 *pSIL9011, 
								CObject *pParent,
								TQP_UINT32 dwAttributes,
								ITuner *pTuner,
								ITVAudio *pTvAudio,
								II2C *pI2C,
								TQP_UINT8 cI2CDevAddr
								)
{
	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSIL9011_Constructor()\n"
				 ));

	if (pSIL9011)
	{
		CVideoDecoder	*pVideoDecoder;

		pVideoDecoder = CVideoDecoder_Constructor(&pSIL9011->m_VideoDecoder, 
												  pParent,
												  dwAttributes,
												  pTuner,
												  pTvAudio
												  );

		if (!pVideoDecoder)
		{
			pSIL9011 = QP_NULL;
		}
		else
		{
			// fill in the function table
			//

			// IGenericDevice
			pSIL9011->m_VideoDecoder.m_iVideoDecoder.InitDevice = CSIL9011_InitDevice;

			// IVideoDecoder
			pSIL9011->m_VideoDecoder.m_iVideoDecoder.SetVideoRes = CSIL9011_SetVideoRes;

			// initialize members
			//
			pSIL9011->m_pI2C = pI2C;
			pSIL9011->m_cI2CDevAddr = cI2CDevAddr;

			// inherited members
			//
			pSIL9011->m_VideoDecoder.m_nBrightnessDef = SIL9011_BRIGHTNESS_DEFAULT;
			pSIL9011->m_VideoDecoder.m_nBrightnessMin = SIL9011_BRIGHTNESS_MIN;
			pSIL9011->m_VideoDecoder.m_nBrightnessMax = SIL9011_BRIGHTNESS_MAX;
			pSIL9011->m_VideoDecoder.m_nBrightness = SIL9011_BRIGHTNESS_DEFAULT;
			pSIL9011->m_VideoDecoder.m_nContrastDef = SIL9011_CONTRAST_DEFAULT; 
			pSIL9011->m_VideoDecoder.m_nContrastMin = SIL9011_CONTRAST_MIN;
			pSIL9011->m_VideoDecoder.m_nContrastMax = SIL9011_CONTRAST_MAX;
			pSIL9011->m_VideoDecoder.m_nContrast = SIL9011_CONTRAST_DEFAULT;
			pSIL9011->m_VideoDecoder.m_nSaturationDef = SIL9011_SATURATION_DEFAULT;
			pSIL9011->m_VideoDecoder.m_nSaturationMin = SIL9011_SATURATION_MIN;
			pSIL9011->m_VideoDecoder.m_nSaturationMax = SIL9011_SATURATION_MAX;
			pSIL9011->m_VideoDecoder.m_nSaturation = SIL9011_SATURATION_DEFAULT;
			pSIL9011->m_VideoDecoder.m_nHueDef = SIL9011_HUE_DEFAULT;
			pSIL9011->m_VideoDecoder.m_nHueMin = SIL9011_HUE_MIN;
			pSIL9011->m_VideoDecoder.m_nHueMax = SIL9011_HUE_MAX;
			pSIL9011->m_VideoDecoder.m_nHue = SIL9011_HUE_DEFAULT;
			pSIL9011->m_VideoDecoder.m_nSharpnessDef = SIL9011_SHARPNESS_DEFAULT;
			pSIL9011->m_VideoDecoder.m_nSharpnessMin = SIL9011_SHARPNESS_MIN;
			pSIL9011->m_VideoDecoder.m_nSharpnessMax = SIL9011_SHARPNESS_MAX;
			pSIL9011->m_VideoDecoder.m_nSharpness = SIL9011_SHARPNESS_DEFAULT;
		}
	}

	return (pSIL9011);
}



// destructor
//
void CSIL9011_Destructor(CSIL9011 *This)
{
    TQP_UINT8     buffer[2];
    TQP_UINT8     ucRegAddr;
    TQP_UINT8     ucValue;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CSIL9011_Destructor()\n"
				 ));

    // Andy 2008.10.06: disable SIL9011
    ucRegAddr = 0x00;
    II2C_WriteThenRead(This->m_pI2C,
                                  SIL9011_DEVICE_ADDR_2,        //device address: SIL9011 through device address 0x68
                                  1,           //write size
                                  &ucRegAddr,  //write buffer
                                  1,           //read size
                                  &ucValue     //read buffer
                                 );
    buffer[0] = 0x00;
    buffer[1] = (ucValue & 0x3F);   // clear register 0x00 bit 6 & 7
    II2C_Write(This->m_pI2C,
                          SIL9011_DEVICE_ADDR_2,           //device address: SIL9011 through device address 0x68
                          2,              //write size
                          buffer          //write buffer
                         );

    QPTMDelayMilliS(10);   // delay 10ms

    CVideoDecoder_Destructor(&This->m_VideoDecoder);
}

