///////////////////////////////////////////////////////////////////////////////
//        
//
// cfifo.c
//
// Description: 
//
//	Header file for the fifo.
//
// History:
//	Aug 4, 2006	(dt) initial check in
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


#include "cfifo.h"
#include "qpmemmgr.h"
#ifndef QP_LINXKER
#include <string.h>
#else //QP_LINXKER
#include <linux/string.h>
#endif //!QP_LINXKER


CFifo * CFifo_Constructor(CFifo *pFifo,
						  CObject *pParent, 
						  TQP_UINT32 dwAttributes,
						  TQP_UINT32 size,
						  TQP_UINT32 sizeEntry
						  )
{
	if (pFifo)
	{
		CObject_Constructor(&pFifo->m_Object,
		 					pParent, 
							dwAttributes
		 					);
		pFifo->m_dwReadPtr = 0;
		pFifo->m_dwWritePtr = 0;
		pFifo->m_dwFifoLevel = 0;
		pFifo->m_size = size;
		pFifo->m_sizeEntry = sizeEntry;

		pFifo->m_Fifo = (PTQP_UINT8)QPMMMalloc2Ex(sizeEntry * size);
		if (!pFifo->m_Fifo)
		{
			CFifo_Destructor(pFifo);
			pFifo = QP_NULL;
		}
	}

	return (pFifo);
}


void CFifo_Destructor(CFifo *This) 
{
	if (This->m_Fifo)
	{
		QPMMFree2Ex(This->m_Fifo);
		This->m_Fifo = QP_NULL;
	}

	CObject_Destructor(&This->m_Object);
}



TQP_BOOL CFifo_PeekFifo(CFifo *This, 
						TQP_PVOID val
						)
{
	TQP_BOOL ret;

	if (!This)
	{
		return (QP_FALSE);
	}

	ENTER_CRITICAL(&This->m_Object)

	ret = (This->m_dwReadPtr != This->m_dwWritePtr);
	if (ret)
	{
		memcpy(val, 
			   (TQP_PVOID)(This->m_Fifo + (This->m_dwReadPtr * This->m_sizeEntry)),
			   This->m_sizeEntry
			   );
	}

	LEAVE_CRITICAL(&This->m_Object)
	return (ret);
}



TQP_BOOL CFifo_GetFifo(CFifo *This, 
					   TQP_PVOID val
					   )
{
	TQP_BOOL ret;

	if (!This)
	{
		return (QP_FALSE);
	}

	ENTER_CRITICAL(&This->m_Object)

	ret = (This->m_dwReadPtr != This->m_dwWritePtr);
	if (ret)
	{
		memcpy(val, 
			   (TQP_PVOID)(This->m_Fifo + (This->m_dwReadPtr * This->m_sizeEntry)),
			   This->m_sizeEntry
			   );

		if (This->m_dwReadPtr >= This->m_dwWritePtr)
		{
			if (This->m_dwReadPtr == (This->m_size - 1))
				This->m_dwReadPtr = 0; // wrap around
			else
				This->m_dwReadPtr++;
		}
		else
			This->m_dwReadPtr++;

		// decrease number in fifo
		This->m_dwFifoLevel--;
	}

	LEAVE_CRITICAL(&This->m_Object)
	return (ret);
}



TQP_BOOL CFifo_SetFifo(CFifo *This, 
					   TQP_PVOID Data
					   )
{
	if (!This ||
		!This->m_Fifo)
	{
		return (QP_FALSE);
	}

	ENTER_CRITICAL(&This->m_Object)

	// sacrifice one fifo entry for efficency
	if (This->m_dwFifoLevel == This->m_size - 1)
	{
		LEAVE_CRITICAL(&This->m_Object)

		return (QP_FALSE);
	}

	// save it in the fifo
	memcpy((TQP_PVOID)(This->m_Fifo + (This->m_dwWritePtr * This->m_sizeEntry)),
		   Data,
		   This->m_sizeEntry
		   );

	// increase number of fifo
	This->m_dwFifoLevel++;

	// update write pointer
	if (This->m_dwWritePtr >= This->m_dwReadPtr)
	{
		if (This->m_dwWritePtr == (This->m_size - 1))
			This->m_dwWritePtr = 0; // wrap around
		else
			This->m_dwWritePtr++;
	}
	else
		This->m_dwWritePtr++;

	LEAVE_CRITICAL(&This->m_Object)
	return (QP_TRUE);
}



void CFifo_FlushFifo(CFifo *This) 
{
	if (This)
	{
		ENTER_CRITICAL(&This->m_Object)

		This->m_dwReadPtr = This->m_dwWritePtr = This->m_dwFifoLevel = 0;

		LEAVE_CRITICAL(&This->m_Object)
	}
}



void CFifo_UpdateReadPtr(CFifo *This)
{
	if (This)
	{
		ENTER_CRITICAL(&This->m_Object)

		// m_dwReadPtr == m_dwWritePtr should never happen in this case
		if (This->m_dwReadPtr >= This->m_dwWritePtr)
		{
			if (This->m_dwReadPtr == (This->m_size - 1))
				This->m_dwReadPtr = 0; // wrap around
			else
				This->m_dwReadPtr++;
		}
		else
			This->m_dwReadPtr++;

		// decrease number in fifo
		This->m_dwFifoLevel--;

		LEAVE_CRITICAL(&This->m_Object)
	}
}



TQP_BOOL CFifo_UpdateEntry(CFifo *This, 
						   TQP_PVOID val
						   )
{
	TQP_BOOL ret;

	if (!This ||
		!This->m_Fifo
		)
	{
		return (QP_FALSE);
	}

	ENTER_CRITICAL(&This->m_Object)

	ret = (This->m_dwReadPtr != This->m_dwWritePtr);
	if (ret)
	{
		memcpy((TQP_PVOID)(This->m_Fifo + (This->m_dwReadPtr * This->m_sizeEntry)),
			   val,
			   This->m_sizeEntry
			   );
	}

	LEAVE_CRITICAL(&This->m_Object)
	return (ret);
}



TQP_UINT32 CFifo_GetFifoLevel(CFifo *This)
{
	return (This? This->m_dwFifoLevel : 0);
}



TQP_BOOL CFifo_IsEmpty(CFifo *This)
{
	return (This ? (0 == This->m_dwFifoLevel) : QP_TRUE);
}

