///////////////////////////////////////////////////////////////////////////////
//        
//
// Cobjectmgr.c
//
// Description: 
//
//	the object that manages multiple objects.
// 
// History:
//	Aug 4, 2006	(dt) initial check in
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


#include "cobjectmgr.h"
#include "util.h"


static void CObjectMgr_RemoveObjects(CObjectMgr *This)
{
	OBJECT_ENTRY	*pthis_entry;
	OBJECT_ENTRY	*ptemp_entry;

	pthis_entry = This->m_pHead;	

	while (pthis_entry)
	{
		ptemp_entry = pthis_entry;
		pthis_entry = pthis_entry->pNext;
		if (This->m_pFuncCallback && (ptemp_entry->pObject))
			This->m_pFuncCallback(ptemp_entry->pObject);
		QPMMFree2Ex(ptemp_entry);
	}
	This->m_dwObjectNb = 0;
	This->m_pHead = QP_NULL;
}



CObjectMgr * CObjectMgr_Constructor(CObjectMgr *pObjectMgr,
									CObject *pParent, 
									TQP_UINT32 dwAttributes,
									OBJECTMGR_CALLBACK pFuncCallback
									)
{
	if (pObjectMgr)
	{
		CObject_Constructor(&pObjectMgr->m_Object, 
							pParent, 
							dwAttributes
							);
		pObjectMgr->m_hCurObject = 0;
		pObjectMgr->m_pHead = QP_NULL;
		pObjectMgr->m_dwObjectNb = 0;
		pObjectMgr->m_pFuncCallback = pFuncCallback;
	}

	return (pObjectMgr);
}



void CObjectMgr_Destructor(CObjectMgr *This)
{
	// release all the objects
	//
	CObjectMgr_RemoveObjects(This);

	CObject_Destructor(&This->m_Object);
}



TQP_OBJECT_HANDLE CObjectMgr_AddObject(CObjectMgr *This,
									   TQP_PVOID pObject
									   )
{
	TQP_OBJECT_HANDLE	hRet = 0;
	OBJECT_ENTRY		*pNewEntry = (OBJECT_ENTRY *)QPMMMalloc2Ex(sizeof(OBJECT_ENTRY));

	ENTER_CRITICAL(&This->m_Object)

	if (QP_NULL == This->m_pHead)
	{
		This->m_pHead = pNewEntry;
		This->m_pHead->hObject = 1;
		This->m_pHead->pObject = pObject;
		This->m_pHead->pNext = (OBJECT_ENTRY *)QP_NULL;
		This->m_dwObjectNb++;
		hRet = This->m_pHead->hObject;
	}
	else
	{		
		OBJECT_ENTRY	*ptemp_entry;
		
		if (This->m_pHead->hObject != 1)
		{
			ptemp_entry = pNewEntry;
			ptemp_entry->hObject = 1;
			ptemp_entry->pObject = pObject;
			ptemp_entry->pNext = This->m_pHead;
			This->m_pHead = ptemp_entry;
			This->m_dwObjectNb++;
			hRet = This->m_pHead->hObject;
		}
		else
		{
			OBJECT_ENTRY	*pthis_entry = This->m_pHead;
			OBJECT_ENTRY	*pnext_entry = This->m_pHead->pNext;
            
			while (pnext_entry && 
				   (pnext_entry->hObject == (pthis_entry->hObject + 1))
				   )
			{
				pthis_entry = pnext_entry;
				pnext_entry = pnext_entry->pNext;
			}
			ptemp_entry = pNewEntry;
			ptemp_entry->hObject = pthis_entry->hObject + 1;
			ptemp_entry->pObject = pObject;
			ptemp_entry->pNext = pnext_entry;
			pthis_entry->pNext = ptemp_entry;
			This->m_dwObjectNb++;
			hRet = ptemp_entry->hObject;
		}
	}

	LEAVE_CRITICAL(&This->m_Object)

	return (hRet);
}



TQP_PVOID CObjectMgr_RemoveObject(CObjectMgr *This,
								  TQP_OBJECT_HANDLE hObject
								  )
{
	OBJECT_ENTRY	*pthis_entry;
	OBJECT_ENTRY	*pprev_entry;
	OBJECT_ENTRY	*pRemove = QP_NULL;
	TQP_PVOID		pObject = QP_NULL;

	ENTER_CRITICAL(&This->m_Object)

	pthis_entry = This->m_pHead;
	if (pthis_entry)
	{
		if (pthis_entry->hObject == hObject)
		{
			pObject = pthis_entry->pObject;
			This->m_pHead = pthis_entry->pNext;
			pRemove = pthis_entry;
			This->m_dwObjectNb--;
		}
		else
		{
			do	{
				pprev_entry = pthis_entry;
				pthis_entry = pthis_entry->pNext;
				if (pthis_entry && (pthis_entry->hObject == hObject))
				{
					pprev_entry->pNext = pthis_entry->pNext;
					pObject = pthis_entry->pObject;
					pRemove = pthis_entry;
					This->m_dwObjectNb--;
					break;
				}		
			} while (pthis_entry);
		}
	}

	LEAVE_CRITICAL(&This->m_Object)

	if (pRemove)
		QPMMFree2Ex(pRemove);

	return (pObject);
}



TQP_PVOID CObjectMgr_GetObjectByHandle(CObjectMgr *This,
									   TQP_OBJECT_HANDLE hObject
									   )
{
	OBJECT_ENTRY	*pthis_entry;
	TQP_PVOID		pObject = QP_NULL;
	TQP_UINT32		i;

	ENTER_CRITICAL(&This->m_Object)

	pthis_entry = This->m_pHead;	

	for (i = 0; i < This->m_dwObjectNb; i++)
	{
		if (pthis_entry->hObject == hObject)
		{
			pObject = pthis_entry->pObject;
		}
		pthis_entry = pthis_entry->pNext;
	}

	LEAVE_CRITICAL(&This->m_Object)

	return (pObject);
}



TQP_OBJECT_HANDLE CObjectMgr_GetHandleByObject(CObjectMgr *This,
											   TQP_PVOID pObject
											   )
{
	OBJECT_ENTRY		*pthis_entry;
	TQP_OBJECT_HANDLE	hObject = TQP_NULL_HANDLE;
	TQP_UINT32			i;

	ENTER_CRITICAL(&This->m_Object)

	pthis_entry = This->m_pHead;	

	for (i = 0; i < This->m_dwObjectNb; i++)
	{
		if (pthis_entry->pObject == pObject)
		{
			hObject = pthis_entry->hObject;
			break;
		}
		pthis_entry = pthis_entry->pNext;
	}

	LEAVE_CRITICAL(&This->m_Object)

	return (hObject);
}



TQP_PVOID CObjectMgr_GetObjectByIndex(CObjectMgr *This,
									  TQP_UINT32 dwIndex
									  )
{
	OBJECT_ENTRY	*pthis_entry;
	TQP_PVOID		pObject = QP_NULL;
	TQP_UINT32		i;

	ENTER_CRITICAL(&This->m_Object)

	pthis_entry = This->m_pHead;	

	for (i = 0; i < This->m_dwObjectNb; i++)
	{
		if (i == dwIndex)
		{
			pObject = pthis_entry->pObject;
			break;
		}
		pthis_entry = pthis_entry->pNext;
	}

	LEAVE_CRITICAL(&This->m_Object)

	return (pObject);
}



TQP_OBJECT_HANDLE CObjectMgr_SetCurrentObject(CObjectMgr *This,
											  TQP_OBJECT_HANDLE hObject
											  )
{
	TQP_OBJECT_HANDLE	hold_object = This->m_hCurObject;

	This->m_hCurObject = hObject;
	return (hold_object);
}



TQP_OBJECT_HANDLE CObjectMgr_GetCurrentObject(CObjectMgr *This) 
{
	return (This->m_hCurObject);
}



TQP_UINT32 CObjectMgr_GetNumberOfObjects(CObjectMgr *This) 
{
	return (This->m_dwObjectNb);
}



