///////////////////////////////////////////////////////////////////////////////
//        
//
// Cqueue.c
//
// Description: 
//
//	Queue implementation.
// 
// History:
//	Aug 4, 2006	(dt) initial check in
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


#include "cobject.h"
#include "cqueue.h"
#include "qpdbgmgr.h"


CQueue * CQueue_Constructor(CQueue *pQueue,
							CObject *pParent, 
							TQP_UINT32 dwAttributes
							)
{
	if (pQueue)
	{
		CObject_Constructor(&pQueue->m_Object, 
							pParent, 
							dwAttributes
							);
		pQueue->m_dwNbInQueue = 0;
		CQueue_ResetQueue(pQueue);
	}

	return (pQueue);
}



void CQueue_Destructor(CQueue *This)
{
	CQueue_FlushQueue(This);

	CObject_Destructor(&This->m_Object);
}



// add one entry to the end of the queue
//
void CQueue_AddEntry(CQueue *This, 
					 QUEUE_ENTRY *pEntry
					 )
{
	ENTER_CRITICAL(&This->m_Object)

	if (pEntry)
	{
		if (This->m_Queue.pTail)
		{
			This->m_Queue.pTail->pNext = pEntry;
			This->m_Queue.pTail = pEntry;
		}
		else
		{
			This->m_Queue.pHead = pEntry;
			This->m_Queue.pTail = pEntry;
		}
		pEntry->pNext = QP_NULL;
		This->m_dwNbInQueue++;
	}

	LEAVE_CRITICAL(&This->m_Object)
}



// add one entry to the top of the queue
//
void CQueue_AddEntryToTop(CQueue *This, 
						  QUEUE_ENTRY *pEntry
						  )
{
	ENTER_CRITICAL(&This->m_Object)

	if (pEntry)
	{
        pEntry->pNext = This->m_Queue.pHead;
        This->m_Queue.pHead = pEntry;

	    if (!This->m_Queue.pTail)
	    {
		    This->m_Queue.pTail = pEntry;
	    }
	    This->m_dwNbInQueue++;
	}

	LEAVE_CRITICAL(&This->m_Object)
}



// remove the specified entry from the queue
//
QUEUE_ENTRY* CQueue_RemoveEntry(CQueue *This, 
								QUEUE_ENTRY *pEntry
								)
{
	QUEUE_ENTRY	*pthis_queue = This->m_Queue.pHead;
	QUEUE_ENTRY	*pprev_queue = QP_NULL;

	ENTER_CRITICAL(&This->m_Object)

	while (pthis_queue)
	{
		if (pthis_queue == pEntry)
		{
			if (!pprev_queue)
			{
				This->m_Queue.pHead = pthis_queue->pNext;
				if (!This->m_Queue.pHead)
					This->m_Queue.pTail = This->m_Queue.pHead;
			}
			else
			{
				pprev_queue->pNext = pthis_queue->pNext;
				if (!pthis_queue->pNext)
					This->m_Queue.pTail = pprev_queue;
			}
			This->m_dwNbInQueue--;
			break;
		}
		pprev_queue = pthis_queue;
		pthis_queue = pthis_queue->pNext;
	}

	LEAVE_CRITICAL(&This->m_Object)

	return (pthis_queue);
}



// get one entry actually remove the first entry from the queue.
//
QUEUE_ENTRY* CQueue_GetOneEntry(CQueue *This)
{
	QUEUE_ENTRY	*pEntry = QP_NULL;

	ENTER_CRITICAL(&This->m_Object)

	if (This->m_Queue.pHead)
	{
		pEntry = This->m_Queue.pHead;
		This->m_Queue.pHead = This->m_Queue.pHead->pNext;
		if (!This->m_Queue.pHead)
			This->m_Queue.pTail = This->m_Queue.pHead;
		This->m_dwNbInQueue--;
	}

	LEAVE_CRITICAL(&This->m_Object)

	return (pEntry);
}



// return the entry without remove it from the queue
//
QUEUE_ENTRY* CQueue_PeekEntryByData(CQueue *This, 
									TQP_PVOID Data
									)
{
	QUEUE_ENTRY	*pthis_queue = This->m_Queue.pHead;
	QUEUE_ENTRY	*pRet = QP_NULL;
		  
	ENTER_CRITICAL(&This->m_Object)

	while (pthis_queue)
	{
		if (pthis_queue->Data == Data)
		{
			pRet = pthis_queue;
			break;
		}
		pthis_queue = pthis_queue->pNext;
	}

	LEAVE_CRITICAL(&This->m_Object)

	return (pRet);
}



// remove the specified entry from the queue
//
QUEUE_ENTRY* CQueue_GetEntryByData(CQueue *This, 
								   TQP_PVOID Data
								   )
{
	QUEUE_ENTRY	*pthis_queue = This->m_Queue.pHead;
	QUEUE_ENTRY	*pprev_queue = QP_NULL;

	ENTER_CRITICAL(&This->m_Object)

	while (pthis_queue)
	{
		if (pthis_queue->Data == Data)
		{
			if (!pprev_queue)
			{
				This->m_Queue.pHead = pthis_queue->pNext;
				if (!This->m_Queue.pHead)
					This->m_Queue.pTail = This->m_Queue.pHead;
			}
			else
			{
				pprev_queue->pNext = pthis_queue->pNext;
				if (!pthis_queue->pNext)
					This->m_Queue.pTail = pprev_queue;
			}
			This->m_dwNbInQueue--;
			break;
		}
		pprev_queue = pthis_queue;
		pthis_queue = pthis_queue->pNext;
	}

	LEAVE_CRITICAL(&This->m_Object)

	return (pthis_queue);
}



QUEUE_ENTRY* CQueue_PeekOneEntry(CQueue *This)
{
	QUEUE_ENTRY	*pEntry;

	ENTER_CRITICAL(&This->m_Object)

	pEntry = This->m_Queue.pHead;

	LEAVE_CRITICAL(&This->m_Object)

	return (pEntry);
}



QUEUE_ENTRY* CQueue_PeekEntryByIndex(CQueue *This, 
									 TQP_UINT32 index
									 )
{
	QUEUE_ENTRY	*pthis_queue = This->m_Queue.pHead;
	QUEUE_ENTRY	*pRet = QP_NULL;
	TQP_UINT32	count = 0;
		  
	ENTER_CRITICAL(&This->m_Object)

	while (pthis_queue)
	{
		if (count++ == index)
		{
			pRet = pthis_queue;
			break;
		}
		pthis_queue = pthis_queue->pNext;
	}

	LEAVE_CRITICAL(&This->m_Object)

	return (pRet);
}



void CQueue_ResetQueue(CQueue *This) 
{
	This->m_Queue.pHead = This->m_Queue.pTail = QP_NULL;
	This->m_dwNbInQueue = 0;
}


// flush the queue
//
void CQueue_FlushQueue(CQueue *This)
{
	QUEUE_ENTRY	*pthis_queue;

	ENTER_CRITICAL(&This->m_Object)

	while (This->m_Queue.pHead)
	{
		pthis_queue = This->m_Queue.pHead;
		This->m_Queue.pHead = pthis_queue->pNext;
		This->m_dwNbInQueue--;
	}

	This->m_Queue.pTail = QP_NULL;

	LEAVE_CRITICAL(&This->m_Object)
}



TQP_BOOL CQueue_IsEmpty(CQueue *This) 
{
	return (This->m_Queue.pHead ? QP_FALSE : QP_TRUE);
}



TQP_UINT32 CQueue_GetQueueLevel(CQueue *This) 
{
	return (This->m_dwNbInQueue);
}



// is data already in queue?
//
TQP_BOOL CQueue_InQueue(CQueue *This, 
						TQP_PVOID Data
						)
{
	QUEUE_ENTRY	*pthis_queue = This->m_Queue.pHead;
	TQP_BOOL	bRet = QP_FALSE;

	ENTER_CRITICAL(&This->m_Object)

	while (pthis_queue)
	{
		if (pthis_queue->Data == Data)
		{
			bRet = QP_TRUE;
			break;
		}
		pthis_queue = pthis_queue->pNext;
	}

	LEAVE_CRITICAL(&This->m_Object)

	return (bRet);
}



void CQueue_WalkQueue(CQueue *This)
{
	QUEUE_ENTRY	*pthis_queue = This->m_Queue.pHead;

	ENTER_CRITICAL(&This->m_Object)

	QPDbgMPrintf((QPDBGM_LEVEL_TRACE, 
				 "CQueue::walkQueue() ######################################\n"
				 ));
	QPDbgMPrintf((QPDBGM_LEVEL_TRACE,
				 "CQueue::walkQueue() head = %lx, tail = %lx\n", 
				 This->m_Queue.pHead, 
				 This->m_Queue.pTail
				 ));
	while (pthis_queue)
	{
		QPDbgMPrintf((QPDBGM_LEVEL_TRACE,
					 "CQueue::walkQueue() ptr=%lx, next=%lx\n", 
					 pthis_queue, 
					 pthis_queue->pNext
					 ));
		pthis_queue = pthis_queue->pNext;
	}

	LEAVE_CRITICAL(&This->m_Object)
}

